Package: TemporalModelR
Type: Package
Title: Temporally Explicit Species Distribution Modelling
Version: 0.3.0
Authors@R: c(
    person(
      "Connor", "Hughes",
      email = "connorhughes926@gmail.com",
      role  = c("aut", "cre"),
      comment = c(ORCID = "0000-0002-3720-0837")
    ),
    person(
      "Mariana", "Castaneda-Guzman",
      email = "mcguzman1997@gmail.com",
      role  = c("aut"), 
      comment = c(ORCID = "0000-0001-6106-4284")
    ),
    person(
      "Luis", "E. Escobar",
      role  = c("aut"),
      comment = c(ORCID = "0000-0001-5735-2750")
    )
  )
Description: Increases the ease of implementing a temporally-explicit modeling
    methodology when building ecological niche and species distribution models.
    Provides functions to assist with three major steps of temporally-explicit
    models: (i) preprocessing species and environmental data and generating
    suitable background or pseudoabsence data, (ii) building a niche model and
    generating temporally-explicit predictions from that model, and (iii) model
    postprocessing to explore spatiotemporal trends in model predictions.
    Methodological and theoretical foundations are described in Ingenloff and
    Peterson (2021) <doi:10.1111/2041-210X.13564>, Franklin (2010, ISBN:9780521700023), 
    Peterson et al. (2011, ISBN:9780691136882), Blonder (2018) <doi:10.1111/ecog.03187>, 
    Senay et al. (2013) <doi:10.1371/journal.pone.0071218>, and Li and Zhang (2024) <doi:10.48550/arXiv.2404.05933>.
URL: https://github.com/CJHughes926/TemporalModelR,
        https://cjhughes926.github.io/TemporalModelR/
BugReports: https://github.com/CJHughes926/TemporalModelR/issues
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Depends: R (>= 4.1)
Imports: deldir, graphics, grDevices, sf, stats, terra, tools, utils,
        exactextractr
Suggests: fastcpd, ggplot2, hypervolume, knitr, mgcv, randomForest,
        RColorBrewer, rmarkdown, scatterpie
VignetteBuilder: knitr
Language: en-US
Config/roxygen2/version: 8.0.0
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2026-06-30 17:03:53 UTC; Conno
Author: Connor Hughes [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-3720-0837>),
  Mariana Castaneda-Guzman [aut] (ORCID:
    <https://orcid.org/0000-0001-6106-4284>),
  Luis E. Escobar [aut] (ORCID: <https://orcid.org/0000-0001-5735-2750>)
Maintainer: Connor Hughes <connorhughes926@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-02 07:20:24 UTC

# Generated by roxygen2: do not edit by hand

export(analyze_temporal_patterns)
export(analyze_trends_by_spatial_unit)
export(build_temporal_gam)
export(build_temporal_glm)
export(build_temporal_hv)
export(build_temporal_rf)
export(generate_absences)
export(generate_spatiotemporal_predictions)
export(plot_model_assessment)
export(raster_align)
export(scale_rasters)
export(spatiotemporal_partition)
export(spatiotemporal_rarefaction)
export(summarize_raster_outputs)
export(temporally_explicit_extraction)
importFrom(deldir,deldir)
importFrom(deldir,tile.list)
importFrom(exactextractr,exact_extract)
importFrom(grDevices,adjustcolor)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,dev.cur)
importFrom(grDevices,dev.off)
importFrom(grDevices,hcl.colors)
importFrom(grDevices,hcl.pals)
importFrom(grDevices,heat.colors)
importFrom(grDevices,png)
importFrom(grDevices,rainbow)
importFrom(grDevices,recordPlot)
importFrom(grDevices,terrain.colors)
importFrom(graphics,abline)
importFrom(graphics,axis)
importFrom(graphics,barplot)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,mtext)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(graphics,rect)
importFrom(graphics,rug)
importFrom(sf,sf_use_s2)
importFrom(sf,st_as_sf)
importFrom(sf,st_as_sfc)
importFrom(sf,st_bbox)
importFrom(sf,st_buffer)
importFrom(sf,st_cast)
importFrom(sf,st_collection_extract)
importFrom(sf,st_contains)
importFrom(sf,st_convex_hull)
importFrom(sf,st_coordinates)
importFrom(sf,st_crs)
importFrom(sf,st_difference)
importFrom(sf,st_distance)
importFrom(sf,st_drop_geometry)
importFrom(sf,st_geometry)
importFrom(sf,st_geometry_type)
importFrom(sf,st_intersection)
importFrom(sf,st_intersects)
importFrom(sf,st_is_empty)
importFrom(sf,st_join)
importFrom(sf,st_point)
importFrom(sf,st_point_on_surface)
importFrom(sf,st_polygon)
importFrom(sf,st_read)
importFrom(sf,st_sample)
importFrom(sf,st_sf)
importFrom(sf,st_sfc)
importFrom(sf,st_transform)
importFrom(sf,st_union)
importFrom(sf,st_voronoi)
importFrom(stats,as.formula)
importFrom(stats,binomial)
importFrom(stats,chisq.test)
importFrom(stats,coef)
importFrom(stats,complete.cases)
importFrom(stats,dbinom)
importFrom(stats,dist)
importFrom(stats,glm)
importFrom(stats,kmeans)
importFrom(stats,logLik)
importFrom(stats,median)
importFrom(stats,pchisq)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,var)
importFrom(terra,app)
importFrom(terra,classify)
importFrom(terra,crop)
importFrom(terra,crs)
importFrom(terra,ext)
importFrom(terra,extract)
importFrom(terra,focal)
importFrom(terra,freq)
importFrom(terra,global)
importFrom(terra,is.related)
importFrom(terra,mask)
importFrom(terra,mosaic)
importFrom(terra,ncell)
importFrom(terra,nlyr)
importFrom(terra,plot)
importFrom(terra,project)
importFrom(terra,rast)
importFrom(terra,res)
importFrom(terra,resample)
importFrom(terra,values)
importFrom(terra,vect)
importFrom(terra,writeRaster)
importFrom(terra,xyFromCell)
importFrom(tools,file_ext)
importFrom(tools,file_path_sans_ext)
importFrom(tools,toTitleCase)
importFrom(utils,capture.output)
importFrom(utils,combn)
importFrom(utils,head)
importFrom(utils,modifyList)
importFrom(utils,read.csv)
importFrom(utils,setTxtProgressBar)
importFrom(utils,txtProgressBar)
importFrom(utils,write.csv)

Temporally Explicit Species Distribution Modeling in R
================
Connor Hughes, Mariana Castaneda-Guzman, Luis E. Escobar

- [Background](#background)
- [Package description](#package-description)
- [Installing the package](#installing-the-package)
- [Workflow in TemporalModelR](#workflow-in-temporalmodelr)
  - [Preprocessing](#preprocessing)
  - [Modeling](#modeling)
  - [Postprocessing](#postprocessing)
- [Citation](#citation)
- [Note on AI usage](#note-on-ai-usage)
- [Contributing](#contributing)
- [Getting help](#getting-help)

<!-- README.md is generated from README.Rmd. Please edit that file -->
<!-- badges: start -->

[![R-CMD-check](https://github.com/CJHughes926/TemporalModelR/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/CJHughes926/TemporalModelR/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/TemporalModelR)](https://CRAN.R-project.org/package=TemporalModelR)
[![downloads](https://cranlogs.r-pkg.org/badges/grand-total/TemporalModelR)](https://cranlogs.r-pkg.org:443/badges/grand-total/TemporalModelR)
[![License:
MIT](https://img.shields.io/badge/license-MIT-lightgrey.svg?style=flat)](https://opensource.org/licenses/MIT)
<!-- badges: end -->

## Background

Ecological niche modeling (ENM) and species distribution modeling (SDM)
are widely used to reconstruct and predict species environmental
tolerances and geographic distributions, expanding ecological and
biogeographic theory. Despite their widespread use, one aspect that has
received relatively little attention is the temporally dynamic
environment.

Environmental conditions shaping species distributions vary over time,
including climate, resource availability, and habitat availability. As a
result, the suitability of any given location may change as key
environmental conditions become more or less favorable for a species.
Most existing implementations of ENM assume a static relationship
between species and environment: occurrences are pooled across years,
predictors are time-averaged, and a single model is fit to the combined
dataset. This works well when relationships are genuinely stable, but it
carries three costs when they aren’t:

1.  **Temporal mismatch.** Pairing each occurrence with long-term
    average conditions can misrepresent the environment the species
    actually experienced at the moment it was recorded, leading to
    over-generalized niche estimates and higher rates of both
    extrapolation and omission.
2.  **Lost training data.** Standard spatial rarefaction retains one
    point per pixel, discarding temporally independent observations from
    the same location. When environments at a location have changed over
    time, those repeat observations carry unique information about the
    species niche.
3.  **No temporal dynamics in the output.** Static models produce a
    single predictive surface and cannot directly answer questions about
    how suitable conditions for a species have changed over time, or
    what environmental drivers are responsible for those changes.

Temporally-explicit ENM offer an alternative practice which preserve the
time stamp on each observation. Each occurrence is paired with the
environmental conditions it experienced *at the time step it was
recorded*, rather than against a long-term average or other static
variable. The resulting models are sensitive to year-to-year (or
season-to-season) variation in the species-environment relationship, and
predictions are projected as time series rather than as a single static
surface.

**TemporalModelR** is an R package for building temporally explicit SDMs
end-to-end, from raw occurrence records to time-resolved trend analyses.
It provides a consistent and replicable workflow for implementing
temporally explicit ENM.

<br>

## Package description

**TemporalModelR** operates across two complementary dimensions to
support temporally explicit modeling:

- **E-space (Environmental Space)** represents a location in terms of
  its environmental conditions, independent of geography. A species
  niche in E-space is its tolerance for a given set of environmental
  variables.
- **G-space (Geographic Space)** represents a location in terms of where
  it physically sits on the landscape. Every point in G-space has a
  corresponding location in E-space.

Because environmental conditions change over time, the same point in
G-space may move through E-space across years, decades, or seasons.
Traditional, temporally-static workflows assume both G-space and E-space
to be stable in time. TemporalModelR assumes only that the *niche*
(i.e., the species tolerance in E-space) is stable across the study
period, while the E-space coordinates of any G-space location may change
over time. Species observations are therefore matched to E-space data
correct in both space and time, the niche is modeled in time-independent
E-space. The resulting niche estimate is projected back onto G-space at
explicit time periods to produce temporally dynamic predictions across
time rather than a single static surface.

The package structures the SDM workflow around three phases:

- **Preprocessing**: aligning environmental rasters to a common grid,
  rarefying occurrence data in space and time, extracting environmental
  values matched to observation time, scaling rasters to a standardized
  range, partitioning into spatially and/or temporally independent
  cross-validation folds, and generating fold-stratified pseudoabsences
  for presence/absence models.
- **Modeling**: fitting one of four supported algorithms (GLM, GAM,
  random forest, hypervolume) per cross-validation fold, with automatic
  threshold selection and per-fold evaluation metrics in both E-space
  (time-independent) and G-space (time-specific).
- **Postprocessing**: combining per-fold predictions into a consensus
  binary stack for more straightforward assessments, classifying each
  pixel’s temporal trajectory via changepoint detection into
  never-suitable, always-suitable, no pattern, increasing, decreasing,
  or fluctuating in suitability, and aggregating patterns across
  user-defined spatial units like administrative regions or watersheds.

Every function in the package is built around the concept of **time
placeholders** in raster filenames (`forest_cover_2005.tif`,
`pr_seas_2003_Spring.tif`). Patterns map clean predictor names to
filenames; `time_cols` defines which columns in your occurrence data
correspond to those placeholders. For example, a column in species
occurrence data which labeled `year` with subsequent values `2001`,
`2002`, `2005`, and a user defined variable name as `forest_cover_year`,
would match to rasters `forest_cover_2001`, `forest_cover_2002`, and
`forest_cover_2005`. Likewise, a databse with both columns `year` and
`month` and a user defined variable name of `pr_m_year_month` would
match to rasters like `pr_m_1997_09`.

This makes the workflow flexible to whether your time scale is annual,
monthly, daily, or anything in between, so long as you have rasters data
at the same temporal level. It also allows different predictors to vary
at different time scales within the same model, flexibly allowing
temporally static variables (e.g. elevation) to be processed in the same
suite and used in the same models as temporally dynamic variables, which
may themselves be measured at different time steps (e.g. annual forest
cover estimates, monthly precipitation).

<br>

## Installing the package

Note: Internet connection is required to install the package.

To install the latest release of TemporalModelR from CRAN use the
following line of code:

``` r
install.packages("TemporalModelR")
```

The development version of TemporalModelR can be installed using the
code below.

``` r
# install.packages("devtools")
devtools::install_github("CJHughes926/TemporalModelR")
```

<br>

## Workflow in TemporalModelR

A typical TemporalModelR workflow follows these steps:

1.  **Preprocess** — align rasters, rarefy occurrences, extract and
    optionally scale environmental values, partition into folds, and
    generate pseudoabsences (if needed)
2.  **Model** — fit a GLM, GAM, random forest, or hypervolume model,
    assess time-independent model performance. Project the fitted model
    across space and time, assess temporally-explicit model performance
3.  **Postprocess** — summarize predictions to consensus surfaces,
    identify temporal patterns, and aggregate by spatial unit

A brief description of each step is presented below. For full
walkthrough of each phase, see the vignettes hosted on the
[TemporalModelR site](https://CJHughes926.github.io/TemporalModelR/)
under the *Articles* menu, also linked inline throughout this README.

The package ships a small synthetic dataset that every vignette runs
against, so the full workflow can be reproduced end-to-end without
downloading any external data. The dataset includes raw and aligned
environmental rasters, an occurrence point database, intermediate point
files, and pre-computed `data()` objects from each phase of the
workflow. Reading the data set vignette first is recommended, as it
provides the shared context every other vignette refers back to.

> For a description of the synthetic data and the bundled `data()`
> objects used in the workflow vignettes, see the [About the Example
> Dataset](https://cjhughes926.github.io/TemporalModelR/articles/V1_dataset.html)
> vignette.

<br>

### Preprocessing

The preprocessing pipeline transforms raw occurrence records and
environmental rasters into the structured inputs that downstream models
expect. Six functions cover this phase. `raster_align()` standardizes
rasters to a common projection, extent, and resolution to prevent
misalignment in future analyses. `spatiotemporal_rarefaction()` subsets
data to one point per pixel per time step, reducing pseudoreplication
from your data but preserving truly unique temporally independent
observations from the same location. `temporally_explicit_extraction()`
extracts environmental values from the raster that matches each point’s
observation time step, producing a table of observations with their
relevant E-space conditions correct in both space and time, along with
the means and standard deviations of each variable across the data set.
`scale_rasters()` optionally uses those scaling parameters to z-score
every raster, ensuring variables share a comparable range and contribute
evenly to subsequent modeling algorithms sensitive to that.
`spatiotemporal_partition()` builds cross-validation folds via one of
four methods: purely spatial, purely temporal, spatiotemporal (combining
both), or random as a baseline. Finally, `generate_absences()` produces
fold-stratified pseudoabsences relevant for presence/absence models.

> For a detailed workflow showcasing each preprocessing function, see
> the [Preprocessing temporally explicit
> data](https://cjhughes926.github.io/TemporalModelR/articles/V2_Preprocessing.html)
> vignette.

<br>

### Modeling

TemporalModelR supports four modeling algorithms, all of which share the
same data requirements (`partition_result`, `pseudoabsence_result` for
presnce/absence models, a `model_formula` or `model_vars`). One model is
fit per cross-validation fold; each is evaluated on its held-out test
set, and for any continuous models a threshold is selected and applied
to produce only binary results.

- **`build_temporal_glm()`** fits one generalized linear model per fold
  with a binomial link function. Supports any combination of linear,
  polynomial, and interactive terms via standard R formula syntax.
- **`build_temporal_gam()`** fits one generalized additive model per
  fold via `mgcv::gam()`. Allows for complex nonlinear relationships
  between environmental gradients and species presence.
- **`build_temporal_rf()`** fits one random forest classifier per fold
  via `randomForest::randomForest()`. Uses ensembles of decision trees
  to capture complex nonlinear relationships and interactions.
- **`build_temporal_hv()`** constructs n-dimensional hypervolumes per
  fold via Gaussian kernel density estimation or one-class support
  vector machine. The presence-only option; no pseudoabsences required.

Each model builder produces evaluation metrics across folds in E-space
(time-independent), capturing overall niche estimation quality
regardless of when the test points were observed. E-space evaluation
gives a stable picture of model performance even in years where
occurrence records are sparse.

Once a model is fit, `generate_spatiotemporal_predictions()` projects it
across user-defined time steps. The function accepts any of the four
model types via the `model_result` argument and writes one prediction
raster per fold per time step alongside per-time-step evaluation
metrics. The same call handles single-timestep and multi-timestep
workflows given the relevant raster data is available: for example
`time_steps = data.frame(year = 1:15)` produces 15 annual maps;
`expand.grid(year = 1:15, season = "Spring")` produces 15 spring
snapshots; `expand.grid(year = 1:15, season = c("Spring", "Summer"))`
produces 30 maps spanning two seasons.

Once predictions are projected into G-space, time-specific assessment
metrics become available, including cumulative binomial probability
(CBP) tests assessing whether predictions in a given time step are
better than random. G-space evaluation complements the E-space metrics
from model fitting and both may be considered in model evaluation:
G-space is robust and specific in years with substantial sample sizes
but loses meaning when few records exist for a given time step, while
E-space gives a more stable overall picture independent of time.
Reporting both gives the most complete view of model performance.

> For a full walkthrough of each modeling algorithm, including formula
> syntax, parameter choices, G-space projection, and diagnostic
> interpretation, see the four modeling vignettes: [Modeling with a
> GLM](https://cjhughes926.github.io/TemporalModelR/articles/V3a_GLM.html),
> [Modeling with a
> GAM](https://cjhughes926.github.io/TemporalModelR/articles/V3b_GAM.html),
> [Modeling with a Random
> Forest](https://cjhughes926.github.io/TemporalModelR/articles/V3c_RF.html),
> and [Modeling with a
> Hypervolume](https://cjhughes926.github.io/TemporalModelR/articles/V3d_HV.html).

<br>

### Postprocessing

The postprocessing phase converts the per-fold prediction rasters into
interpretable summaries. `summarize_raster_outputs()` collapses
fold-level predictions into a binary consensus stack (where a
user-specified number of folds agree on suitability) and a frequency
raster (proportion of time steps each pixel was suitable) which explored
the stability of G-space suitability projections across time.
`analyze_temporal_patterns()` uses the `fastcpd` package to detect
structured changes in each pixel’s suitability across the time series,
classifying trajectories as never-suitable, always-suitable, no pattern,
increasing, decreasing, or fluctuating. The changepoint analysis
accounts for temporal autocorrelation by including each pixel’s
previous-time-step value as a predictor in the change point detection
model; users can optionally enable spatial autocorrelation, which adds
the proportion of a pixel’s first-order (queen’s case) neighbors
predicted as suitable as an additional covariate.
`analyze_trends_by_spatial_unit()` aggregates pixel-level patterns
across user-defined polygons (administrative units, watersheds,
ecoregions), producing useful tables summarizing losses and gains in
suitability over time across a these regions, and is useful for
identifying regional variation in suitability trends.

> For the full postprocessing workflow including consensus threshold
> selection, changepoint diagnostics, and zone aggregation, see the
> [Postprocessing
> predictions](https://cjhughes926.github.io/TemporalModelR/articles/V4_Postprocessing.html)
> vignette.

<br>

## Citation

If you use TemporalModelR in your research, please cite:

Hughes, C., Castaneda-Guzman, M., & Escobar, L.E. (2026).
*TemporalModelR: An R Package for temporally-explicit species
distribution modeling.* \[Journal information to be added\]

<br>

## Note on AI usage

To maintain high standards of code quality and documentation, we have
used AI LLM tools while constructing code for this package. We used
these tools for grammatical polishing and exploring technical
implementation strategies for specialized functions. We manually checked
and tested all code and documentation refined with these tools.

<br>

## Contributing

We welcome contributions to improve `TemporalModelR`. To maintain the
integrity and performance of the package, we follow a few core
principles:

- **Quality over quantity**: we prioritize well-thought-out, stable
  improvements over frequent, minor changes. Please ensure your code is
  well-documented and follows the existing style of the package.
- **Minimal dependencies**: one of the goals of TemporalModelR is to
  remain efficient. We prefer solutions that use base R or existing
  dependencies. Proposals that introduce new package dependencies will
  be evaluated for their necessity.
- **AI-assisted code**: if you use AI tools to generate code
  alternatives or improvements, please manually verify the logic and
  accuracy of the output and demonstrate the benefit in your Pull
  Request.
- **Testing**: new features should include examples, and tests should be
  performed to ensure they work as intended and do not break existing
  workflows.

If you have an idea for a major change, please open an Issue first to
discuss it with the maintainers.

<br>

## Getting help

- Report bugs or request features: [GitHub
  Issues](https://github.com/CJHughes926/TemporalModelR/issues)
- Ask questions: [GitHub
  Discussions](https://github.com/CJHughes926/TemporalModelR/discussions)

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 5,
  dpi       = 100,
  out.width = "95%"
)

## -----------------------------------------------------------------------------
library(TemporalModelR)
library(terra)
library(sf)

raw_dir  <- system.file("extdata/rasters_raw", package = "TemporalModelR")

pts_file <- system.file("extdata/points/synthetic_occurrence_points.csv",
                        package = "TemporalModelR")

ref_file <- file.path(raw_dir, "elevation.tif")
study_crs <- sf::st_crs(terra::rast(ref_file))

## ----eval=FALSE---------------------------------------------------------------
# aligned_dir <- file.path(tempdir(), "rasters_aligned")
# 
# raster_align(
#   input_dir        = raw_dir,
#   output_dir       = aligned_dir,
#   reference_raster = ref_file,
#   resample_method  = "bilinear",
#   overwrite        = TRUE
# )

## -----------------------------------------------------------------------------
aligned_dir <- system.file("extdata/rasters_aligned", package = "TemporalModelR")

list.files(aligned_dir, pattern = "\\.tif$")[1:6]

## -----------------------------------------------------------------------------
rare_dir <- file.path(tempdir(), "rarefied")

rare_out <- spatiotemporal_rarefaction(
  points_sp        = pts_file,
  output_dir       = rare_dir,
  reference_raster = ref_file,
  time_cols        = c("year", "season"),
  xcol             = "x",
  ycol             = "y",
  points_crs       = study_crs,
  output_prefix    = "Pts_seasonal",
  verbose          = FALSE
)

rare_out$input_points

rare_out$spatial_points

rare_out$spatiotemporal_points

## -----------------------------------------------------------------------------
rare_out$files_created

## -----------------------------------------------------------------------------
rare_dir <- file.path(tempdir(), "rarefied")

rare_out_annual <- spatiotemporal_rarefaction(
  points_sp        = pts_file,
  output_dir       = rare_dir,
  reference_raster = ref_file,
  time_cols        = "year",
  xcol             = "x",
  ycol             = "y",
  points_crs       = study_crs,
  output_prefix    = "Pts_ann",
  verbose          = FALSE
)

rare_out_annual$input_points

rare_out_annual$spatial_points

rare_out_annual$spatiotemporal_points

## -----------------------------------------------------------------------------
ext_dir <- file.path(tempdir(), "extracted")

ext_out <- temporally_explicit_extraction(
  points_sp           = rare_out$files_created$spatiotemporal,
  raster_dir          = aligned_dir,
  variable_patterns   = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR",
    "prseas"       = "prseas_YEAR_SEASON"
  ),
  time_cols           = c("year", "season"),
  xcol                = "X",
  ycol                = "Y",
  points_crs          = study_crs,
  output_dir          = ext_dir,
  output_prefix       = "extracted_seasonal",
  save_raw            = TRUE,
  save_scaled         = TRUE,
  save_scaling_params = TRUE,
  verbose             = FALSE
)

head(ext_out$raw_values)

## -----------------------------------------------------------------------------
ext_out$files_created

## -----------------------------------------------------------------------------
ext_out$scaling_params

## ----eval=FALSE---------------------------------------------------------------
# scaled_dir <- file.path(tempdir(), "rasters_scaled")
# 
# scale_rasters(
#   input_dir           = aligned_dir,
#   output_dir          = scaled_dir,
#   scaling_params_file = ext_out$files_created$scaling_params,
#   variable_patterns   = c(
#     "elevation"    = "elevation",
#     "forest_cover" = "forest_cover_YEAR",
#     "prseas"       = "prseas_YEAR_SEASON"
#   ),
#   time_cols           = c("year", "season"),
#   overwrite           = TRUE,
#   verbose             = FALSE
# )

## -----------------------------------------------------------------------------
scaled_dir <- system.file("extdata/rasters_scaled", package = "TemporalModelR")

list.files(scaled_dir, pattern = "\\.tif$")[1:6]

## -----------------------------------------------------------------------------
ext_scaled_file <- system.file(
  "extdata/points/extracted_seasonal_Scaled_Values.csv",
  package = "TemporalModelR"
)

study_area_sf <- sf::st_as_sf(sf::st_as_sfc(
  sf::st_bbox(c(xmin = 0, xmax = 3000, ymin = 0, ymax = 1500),
              crs = study_crs)
))

partition <- spatiotemporal_partition(
  reference_shapefile_path = study_area_sf,
  points_file_path         = ext_scaled_file,
  xcol                     = "x",
  ycol                     = "y",
  points_crs               = study_crs,
  time_cols                = "year",
  n_spatial_folds          = 2,
  n_temporal_folds         = 2,
  max_attempts = 10,
  max_imbalance = 0.15,
  create_plot              = TRUE,
  verbose                  = FALSE
)

partition$summary

## -----------------------------------------------------------------------------
absences <- generate_absences(
  partition_result         = partition,
  reference_shapefile_path = study_area_sf,
  raster_dir               = scaled_dir,
  variable_patterns        = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR",
    "prseas"       = "prseas_YEAR_SEASON"
  ),
  method                   = "buffer",
  buffer_distance          = 300,
  ratio                    = 2,
  time_cols                = c("year", "season"),
  create_plot              = TRUE,
  plot_by_fold = TRUE,
  verbose                  = FALSE
)


absences$summary

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 5,
  dpi       = 100,
  out.width = "95%"
)

## ----eval=FALSE---------------------------------------------------------------
# install.packages("mgcv")

## -----------------------------------------------------------------------------
library(TemporalModelR)
library(terra)

data(tmr_partition, package = "TemporalModelR")

data(tmr_absences,  package = "TemporalModelR")

## -----------------------------------------------------------------------------
gam_out <- build_temporal_gam(
  partition_result     = tmr_partition,
  pseudoabsence_result = tmr_absences,
  model_formula        = ~ s(elevation) + s(forest_cover) + s(prseas),
  link                 = "logit",
  gam_params           = list(method = "REML"),
  threshold_method     = "tss",
  output_dir           = file.path(tempdir(), "GAM_Models"),
  create_plot          = TRUE,
  time_cols            = c("year", "season"),
  verbose              = FALSE
)

## -----------------------------------------------------------------------------
class(gam_out)

names(gam_out)

## ----eval=FALSE---------------------------------------------------------------
# ### Univariate smooths
# ~ s(elevation) + s(forest_cover) + s(prseas)
# 
# ### Mix smooth and parametric terms
# ~ s(elevation) + forest_cover + prseas
# 
# ### Tensor product smooth between forest cover and precipitation
# ~ s(elevation) + te(forest_cover, prseas)
# 
# ### Tensor interaction with main effects also kept
# ~ s(elevation) + s(forest_cover) + s(prseas) + ti(forest_cover, prseas)
# 
# ### Constrained basis dimension and explicit basis type
# ~ s(elevation, k = 5, bs = "tp") + s(forest_cover, k = 5) + s(prseas, k = 5)

## -----------------------------------------------------------------------------
gam_out$thresholds

## -----------------------------------------------------------------------------
gam_out$fold_test_metrics

## ----eval=FALSE---------------------------------------------------------------
# scaled_dir <- system.file("extdata/rasters_scaled", package = "TemporalModelR")
# 
# time_steps <- expand.grid(
#   year             = 1:15,
#   season           = c("Spring", "Summer", "Autumn", "Winter"),
#   stringsAsFactors = FALSE
# )
# 
# preds <- generate_spatiotemporal_predictions(
#   partition_result     = tmr_partition,
#   model_result         = gam_out,
#   pseudoabsence_result = tmr_absences,
#   raster_dir           = scaled_dir,
#   variable_patterns    = c(
#     "elevation"    = "elevation",
#     "forest_cover" = "forest_cover_YEAR",
#     "prseas"       = "prseas_YEAR_SEASON"
#   ),
#   time_cols            = c("year", "season"),
#   time_steps           = time_steps,
#   output_dir           = file.path(tempdir(), "GAM_Predictions"),
#   overwrite            = TRUE,
#   verbose              = FALSE
# )

## ----echo=FALSE---------------------------------------------------------------
preds <- readRDS(system.file("extdata/precomputed/gam/preds.rds",
                             package = "TemporalModelR"))
preds$prediction_files <- list.files(
  system.file("extdata/precomputed/gam/pred_tifs", package = "TemporalModelR"),
  pattern = "\\.tif$", full.names = TRUE
)

## ----fig.width=10, fig.height=5-----------------------------------------------
pred_stack <- terra::rast(preds$prediction_files)

pred_names    <- basename(preds$prediction_files)
pred_seasons  <- sub(".*_(Spring|Summer|Autumn|Winter)\\.tif$", "\\1", pred_names)
pred_years    <- as.numeric(sub(".*_(\\d+)_(Spring|Summer|Autumn|Winter)\\.tif$",
                                "\\1", pred_names))
season_levels <- c("Spring", "Summer", "Autumn", "Winter")
stack_order   <- order(pred_years, match(pred_seasons, season_levels))

pred_stack <- pred_stack[[stack_order]]
ordered_years   <- pred_years[stack_order]
ordered_seasons <- pred_seasons[stack_order]
names(pred_stack) <- paste0("Y", ordered_years, "_", ordered_seasons)

block1 <- which(ordered_years %in%  1:4)
block2 <- which(ordered_years %in%  5:8)
block3 <- which(ordered_years %in%  9:12)
block4 <- which(ordered_years %in% 13:15)

## ----fig.width=10, fig.height=5-----------------------------------------------
terra::plot(pred_stack[[block1]], nr = 4, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0), legend = FALSE)

## ----fig.width=10, fig.height=5-----------------------------------------------
terra::plot(pred_stack[[block2]], nr = 4, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0), legend = FALSE)

## ----fig.width=10, fig.height=5-----------------------------------------------
terra::plot(pred_stack[[block3]], nr = 4, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0), legend = FALSE)

## ----fig.width=10, fig.height=4-----------------------------------------------
terra::plot(pred_stack[[block4]], nr = 3, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0), legend = FALSE)

## -----------------------------------------------------------------------------
head(preds$timestep_metrics)

## -----------------------------------------------------------------------------
preds$overall_summary

## -----------------------------------------------------------------------------
plot_model_assessment(
  predictions         = preds,
  time_column         = c("year", "season"),
  secondary_time_mode = "combine",
  model_result        = gam_out
)

## ----fig.height=20------------------------------------------------------------
plot_model_assessment(
  predictions         = preds,
  time_column         = c("year", "season"),
  secondary_time_mode = "facet",
  model_result        = gam_out,
  cbp_threshold       = 0.001
)

#' @keywords internal
#' @noRd
.natural_sort <- function(paths) {
  keys <- lapply(basename(paths), function(f) {
    parts <- strsplit(f, "[^0-9]+", perl = TRUE)[[1]]
    parts <- suppressWarnings(as.numeric(parts))
    parts[!is.na(parts)]
  })
  paths[order(vapply(keys,
                     function(k) paste(sprintf("%020.0f", k), collapse = "_"),
                     character(1)))]
}

#' @keywords internal
#' @noRd
#' @importFrom sf sf_use_s2 st_as_sf st_as_sfc st_bbox st_buffer st_cast
#' @importFrom sf st_collection_extract st_contains st_convex_hull st_coordinates
#' @importFrom sf st_crs st_distance st_drop_geometry st_geometry st_geometry_type
#' @importFrom sf st_intersection st_intersects st_is_empty st_join st_polygon
#' @importFrom sf st_read st_sample st_sf st_sfc st_union st_voronoi
#' @importFrom deldir deldir tile.list
#' @importFrom stats as.formula binomial chisq.test coef dist glm kmeans setNames
#' @importFrom stats logLik pchisq pnorm
#' @importFrom grDevices adjustcolor colorRampPalette hcl.colors hcl.pals recordPlot
#' @importFrom graphics abline barplot legend mtext par plot points
#' @importFrom tools file_ext toTitleCase
#' @importFrom utils read.csv
.test_cp_likelihood <- function(data, cp, alpha = 0.05, use_neighbor = TRUE) {
  n <- nrow(data)
  seg1_data <- data[1:cp, ]
  seg2_data <- data[(cp + 1):n, ]

  if (nrow(seg1_data) < 3 || nrow(seg2_data) < 3) {
    return(list(p_value = NA, significant = FALSE, test_statistic = NA))
  }

  tryCatch({
    if (use_neighbor) {
      model_full <- stats::glm(y ~ lag1 + neighbor, data = data, family = stats::binomial())
      model_seg1 <- stats::glm(y ~ lag1 + neighbor, data = seg1_data, family = stats::binomial())
      model_seg2 <- stats::glm(y ~ lag1 + neighbor, data = seg2_data, family = stats::binomial())
    } else {
      model_full <- stats::glm(y ~ lag1, data = data, family = stats::binomial())
      model_seg1 <- stats::glm(y ~ lag1, data = seg1_data, family = stats::binomial())
      model_seg2 <- stats::glm(y ~ lag1, data = seg2_data, family = stats::binomial())
    }

    loglik_full <- stats::logLik(model_full)[1]
    loglik_seg <- stats::logLik(model_seg1)[1] + stats::logLik(model_seg2)[1]
    lr_stat <- 2 * (loglik_seg - loglik_full)
    df <- length(stats::coef(model_seg1)) + length(stats::coef(model_seg2)) - length(stats::coef(model_full))
    p_value <- 1 - stats::pchisq(lr_stat, df = max(df, 1))

    list(p_value = p_value, significant = p_value < alpha, test_statistic = lr_stat, df = df)
  }, error = function(e) {
    list(p_value = NA, significant = FALSE, test_statistic = NA)
  })
}

#' @keywords internal
#' @noRd
.test_cp_permutation <- function(data, cp, n_perm = 1000, alpha = 0.05) {
  n <- nrow(data)
  seg1_mean <- mean(data$y[1:cp])
  seg2_mean <- mean(data$y[(cp + 1):n])
  obs_stat <- abs(seg1_mean - seg2_mean)

  perm_stats <- replicate(n_perm, {
    perm_y <- sample(data$y)
    abs(mean(perm_y[1:cp]) - mean(perm_y[(cp + 1):n]))
  })

  p_value <- mean(perm_stats >= obs_stat)
  list(p_value = p_value, significant = p_value < alpha,
       test_statistic = obs_stat, effect_size = obs_stat)
}

#' @keywords internal
#' @noRd
#' @importFrom stats pnorm
.test_cp_proportion <- function(data, cp, alpha = 0.05) {
  n <- nrow(data)
  seg1_y <- data$y[1:cp]
  seg2_y <- data$y[(cp + 1):n]

  n1 <- length(seg1_y)
  n2 <- length(seg2_y)
  p1 <- mean(seg1_y)
  p2 <- mean(seg2_y)
  p_pooled <- (sum(seg1_y) + sum(seg2_y)) / (n1 + n2)

  if (p_pooled == 0 || p_pooled == 1 || p1 == p2) {
    return(list(p_value = 1, significant = FALSE, test_statistic = 0, effect_size = abs(p1 - p2)))
  }

  se <- sqrt(p_pooled * (1 - p_pooled) * (1 / n1 + 1 / n2))
  if (se == 0) {
    return(list(p_value = 1, significant = FALSE, test_statistic = 0, effect_size = abs(p1 - p2)))
  }

  z_stat <- (p1 - p2) / se
  p_value <- 2 * (1 - stats::pnorm(abs(z_stat)))

  list(p_value = p_value, significant = p_value < alpha,
       test_statistic = z_stat, effect_size = abs(p1 - p2))
}

#' @keywords internal
#' @noRd
#' @importFrom stats chisq.test
.test_cp_chisquare <- function(data, cp, alpha = 0.05) {
  n <- nrow(data)
  seg1_y <- data$y[1:cp]
  seg2_y <- data$y[(cp + 1):n]

  cont_table <- matrix(c(
    sum(seg1_y == 1), sum(seg1_y == 0),
    sum(seg2_y == 1), sum(seg2_y == 0)
  ), nrow = 2, byrow = TRUE)

  test_result <- stats::chisq.test(cont_table, correct = FALSE)

  list(p_value = test_result$p.value, significant = test_result$p.value < alpha,
       test_statistic = test_result$statistic, effect_size = abs(mean(seg1_y) - mean(seg2_y)))
}

#' @keywords internal
#' @noRd
.assess_changepoint_significance <- function(data, cp_set, alpha = 0.05, n_perm = 1000, use_neighbor = TRUE) {
  if (length(cp_set) == 0) return(data.frame())

  cp_set <- sort(cp_set)
  significant_cps <- integer(0)
  results_list <- list()

  for (i in seq_along(cp_set)) {
    cp <- cp_set[i]

    baseline_start <- if (length(significant_cps) == 0) 1 else max(significant_cps) + 1
    current_end <- if (i < length(cp_set)) cp_set[i + 1] else nrow(data)

    baseline_segment <- baseline_start:cp
    current_segment <- (cp + 1):current_end

    if (length(baseline_segment) < 3 || length(current_segment) < 3) {
      results_list[[i]] <- data.frame(
        ChangePoint = cp,
        LR_PValue = NA, LR_Significant = FALSE,
        Perm_PValue = NA, Perm_Significant = FALSE, Perm_EffectSize = NA,
        Prop_PValue = NA, Prop_Significant = FALSE,
        Chi_PValue = NA, Chi_Significant = FALSE,
        Seg1_Proportion = NA, Seg2_Proportion = NA,
        Overall_Significant = FALSE
      )
      next
    }

    baseline_prop <- mean(data$y[baseline_segment])
    current_prop <- mean(data$y[current_segment])

    temp_combined <- data[c(baseline_segment, current_segment), ]
    cp_temp <- length(baseline_segment)

    lr_test <- suppressWarnings(.test_cp_likelihood(temp_combined, cp_temp, alpha, use_neighbor))
    lr_p_value <- ifelse(is.null(lr_test$p.value), lr_test$p_value, lr_test$p.value)
    lr_significant <- ifelse(is.na(lr_p_value), FALSE, lr_p_value < alpha)

    perm_test <- suppressWarnings(.test_cp_permutation(temp_combined, cp_temp, n_perm, alpha))
    prop_test <- suppressWarnings(.test_cp_proportion(temp_combined, cp_temp, alpha))

    chi_test <- suppressWarnings(tryCatch({
      .test_cp_chisquare(temp_combined, cp_temp, alpha)
    }, error = function(e) {
      list(p_value = NA, significant = FALSE, test_statistic = NA)
    }))
    chi_significant <- ifelse(is.na(chi_test$p_value), FALSE, chi_test$p_value < alpha)

    sig_votes <- sum(c(lr_significant, perm_test$significant, prop_test$significant, chi_significant), na.rm = TRUE)
    overall_significant <- sig_votes >= 2

    if (overall_significant) significant_cps <- c(significant_cps, cp)

    results_list[[i]] <- data.frame(
      ChangePoint = cp,
      LR_PValue = lr_p_value, LR_Significant = lr_significant,
      Perm_PValue = perm_test$p_value, Perm_Significant = perm_test$significant,
      Perm_EffectSize = perm_test$effect_size,
      Prop_PValue = prop_test$p_value, Prop_Significant = prop_test$significant,
      Chi_PValue = chi_test$p_value, Chi_Significant = chi_significant,
      Seg1_Proportion = baseline_prop, Seg2_Proportion = current_prop,
      Overall_Significant = overall_significant
    )
  }

  do.call(rbind, results_list)
}

#' @keywords internal
#' @noRd
.classify_pattern <- function(sig_results) {
  sig <- sig_results[sig_results$Overall_Significant == TRUE, ]
  if (nrow(sig) == 0) return("No Pattern")

  tolerance <- 1e-10
  prop_diffs <- sig$Seg2_Proportion - sig$Seg1_Proportion
  has_inc <- any(prop_diffs > tolerance)
  has_dec <- any(prop_diffs < -tolerance)

  if (has_inc & has_dec) return("Fluctuating/Intermittent")
  if (has_inc) return("Increasing")
  if (has_dec) return("Decreasing")
  return("Failed Classification")
}

#' @keywords internal
#' @noRd
.classify_pixel_with_times <- function(pixel_vals, n_middle, time_steps,
                                       fastcpd_params = list(), alpha = 0.05,
                                       n_perm = 1000, use_neighbor = TRUE) {

  y <- pixel_vals[1:n_middle]
  lag <- pixel_vals[(n_middle + 1):(2 * n_middle)]

  if (use_neighbor) {
    neighbor <- pixel_vals[(2 * n_middle + 1):(3 * n_middle)]
    mean_val <- pixel_vals[3 * n_middle + 1]
  } else {
    neighbor <- NULL
    mean_val <- pixel_vals[2 * n_middle + 1]
  }

  if (is.na(mean_val)) return(c(NA, NA, NA))
  if (mean_val < 0.01) return(c(1, NA, NA))
  if (mean_val > 0.99) return(c(2, NA, NA))
  if (any(is.na(c(y, lag)))) return(c(NA, NA, NA))
  if (use_neighbor && any(is.na(neighbor))) return(c(NA, NA, NA))

  data_matrix <- if (use_neighbor) cbind(y, lag, neighbor) else cbind(y, lag)

  cp_result <- tryCatch({
    suppressWarnings(do.call(
      fastcpd::fastcpd.binomial,
      c(list(data = data_matrix, r.progress = FALSE), fastcpd_params)
    ))
  }, error = function(e) NULL)

  if (is.null(cp_result)) return(c(7, NA, NA))

  cp_set <- cp_result@cp_set
  if (length(cp_set) == 0) return(c(3, NA, NA))

  data_df <- if (use_neighbor) {
    data.frame(y = y, lag1 = lag, neighbor = neighbor)
  } else {
    data.frame(y = y, lag1 = lag)
  }

  sig_results <- tryCatch({
    suppressWarnings(.assess_changepoint_significance(data_df, cp_set, alpha, n_perm, use_neighbor))
  }, error = function(e) NULL)

  if (is.null(sig_results) || nrow(sig_results) == 0) return(c(7, NA, NA))

  pattern <- .classify_pattern(sig_results)
  pattern_codes <- c("Always Absent" = 1, "Always Present" = 2, "No Pattern" = 3,
                     "Increasing" = 4, "Decreasing" = 5, "Fluctuating/Intermittent" = 6,
                     "Failed Classification" = 7)
  classification_code <- pattern_codes[pattern]

  time_decrease <- NA
  time_increase <- NA
  sig_cps <- sig_results[sig_results$Overall_Significant == TRUE, ]

  if (nrow(sig_cps) > 0) {
    if (pattern == "Decreasing") {
      dec_cps <- sig_cps[sig_cps$Seg2_Proportion < sig_cps$Seg1_Proportion, ]
      if (nrow(dec_cps) > 0) time_decrease <- time_steps[min(dec_cps$ChangePoint) + 1]
    }
    if (pattern == "Increasing") {
      inc_cps <- sig_cps[sig_cps$Seg2_Proportion > sig_cps$Seg1_Proportion, ]
      if (nrow(inc_cps) > 0) time_increase <- time_steps[min(inc_cps$ChangePoint) + 1]
    }
  }

  c(classification_code, time_decrease, time_increase)
}

#' @keywords internal
#' @noRd
.load_points_data <- function(points_file_path, xcol, ycol, points_crs,
                              verbose = TRUE) {

  if (is.character(points_file_path)) {
    if (!file.exists(points_file_path)) stop(paste("ERROR: File does not exist:", points_file_path))
    file_ext <- tolower(tools::file_ext(points_file_path))

    if (file_ext == "csv") {
      if (is.null(xcol)) stop("ERROR: 'xcol' is required when reading CSV files.")
      if (is.null(ycol)) stop("ERROR: 'ycol' is required when reading CSV files.")
      if (is.null(points_crs)) stop("ERROR: 'points_crs' is required when reading CSV files.")
      if (verbose) message(paste("Reading CSV file:", basename(points_file_path)))
      pts <- utils::read.csv(points_file_path, stringsAsFactors = FALSE)
      if (!xcol %in% names(pts)) stop(paste0("ERROR: Column '", xcol, "' not found in CSV."))
      if (!ycol %in% names(pts)) stop(paste0("ERROR: Column '", ycol, "' not found in CSV."))

    } else if (file_ext %in% c("shp", "geojson", "gpkg")) {
      if (verbose) message(paste("Reading spatial file:", basename(points_file_path)))
      pts_sf_raw <- sf::st_read(points_file_path, quiet = TRUE)
      pts <- sf::st_drop_geometry(pts_sf_raw)
      coords <- sf::st_coordinates(pts_sf_raw)
      if (is.null(xcol)) xcol <- "X"
      if (is.null(ycol)) ycol <- "Y"
      pts[[xcol]] <- coords[, 1]
      pts[[ycol]] <- coords[, 2]
      if (is.null(points_crs)) points_crs <- sf::st_crs(pts_sf_raw)
    } else {
      stop(paste("ERROR: Unsupported file format:", file_ext))
    }

  } else if (inherits(points_file_path, "sf")) {
    if (verbose) message("Using provided sf object...")
    pts <- sf::st_drop_geometry(points_file_path)
    coords <- sf::st_coordinates(points_file_path)
    if (is.null(xcol)) xcol <- "X"
    if (is.null(ycol)) ycol <- "Y"
    pts[[xcol]] <- coords[, 1]
    pts[[ycol]] <- coords[, 2]
    if (is.null(points_crs)) points_crs <- sf::st_crs(points_file_path)

  } else if (inherits(points_file_path, "sfc")) {
    if (verbose) message("Converting sfc to sf...")
    pts_sf_tmp <- sf::st_sf(geometry = points_file_path)
    pts <- data.frame(row.names = seq_along(points_file_path))
    coords <- sf::st_coordinates(pts_sf_tmp)
    if (is.null(xcol)) xcol <- "X"
    if (is.null(ycol)) ycol <- "Y"
    pts[[xcol]] <- coords[, 1]
    pts[[ycol]] <- coords[, 2]
    if (is.null(points_crs)) points_crs <- sf::st_crs(pts_sf_tmp)

  } else if (inherits(points_file_path, "Spatial")) {
    if (verbose) message("Converting Spatial object to sf...")
    pts_sf_tmp <- sf::st_as_sf(points_file_path)
    pts <- sf::st_drop_geometry(pts_sf_tmp)
    coords <- sf::st_coordinates(pts_sf_tmp)
    if (is.null(xcol)) xcol <- "X"
    if (is.null(ycol)) ycol <- "Y"
    pts[[xcol]] <- coords[, 1]
    pts[[ycol]] <- coords[, 2]
    if (is.null(points_crs)) points_crs <- sf::st_crs(pts_sf_tmp)

  } else if (is.data.frame(points_file_path)) {
    if (is.null(xcol)) stop("ERROR: 'xcol' is required when providing a data frame.")
    if (is.null(ycol)) stop("ERROR: 'ycol' is required when providing a data frame.")
    if (is.null(points_crs)) stop("ERROR: 'points_crs' is required when providing a data frame.")
    if (verbose) message("Using provided data frame...")
    pts <- points_file_path
    if (!xcol %in% names(pts)) stop(paste0("ERROR: Column '", xcol, "' not found."))
    if (!ycol %in% names(pts)) stop(paste0("ERROR: Column '", ycol, "' not found."))
  } else {
    stop("ERROR: points_file_path must be an sf object, sfc, Spatial, data frame, or file path.")
  }

  attr(pts, "xcol") <- xcol
  attr(pts, "ycol") <- ycol
  attr(pts, "crs") <- points_crs
  pts
}

#' @keywords internal
#' @noRd
#' @importFrom deldir deldir tile.list
.build_spatial_blocks <- function(pts_sf, reference_shapefile, n_spatial) {

  coords <- sf::st_coordinates(pts_sf)

  if (n_spatial == 1) {
    centers <- data.frame(spatial_block = 1,
                          center_x = mean(coords[, 1]),
                          center_y = mean(coords[, 2]))
    voronoi_sf <- sf::st_sf(fold_group = 1,
                            geometry = sf::st_geometry(sf::st_union(reference_shapefile)))
    return(list(clusters = rep(1, nrow(pts_sf)), voronoi_sf = voronoi_sf,
                adjacency = matrix(TRUE, 1, 1), centers = centers, dist_matrix = matrix(0, 1, 1)))
  }

  km <- stats::kmeans(coords, centers = n_spatial, nstart = 50, iter.max = 100)
  centers <- data.frame(spatial_block = seq_len(n_spatial),
                        center_x = km$centers[, 1],
                        center_y = km$centers[, 2])

  bbox <- sf::st_bbox(reference_shapefile)
  vor <- deldir::deldir(centers$center_x, centers$center_y,
                        rw = c(bbox["xmin"], bbox["xmax"], bbox["ymin"], bbox["ymax"]))
  tiles <- deldir::tile.list(vor)
  polys <- lapply(seq_along(tiles), function(i) {
    tile <- tiles[[i]]
    sf::st_polygon(list(cbind(c(tile$x, tile$x[1]), c(tile$y, tile$y[1]))))
  })

  voronoi_sf <- sf::st_sf(fold_group = seq_len(n_spatial),
                          geometry = sf::st_sfc(polys, crs = sf::st_crs(pts_sf)))
  voronoi_sf <- suppressWarnings(sf::st_intersection(voronoi_sf, sf::st_union(reference_shapefile)))

  n_blocks <- nrow(voronoi_sf)
  adj <- matrix(FALSE, n_blocks, n_blocks)
  for (i in seq_len(n_blocks - 1)) {
    for (j in (i + 1):n_blocks) {
      isect <- suppressWarnings(sf::st_intersection(voronoi_sf$geometry[i], voronoi_sf$geometry[j]))
      if (length(isect) > 0 && !sf::st_is_empty(isect)) {
        gtype <- sf::st_geometry_type(isect)
        if (gtype %in% c("LINESTRING", "MULTILINESTRING", "GEOMETRYCOLLECTION")) {
          adj[i, j] <- TRUE
          adj[j, i] <- TRUE
        }
      }
    }
  }

  dm <- as.matrix(stats::dist(cbind(centers$center_x, centers$center_y)))
  list(clusters    = km$cluster,
       voronoi_sf  = voronoi_sf,
       adjacency   = adj,
       centers     = centers,
       dist_matrix = dm)
}

#' @keywords internal
#' @noRd
.select_dispersed_blocks <- function(distance_matrix, n_select, all_blocks) {
  if (n_select == 1) return(all_blocks[1])

  best_min_dist <- -Inf
  best_selection <- NULL
  for (first in all_blocks) {
    selected <- integer(n_select)
    selected[1] <- first
    for (k in 2:n_select) {
      remaining <- setdiff(all_blocks, selected[seq_len(k - 1)])
      dists <- apply(distance_matrix[remaining, selected[seq_len(k - 1)], drop = FALSE], 1, min)
      selected[k] <- remaining[which.max(dists)]
    }
    min_pw <- min(distance_matrix[selected, selected][upper.tri(distance_matrix[selected, selected])])
    if (min_pw > best_min_dist) {
      best_min_dist  <- min_pw
      best_selection <- selected
    }
  }
  best_selection
}

#' @keywords internal
#' @noRd
.assign_balanced_folds <- function(pts_work, n_folds, n_spatial, spatial_centers,
                                   distance_matrix, adjacency_matrix, max_imbalance, total_folds) {

  pts_work$fold       <- NA_integer_
  pts_work$block_type <- NA_character_
  all_blocks          <- seq_len(n_spatial)

  core_blocks <- .select_dispersed_blocks(distance_matrix, n_folds, all_blocks)
  assigned_blocks <- core_blocks
  fold_block_pairs <- vector("list", n_folds)

  for (fold_id in seq_len(n_folds)) {
    core <- core_blocks[fold_id]
    adjacent <- which(adjacency_matrix[core, ])
    available_adj <- setdiff(adjacent, assigned_blocks)

    second <- if (length(available_adj) > 0) {
      available_adj[1]
    } else {
      available_any <- setdiff(all_blocks, assigned_blocks)
      if (length(available_any) > 0) available_any[which.min(distance_matrix[core, available_any])] else core
    }
    fold_block_pairs[[fold_id]] <- c(core, second)
    assigned_blocks             <- c(assigned_blocks, second)

    for (block_id in fold_block_pairs[[fold_id]]) {
      for (temp_id in seq_len(2)) {
        idx                      <- which(pts_work$spatial_block == block_id & pts_work$temporal_block == temp_id)
        pts_work$fold[idx]       <- fold_id
        pts_work$block_type[idx] <- "balanced_core"
      }
    }
  }

  unassigned <- setdiff(all_blocks, assigned_blocks)
  for (block_id in unassigned) {
    bc <- c(spatial_centers$center_x[block_id], spatial_centers$center_y[block_id])
    fold_dists <- vapply(seq_len(n_folds), function(f) {
      fb <- fold_block_pairs[[f]]
      fc <- c(mean(spatial_centers$center_x[fb]), mean(spatial_centers$center_y[fb]))
      sqrt(sum((bc - fc)^2))
    }, numeric(1))
    sorted <- order(fold_dists)

    idx_t1 <- which(pts_work$spatial_block == block_id & pts_work$temporal_block == 1)
    idx_t2 <- which(pts_work$spatial_block == block_id & pts_work$temporal_block == 2)
    if (length(idx_t1) > 0) {
      pts_work$fold[idx_t1]       <- sorted[1]
      pts_work$block_type[idx_t1] <- "balanced_shared"
    }
    if (length(idx_t2) > 0) {
      pts_work$fold[idx_t2]       <- sorted[min(2, length(sorted))]
      pts_work$block_type[idx_t2] <- "balanced_shared"
    }
  }

  .rebalance_folds(pts_work, total_folds, max_imbalance, max_iterations = 25, moveable_type = "balanced_shared")
}

#' @keywords internal
#' @noRd
.kd_split_st <- function(idx, n_folds, target_sizes, coords) {
  if (n_folds == 1) return(list(idx))
  if (length(idx) <= 1) return(list(idx))
  sub_coords  <- coords[idx, , drop = FALSE]
  x_range     <- diff(range(sub_coords[, 1]))
  y_range     <- diff(range(sub_coords[, 2]))
  axis        <- if (x_range >= y_range) 1 else 2
  order_along <- order(sub_coords[, axis])
  idx_sorted  <- idx[order_along]
  left_n      <- floor(n_folds / 2)
  right_n     <- n_folds - left_n
  split_at    <- sum(target_sizes[seq_len(left_n)])
  split_at    <- max(1, min(split_at, length(idx_sorted) - 1))
  left_idx    <- idx_sorted[seq_len(split_at)]
  right_idx   <- idx_sorted[(split_at + 1):length(idx_sorted)]
  c(.kd_split_st(left_idx,  left_n,  target_sizes[seq_len(left_n)], coords),
    .kd_split_st(right_idx, right_n, target_sizes[(left_n + 1):n_folds], coords))
}

#' @keywords internal
#' @noRd
.assign_spatiotemporal_folds <- function(pts_work, n_spatial_folds, n_temporal_folds,
                                         total_folds, n_spatial, n_temporal,
                                         spatial_centers, distance_matrix,
                                         adjacency_matrix, temporal_partitioning,
                                         max_imbalance, use_balanced) {

  pts_work$fold       <- NA_integer_
  pts_work$block_type <- NA_character_
  coords              <- sf::st_coordinates(pts_work)
  n_pts  <- nrow(coords)

  if (n_spatial_folds > 0 && n_temporal_folds == 0) {

    target_size  <- floor(n_pts / n_spatial_folds)
    n_larger     <- n_pts - target_size * n_spatial_folds
    target_sizes <- rep(target_size, n_spatial_folds)
    if (n_larger > 0) target_sizes[seq_len(n_larger)] <- target_size + 1

    groups   <- .kd_split_st(seq_len(n_pts), n_spatial_folds, target_sizes, coords)
    fold_vec <- integer(n_pts)
    for (f in seq_along(groups)) fold_vec[groups[[f]]] <- f

    group_centroids <- do.call(rbind, lapply(seq_along(groups), function(f) {
      colMeans(coords[groups[[f]], , drop = FALSE])
    }))
    dist_to_centroids <- matrix(NA_real_, nrow = n_pts, ncol = n_spatial_folds)
    for (f in seq_len(n_spatial_folds)) {
      dx <- coords[, 1] - group_centroids[f, 1]
      dy <- coords[, 2] - group_centroids[f, 2]
      dist_to_centroids[, f] <- sqrt(dx^2 + dy^2)
    }
    fold_vec <- apply(dist_to_centroids, 1, which.min)

    for (iter in seq_len(n_pts * 2)) {
      counts <- tabulate(fold_vec, nbins = n_spatial_folds)
      over   <- which(counts > target_sizes)
      under  <- which(counts < target_sizes)
      if (length(over) == 0 || length(under) == 0) break
      of <- over[1]; uf <- under[1]
      candidates <- which(fold_vec == of)
      best <- candidates[which.min(dist_to_centroids[candidates, uf])]
      fold_vec[best] <- uf
    }

    pts_work$fold       <- fold_vec
    pts_work$block_type <- "spatial_exclusive"
    return(pts_work)
  }

  temporal_folds <- as.integer((n_spatial_folds + 1):total_folds)
  n_groups       <- total_folds

  fold_size  <- floor(n_pts / n_groups)
  remainder  <- n_pts - fold_size * n_groups
  group_target_sizes <- rep(fold_size, n_groups)
  if (remainder > 0) group_target_sizes[seq_len(remainder)] <- fold_size + 1

  groups    <- .kd_split_st(seq_len(n_pts), n_groups, group_target_sizes, coords)
  group_vec <- integer(n_pts)
  for (g in seq_along(groups)) group_vec[groups[[g]]] <- g

  group_centroids <- do.call(rbind, lapply(seq_len(n_groups), function(g) {
    idx <- which(group_vec == g)
    if (length(idx) == 0) return(c(NA_real_, NA_real_))
    colMeans(coords[idx, , drop = FALSE])
  }))

  dist_to_groups <- matrix(NA_real_, nrow = n_pts, ncol = n_groups)
  for (g in seq_len(n_groups)) {
    if (any(is.na(group_centroids[g, ]))) next
    dx <- coords[, 1] - group_centroids[g, 1]
    dy <- coords[, 2] - group_centroids[g, 2]
    dist_to_groups[, g] <- sqrt(dx^2 + dy^2)
  }

  if (!any(is.na(dist_to_groups))) {
    group_vec <- apply(dist_to_groups, 1, which.min)
  }

  for (iter in seq_len(n_pts * 2)) {
    counts <- tabulate(group_vec, nbins = n_groups)
    over   <- which(counts > group_target_sizes)
    under  <- which(counts < group_target_sizes)
    if (length(over) == 0 || length(under) == 0) break
    of <- over[1]; uf <- under[1]
    candidates <- which(group_vec == of)
    if (length(candidates) == 0) break
    best <- candidates[which.min(dist_to_groups[candidates, uf])]
    group_vec[best] <- uf
  }
  centroid_dists <- as.matrix(stats::dist(group_centroids))

  if (temporal_partitioning && "temporal_block" %in% names(pts_work)) {

    group_tb_score <- vapply(seq_len(n_groups), function(g) {
      idx <- which(group_vec == g)
      tb  <- pts_work$temporal_block[idx]
      tb  <- tb[!is.na(tb)]
      if (length(tb) == 0) return(Inf)
      props <- as.numeric(table(factor(tb, levels = seq_len(n_temporal)))) / length(tb)
      sum((props - 1 / n_temporal) ^ 2)
    }, numeric(1))

    if (n_temporal_folds == 1) {
      pool_groups <- which.min(group_tb_score)
    } else {
      seed        <- which.min(group_tb_score)
      pool_groups <- seed
      remaining   <- setdiff(seq_len(n_groups), seed)
      for (step in seq_len(n_temporal_folds - 1)) {
        if (length(remaining) == 0) break
        if (length(pool_groups) == 1) {
          dists_to_pool <- centroid_dists[remaining, pool_groups]
        } else {
          dists_to_pool <- apply(centroid_dists[remaining, pool_groups, drop = FALSE], 1, min)
        }
        next_group  <- remaining[which.min(dists_to_pool)]
        pool_groups <- c(pool_groups, next_group)
        remaining   <- setdiff(remaining, next_group)
      }
    }

  } else {
    all_centroid <- colMeans(group_centroids, na.rm = TRUE)
    dists_to_center <- sqrt((group_centroids[, 1] - all_centroid[1])^2 +
                              (group_centroids[, 2] - all_centroid[2])^2)
    seed        <- which.min(dists_to_center)
    pool_groups <- seed
    remaining   <- setdiff(seq_len(n_groups), seed)
    for (step in seq_len(n_temporal_folds - 1)) {
      if (length(remaining) == 0) break
      if (length(pool_groups) == 1) {
        dists_to_pool <- centroid_dists[remaining, pool_groups]
      } else {
        dists_to_pool <- apply(centroid_dists[remaining, pool_groups, drop = FALSE], 1, min)
      }
      next_group  <- remaining[which.min(dists_to_pool)]
      pool_groups <- c(pool_groups, next_group)
      remaining   <- setdiff(remaining, next_group)
    }
  }

  spatial_groups <- setdiff(seq_len(n_groups), pool_groups)
  for (f in seq_len(n_spatial_folds)) {
    g   <- spatial_groups[f]
    idx <- which(group_vec == g)
    pts_work$fold[idx]       <- f
    pts_work$block_type[idx] <- "spatial_exclusive"
  }

  temporal_pool_idx <- which(group_vec %in% pool_groups)

  if (length(temporal_pool_idx) > 0) {

    if (temporal_partitioning &&
        "temporal_block" %in% names(pts_work) &&
        any(!is.na(pts_work$temporal_block[temporal_pool_idx]))) {

      t_blocks  <- pts_work$temporal_block[temporal_pool_idx]
      unique_tb <- sort(unique(t_blocks[!is.na(t_blocks)]))
      n_unique  <- length(unique_tb)

      tb_fold_map <- stats::setNames(
        temporal_folds[((seq_len(n_unique) - 1) %% n_temporal_folds) + 1],
        as.character(unique_tb)
      )

      for (k in seq_along(temporal_pool_idx)) {
        gi <- temporal_pool_idx[k]
        tb <- pts_work$temporal_block[gi]
        if (!is.na(tb)) {
          pts_work$fold[gi]       <- tb_fold_map[as.character(tb)]
          pts_work$block_type[gi] <- "temporal_exclusive"
        } else {
          t_counts <- tabulate(
            pts_work$fold[pts_work$fold %in% temporal_folds & !is.na(pts_work$fold)],
            nbins = max(temporal_folds)
          )[temporal_folds]
          pts_work$fold[gi]       <- temporal_folds[which.min(t_counts)]
          pts_work$block_type[gi] <- "temporal_exclusive"
        }
      }

    } else {
      pool_coords <- coords[temporal_pool_idx, , drop = FALSE]
      x_rng       <- diff(range(pool_coords[, 1]))
      y_rng       <- diff(range(pool_coords[, 2]))
      sort_axis   <- if (x_rng >= y_rng) 1 else 2
      pool_order  <- order(pool_coords[, sort_axis])
      n_pool      <- length(temporal_pool_idx)

      chunk_size      <- floor(n_pool / n_temporal_folds)
      n_larger_chunks <- n_pool - chunk_size * n_temporal_folds
      chunk_sizes     <- rep(chunk_size, n_temporal_folds)
      if (n_larger_chunks > 0)
        chunk_sizes[seq_len(n_larger_chunks)] <- chunk_size + 1

      pos <- 1
      for (ti in seq_len(n_temporal_folds)) {
        end     <- pos + chunk_sizes[ti] - 1
        sel_idx <- temporal_pool_idx[pool_order[pos:end]]
        pts_work$fold[sel_idx]       <- temporal_folds[ti]
        pts_work$block_type[sel_idx] <- "temporal_exclusive"
        pos <- end + 1
      }
    }
  }

  still_na <- which(is.na(pts_work$fold))
  if (length(still_na) > 0) {
    counts_all <- tabulate(pts_work$fold[!is.na(pts_work$fold)], nbins = total_folds)
    for (idx in still_na) {
      pts_work$fold[idx]       <- as.integer(which.min(counts_all))
      pts_work$block_type[idx] <- "remainder"
      counts_all[pts_work$fold[idx]] <- counts_all[pts_work$fold[idx]] + 1
    }
  }

  pts_work
}

#' @keywords internal
#' @noRd
.rebalance_folds <- function(pts_work, total_folds, max_imbalance,
                             max_iterations = 25, moveable_type = NULL) {

  point_coords <- sf::st_coordinates(pts_work)

  for (iter in seq_len(max_iterations)) {
    counts <- table(factor(pts_work$fold, levels = seq_len(total_folds)))
    mean_pf <- mean(counts)
    imb <- max(abs(counts - mean_pf)) / mean_pf
    if (imb <= max_imbalance) break

    largest <- as.integer(names(which.max(counts)))
    smallest <- as.integer(names(which.min(counts)))
    if (counts[largest] <= counts[smallest]) break

    candidates <- if (!is.null(moveable_type)) {
      which(pts_work$fold == largest & pts_work$block_type == moveable_type)
    } else {
      which(pts_work$fold == largest)
    }
    if (length(candidates) == 0) break

    smallest_pts <- which(pts_work$fold == smallest)
    if (length(smallest_pts) == 0) break
    centroid <- c(mean(point_coords[smallest_pts, 1]), mean(point_coords[smallest_pts, 2]))
    cc <- point_coords[candidates, , drop = FALSE]
    dists <- sqrt((cc[, 1] - centroid[1])^2 + (cc[, 2] - centroid[2])^2)
    pts_work$fold[candidates[which.min(dists)]] <- smallest
  }
  pts_work
}

#' @keywords internal
#' @noRd
.resolve_palette <- function(n, palette) {
  if (palette %in% grDevices::hcl.pals()) {
    return(grDevices::hcl.colors(n, palette = palette))
  }

  if (requireNamespace("RColorBrewer", quietly = TRUE)) {
    brewer_pals <- rownames(RColorBrewer::brewer.pal.info)
    if (palette %in% brewer_pals) {
      max_n <- RColorBrewer::brewer.pal.info[palette, "maxcolors"]
      pal_colors <- RColorBrewer::brewer.pal(min(n, max_n), palette)
      if (n > max_n) {
        pal_colors <- grDevices::colorRampPalette(pal_colors)(n)
      }
      return(pal_colors)
    }
  }

  stop(paste0(
    "ERROR: Palette '", palette, "' not recognized. ",
    "Use grDevices::hcl.pals() to see available HCL palettes, or install ",
    "RColorBrewer for additional palette options."
  ))
}

#' @keywords internal
#' @noRd
.build_fold_polygons <- function(pts_sf, reference_shapefile, total_folds, partition_mode) {

  if (is.null(reference_shapefile) || partition_mode == "random") return(NULL)

  pts_coords_map <- sf::st_coordinates(pts_sf)
  fold_vec_pts   <- pts_sf$fold
  ref_sf_union   <- sf::st_union(reference_shapefile)
  bbox           <- sf::st_bbox(reference_shapefile)

  is_temporal_mode <- partition_mode %in% c("spatiotemporal", "balanced", "temporal_only")

  if (is_temporal_mode && "block_type" %in% names(pts_sf)) {
    pts_df_bt <- sf::st_drop_geometry(pts_sf)
    spatial_folds <- sort(unique(pts_df_bt$fold[
      pts_df_bt$block_type %in% c("spatial_exclusive", "rebalanced")
    ]))
    temporal_folds <- sort(unique(pts_df_bt$fold[
      pts_df_bt$block_type == "temporal_exclusive"
    ]))
    fold_vec_render <- fold_vec_pts
    fold_vec_render[fold_vec_pts %in% temporal_folds] <- total_folds + 1
    render_folds <- as.integer(c(spatial_folds, total_folds + 1))
  } else {
    spatial_folds   <- as.integer(seq_len(total_folds))
    temporal_folds  <- integer(0)
    fold_vec_render <- fold_vec_pts
    render_folds    <- as.integer(seq_len(total_folds))
  }

  max_render_idx <- max(render_folds)
  fold_mcps      <- vector("list", max_render_idx)
  valid_render   <- integer(0)

  for (f in render_folds) {
    fold_pts <- pts_coords_map[fold_vec_render == f, , drop = FALSE]
    if (nrow(fold_pts) == 0) next
    fold_sfc <- sf::st_as_sf(as.data.frame(fold_pts), coords = c("X", "Y"),
                             crs = sf::st_crs(pts_sf))
    fold_mcps[[f]] <- if (nrow(fold_pts) >= 3) {
      sf::st_convex_hull(sf::st_union(fold_sfc))
    } else {
      sf::st_buffer(sf::st_union(fold_sfc), dist = 1e-6)
    }
    valid_render <- c(valid_render, f)
  }
  valid_render <- as.integer(valid_render)

  fold_real_pts <- lapply(seq_len(max_render_idx), function(f) {
    if (!f %in% valid_render) return(NULL)
    fold_pts <- pts_coords_map[fold_vec_render == f, , drop = FALSE]
    if (nrow(fold_pts) == 0) return(NULL)
    sf::st_as_sf(as.data.frame(fold_pts), coords = c("X", "Y"),
                 crs = sf::st_crs(pts_sf))
  })

  set.seed(42)
  sample_pts <- sf::st_sample(ref_sf_union, size = 1000, type = "random")
  sample_pts <- sf::st_as_sf(sample_pts)
  sample_pts <- sample_pts[!sf::st_is_empty(sample_pts), ]
  sample_xy  <- sf::st_coordinates(sample_pts)
  n_sample   <- nrow(sample_xy)

  mcp_geoms      <- do.call(c, lapply(valid_render, function(f) fold_mcps[[f]]))
  contain_sparse <- suppressMessages(suppressWarnings(
    sf::st_intersects(sample_pts, mcp_geoms, sparse = TRUE)
  ))

  n_hits      <- lengths(contain_sparse)
  idx_one     <- which(n_hits == 1)
  idx_multi   <- which(n_hits > 1)
  idx_outside <- which(n_hits == 0)
  sample_fold <- integer(n_sample)

  if (length(idx_one) > 0) {
    sample_fold[idx_one] <- valid_render[vapply(contain_sparse[idx_one],
                                                `[[`, integer(1), 1)]
  }

  if (length(idx_multi) > 0) {
    sample_fold[idx_multi] <- vapply(idx_multi, function(i) {
      hit_folds <- valid_render[contain_sparse[[i]]]
      best_fold <- hit_folds[1]
      best_dist <- Inf
      pt_i      <- sample_pts[i, ]
      for (hf in hit_folds) {
        if (is.null(fold_real_pts[[hf]])) next
        d <- min(as.numeric(sf::st_distance(pt_i, fold_real_pts[[hf]])))
        if (d < best_dist) {
          best_dist <- d
          best_fold <- hf
        }
      }
      as.integer(best_fold)
    }, integer(1))
  }

  if (length(idx_outside) > 0) {
    dist_mat <- suppressMessages(suppressWarnings(
      sf::st_distance(sample_pts[idx_outside, ], mcp_geoms)
    ))
    nearest_col              <- apply(dist_mat, 1, which.min)
    sample_fold[idx_outside] <- valid_render[nearest_col]
  }

  sample_assigned <- sf::st_as_sf(
    data.frame(fold = sample_fold, sample_xy),
    coords = c("X", "Y"), crs = sf::st_crs(pts_sf)
  )

  s2_state <- sf::sf_use_s2()
  suppressMessages(sf::sf_use_s2(FALSE))
  vor_raw <- tryCatch({
    suppressMessages(suppressWarnings(
      sf::st_voronoi(sf::st_union(sample_assigned),
                     envelope = sf::st_as_sfc(bbox))
    ))
  }, error = function(e) NULL)
  suppressMessages(sf::sf_use_s2(s2_state))

  if (is.null(vor_raw)) return(NULL)

  vor_polys <- sf::st_collection_extract(vor_raw, "POLYGON")
  vor_sf    <- sf::st_sf(geometry = vor_polys, crs = sf::st_crs(pts_sf))

  joined <- suppressMessages(suppressWarnings(
    sf::st_join(vor_sf, sample_assigned, join = sf::st_contains)
  ))

  real_pts_render <- sf::st_as_sf(
    data.frame(fold = fold_vec_render, pts_coords_map),
    coords = c("X", "Y"), crs = sf::st_crs(pts_sf)
  )

  sentinel     <- total_folds + 1
  orphan_geoms <- NULL
  keep_list    <- list()

  unique_render_folds <- unique(joined$fold[!is.na(joined$fold)])

  for (f in unique_render_folds) {
    sub <- joined[!is.na(joined$fold) & joined$fold == f, ]
    if (nrow(sub) == 0) next

    merged  <- suppressWarnings(sf::st_union(sub))
    clipped <- suppressWarnings(sf::st_intersection(merged, ref_sf_union))
    clipped <- clipped[!sf::st_is_empty(clipped)]
    if (length(clipped) == 0) next

    parts_sf <- suppressWarnings(
      sf::st_cast(sf::st_sf(fold = f, geometry = clipped), "POLYGON")
    )
    parts_sf <- parts_sf[!sf::st_is_empty(parts_sf), ]
    if (nrow(parts_sf) == 0) next

    if (f %in% spatial_folds) {
      fold_real <- real_pts_render[real_pts_render$fold == f, ]

      if (nrow(fold_real) == 0) {
        orphan_geoms <- c(orphan_geoms, sf::st_geometry(parts_sf))
        next
      }

      bbox_diag <- sqrt((bbox["xmax"] - bbox["xmin"])^2 +
                          (bbox["ymax"] - bbox["ymin"])^2)
      proximity_threshold <- bbox_diag * 0.001

      part_min_dists <- vapply(seq_len(nrow(parts_sf)), function(p) {
        suppressMessages(suppressWarnings(
          min(as.numeric(sf::st_distance(fold_real, parts_sf[p, ])))
        ))
      }, numeric(1))

      keep_mask <- part_min_dists <= proximity_threshold

      if (any(keep_mask)) {
        kept_geom <- suppressWarnings(sf::st_union(parts_sf[keep_mask, ]))
        keep_list[[length(keep_list) + 1]] <- sf::st_sf(
          fold     = f,
          geometry = sf::st_sfc(kept_geom, crs = sf::st_crs(pts_sf))
        )
      }
      if (any(!keep_mask)) {
        orphan_geoms <- c(orphan_geoms, sf::st_geometry(parts_sf[!keep_mask, ]))
      }

    } else {
      keep_list[[length(keep_list) + 1]] <- sf::st_sf(
        fold     = f,
        geometry = sf::st_sfc(suppressWarnings(sf::st_union(parts_sf)),
                              crs = sf::st_crs(pts_sf))
      )
    }
  }

  dissolved <- if (length(keep_list) > 0) do.call(rbind, keep_list) else NULL

  if (!is.null(orphan_geoms) && length(orphan_geoms) > 0 &&
      length(temporal_folds) > 0 && !is.null(dissolved)) {
    orphan_sfc <- sf::st_sfc(orphan_geoms, crs = sf::st_crs(pts_sf))
    existing_t <- dissolved[!is.na(dissolved$fold) & dissolved$fold == sentinel, ]
    if (nrow(existing_t) > 0) {
      merged_t <- suppressWarnings(
        sf::st_union(c(sf::st_geometry(existing_t), orphan_sfc))
      )
      dissolved <- dissolved[is.na(dissolved$fold) | dissolved$fold != sentinel, ]
    } else {
      merged_t <- suppressWarnings(sf::st_union(orphan_sfc))
    }
    dissolved <- rbind(
      dissolved,
      sf::st_sf(fold     = sentinel,
                geometry = sf::st_sfc(merged_t, crs = sf::st_crs(pts_sf)))
    )
  }

  dissolved
}

#' @keywords internal
#' @noRd
.plot_partitions_base <- function(pts_sf, reference_shapefile, final_fold_counts,
                                  mean_per_fold, total_folds, partition_mode,
                                  time_cols, temporal_partitioning, n_temporal,
                                  voronoi_sf, plot_palette = "Dark 2") {

  fold_colors <- .resolve_palette(total_folds, plot_palette)
  plot_list <- list()

  graphics::barplot(
    as.numeric(final_fold_counts), names.arg = names(final_fold_counts),
    col = fold_colors, main = "Fold Balance", xlab = "Fold", ylab = "Points", border = "gray30"
  )
  graphics::abline(h = mean_per_fold, lty = 2, col = "red", lwd = 2)
  plot_list$balance <- grDevices::recordPlot()

  if (!is.null(time_cols) && time_cols %in% names(pts_sf)) {
    time_vals <- sf::st_drop_geometry(pts_sf)[[time_cols]]
    fold_vals <- pts_sf$fold
    time_range <- range(time_vals, na.rm = TRUE)
    breaks_seq <- seq(time_range[1], time_range[2], length.out = 31)
    bin_labels <- round(breaks_seq[-length(breaks_seq)])

    bin_ids <- findInterval(time_vals, breaks_seq, rightmost.closed = TRUE)
    bin_ids[bin_ids == 0] <- 1
    n_bins <- length(breaks_seq) - 1

    count_matrix <- matrix(0, nrow = total_folds, ncol = n_bins)
    for (f in seq_len(total_folds)) {
      fold_bins <- bin_ids[fold_vals == f]
      tab <- table(factor(fold_bins, levels = seq_len(n_bins)))
      count_matrix[f, ] <- as.numeric(tab)
    }
    rownames(count_matrix) <- paste("Fold", seq_len(total_folds))

    label_interval <- max(1, round(n_bins / 15))
    display_labels <- rep("", n_bins)
    display_labels[seq(1, n_bins, by = label_interval)] <- bin_labels[seq(1, n_bins, by = label_interval)]

    graphics::barplot(
      count_matrix, beside = FALSE, col = fold_colors,
      names.arg = display_labels, las = 2, border = NA,
      main = "Temporal Distribution by Fold", xlab = time_cols, ylab = "Count"
    )
    graphics::legend("topright", legend = paste("Fold", seq_len(total_folds)),
                     fill = fold_colors, border = NA, cex = 0.8, bty = "n")
    plot_list$temporal <- grDevices::recordPlot()
  }

  if (!is.null(reference_shapefile)) {
    map_title <- paste(gsub("_", " ", tools::toTitleCase(partition_mode)), "Partitioning")
    graphics::plot(sf::st_geometry(reference_shapefile), col = "gray98", border = "gray40",
                   main = map_title)

    if (partition_mode != "random") {

      dissolved <- .build_fold_polygons(pts_sf, reference_shapefile, total_folds, partition_mode)

      is_temporal_mode <- partition_mode %in% c("spatiotemporal", "balanced", "temporal_only")

      if (is_temporal_mode && "block_type" %in% names(pts_sf)) {
        pts_df_bt      <- sf::st_drop_geometry(pts_sf)
        spatial_folds  <- sort(unique(pts_df_bt$fold[
          pts_df_bt$block_type %in% c("spatial_exclusive", "rebalanced")
        ]))
        temporal_folds <- sort(unique(pts_df_bt$fold[
          pts_df_bt$block_type == "temporal_exclusive"
        ]))
      } else {
        spatial_folds  <- as.integer(seq_len(total_folds))
        temporal_folds <- integer(0)
      }
      sentinel <- total_folds + 1

      if (!is.null(dissolved)) {
        for (f in spatial_folds) {
          sub <- dissolved[!is.na(dissolved$fold) & dissolved$fold == f, ]
          if (nrow(sub) == 0) next
          graphics::plot(sf::st_geometry(sub), add = TRUE,
                         col = grDevices::adjustcolor(fold_colors[f], alpha.f = 0.25),
                         border = fold_colors[f], lwd = 0.8)
        }

        if (length(temporal_folds) > 0) {
          sub_t <- dissolved[!is.na(dissolved$fold) & dissolved$fold == sentinel, ]
          if (nrow(sub_t) > 0) {
            graphics::plot(sf::st_geometry(sub_t), add = TRUE,
                           col = grDevices::adjustcolor("gray60", alpha.f = 0.20),
                           border = "gray50", lwd = 0.8, lty = 2)
          }
        }
      }

      graphics::plot(sf::st_geometry(reference_shapefile), add = TRUE,
                     col = NA, border = "gray40")
    }

    pts_coords <- sf::st_coordinates(pts_sf)

    use_temporal_shapes <- n_temporal > 1 && !any(is.na(pts_sf$temporal_block))
    pch_options <- c(16, 17, 15, 18)
    point_pch <- if (use_temporal_shapes) {
      pch_options[pmin(pts_sf$temporal_block, length(pch_options))]
    } else {
      16
    }

    graphics::points(pts_coords[, 1], pts_coords[, 2], col = fold_colors[pts_sf$fold],
                     pch = point_pch, cex = 0.8)

    legend_labels <- paste("Fold", seq_len(total_folds))
    graphics::legend("topright", legend = legend_labels,
                     col = fold_colors, pch = 16, cex = 0.8, bty = "n")
    plot_list$combined <- grDevices::recordPlot()
  }

  list(plots = plot_list, fold_polygons = if (exists("dissolved")) dissolved else NULL)
}

#' @keywords internal
#' @noRd
.build_partition_result <- function(pts_sf, voronoi_sf,
                                    final_fold_counts, mean_per_fold, imbalance,
                                    n_spatial, n_temporal, total_folds, partition_mode,
                                    temporal_partitioning, use_balanced,
                                    use_random, n_spatial_folds, n_temporal_folds,
                                    n_balanced_folds, n_random_folds,
                                    n_removed, n_original, plot_list, output_file,
                                    reference_shapefile = NULL,
                                    verbose = TRUE) {

  total_points <- nrow(pts_sf)

  if (verbose) message(paste0("--- Fold Structure (",
                              paste(gsub("_", " ", tools::toTitleCase(partition_mode))), ") ---"))
  if (verbose) message(paste0("  ", total_folds, " fold",
                              if (total_folds != 1) "s" else ""))
  for (f in seq_len(total_folds)) {
    n_pts <- as.numeric(final_fold_counts[f])
    if (verbose) message(paste0("  Fold ", f, ": ", n_pts, " (", round(n_pts / total_points * 100, 2), "%)"))
  }

  summary_stats <- data.frame(
    parameter = c("total_folds", "n_spatial_folds", "n_temporal_folds",
                  "n_balanced_folds", "n_random_folds", "partition_mode",
                  "total_points", "points_removed", "pct_rows_removed",
                  "final_imbalance_pct", "temporal_partitioning_enabled"),
    value = c(total_folds,
              n_spatial_folds, n_temporal_folds, n_balanced_folds,
              n_random_folds, partition_mode,
              total_points, n_removed,
              ifelse(n_removed > 0, round(n_removed / n_original * 100, 2), 0),
              round(imbalance * 100, 2),
              as.character(temporal_partitioning)),
    stringsAsFactors = FALSE
  )

  internal_cols <- c("spatial_block", "temporal_block", "block_type")
  pts_sf_public <- pts_sf[, !names(pts_sf) %in% internal_cols, drop = FALSE]

  fold_polygons <- .build_fold_polygons(pts_sf, reference_shapefile, total_folds, partition_mode)

  results <- list(
    folds         = sf::st_drop_geometry(pts_sf)[, "fold", drop = FALSE],
    points_sf     = pts_sf_public,
    voronoi_folds = if (!is.null(fold_polygons)) fold_polygons else voronoi_sf,
    summary       = summary_stats,
    plots         = plot_list
  )

  if (!is.null(output_file)) {
    out_dir <- dirname(output_file)
    if (!dir.exists(out_dir)) dir.create(out_dir, recursive = TRUE, showWarnings = FALSE)
    saveRDS(results, output_file)
    if (verbose) message(paste0("Results saved to: ", output_file))
  }

  invisible(results)
}

#' @keywords internal
#' @noRd
.plot_trend_timeseries <- function(timestep_summary, all_units, unit_colors,
                                   ts_x_rng, ts_y_max, x_ticks) {
  right_mar <- max(2, ceiling(max(nchar(all_units)) / 2.5) + 1)
  opar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(opar), add = TRUE)
  graphics::par(mar = c(4, 7.5, 3.5, right_mar),
                mgp = c(1.8, 0.7, 0))
  graphics::plot(NULL,
                 xlim = ts_x_rng,
                 ylim = c(0, ts_y_max * 1.1),
                 xlab = "Time Step", ylab = "",
                 main = "Suitable Habitat Pixels Over Time",
                 las = 1, xaxt = "n")
  graphics::mtext("Suitable Pixels", side = 2, line = 4.5, cex = 1)
  graphics::axis(1, at = x_ticks)
  for (u in all_units) {
    ud <- timestep_summary[timestep_summary$Spatial_Unit == u, ]
    ud <- ud[order(ud$Time_Step), ]
    graphics::lines(ud$Time_Step, ud$Pixels_Suitable,
                    col = unit_colors[u], lwd = 1.8)
    graphics::points(ud$Time_Step, ud$Pixels_Suitable,
                     col = unit_colors[u], pch = 19, cex = 0.7)
  }
  usr <- graphics::par("usr")
  x_legend <- usr[2] + (usr[2] - usr[1]) * 0.02
  y_legend <- usr[4]
  graphics::legend(x_legend, y_legend,
                   legend = all_units,
                   col    = unit_colors[all_units],
                   lwd    = 2, pch = 19,
                   bty    = "n", cex = 0.78,
                   xpd    = TRUE)
}

#' @keywords internal
#' @noRd
.plot_trend_change_per_step <- function(gain_v, loss_v, yrs_chg, ac_ylim) {
  opar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(opar), add = TRUE)
  graphics::par(mar = c(5, 8, 3.5, 2),
                mgp = c(1.8, 0.7, 0))
  graphics::barplot(gain_v,
                    names.arg = yrs_chg, las = 2,
                    ylim      = ac_ylim,
                    col       = "#267300",
                    border    = NA,
                    ylab      = "",
                    main      = "Annual Habitat Gains and Losses")
  graphics::mtext("Pixels", side = 2, line = 5, cex = 1)
  graphics::barplot(-loss_v,
                    add    = TRUE,
                    col    = "#E31A1C",
                    border = NA,
                    axes   = FALSE,
                    names.arg = rep("", length(yrs_chg)))
  graphics::abline(h = 0, col = "black", lwd = 0.8)
  graphics::legend("topright",
                   legend = c("Gain", "Loss"),
                   fill   = c("#267300", "#E31A1C"),
                   bty    = "n", cex = 0.85)
}

#' @keywords internal
#' @noRd
.plot_trend_total_unit <- function(g_tot, l_tot, shared_units, time_steps, tcu_ymax) {
  left_mar <- max(4, ceiling(max(nchar(shared_units)) / 2.5))
  opar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(opar), add = TRUE)
  graphics::par(mar = c(4, left_mar, 3.5, 1),
                mgp = c(1.8, 0.7, 0))

  bar_mat <- matrix(c(g_tot, l_tot), nrow = 2, byrow = TRUE,
                    dimnames = list(c("Gain", "Loss"), shared_units))

  graphics::barplot(
    bar_mat,
    beside    = TRUE,
    horiz     = TRUE,
    col       = c("#267300", "#E31A1C"),
    border    = NA,
    xlim      = c(0, tcu_ymax * 1.25),
    las       = 1,
    xlab      = "Total Pixels",
    main      = paste0("Total Gains and Losses by Unit (",
                       min(time_steps), "-", max(time_steps), ")")
  )
  graphics::legend("bottomright",
                   legend = c("Gain", "Loss"),
                   fill   = c("#267300", "#E31A1C"),
                   bty    = "n", cex = 0.9)
}

#' @keywords internal
#' @noRd
.plot_trend_facet_units <- function(change_by_timestep, all_units, n_rows, n_cols) {
  opar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(opar), add = TRUE)
  graphics::par(mfrow = c(n_rows, n_cols),
                mar   = c(3.5, 7, 2.5, 0.5),
                oma   = c(0, 0, 3, 0),
                mgp   = c(2.2, 0.6, 0))
  for (u in all_units) {
    ud    <- change_by_timestep[change_by_timestep$Spatial_Unit == u, ]
    ud    <- ud[order(ud$Time_Step), ]
    g_v   <- as.numeric(ud$Increase_Pixels)
    l_v   <- as.numeric(ud$Decrease_Pixels)
    y_max_raw <- suppressWarnings(max(c(g_v, l_v), na.rm = TRUE))
    y_max <- if (is.finite(y_max_raw) && y_max_raw > 0) y_max_raw * 1.2 else 1
    y_lim <- c(-y_max, y_max)

    graphics::barplot(g_v,
                      names.arg = ud$Time_Step,
                      ylim      = y_lim,
                      col       = "#267300",
                      border    = NA,
                      las       = 2,
                      cex.names = 0.55,
                      main      = u,
                      ylab      = "")
    graphics::mtext("Pixels", side = 2, line = 4.5, cex = 0.7)
    graphics::barplot(-l_v,
                      add       = TRUE,
                      col       = "#E31A1C",
                      border    = NA,
                      axes      = FALSE,
                      names.arg = rep("", nrow(ud)))
    graphics::abline(h = 0, col = "gray40", lwd = 0.6)
  }
  graphics::mtext("Gains and Losses by Unit",
                  outer = TRUE, side = 3, line = 1, cex = 1.1, font = 2)
}

#' @keywords internal
#' @noRd
.validate_variable_patterns <- function(variable_patterns) {
  if (!is.vector(variable_patterns) || is.null(names(variable_patterns))) {
    stop(paste0(
      "ERROR: 'variable_patterns' must be a named character vector mapping variable names ",
      "to filename patterns (e.g. c(\"Var1\" = \"Var1_YEAR\", \"StaticVar\" = \"StaticVar\"))."
    ))
  }
  if (any(names(variable_patterns) == "")) {
    stop("ERROR: All elements in 'variable_patterns' must be named.")
  }
  invisible(NULL)
}

#' @keywords internal
#' @noRd
.classify_variable_patterns <- function(variable_patterns, time_cols) {
  if (is.null(time_cols)) time_cols <- character(0)
  time_cols <- as.character(time_cols)

  dynamic_vars <- character(0)
  for (v in names(variable_patterns)) {
    pat <- variable_patterns[v]
    is_dynamic <- FALSE
    for (tc in time_cols) {
      if (grepl(tc, pat, ignore.case = TRUE)) {
        is_dynamic <- TRUE
        break
      }
    }
    if (is_dynamic) dynamic_vars <- c(dynamic_vars, v)
  }
  static_vars <- setdiff(names(variable_patterns), dynamic_vars)

  if (length(dynamic_vars) > 0 && length(time_cols) > 0) {
    placeholders <- character(0)
    for (v in dynamic_vars) {
      parts <- strsplit(variable_patterns[v], "_")[[1]]
      placeholders <- c(placeholders, parts[toupper(parts) %in% toupper(time_cols)])
    }
    placeholders <- unique(placeholders)
    missing_in_patterns <- toupper(time_cols)[!toupper(time_cols) %in% toupper(placeholders)]
    if (length(missing_in_patterns) > 0) {
      warning("time_cols includes columns not found in variable_patterns: ",
              paste(missing_in_patterns, collapse = ", "))
    }
  }

  list(dynamic = dynamic_vars, static = static_vars)
}

#' @keywords internal
#' @noRd
.load_raster_input <- function(x, arg_name = "raster") {
  if (is.character(x)) {
    if (length(x) == 1 && dir.exists(x)) {
      tif_files <- list.files(x, pattern = "\\.tif$", full.names = TRUE)
      if (length(tif_files) == 0) {
        stop(paste0("ERROR: No .tif files found in directory '", x, "' for '", arg_name, "'."))
      }
      tif_files <- .natural_sort(tif_files)
      out <- terra::rast(tif_files)
      names(out) <- tools::file_path_sans_ext(basename(tif_files))
      return(out)
    }
    if (!file.exists(x)) {
      stop(paste0("ERROR: '", arg_name, "' file does not exist: '", x, "'."))
    }
    return(terra::rast(x))
  }
  if (inherits(x, "RasterLayer")) {
    warning(paste0("Converting '", arg_name, "' from 'RasterLayer' to 'SpatRaster' for terra compatibility."))
    return(terra::rast(x))
  }
  if (inherits(x, "SpatRaster")) {
    return(x)
  }
  stop(paste0("ERROR: '", arg_name, "' must be a file path, 'RasterLayer', or 'SpatRaster'. ",
              "Provided object is of class: ", class(x)[1]))
}

#' @keywords internal
#' @noRd
.load_shapefile_input <- function(x, arg_name = "shapefile") {
  if (inherits(x, "sf")) return(x)
  if (inherits(x, "sfc")) return(sf::st_sf(geometry = x))
  if (inherits(x, "Spatial")) return(sf::st_as_sf(x))
  if (is.character(x)) {
    if (!file.exists(x)) {
      stop(paste0("ERROR: '", arg_name, "' file not found: '", x, "'."))
    }
    return(sf::st_read(x, quiet = TRUE))
  }
  stop(paste0("ERROR: '", arg_name, "' must be an sf object, sfc, Spatial object, or file path. ",
              "Provided object is of class: ", class(x)[1]))
}

#' @keywords internal
#' @noRd
.resolve_threshold_method <- function(threshold_method) {
  if (is.numeric(threshold_method)) {
    if (length(threshold_method) != 1 || threshold_method <= 0 || threshold_method >= 1) {
      stop("ERROR: When supplying a numeric threshold, it must be a single value between 0 and 1.")
    }
    return(list(method = "manual", manual_value = threshold_method))
  }
  if (!threshold_method %in% c("prevalence", "tss", "manual")) {
    stop(paste0("ERROR: 'threshold_method' must be 'prevalence', 'tss', ",
                "or a numeric value between 0 and 1."))
  }
  list(method = threshold_method, manual_value = NULL)
}

#' @keywords internal
#' @noRd
.normalize_time_steps <- function(time_steps, fn_name) {
  secondary_filters <- list()
  if (is.data.frame(time_steps) || is.matrix(time_steps)) {
    time_steps_df <- as.data.frame(time_steps, stringsAsFactors = FALSE)
    if (ncol(time_steps_df) == 1) {
      time_steps <- time_steps_df[[1]]
    } else {
      primary_col    <- names(time_steps_df)[1]
      secondary_cols <- names(time_steps_df)[-1]
      for (sc in secondary_cols) {
        u_vals <- unique(time_steps_df[[sc]])
        if (length(u_vals) > 1) {
          stop(paste0(
            "ERROR: Secondary time column '", sc, "' has ", length(u_vals),
            " unique values (", paste(u_vals, collapse = ", "), "). ",
            fn_name, "() requires each secondary column to have ",
            "exactly one unique value so that raster layers are unambiguous. ",
            "Filter time_steps to a single value of '", sc, "' before calling this function. ",
            "For example: time_steps <- expand.grid(time_step = 1:15, season = \"Spring\", ",
            "stringsAsFactors = FALSE)"
          ))
        }
        secondary_filters[[sc]] <- u_vals
      }
      time_steps <- time_steps_df[[primary_col]]
    }
  }
  if (!is.atomic(time_steps) || is.list(time_steps)) {
    stop("ERROR: 'time_steps' must be a plain atomic vector of time step values.")
  }
  if (anyDuplicated(time_steps) > 0) {
    stop(paste0(
      "ERROR: 'time_steps' contains duplicate values. ",
      "Each raster layer must correspond to a unique time step. ",
      "If your predictions have repeated time steps across a secondary time column (e.g. season), ",
      "filter time_steps to a single value of that column before calling this function. ",
      "For example: time_steps <- expand.grid(time_step = 1:15, season = \"Spring\", ",
      "stringsAsFactors = FALSE)"
    ))
  }
  list(time_steps = time_steps, secondary_filters = secondary_filters)
}

#' @keywords internal
#' @noRd
.parse_model_formula <- function(f) {
  if (inherits(f, "formula")) {
    if (length(f) == 3) {
      warning(paste0(
        "model_formula appears to include a response variable on the left-hand side. ",
        "Only the right-hand side is used; 'presence' is always added automatically."
      ))
      return(stats::as.formula(
        paste("~", gsub("\\s+", " ", paste(deparse(f[[3]]), collapse = " ")))
      ))
    }
    return(f)
  }
  if (is.character(f) && length(f) == 1) {
    f <- trimws(f)
    if (!startsWith(f, "~")) {
      stop(paste0(
        "ERROR: model_formula as a character string must start with '~', ",
        "e.g. '~ Var1 + Var2'. Got: '", f, "'."
      ))
    }
    return(tryCatch(
      stats::as.formula(f),
      error = function(e) stop(paste0("ERROR: Could not parse model_formula '", f, "': ", e$message))
    ))
  }
  stop("ERROR: model_formula must be a formula object or a character string starting with '~'.")
}

#' @keywords internal
#' @noRd
.select_threshold_prob <- function(pred_prob, observed, method, manual_value) {
  if (method == "manual") return(manual_value)
  if (method == "prevalence") return(mean(observed, na.rm = TRUE))
  candidates <- seq(0.01, 0.99, by = 0.01)
  tss_vals <- vapply(candidates, function(thr) {
    pred_bin <- as.integer(pred_prob >= thr)
    tp   <- sum(pred_bin == 1 & observed == 1, na.rm = TRUE)
    fn   <- sum(pred_bin == 0 & observed == 1, na.rm = TRUE)
    tn   <- sum(pred_bin == 0 & observed == 0, na.rm = TRUE)
    fp   <- sum(pred_bin == 1 & observed == 0, na.rm = TRUE)
    sens <- if ((tp + fn) > 0) tp / (tp + fn) else 0
    spec <- if ((tn + fp) > 0) tn / (tn + fp) else 0
    sens + spec - 1
  }, numeric(1))
  candidates[which.max(tss_vals)]
}

#' @keywords internal
#' @noRd
.compute_confusion_metrics <- function(observed, predicted, pred_prob) {
  tp <- sum(predicted == 1 & observed == 1, na.rm = TRUE)
  fn <- sum(predicted == 0 & observed == 1, na.rm = TRUE)
  tn <- sum(predicted == 0 & observed == 0, na.rm = TRUE)
  fp <- sum(predicted == 1 & observed == 0, na.rm = TRUE)
  sensitivity <- if ((tp + fn) > 0) tp / (tp + fn) else NA
  specificity <- if ((tn + fp) > 0) tn / (tn + fp) else NA
  tss         <- if (!is.na(sensitivity) && !is.na(specificity)) sensitivity + specificity - 1 else NA
  n     <- tp + tn + fp + fn
  po    <- (tp + tn) / n
  pe    <- ((tp + fp) * (tp + fn) + (tn + fn) * (tn + fp)) / (n * n)
  kappa <- if ((1 - pe) > 0) (po - pe) / (1 - pe) else NA
  auc <- tryCatch({
    ord   <- order(pred_prob, decreasing = TRUE)
    o_obs <- observed[ord]
    n_pos <- sum(observed == 1, na.rm = TRUE)
    n_neg <- sum(observed == 0, na.rm = TRUE)
    if (n_pos == 0 || n_neg == 0) return(NA)
    tpr_v <- cumsum(o_obs == 1) / n_pos
    fpr_v <- cumsum(o_obs == 0) / n_neg
    sum(diff(fpr_v) * (tpr_v[-1] + tpr_v[-length(tpr_v)]) / 2)
  }, error = function(e) NA)
  list(tp = tp, fn = fn, tn = tn, fp = fp,
       sensitivity = sensitivity, specificity = specificity,
       tss = tss, kappa = kappa, auc = auc)
}

#' @keywords internal
#' @noRd
.compute_roc_curve <- function(observed, predicted) {
  thresholds <- sort(unique(predicted), decreasing = TRUE)
  tpr   <- numeric(length(thresholds))
  fpr   <- numeric(length(thresholds))
  n_pos <- sum(observed == 1, na.rm = TRUE)
  n_neg <- sum(observed == 0, na.rm = TRUE)
  for (i in seq_along(thresholds)) {
    pred_bin <- as.integer(predicted >= thresholds[i])
    tp <- sum(pred_bin == 1 & observed == 1, na.rm = TRUE)
    fp <- sum(pred_bin == 1 & observed == 0, na.rm = TRUE)
    tpr[i] <- if (n_pos > 0) tp / n_pos else 0
    fpr[i] <- if (n_neg > 0) fp / n_neg else 0
  }
  list(tpr = tpr, fpr = fpr, thresholds = thresholds)
}

#' @keywords internal
#' @noRd
.plot_response_curves_parametric <- function(model_fit, train_data, rhs_vars,
                                             var_means, fold_color, threshold,
                                             fold, n_rows, n_cols, model_label) {
  opar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(opar), add = TRUE)
  graphics::par(mfrow = c(n_rows, n_cols), mar = c(4, 4, 2.5, 1),
                oma = c(0, 0, 3, 0))

  pr_data <- train_data[train_data$presence == 1, , drop = FALSE]
  pa_data <- train_data[train_data$presence == 0, , drop = FALSE]
  for (v in rhs_vars) {
    v_range   <- seq(min(train_data[[v]], na.rm = TRUE),
                     max(train_data[[v]], na.rm = TRUE), length.out = 100)
    pred_grid <- as.data.frame(matrix(
      rep(var_means, each = 100), nrow = 100,
      dimnames = list(NULL, rhs_vars)
    ))
    pred_grid[[v]] <- v_range
    pred_prob <- stats::predict(model_fit, newdata = pred_grid, type = "response")
    graphics::plot(v_range, pred_prob, type = "l", lwd = 2,
                   col = fold_color, ylim = c(0, 1),
                   xlab = v, ylab = "P(presence)", main = v, las = 1)
    graphics::rug(pr_data[[v]], side = 3,
                  col = grDevices::adjustcolor("steelblue", alpha.f = 0.5),
                  ticksize = 0.03, lwd = 0.8)
    graphics::rug(pa_data[[v]], side = 1,
                  col = grDevices::adjustcolor("tomato", alpha.f = 0.5),
                  ticksize = 0.03, lwd = 0.8)
    if (!is.null(threshold)) {
      graphics::abline(h = threshold, lty = 2, col = "gray40", lwd = 1)
      graphics::legend("topright", legend = paste0("thr = ", round(threshold, 2)),
                       lty = 2, col = "gray40", bty = "n", cex = 0.8)
    }
  }
  graphics::mtext(
    paste0(model_label, " Fold ", fold, " Response Curves"),
    side = 3, outer = TRUE, line = 1, cex = 1.1, font = 2
  )
}

#' @keywords internal
#' @noRd
.plot_roc_summary_parametric <- function(model_list, train_data_list,
                                         thresholds, fold_metrics_df,
                                         all_folds, fold_colors, model_label) {
  opar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(opar), add = TRUE)
  graphics::par(mar = c(4.5, 4.5, 3, 1.5))

  roc_curves <- lapply(all_folds, function(fold) {
    fold_key  <- paste0("fold", fold)
    pred_prob <- stats::predict(model_list[[fold_key]],
                                newdata = train_data_list[[fold_key]],
                                type = "response")
    .compute_roc_curve(observed = train_data_list[[fold_key]]$presence,
                       predicted = pred_prob)
  })
  graphics::plot(NULL, xlim = c(0, 1), ylim = c(0, 1),
                 xlab = "1 - Specificity (FPR)", ylab = "Sensitivity (TPR)",
                 main = paste0(model_label, " ROC Curves - All Folds"), las = 1)
  graphics::abline(a = 0, b = 1, lty = 2, col = "gray60")
  for (i in seq_along(all_folds)) {
    rc       <- roc_curves[[i]]
    fold_key <- paste0("fold", all_folds[i])
    graphics::lines(rc$fpr, rc$tpr, col = fold_colors[fold_key], lwd = 2)
    thr         <- thresholds[fold_key]
    closest_idx <- which.min(abs(rc$thresholds - thr))
    graphics::points(rc$fpr[closest_idx], rc$tpr[closest_idx],
                     pch = 19, cex = 1.3, col = fold_colors[fold_key])
  }
  graphics::legend(
    "bottomright",
    legend = paste0("Fold ", all_folds,
                    " (AUC = ", round(fold_metrics_df$AUC, 2), ")"),
    col = fold_colors, lwd = 2, pch = 19,
    bty = "n", cex = 0.85
  )
}

#' @keywords internal
#' @noRd
.plot_rf_importance <- function(model_vars, all_folds, importance_list, fold_colors) {
  opar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(opar), add = TRUE)

  n_vars_loc  <- length(model_vars)
  n_cols_loc  <- min(n_vars_loc, 3)
  n_rows_loc  <- ceiling(n_vars_loc / n_cols_loc)
  bot_mar     <- max(nchar(paste0("Fold ", all_folds))) * 0.5 + 1.5

  graphics::par(mfrow = c(n_rows_loc, n_cols_loc),
                mar   = c(bot_mar, 4, 3, 1),
                oma   = c(0, 0, 3, 0))

  for (v in model_vars) {
    imp_vals <- vapply(all_folds, function(fold) {
      fk  <- paste0("fold", fold)
      df  <- importance_list[[fk]]
      if (is.null(df)) return(NA_real_)
      idx <- match(v, df$variable)
      if (is.na(idx)) NA_real_ else df$mean_decr_acc[idx]
    }, numeric(1))

    bar_cols  <- fold_colors[paste0("fold", all_folds)]
    fold_labs <- paste0("Fold ", all_folds)
    y_rng     <- c(0, max(imp_vals, na.rm = TRUE) * 1.15)

    graphics::barplot(
      imp_vals,
      names.arg = fold_labs,
      col       = bar_cols,
      border    = NA,
      horiz     = FALSE,
      las       = 2,
      ylim      = y_rng,
      ylab      = "Mean Decrease in Accuracy",
      main      = v,
      cex.names = 0.8,
      cex.main  = 0.95
    )
  }
  graphics::mtext(
    "RF Variable Importance by Fold",
    side = 3, outer = TRUE, line = 1, cex = 1.1, font = 2
  )
}

#' @keywords internal
#' @noRd
.plot_rf_marginal_curves <- function(model_fit, train_data, model_vars,
                                     var_means, fold_color, threshold,
                                     fold, n_rows, n_cols) {
  opar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(opar), add = TRUE)
  graphics::par(mfrow = c(n_rows, n_cols), mar = c(4, 4, 2.5, 1),
                oma = c(0, 0, 3, 0))

  pr_data <- train_data[train_data$presence == 1, , drop = FALSE]
  pa_data <- train_data[train_data$presence == 0, , drop = FALSE]
  for (v in model_vars) {
    v_range   <- seq(min(train_data[[v]], na.rm = TRUE),
                     max(train_data[[v]], na.rm = TRUE), length.out = 100)
    pred_grid <- as.data.frame(matrix(
      rep(var_means, each = 100), nrow = 100,
      dimnames = list(NULL, model_vars)
    ))
    pred_grid[[v]]     <- v_range
    pred_grid$presence <- factor(0, levels = c(0, 1))

    pred_prob <- tryCatch(
      stats::predict(model_fit, newdata = pred_grid, type = "prob")[, "1"],
      error = function(e) rep(NA_real_, 100)
    )

    graphics::plot(v_range, pred_prob, type = "l", lwd = 2,
                   col = fold_color, ylim = c(0, 1),
                   xlab = v, ylab = "P(presence)", main = v, las = 1)
    graphics::rug(pr_data[[v]], side = 3,
                  col = grDevices::adjustcolor("steelblue", alpha.f = 0.5),
                  ticksize = 0.03, lwd = 0.8)
    graphics::rug(pa_data[[v]], side = 1,
                  col = grDevices::adjustcolor("tomato", alpha.f = 0.5),
                  ticksize = 0.03, lwd = 0.8)
    if (!is.null(threshold)) {
      graphics::abline(h = threshold, lty = 2, col = "gray40", lwd = 1)
      graphics::legend("topright", legend = paste0("thr = ", round(threshold, 2)),
                       lty = 2, col = "gray40", bty = "n", cex = 0.8)
    }
  }
  graphics::mtext(
    paste0("RF Fold ", fold, " Marginal Prediction Curves"),
    side = 3, outer = TRUE, line = 1, cex = 1.1, font = 2
  )
}

#' @keywords internal
#' @noRd
#' @importFrom tools file_ext
.load_pseudoabsence_result <- function(pseudoabsence_result, verbose = TRUE) {
  if (is.character(pseudoabsence_result)) {
    if (!file.exists(pseudoabsence_result)) {
      stop(paste0("ERROR: File not found: ", pseudoabsence_result))
    }
    if (tolower(tools::file_ext(pseudoabsence_result)) != "rds") {
      stop("ERROR: pseudoabsence_result file must be .rds format.")
    }
    if (verbose) message(paste("Reading pseudoabsence results from:", basename(pseudoabsence_result)))
    pseudoabsence_result <- tryCatch(
      readRDS(pseudoabsence_result),
      error = function(e) stop(paste0("ERROR reading pseudoabsence_result .rds: ", e$message))
    )
  }
  if (!is.list(pseudoabsence_result) || !"pseudoabsences" %in% names(pseudoabsence_result)) {
    stop(paste0("ERROR: pseudoabsence_result must contain a 'pseudoabsences' element. ",
                "Re-run generate_absences()."))
  }
  pseudoabsence_result
}

#' @keywords internal
#' @noRd
#' @importFrom tools file_ext
.load_partition_result <- function(partition_result, verbose = TRUE) {
  if (is.character(partition_result)) {
    if (!file.exists(partition_result)) {
      stop(paste0("ERROR: File not found: ", partition_result))
    }
    if (tolower(tools::file_ext(partition_result)) != "rds") {
      stop("ERROR: partition_result file must be .rds format.")
    }
    if (verbose) message(paste("Reading partition results from:", basename(partition_result)))
    partition_result <- tryCatch(
      readRDS(partition_result),
      error = function(e) stop(paste0("ERROR reading partition_result .rds: ", e$message))
    )
  }
  if (!is.list(partition_result) || !"points_sf" %in% names(partition_result)) {
    stop(paste0("ERROR: partition_result must contain a 'points_sf' element. ",
                "Re-run spatiotemporal_partition()."))
  }
  partition_result
}

#' @keywords internal
#' @noRd
.prepare_combined_modeling_data <- function(partition_result, pseudoabsence_result,
                                            model_vars, time_cols = NULL,
                                            verbose = TRUE) {

  presence_sf <- partition_result$points_sf
  if (is.null(presence_sf) || nrow(presence_sf) == 0) {
    stop("ERROR: 'partition_result$points_sf' is empty.")
  }
  if (!"fold" %in% names(presence_sf)) {
    stop("ERROR: 'partition_result$points_sf' is missing 'fold' column. Re-run spatiotemporal_partition().")
  }
  presence_df <- sf::st_drop_geometry(presence_sf)
  if (!"presence" %in% names(presence_df)) {
    if (verbose) message("No 'presence' column found in partition_result$points_sf. Adding presence = 1 for all rows.")
    presence_df$presence <- 1
  }

  pseudoabs_sf <- pseudoabsence_result$pseudoabsences
  if (is.null(pseudoabs_sf) || nrow(pseudoabs_sf) == 0) {
    stop("ERROR: 'pseudoabsence_result$pseudoabsences' is empty.")
  }
  pseudoabs_df <- sf::st_drop_geometry(pseudoabs_sf)
  if (!"presence" %in% names(pseudoabs_df)) {
    pseudoabs_df$presence <- 0
  }
  if (!"fold" %in% names(pseudoabs_df)) {
    stop("ERROR: 'pseudoabsence_result$pseudoabsences' is missing 'fold' column. Re-run generate_absences().")
  }

  missing_pr <- model_vars[!model_vars %in% names(presence_df)]
  missing_pa <- model_vars[!model_vars %in% names(pseudoabs_df)]
  if (length(missing_pr) > 0) {
    stop(paste0(
      "ERROR: The following predictors are missing from the presence data: ",
      paste(missing_pr, collapse = ", "),
      ". Run temporally_explicit_extraction() before partitioning, or check variable names."
    ))
  }
  if (length(missing_pa) > 0) {
    stop(paste0(
      "ERROR: The following predictors are missing from the pseudoabsence data: ",
      paste(missing_pa, collapse = ", "),
      ". Ensure generate_absences() was run with a raster_dir."
    ))
  }

  keep_cols    <- unique(c("fold", "presence", model_vars, time_cols))
  keep_cols_pr <- keep_cols[keep_cols %in% names(presence_df)]
  keep_cols_pa <- keep_cols[keep_cols %in% names(pseudoabs_df)]

  combined_df <- rbind(
    presence_df[, keep_cols_pr, drop = FALSE],
    pseudoabs_df[, keep_cols_pa, drop = FALSE]
  )

  combined_df          <- combined_df[!is.na(combined_df$fold), ]
  combined_df$presence <- as.integer(combined_df$presence)

  complete_rows <- stats::complete.cases(combined_df[, model_vars, drop = FALSE])
  if (sum(!complete_rows) > 0) {
    if (verbose) message(paste0("Removing ", sum(!complete_rows), " rows with NA values in model predictors."))
    combined_df <- combined_df[complete_rows, ]
  }
  if (nrow(combined_df) == 0) {
    stop("ERROR: No complete cases remain after removing NA values in model predictors.")
  }

  all_folds <- sort(unique(combined_df$fold))
  if (length(all_folds) == 0) {
    stop("ERROR: No valid fold assignments found. Re-run spatiotemporal_partition().")
  }
  single_fold_mode <- length(all_folds) == 1
  if (single_fold_mode && verbose) {
    message("Single-fold mode: all points used for both training and testing.")
  }

  list(combined_df = combined_df, all_folds = all_folds,
       single_fold_mode = single_fold_mode)
}

#' @keywords internal
#' @noRd
.build_fold_metric_row <- function(fold, thr, metrics) {
  data.frame(
    Fold          = fold,
    Threshold     = round(thr, 4),
    Testing_TP    = metrics$tp,
    Testing_FN    = metrics$fn,
    Testing_TN    = metrics$tn,
    Testing_FP    = metrics$fp,
    Sensitivity   = round(metrics$sensitivity, 4),
    Specificity   = round(metrics$specificity, 4),
    TSS           = round(metrics$tss,         4),
    Kappa         = round(metrics$kappa,       4),
    AUC           = round(metrics$auc,         4),
    stringsAsFactors = FALSE
  )
}

#' @keywords internal
#' @noRd
.plot_fold_metrics_bars <- function(metrics_df, rate_cols, fold_colors, title) {

  n_metrics <- length(rate_cols)

  ref_h <- vapply(rate_cols, function(m) {
    if (m %in% c("TSS", "Kappa")) 0 else NA_real_
  }, numeric(1))

  fold_ids <- metrics_df$Fold

  bar_cols <- vapply(fold_ids, function(f) {
    key <- paste0("fold", f)
    if (key %in% names(fold_colors)) fold_colors[[key]]
    else if (as.character(f) %in% names(fold_colors)) fold_colors[[as.character(f)]]
    else "gray70"
  }, character(1))

  opar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(opar), add = TRUE)

  graphics::par(
    mfrow  = c(1, n_metrics),
    mar    = c(4, 4, 4, 1),
    oma    = c(0, 0, 3, 0)
  )

  for (mi in seq_along(rate_cols)) {
    col_name <- rate_cols[mi]
    vals     <- metrics_df[[col_name]]
    ref      <- ref_h[mi]

    y_lim <- c(
      max(min(c(0, if (!is.na(ref)) ref, vals), na.rm = TRUE) - 0.05, -0.1),
      min(max(c(1, if (!is.na(ref)) ref, vals), na.rm = TRUE) + 0.05,  1.1)
    )

    graphics::barplot(
      vals,
      names.arg = paste0("Fold ", fold_ids),
      col       = bar_cols,
      border    = "gray30",
      ylim      = y_lim,
      las       = 1,
      main      = col_name,
      ylab      = if (mi == 1) "Value" else "",
      cex.names = 0.85,
      cex.main  = 1.1,
      cex.axis  = 0.9
    )

    if (!is.na(ref)) {
      graphics::abline(h = ref, col = "gray50", lty = 3, lwd = 0.9)
    }

    mean_val <- mean(vals, na.rm = TRUE)
    graphics::abline(h = mean_val, col = "black", lty = 2, lwd = 1.8)
    graphics::mtext(paste0("mean = ", round(mean_val, 3)),
                    side = 3, line = 0, adj = 1, cex = 0.72, col = "gray30")
  }

  graphics::mtext(title, side = 3, outer = TRUE, line = 1, cex = 1.05, font = 2)
}

#' @keywords internal
#' @noRd
.compute_timestep_metrics <- function(binary_raster, test_pts_time,
                                      pseudo_pts_time = NULL,
                                      model = NULL, model_vars = NULL,
                                      threshold = NULL, model_type = NULL,
                                      time_values, time_cols, fold) {

  total_cells    <- terra::ncell(binary_raster) -
    sum(is.na(terra::values(binary_raster)), na.rm = TRUE)
  suitable_cells <- sum(terra::values(binary_raster) == 1, na.rm = TRUE)
  pct_suitable   <- if (total_cells > 0) suitable_cells / total_cells else NA_real_

  n_pres <- nrow(test_pts_time)
  n_abs  <- if (!is.null(pseudo_pts_time)) nrow(pseudo_pts_time) else 0
  has_pa <- !is.null(model) && !is.null(threshold) && !is.null(model_vars) &&
    !is.null(pseudo_pts_time) && n_abs > 0

  tp <- NA_integer_; fn <- NA_integer_
  tn <- NA_integer_; fp <- NA_integer_
  sensitivity <- NA_real_; specificity <- NA_real_
  tss <- NA_real_;  cbp <- NA_real_

  if (n_pres > 0 || has_pa) {

    if (has_pa) {

      pr_df <- sf::st_drop_geometry(test_pts_time)
      pa_df <- sf::st_drop_geometry(pseudo_pts_time)

      pr_has_vars <- all(model_vars %in% names(pr_df))
      pa_has_vars <- all(model_vars %in% names(pa_df))

      if (pr_has_vars || pa_has_vars) {
        rows <- list()
        if (pr_has_vars && n_pres > 0) {
          r <- pr_df[, model_vars, drop = FALSE]; r$obs <- 1; rows[["pr"]] <- r
        }
        if (pa_has_vars && n_abs > 0) {
          r <- pa_df[, model_vars, drop = FALSE]; r$obs <- 0; rows[["pa"]] <- r
        }
        combined      <- do.call(rbind, rows)
        complete_rows <- stats::complete.cases(combined[, model_vars, drop = FALSE])
        combined      <- combined[complete_rows, , drop = FALSE]

        if (nrow(combined) > 0) {
          pred_sub <- combined[, model_vars, drop = FALSE]
          raw <- tryCatch({
            if (model_type == "rf") {
              pred_sub$presence <- factor(0, levels = c(0, 1))
              stats::predict(model, newdata = pred_sub, type = "prob")[, "1"]
            } else {
              stats::predict(model, newdata = pred_sub, type = "response")
            }
          }, error = function(e) {
            warning(paste("Timestep prediction failed:", e$message))
            NULL
          })

          if (!is.null(raw)) {
            pred_bin <- as.integer(raw >= threshold)
            obs      <- combined$obs
            tp  <- sum(pred_bin == 1 & obs == 1, na.rm = TRUE)
            fn  <- sum(pred_bin == 0 & obs == 1, na.rm = TRUE)
            tn  <- sum(pred_bin == 0 & obs == 0, na.rm = TRUE)
            fp  <- sum(pred_bin == 1 & obs == 0, na.rm = TRUE)
            n_p <- tp + fn; n_n <- tn + fp
            sensitivity <- if (n_p > 0) tp / n_p else NA_real_
            specificity <- if (n_n > 0) tn / n_n else NA_real_
            tss         <- if (!is.na(sensitivity) && !is.na(specificity)) {
              sensitivity + specificity - 1
            } else NA_real_
            cbp <- if (n_p > 0 && total_cells > 0) {
              stats::dbinom(tp, size = n_p, prob = suitable_cells / total_cells)
            } else NA_real_
          }
        }
      }

    } else if (n_pres > 0) {
      pts_vect  <- tryCatch(suppressWarnings(terra::vect(test_pts_time)), error = function(e) NULL)
      extracted <- if (!is.null(pts_vect)) {
        tryCatch(terra::extract(binary_raster, pts_vect)[, 2],
                 error = function(e) rep(NA_real_, n_pres))
      } else rep(NA_real_, n_pres)

      extracted[is.na(extracted)] <- 0
      tp  <- sum(extracted == 1, na.rm = TRUE)
      fn  <- sum(extracted == 0, na.rm = TRUE)
      tot <- tp + fn
      sensitivity <- if (tot > 0) tp / tot else NA_real_
      cbp <- if (tot > 0 && total_cells > 0) {
        stats::dbinom(tp, size = tot, prob = suitable_cells / total_cells)
      } else NA_real_
    }
  }

  row <- data.frame(
    Fold         = fold,
    Pct_Suitable = round(pct_suitable, 4),
    N_Pres       = n_pres,
    TP           = tp,
    FN           = fn,
    Sensitivity  = round(sensitivity, 4),
    CBP          = cbp,
    N_Abs        = if (has_pa) n_abs                else NA_integer_,
    TN           = if (has_pa) tn                   else NA_integer_,
    FP           = if (has_pa) fp                   else NA_integer_,
    Specificity  = if (has_pa) round(specificity, 4) else NA_real_,
    TSS          = if (has_pa) round(tss,          4) else NA_real_,
    stringsAsFactors = FALSE
  )

  for (tc in time_cols) row[[tc]] <- time_values[[tc]]
  row
}

#' @keywords internal
#' @noRd
.resolve_raster_paths <- function(variable_patterns, dynamic_vars, static_vars,
                                  time_cols, time_values, raster_dir) {
  all_files <- list.files(raster_dir, pattern = "\\.tif$", full.names = TRUE)
  paths     <- character(0)

  for (var in dynamic_vars) {
    fname <- variable_patterns[[var]]
    for (tc in time_cols) {
      if (grepl(tc, fname, ignore.case = TRUE)) {
        fname <- gsub(tc, as.character(time_values[[tc]]), fname, ignore.case = TRUE)
      }
    }

    anchored_pat <- paste0(fname, "(?=[^0-9]|$)")
    matches <- all_files[grepl(anchored_pat, basename(all_files),
                               ignore.case = TRUE, perl = TRUE)]
    if (length(matches) == 0) {
      matches <- all_files[grepl(fname, basename(all_files), ignore.case = TRUE)]
    }
    if (length(matches) == 0) {
      warning(paste0("Missing raster file for dynamic variable '", var,
                     "' matching pattern '", fname, "'."))
      return(NULL)
    }
    if (length(matches) > 1) {
      stop(paste0(
        "ERROR: Multiple raster files found for dynamic variable '", var,
        "' matching pattern '", fname, "'.",
        "\nMatching files:\n", paste(matches, collapse = "\n")
      ))
    }
    paths <- c(paths, matches)
  }

  for (var in static_vars) {
    fname   <- variable_patterns[[var]]
    matches <- all_files[grepl(fname, basename(all_files), ignore.case = TRUE)]
    if (length(matches) == 0) {
      warning(paste0("Missing raster file for static variable '", var,
                     "' matching pattern '", fname, "'."))
      return(NULL)
    }
    if (length(matches) > 1) {
      stop(paste0(
        "ERROR: Multiple raster files found for static variable '", var,
        "' matching pattern '", fname, "'.",
        "\nMatching files:\n", paste(matches, collapse = "\n")
      ))
    }
    paths <- c(paths, matches)
  }

  paths
}

#' @keywords internal
#' @noRd
.predict_hypervolume_raster <- function(model, rast_stack, fold, time_label) {
  dims     <- model@Dimensionality
  n_layers <- terra::nlyr(rast_stack)
  if (dims != n_layers) {
    warning(paste0("Fold ", fold, ": hypervolume dimensionality (", dims,
                   ") != raster layers (", n_layers, ") at ", time_label, " -- skipping."))
    return(NULL)
  }
  tryCatch({
    proj <- suppressMessages(suppressWarnings(hypervolume::hypervolume_project(
      model, rasters = rast_stack, type = "inclusion",
      fast.or.accurate = "fast", verbose = FALSE
    )))
    out <- rast_stack[[1]]
    v   <- as.integer(as.vector(proj))
    if (length(v) == terra::ncell(out)) {
      terra::values(out) <- v
    } else {
      terra::values(out) <- NA_integer_
    }
    out
  }, error = function(e) {
    warning(paste0("Fold ", fold, ": hypervolume projection failed at ",
                   time_label, ": ", e$message))
    NULL
  })
}

#' @keywords internal
#' @noRd
.bstart_match <- function(nms, pfx) {
  which(startsWith(nms, pfx) &
          (nchar(nms) == nchar(pfx) |
             !grepl("^[A-Za-z0-9]", substr(nms, nchar(pfx) + 1, nchar(pfx) + 1))))
}

#' @keywords internal
#' @noRd
.predict_parametric_raster <- function(model, rast_stack, model_vars, threshold,
                                       model_type, variable_patterns = NULL,
                                       time_cols = NULL, time_values = NULL) {
  cell_vals  <- as.data.frame(terra::values(rast_stack))
  rast_names <- names(cell_vals)

  name_map <- vapply(model_vars, function(v) {
    if (!is.null(variable_patterns) && v %in% names(variable_patterns) &&
        !is.null(time_cols) && !is.null(time_values)) {
      fname <- variable_patterns[[v]]
      for (tc in time_cols) {
        if (grepl(tc, fname, ignore.case = TRUE)) {
          fname <- gsub(tc, as.character(time_values[[tc]]), fname, ignore.case = TRUE)
        }
      }
      for (candidate in c(paste0(fname, "_Scaled"), fname)) {
        m <- which(rast_names == candidate)
        if (length(m) > 0) return(rast_names[m[1]])
      }
      m <- .bstart_match(rast_names, paste0(fname, "_Scaled"))
      if (length(m) > 0) return(rast_names[m[1]])
      m <- .bstart_match(rast_names, fname)
      if (length(m) > 0) return(rast_names[m[1]])
    }
    m <- which(rast_names == v)
    if (length(m) > 0) return(rast_names[m[1]])
    m <- .bstart_match(rast_names, v)
    if (length(m) > 0) return(rast_names[m[1]])
    NA_character_
  }, character(1))

  unmatched <- model_vars[is.na(name_map)]
  if (length(unmatched) > 0) {
    warning(paste0("Could not match model variables to raster layers: ",
                   paste(unmatched, collapse = ", "),
                   ". Available: ", paste(rast_names, collapse = ", ")))
    return(NULL)
  }

  pred_df        <- cell_vals[, name_map, drop = FALSE]
  names(pred_df) <- model_vars
  complete_idx   <- which(stats::complete.cases(pred_df))
  pred_vals      <- rep(NA_real_, nrow(pred_df))

  if (length(complete_idx) > 0) {
    pred_sub <- pred_df[complete_idx, , drop = FALSE]
    raw_pred <- tryCatch({
      if (model_type == "rf") {
        pred_sub$presence <- factor(0, levels = c(0, 1))
        stats::predict(model, newdata = pred_sub, type = "prob")[, "1"]
      } else {
        stats::predict(model, newdata = pred_sub, type = "response")
      }
    }, error = function(e) { warning(paste("Raster prediction failed:", e$message)); NULL })

    if (!is.null(raw_pred)) {
      pred_vals[complete_idx] <- as.integer(raw_pred >= threshold)
    }
  }

  out_rast <- rast_stack[[1]]
  terra::values(out_rast) <- pred_vals
  out_rast
}

#' @keywords internal
#' @noRd
.tseries <- function(x_all, y_all, folds_all, x_sum, y_sum,
                     all_folds, fold_colors, col_mean,
                     col_cbp_good = NULL, col_cbp_bad = NULL,
                     fold_alpha = 0.3, fold_lwd = 0.8,
                     sum_col = col_mean, sum_lwd = 2,
                     cbp_good_sum = NULL) {
  for (f in all_folds) {
    idx <- !is.na(folds_all) & folds_all == f
    if (sum(idx) == 0) next
    fc  <- grDevices::adjustcolor(fold_colors[as.character(f)], alpha.f = fold_alpha)
    xf  <- x_all[idx]
    yf  <- y_all[idx]
    v   <- !is.na(xf) & !is.na(yf)
    if (sum(v) > 0) {
      ord <- order(xf[v])
      graphics::lines(xf[v][ord], yf[v][ord], col = fc, lwd = fold_lwd)
      graphics::points(xf[v][ord], yf[v][ord], col = fc, pch = 16, cex = 0.5)
    }
  }
  v2 <- !is.na(x_sum) & !is.na(y_sum)
  if (sum(v2) > 0) {
    ord2 <- order(x_sum[v2])
    graphics::lines(x_sum[v2][ord2], y_sum[v2][ord2], col = sum_col, lwd = sum_lwd)
    if (!is.null(cbp_good_sum)) {
      pt_col <- ifelse(cbp_good_sum[v2], col_cbp_good, col_cbp_bad)
      graphics::points(x_sum[v2][ord2], y_sum[v2][ord2], col = pt_col, pch = 21, bg = "white",
                       cex = 1.1, lwd = 1.2)
    } else {
      graphics::points(x_sum[v2][ord2], y_sum[v2][ord2], col = sum_col, pch = 19, cex = 1.0)
    }
  }
}

#' @keywords internal
#' @noRd
.ref_lines <- function(lookup_df, val_col, all_folds, fold_colors,
                       lty = 2, lwd = 1.2, alpha = 0.65) {
  if (is.null(lookup_df) || !val_col %in% names(lookup_df)) return(invisible(NULL))
  for (f in all_folds) {
    rows <- lookup_df[!is.na(lookup_df$Fold) & lookup_df$Fold == f, ]
    if (nrow(rows) == 0) next
    val <- rows[[val_col]][1]
    if (is.na(val)) next
    graphics::abline(h = val,
                     col = grDevices::adjustcolor(fold_colors[as.character(f)], alpha.f = alpha),
                     lty = lty, lwd = lwd)
  }
}

#' @keywords internal
#' @noRd
.draw_xaxis <- function(x_tck, x_labels_attr) {
  if (!is.null(x_labels_attr)) {
    graphics::axis(1, at = x_tck,
                   labels = x_labels_attr[x_tck],
                   las = 2, cex.axis = 0.7)
  } else {
    graphics::axis(1, at = x_tck)
  }
}

#' @keywords internal
#' @noRd
.new_plot <- function(y_vals, y_ref_vals = NULL, title, ylab,
                      x_lim, x_axis_label, x_tck, x_labels_attr,
                      y_floor = NULL, y_ceil = NULL, right_mar = 12,
                      draw_fn = NULL) {
  all_y <- c(y_vals, y_ref_vals)
  y_rng <- range(all_y, na.rm = TRUE)
  y_pad <- diff(y_rng) * 0.08
  y_lo  <- if (!is.null(y_floor)) max(y_floor, y_rng[1] - y_pad) else y_rng[1] - y_pad
  y_hi  <- if (!is.null(y_ceil))  min(y_ceil,  y_rng[2] + y_pad) else y_rng[2] + y_pad
  if (is.na(y_lo) || is.na(y_hi) || y_lo >= y_hi) { y_lo <- 0; y_hi <- 1 }
  bot_mar <- if (!is.null(x_labels_attr)) {
    max(nchar(as.character(x_labels_attr[x_tck])), na.rm = TRUE) * 0.4 + 2.5
  } else {
    4
  }
  opar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(opar), add = TRUE)
  graphics::par(mar = c(bot_mar, 5, 3.5, right_mar), xpd = FALSE)
  if (!is.null(x_labels_attr)) {
    graphics::plot(NULL, xlim = x_lim, ylim = c(y_lo, y_hi),
                   xlab = "", ylab = ylab, main = title, las = 1, xaxt = "n")
    graphics::mtext(x_axis_label, side = 1, line = bot_mar - 1.2, cex = 0.9)
  } else {
    graphics::plot(NULL, xlim = x_lim, ylim = c(y_lo, y_hi),
                   xlab = x_axis_label, ylab = ylab, main = title, las = 1, xaxt = "n")
  }
  .draw_xaxis(x_tck, x_labels_attr)
  if (!is.null(draw_fn)) draw_fn()
}

#' @keywords internal
#' @noRd
.plot_facet_metric <- function(an, metric_col, ylab, title_prefix,
                               secondary_combos, primary_col, secondary_cols,
                               fold_colors, col_mean,
                               y_floor = NULL, y_ceil = NULL) {
  n_combos <- nrow(secondary_combos)

  all_y <- numeric(0)
  for (ci in seq_len(n_combos)) {
    combo <- secondary_combos[ci, , drop = FALSE]
    idx   <- rep(TRUE, nrow(an))
    for (sc in secondary_cols) idx <- idx & an[[sc]] == combo[[sc]]
    sub   <- an[idx, , drop = FALSE]
    if (metric_col %in% names(sub)) all_y <- c(all_y, sub[[metric_col]])
  }
  if (length(all_y) == 0 || all(is.na(all_y))) return(invisible(NULL))
  y_rng <- range(all_y, na.rm = TRUE)
  y_pad <- diff(y_rng) * 0.08
  y_lo  <- if (!is.null(y_floor)) max(y_floor, y_rng[1] - y_pad) else y_rng[1] - y_pad
  y_hi  <- if (!is.null(y_ceil))  min(y_ceil,  y_rng[2] + y_pad) else y_rng[2] + y_pad
  if (is.na(y_lo) || is.na(y_hi) || y_lo >= y_hi) { y_lo <- 0; y_hi <- 1 }

  opar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(opar), add = TRUE)
  graphics::par(mfrow = c(n_combos, 1), mar = c(4, 5, 2.5, 1), oma = c(0, 0, 2.5, 0))
  for (ci in seq_len(n_combos)) {
    combo       <- secondary_combos[ci, , drop = FALSE]
    combo_label <- paste(unlist(combo), collapse = "_")
    idx   <- rep(TRUE, nrow(an))
    for (sc in secondary_cols) idx <- idx & an[[sc]] == combo[[sc]]
    sub   <- an[idx, , drop = FALSE]
    if (nrow(sub) == 0 || !metric_col %in% names(sub)) next
    x_lim_f <- range(sub[[primary_col]], na.rm = TRUE)
    graphics::plot(NULL, xlim = x_lim_f, ylim = c(y_lo, y_hi),
                   xlab = primary_col, ylab = ylab,
                   main = combo_label, las = 1)
    for (fld in sort(unique(sub$Fold))) {
      fld_key <- as.character(fld)
      fd      <- sub[sub$Fold == fld, , drop = FALSE]
      fd      <- fd[order(fd[[primary_col]]), ]
      if (nrow(fd) == 0) next
      graphics::lines(fd[[primary_col]], fd[[metric_col]],
                      col = fold_colors[fld_key], lwd = 1.5)
      graphics::points(fd[[primary_col]], fd[[metric_col]],
                       col = fold_colors[fld_key], pch = 19, cex = 0.6)
    }
    mn_vals <- tapply(sub[[metric_col]], sub[[primary_col]], mean, na.rm = TRUE)
    graphics::lines(as.numeric(names(mn_vals)), as.numeric(mn_vals),
                    col = col_mean, lwd = 2.5, lty = 1)
  }
  graphics::mtext(paste(title_prefix, "by", paste(secondary_cols, collapse = "/")),
                  side = 3, outer = TRUE, line = 1, cex = 1, font = 2)
}

#' @keywords internal
#' @noRd
.line_legend <- function(lab, col, lty, lwd) {
  usr <- graphics::par("usr")
  x_leg <- usr[2] + diff(usr[1:2]) * 0.03
  y_leg <- usr[4]
  old_xpd <- graphics::par(xpd = NA)
  on.exit(graphics::par(xpd = old_xpd$xpd), add = TRUE)
  graphics::legend(
    x         = x_leg,
    y         = y_leg,
    legend    = lab,
    col       = col,
    lty       = lty,
    lwd       = lwd,
    bty       = "n",
    bg        = "transparent",
    cex       = 0.82,
    seg.len   = 1.2,
    xpd       = NA,
    yjust     = 1,
    x.intersp = 0.5
  )
}

#' @keywords internal
#' @noRd
.bar_legend <- function(bar_labs, bar_cols, pt_lab, pt_col) {
  usr   <- graphics::par("usr")
  x_leg <- usr[2] + diff(usr[1:2]) * 0.03
  y_leg <- usr[4]
  old_xpd <- graphics::par(xpd = NA)
  on.exit(graphics::par(xpd = old_xpd$xpd), add = TRUE)

  graphics::legend(
    x         = x_leg,
    y         = y_leg,
    legend    = bar_labs,
    fill      = bar_cols,
    border    = "gray30",
    bty       = "n",
    bg        = "transparent",
    cex       = 0.82,
    xpd       = NA,
    yjust     = 1,
    x.intersp = 0.5
  )

  n_bar  <- length(bar_labs)
  lh_est <- diff(usr[3:4]) * 0.065
  y_pt   <- y_leg - n_bar * lh_est
  graphics::legend(
    x         = x_leg,
    y         = y_pt,
    legend    = pt_lab,
    col       = "gray20",
    pch       = 21,
    pt.bg     = pt_col,
    bty       = "n",
    bg        = "transparent",
    cex       = 0.82,
    xpd       = NA,
    yjust     = 1,
    x.intersp = 0.5
  )
}

#' @keywords internal
#' @noRd
.format_time_estimate <- function(seconds) {
  hours   <- seconds / 3600
  minutes <- seconds / 60
  if (hours > 1) {
    paste0(round(hours,   1), " hours")
  } else if (minutes > 1) {
    paste0(round(minutes, 1), " minutes")
  } else {
    paste0(round(seconds, 1), " seconds")
  }
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 5,
  dpi       = 100,
  out.width = "95%"
)

## -----------------------------------------------------------------------------
library(TemporalModelR)
library(terra)
library(sf)

raw_dir <- system.file("extdata/rasters_raw",
                       package = "TemporalModelR")

## ----fig.height=3-------------------------------------------------------------
elev <- rast(file.path(raw_dir, "elevation.tif"))

plot(elev, main = "Elevation (m)")

## ----fig.height=14------------------------------------------------------------
years_to_plot <- seq(1, 15, by = 2)

forest_files  <- file.path(raw_dir,
                           paste0("forest_cover_", years_to_plot, ".tif"))
pr_ann_files  <- file.path(raw_dir,
                           paste0("pr_ann_",      years_to_plot, ".tif"))

### Interleave forest and precip so each row of the plot grid is one year
forest_pr_paths        <- c(rbind(forest_files, pr_ann_files))
forest_pr_stack        <- rast(forest_pr_paths)
names(forest_pr_stack) <- c(rbind(paste("Forest_yr", years_to_plot),
                                  paste("Pr_ann_yr", years_to_plot)))

plot(forest_pr_stack, nc = 2)

## ----fig.height=6-------------------------------------------------------------
season_names <- c("Spring", "Summer", "Autumn", "Winter")

prseas_y1_stack <- rast(file.path(raw_dir,
                                  paste0("prseas_1_",
                                         season_names, ".tif")))

names(prseas_y1_stack) <- season_names

plot(prseas_y1_stack,
     range = c(0, max(values(prseas_y1_stack), na.rm = TRUE)))

## -----------------------------------------------------------------------------
pts_file <- system.file("extdata/points/synthetic_occurrence_points.csv",
                        package = "TemporalModelR")
pts <- utils::read.csv(pts_file)

head(pts)


nrow(pts)


table(pts$year, pts$season)

## ----fig.height=20------------------------------------------------------------
seasons <- c("Spring", "Summer", "Autumn")
study_extent <- ext(0, 3000, 0, 1500)

opar <- par(no.readonly = TRUE)

par(mfrow = c(15, 3),
    mar   = c(1.5, 1.5, 1.5, 0.5),
    oma   = c(2, 2, 2, 1))

for (yr in 1:15) {
  for (sea in seasons) {
    sub <- pts[pts$year == yr & pts$season == sea, ]

    plot(NULL,
         xlim = c(0, 3000), ylim = c(0, 1500),
         asp  = 1, xaxt = "n", yaxt = "n",
         xlab = "", ylab = "",
         main = paste0("Year ", yr, " - ", sea),
         cex.main = 0.9)

    rect(0, 0, 3000, 1500, border = "grey70")

    if (nrow(sub) > 0) {
      points(sub$x, sub$y, pch = 19, cex = 0.7, col = "darkblue")
    }
  }
}

par(opar)

#' Visualize Model Assessment Metrics Across Time
#'
#' Generates diagnostic plots from the per-timestep assessment table produced
#' by \code{\link{generate_spatiotemporal_predictions}}, optionally overlaying
#' overall reference values from the model result object.
#'
#' @usage
#' plot_model_assessment(predictions, time_column,
#'                       secondary_time_mode = "combine", model_result = NULL,
#'                       cbp_threshold = 0.05, plot_palette = "Dark 2",
#'                       verbose = TRUE)
#'
#' @param predictions List returned by
#'   \code{\link{generate_spatiotemporal_predictions}}, or a named list with
#'   at least a \code{timestep_metrics} element. The \code{timestep_metrics}
#'   element may also be a path to a \code{Timestep_Assessment_Metrics.csv}
#'   file produced by \code{\link{generate_spatiotemporal_predictions}}.
#' @param time_column Character. Name of the primary time column in
#'   \code{timestep_metrics} to use as the x axis (e.g. \code{"year"}).
#'   When predictions span multiple time columns (e.g. \code{"year"} and
#'   \code{"season"}), provide all relevant column names as a character
#'   vector and control how secondary columns are handled via
#'   \code{secondary_time_mode}.
#' @param secondary_time_mode Character. How to handle secondary time columns
#'   when \code{time_column} has length > 1. One of:
#'   \itemize{
#'     \item \code{"combine"} (default): secondary time values are appended to
#'       the primary value to form a single ordered x-axis label
#'       (e.g. \code{1_Spring}, \code{1_Summer}, \code{2_Spring}, ...).
#'     \item \code{"facet"}: a separate plot is produced for each unique
#'       combination of secondary time values, with the primary time column
#'       as the x axis on every panel.
#'   }
#' @param model_result List or character. Optional. Output from a
#'   \code{build_temporal_*()} function or path to its \code{.rds} file. When
#'   supplied, overall sensitivity and specificity from
#'   \code{model_result$fold_test_metrics} are added as per-fold reference
#'   lines. Default is \code{NULL}.
#' @param cbp_threshold Numeric. Significance threshold for CBP. Default is
#'   \code{0.05}.
#' @param plot_palette Character. Name of an HCL or RColorBrewer palette used
#'   to color folds in diagnostic plots. Accepts any HCL palette name (see
#'   \code{\link[grDevices]{hcl.pals}}) or, if \pkg{RColorBrewer} is installed,
#'   any Brewer palette name. Default is \code{"Dark 2"}.
#' @param verbose Logical. If \code{TRUE} (default), prints progress
#'   messages during processing.
#'
#' @return Invisibly returns a named list containing:
#' \itemize{
#'   \item \code{pct_suitable}: Recorded plot of proportion of study area
#'     predicted suitable per time step.
#'   \item \code{sensitivity}: Recorded plot of per-timestep sensitivity.
#'   \item \code{specificity}: Recorded plot of per-timestep specificity
#'     (only present when pseudoabsence data were used).
#'   \item \code{cbp}: Recorded plot of cumulative binomial probability per
#'     time step on a log scale.
#'   \item \code{tp_fn}: Recorded plot of true positives and false negatives
#'     per time step.
#'   \item \code{tn_fp}: Recorded plot of true negatives and false positives
#'     per time step (only present when pseudoabsence data were used).
#'   \item \code{timestep_summary}: Data frame of per-time-step cross-fold
#'     mean and SD for each metric.
#'   \item \code{overall_summary}: Data frame from
#'     \code{predictions$overall_summary}, when present.
#' }
#'
#'@details
#' Plots per-fold and per-timestep diagnostic plots for data produced by
#' \code{\link{generate_spatiotemporal_predictions}}. These quick visuals can
#' be used by users to assess model performance and significance and decide
#' if the model's performance warrants further interpretation of the results
#' through post-processing analyses.
#'
#' @seealso
#' Preprocessing: \code{\link{spatiotemporal_partition}},
#'   \code{\link{generate_absences}}
#'
#' Modeling: \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
#'   \code{\link{build_temporal_rf}}, \code{\link{build_temporal_hv}},
#'
#' Post-processing: \code{\link{generate_spatiotemporal_predictions}},
#'   \code{\link{summarize_raster_outputs}}
#'
#' @examples
#' data(tmr_predictions, package = "TemporalModelR")
#'
#' plot_model_assessment(
#'   predictions         = tmr_predictions,
#'   time_column         = c("year", "season"),
#'   secondary_time_mode = "combine",
#'   verbose             = FALSE
#' )

#' @export
#' @importFrom graphics abline axis legend lines mtext par plot points rect
#' @importFrom grDevices adjustcolor recordPlot
#' @importFrom stats sd
#' @importFrom utils capture.output read.csv
plot_model_assessment <- function(predictions,
                                  time_column,
                                  secondary_time_mode = "combine",
                                  model_result  = NULL,
                                  cbp_threshold = 0.05,
                                  plot_palette  = "Dark 2",
                                  verbose       = TRUE) {

  if (missing(predictions)) {
    stop(paste0("ERROR: 'predictions' is required. Please provide the list ",
                "output from generate_spatiotemporal_predictions()."))
  }
  if (missing(time_column) || !is.character(time_column) || length(time_column) == 0) {
    stop(paste0("ERROR: 'time_column' must be a character string (or vector of strings) ",
                "naming the time column(s) in timestep_metrics."))
  }

  if (is.character(predictions) && length(predictions) == 1) {
    if (file.exists(predictions)) {
      if (verbose) message("Reading timestep metrics from CSV: ", basename(predictions))
      an <- tryCatch(
        utils::read.csv(predictions, stringsAsFactors = FALSE),
        error = function(e) stop(paste0("ERROR reading 'predictions' CSV: ", e$message))
      )
      ov_sum <- NULL
    } else {
      stop(paste0("ERROR: 'predictions' CSV file not found: ", predictions))
    }
  } else if (is.data.frame(predictions)) {
    an <- predictions; ov_sum <- NULL
  } else if (is.list(predictions)) {
    if (!"timestep_metrics" %in% names(predictions))
      stop("ERROR: 'predictions' must contain 'timestep_metrics'.")
    tm <- predictions$timestep_metrics
    if (is.data.frame(tm)) {
      an <- tm
    } else if (is.character(tm) && file.exists(tm)) {
      an <- tryCatch(utils::read.csv(tm, stringsAsFactors = FALSE),
                     error = function(e) stop(paste0("ERROR reading 'timestep_metrics': ", e$message)))
    } else {
      stop(paste0("ERROR: 'timestep_metrics' file not found: ", tm))
    }
    ov_sum <- if ("overall_summary" %in% names(predictions) &&
                  !is.null(predictions$overall_summary) &&
                  nrow(predictions$overall_summary) > 0)
      predictions$overall_summary else NULL
  } else stop("ERROR: 'predictions' must be the list from generate_spatiotemporal_predictions(), a data frame, or a path to a Timestep_Assessment_Metrics CSV.")

  if (nrow(an) == 0) stop("ERROR: timestep_metrics is empty.")

  an_aliases <- list(
    Pct_Suitable = c("Pct_Suitable", "G_Volume"),
    TP           = c("TP",  "TP_G"),
    FN           = c("FN",  "FN_G"),
    TN           = c("TN",  "TN_G"),
    FP           = c("FP",  "FP_G"),
    Sensitivity  = c("Sensitivity",  "Sensitivity_G"),
    Specificity  = c("Specificity",  "Specificity_G"),
    CBP          = c("CBP", "CBP_G")
  )
  for (canon in names(an_aliases)) {
    if (!canon %in% names(an)) {
      for (alias in an_aliases[[canon]]) {
        if (alias %in% names(an)) { an[[canon]] <- an[[alias]]; break }
      }
    }
  }

  missing_cols <- setdiff(time_column, names(an))
  if (length(missing_cols) > 0)
    stop(paste0("ERROR: time_column value(s) not found in metrics: ",
                paste(missing_cols, collapse = ", "),
                ". Available columns: ", paste(names(an), collapse = ", ")))

  secondary_time_mode <- match.arg(secondary_time_mode, c("combine", "facet"))

  plot_time_col  <- time_column[1]
  secondary_cols <- if (length(time_column) > 1) time_column[-1] else character(0)

  secondary_combos <- if (length(secondary_cols) > 0) {
    unique(an[, secondary_cols, drop = FALSE])
  } else {
    NULL
  }

  if (length(secondary_cols) > 0 && secondary_time_mode == "combine") {

    an$..time_label.. <- apply(
      an[, time_column, drop = FALSE], 1,
      function(r) paste(r, collapse = "_")
    )
    all_labels  <- unique(an$..time_label..)
    label_order <- order(
      as.numeric(an[match(all_labels, an$..time_label..), plot_time_col]),
      as.character(an[match(all_labels, an$..time_label..), secondary_cols[1]])
    )
    an$..x_val..  <- match(an$..time_label.., all_labels[label_order])
    x_labels      <- all_labels[label_order]
    plot_time_col <- "..x_val.."
    secondary_cols <- character(0)
    attr(an, "x_labels") <- x_labels
  }

  has_spec <- "Specificity" %in% names(an) && !all(is.na(an$Specificity))
  has_tnfp <- all(c("TN","FP") %in% names(an)) &&
    !all(is.na(an$TN)) && !all(is.na(an$FP))

  fold_test <- NULL
  if (!is.null(model_result)) {
    if (is.character(model_result) && file.exists(model_result))
      model_result <- tryCatch(readRDS(model_result),
                               error = function(e) { warning(paste("Could not read model_result:", e$message)); NULL })
    if (is.list(model_result) && "fold_test_metrics" %in% names(model_result) &&
        nrow(model_result$fold_test_metrics) > 0) {
      fold_test <- model_result$fold_test_metrics
      if (verbose) message(paste("Loaded fold_test_metrics for", nrow(fold_test), "fold(s)."))
    }
  }

  all_folds   <- sort(unique(an$Fold))
  fold_colors <- .resolve_palette(length(all_folds), plot_palette)
  names(fold_colors) <- as.character(all_folds)

  col_tp <- "#0072B2"
  col_fn <- "#D55E00"
  col_tn <- "#009E73"
  col_fp <- "#CC79A7"
  col_cbp_good <- col_tp
  col_cbp_bad  <- col_fn
  col_mean <- "black"

  opar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(opar), add = TRUE)

  if (secondary_time_mode == "facet" && !is.null(secondary_combos) &&
      nrow(secondary_combos) > 0) {
    orig_secondary_cols <- time_column[-1]
    primary_col         <- time_column[1]

    facet_results <- list()

    if ("Pct_Suitable" %in% names(an)) {
      .plot_facet_metric(an, "Pct_Suitable", "Prop. Suitable", "G-Space Suitability",
                         secondary_combos, primary_col, orig_secondary_cols,
                         fold_colors, col_mean, y_floor = 0, y_ceil = 1)
      facet_results$pct_suitable <- grDevices::recordPlot()
    }
    if ("Sensitivity" %in% names(an)) {
      .plot_facet_metric(an, "Sensitivity", "Sensitivity", "Sensitivity",
                         secondary_combos, primary_col, orig_secondary_cols,
                         fold_colors, col_mean, y_floor = 0, y_ceil = 1)
      facet_results$sensitivity <- grDevices::recordPlot()
    }
    if (has_spec && "Specificity" %in% names(an)) {
      .plot_facet_metric(an, "Specificity", "Specificity", "Specificity",
                         secondary_combos, primary_col, orig_secondary_cols,
                         fold_colors, col_mean, y_floor = 0, y_ceil = 1)
      facet_results$specificity <- grDevices::recordPlot()
    }
    if ("CBP" %in% names(an)) {
      .plot_facet_metric(an, "CBP", "CBP", "CBP",
                         secondary_combos, primary_col, orig_secondary_cols,
                         fold_colors, col_mean)
      facet_results$cbp <- grDevices::recordPlot()
    }
    facet_results$timestep_summary <- data.frame()
    n_plots <- length(facet_results) - 1
    if (verbose) message(paste0("Facet mode: produced ", n_plots, " stacked plot(s) across ",
                                nrow(secondary_combos), " ",
                                paste(orig_secondary_cols, collapse = "/"),
                                " value(s)."))
    return(invisible(facet_results))
  }

  agg_cols <- c("Pct_Suitable","Sensitivity","CBP","TP","FN")
  if (has_spec) agg_cols <- c(agg_cols, "Specificity")
  if (has_tnfp) agg_cols <- c(agg_cols, "TN","FP")
  agg_cols <- agg_cols[agg_cols %in% names(an)]

  time_vals <- sort(unique(an[[plot_time_col]]))
  an_summary_rows <- lapply(time_vals, function(tv) {
    grp <- an[!is.na(an[[plot_time_col]]) & an[[plot_time_col]] == tv, , drop = FALSE]
    row <- data.frame(tv, stringsAsFactors = FALSE); names(row) <- plot_time_col
    for (col in agg_cols) {
      row[[paste0("Mean_", col)]] <- mean(grp[[col]], na.rm = TRUE)
      row[[paste0("SD_",   col)]] <- stats::sd(grp[[col]], na.rm = TRUE)
    }
    row
  })
  an_summary <- do.call(rbind.data.frame, an_summary_rows)
  rownames(an_summary) <- NULL
  an_summary$CBP_good <- if ("Mean_CBP" %in% names(an_summary))
    !is.na(an_summary$Mean_CBP) & an_summary$Mean_CBP < cbp_threshold
  else rep(FALSE, nrow(an_summary))

  t_all <- an[[plot_time_col]]
  t_sum <- an_summary[[plot_time_col]]
  x_lim <- range(t_all, na.rm = TRUE)

  x_labels_attr <- attr(an, "x_labels")

  t_uniq <- sort(unique(t_all[!is.na(t_all)]))
  x_tck  <- if (length(t_uniq) <= 15) {
    t_uniq
  } else {
    t_uniq[seq(1, length(t_uniq), by = ceiling(length(t_uniq) / 15))]
  }

  x_axis_label <- if (!is.null(x_labels_attr)) {
    paste(time_column, collapse = " / ")
  } else {
    plot_time_col
  }

  plot_list <- list()

  if ("Pct_Suitable" %in% names(an) && "Mean_Pct_Suitable" %in% names(an_summary)) {

    ref_vals <- if (!is.null(ov_sum) && "Mean_Pct_Suitable" %in% names(ov_sum))
      ov_sum$Mean_Pct_Suitable else NULL

    leg_lab <- "Cross-fold mean"
    leg_col <- col_mean
    leg_lty <- 1
    leg_lwd <- 3
    if (!is.null(ov_sum) && "Mean_Pct_Suitable" %in% names(ov_sum)) {
      leg_lab <- c(leg_lab, "Overall mean (per fold)")
      leg_col <- c(leg_col, "gray50")
      leg_lty <- c(leg_lty, 2)
      leg_lwd <- c(leg_lwd, 1.5)
    }

    .new_plot(an$Pct_Suitable, ref_vals,
              title = "G-Space Predictions per Time Step",
              ylab  = "Proportion of Study Area Predicted Suitable",
              x_lim = x_lim, x_axis_label = x_axis_label,
              x_tck = x_tck, x_labels_attr = x_labels_attr,
              y_floor = 0, y_ceil = 1,
              draw_fn = function() {
                .ref_lines(ov_sum, "Mean_Pct_Suitable", all_folds, fold_colors)
                .tseries(t_all, an$Pct_Suitable, an$Fold, t_sum, an_summary$Mean_Pct_Suitable,
                         all_folds, fold_colors, col_mean)
                .line_legend(leg_lab, leg_col, leg_lty, leg_lwd)
              })
    plot_list$pct_suitable <- grDevices::recordPlot()
  }

  if ("Sensitivity" %in% names(an) && "Mean_Sensitivity" %in% names(an_summary)) {

    ref_vals <- c(
      if (!is.null(ov_sum)    && "Overall_Sensitivity" %in% names(ov_sum))    ov_sum$Overall_Sensitivity,
      if (!is.null(fold_test) && "Sensitivity"          %in% names(fold_test)) fold_test$Sensitivity
    )

    leg_lab <- "Cross-fold mean"
    leg_col <- col_mean
    leg_lty <- 1
    leg_lwd <- 3
    if (!is.null(ov_sum) && "Overall_Sensitivity" %in% names(ov_sum)) {
      leg_lab <- c(leg_lab, "Overall sensitivity (per fold)")
      leg_col <- c(leg_col, "gray50")
      leg_lty <- c(leg_lty, 2)
      leg_lwd <- c(leg_lwd, 1.5)
    }
    if (!is.null(fold_test) && "Sensitivity" %in% names(fold_test)) {
      leg_lab <- c(leg_lab, "Model sensitivity incl. pseudoabsences (per fold)")
      leg_col <- c(leg_col, "gray50")
      leg_lty <- c(leg_lty, 3)
      leg_lwd <- c(leg_lwd, 1.5)
    }

    .new_plot(an$Sensitivity, ref_vals,
              title = "Sensitivity per Time Step", ylab = "Sensitivity",
              x_lim = x_lim, x_axis_label = x_axis_label,
              x_tck = x_tck, x_labels_attr = x_labels_attr,
              y_floor = 0, y_ceil = 1,
              draw_fn = function() {
                .ref_lines(ov_sum,    "Overall_Sensitivity", all_folds, fold_colors, lty = 2, lwd = 1.2)
                .ref_lines(fold_test, "Sensitivity",          all_folds, fold_colors, lty = 3, lwd = 1.5)
                .tseries(t_all, an$Sensitivity, an$Fold, t_sum, an_summary$Mean_Sensitivity,
                         all_folds, fold_colors, col_mean)
                .line_legend(leg_lab, leg_col, leg_lty, leg_lwd)
              })
    plot_list$sensitivity <- grDevices::recordPlot()
  }

  if (has_spec && "Mean_Specificity" %in% names(an_summary)) {

    ref_vals <- c(
      if (!is.null(ov_sum)    && "Overall_Specificity" %in% names(ov_sum))    ov_sum$Overall_Specificity,
      if (!is.null(fold_test) && "Specificity"          %in% names(fold_test)) fold_test$Specificity
    )

    leg_lab <- "Cross-fold mean"
    leg_col <- col_mean
    leg_lty <- 1
    leg_lwd <- 3
    if (!is.null(ov_sum) && "Overall_Specificity" %in% names(ov_sum)) {
      leg_lab <- c(leg_lab, "Overall specificity (per fold)")
      leg_col <- c(leg_col, "gray50")
      leg_lty <- c(leg_lty, 2)
      leg_lwd <- c(leg_lwd, 1.5)
    }
    if (!is.null(fold_test) && "Specificity" %in% names(fold_test)) {
      leg_lab <- c(leg_lab, "Model specificity incl. pseudoabsences (per fold)")
      leg_col <- c(leg_col, "gray50")
      leg_lty <- c(leg_lty, 3)
      leg_lwd <- c(leg_lwd, 1.5)
    }

    .new_plot(an$Specificity, ref_vals,
              title = "Specificity per Time Step", ylab = "Specificity",
              x_lim = x_lim, x_axis_label = x_axis_label,
              x_tck = x_tck, x_labels_attr = x_labels_attr,
              y_floor = 0, y_ceil = 1,
              draw_fn = function() {
                .ref_lines(ov_sum,    "Overall_Specificity", all_folds, fold_colors, lty = 2, lwd = 1.2)
                .ref_lines(fold_test, "Specificity",          all_folds, fold_colors, lty = 3, lwd = 1.5)
                .tseries(t_all, an$Specificity, an$Fold, t_sum, an_summary$Mean_Specificity,
                         all_folds, fold_colors, col_mean)
                .line_legend(leg_lab, leg_col, leg_lty, leg_lwd)
              })
    plot_list$specificity <- grDevices::recordPlot()
  }

  if ("CBP" %in% names(an) && "Mean_CBP" %in% names(an_summary)) {

    all_cbp <- c(an$CBP,
                 if (!is.null(ov_sum) && "Overall_CBP" %in% names(ov_sum)) ov_sum$Overall_CBP)
    pos_cbp <- all_cbp[!is.na(all_cbp) & all_cbp > 0]

    if (length(pos_cbp) > 0) {
      lo   <- min(pos_cbp); hi <- max(pos_cbp)
      lim  <- c(min(lo, cbp_threshold) * 0.8, max(hi, cbp_threshold) * 1.5)
      lg   <- floor(log10(lim[1])):ceiling(log10(lim[2]))
      brks <- 10^lg

      bot_mar <- if (!is.null(x_labels_attr)) {
        max(nchar(as.character(x_labels_attr[x_tck])), na.rm = TRUE) * 0.4 + 2.5
      } else {
        4
      }
      opar_cbp <- graphics::par(no.readonly = TRUE)
      on.exit(graphics::par(opar_cbp), add = TRUE)
      graphics::par(mar = c(bot_mar, 5, 3.5, 12), xpd = FALSE)
      if (!is.null(x_labels_attr)) {
        graphics::plot(NULL, xlim = x_lim, ylim = lim, log = "y",
                       xlab = "", ylab = "CBP (log scale)",
                       main = "Cumulative Binomial Probability per Time Step",
                       las = 1, xaxt = "n", yaxt = "n")
        graphics::mtext(x_axis_label, side = 1, line = bot_mar - 1.2, cex = 0.9)
      } else {
        graphics::plot(NULL, xlim = x_lim, ylim = lim, log = "y",
                       xlab = x_axis_label, ylab = "CBP (log scale)",
                       main = "Cumulative Binomial Probability per Time Step",
                       las = 1, xaxt = "n", yaxt = "n")
      }
      .draw_xaxis(x_tck, x_labels_attr)

      log_brks <- brks[brks > 0]
      log_labs <- vapply(log_brks, function(b) {
        if (b >= 0.001) formatC(b, format = "f", digits = max(0, -floor(log10(b))))
        else            formatC(b, format = "e", digits = 0)
      }, character(1))
      graphics::axis(2, at = log_brks, labels = log_labs, las = 1, cex.axis = 0.85)

      graphics::abline(h = cbp_threshold, lty = 1, col = "black", lwd = 1.8)

      .ref_lines(ov_sum, "Overall_CBP", all_folds, fold_colors, lty = 4, lwd = 1.0)
      .tseries(t_all, an$CBP, an$Fold, t_sum, an_summary$Mean_CBP,
               all_folds, fold_colors, col_mean,
               col_cbp_good = col_cbp_good, col_cbp_bad = col_cbp_bad,
               cbp_good_sum = an_summary$CBP_good)

      leg_lab <- c(paste0("p < ", cbp_threshold), paste0("p >= ", cbp_threshold), "Threshold")
      leg_col <- c(col_cbp_good, col_cbp_bad, "black")
      leg_lty <- c(1, 1, 1)
      leg_lwd <- c(3, 3, 1.8)
      if (!is.null(ov_sum) && "Overall_CBP" %in% names(ov_sum)) {
        leg_lab <- c(leg_lab, "Overall CBP (per fold)")
        leg_col <- c(leg_col, "gray50")
        leg_lty <- c(leg_lty, 4)
        leg_lwd <- c(leg_lwd, 1.2)
      }
      .line_legend(leg_lab, leg_col, leg_lty, leg_lwd)
      plot_list$cbp <- grDevices::recordPlot()
    }
  }

  if (all(c("TP","FN","Mean_TP","Mean_FN") %in% c(names(an), names(an_summary)))) {

    tp_all <- an$TP; fn_all <- an$FN
    tp_sum <- an_summary$Mean_TP; fn_sum <- an_summary$Mean_FN
    valid  <- !is.na(tp_all) & !is.na(fn_all) & (tp_all > 0 | fn_all > 0)

    if (any(valid)) {
      y_max <- max(c(tp_all[valid], tp_sum), na.rm = TRUE) * 1.15
      y_min <- -max(c(fn_all[valid], fn_sum), na.rm = TRUE) * 1.15

      bot_mar <- if (!is.null(x_labels_attr)) {
        max(nchar(as.character(x_labels_attr[x_tck])), na.rm = TRUE) * 0.4 + 2.5
      } else {
        4
      }
      opar_tp <- graphics::par(no.readonly = TRUE)
      on.exit(graphics::par(opar_tp), add = TRUE)
      graphics::par(mar = c(bot_mar, 5, 3.5, 12), xpd = FALSE)
      if (!is.null(x_labels_attr)) {
        graphics::plot(NULL, xlim = x_lim, ylim = c(y_min, y_max),
                       xlab = "", ylab = "Count  (TP above 0 | FN below 0)",
                       main = "True Positives and False Negatives per Time Step",
                       las = 1, xaxt = "n")
        graphics::mtext(x_axis_label, side = 1, line = bot_mar - 1.2, cex = 0.9)
      } else {
        graphics::plot(NULL, xlim = x_lim, ylim = c(y_min, y_max),
                       xlab = x_axis_label, ylab = "Count  (TP above 0 | FN below 0)",
                       main = "True Positives and False Negatives per Time Step",
                       las = 1, xaxt = "n")
      }
      .draw_xaxis(x_tck, x_labels_attr)
      graphics::abline(h = 0, col = "gray40", lwd = 0.8)

      t_uniq <- sort(unique(t_sum))
      bw <- if (length(t_uniq) > 1) diff(range(t_uniq)) / (length(t_uniq) * 2.2) else 0.4

      for (ti in seq_along(t_sum)) {
        tx <- t_sum[ti]
        if (!is.na(tp_sum[ti])) graphics::rect(tx - bw, 0, tx + bw,  tp_sum[ti],
                                               col = grDevices::adjustcolor(col_tp, 0.7), border = "gray30", lwd = 0.4)
        if (!is.na(fn_sum[ti])) graphics::rect(tx - bw, 0, tx + bw, -fn_sum[ti],
                                               col = grDevices::adjustcolor(col_fn, 0.7), border = "gray30", lwd = 0.4)
      }

      pt_tp <- grDevices::adjustcolor(col_tp, 0.7)
      pt_fn <- grDevices::adjustcolor(col_fn, 0.7)
      graphics::points(t_all[valid],  tp_all[valid], pch = 21, bg = pt_tp, cex = 0.9, col = "gray30")
      graphics::points(t_all[valid], -fn_all[valid], pch = 21, bg = pt_fn, cex = 0.9, col = "gray30")

      .bar_legend(
        bar_labs = c("Mean TP", "Mean FN"),
        bar_cols = c(pt_tp, pt_fn),
        pt_lab   = "Individual folds",
        pt_col   = grDevices::adjustcolor("gray30", 0.6)
      )

      plot_list$tp_fn <- grDevices::recordPlot()
    }
  }

  if (has_tnfp && all(c("Mean_TN","Mean_FP") %in% names(an_summary))) {

    tn_all <- an$TN; fp_all <- an$FP
    tn_sum <- an_summary$Mean_TN; fp_sum <- an_summary$Mean_FP
    valid  <- !is.na(tn_all) & !is.na(fp_all) & (tn_all > 0 | fp_all > 0)

    if (any(valid)) {
      y_max <- max(c(tn_all[valid], tn_sum), na.rm = TRUE) * 1.15
      y_min <- -max(c(fp_all[valid], fp_sum), na.rm = TRUE) * 1.15

      bot_mar <- if (!is.null(x_labels_attr)) {
        max(nchar(as.character(x_labels_attr[x_tck])), na.rm = TRUE) * 0.4 + 2.5
      } else {
        4
      }
      opar_tn <- graphics::par(no.readonly = TRUE)
      on.exit(graphics::par(opar_tn), add = TRUE)
      graphics::par(mar = c(bot_mar, 5, 3.5, 12), xpd = FALSE)
      if (!is.null(x_labels_attr)) {
        graphics::plot(NULL, xlim = x_lim, ylim = c(y_min, y_max),
                       xlab = "", ylab = "Count  (TN above 0 | FP below 0)",
                       main = "True Negatives and False Positives per Time Step",
                       las = 1, xaxt = "n")
        graphics::mtext(x_axis_label, side = 1, line = bot_mar - 1.2, cex = 0.9)
      } else {
        graphics::plot(NULL, xlim = x_lim, ylim = c(y_min, y_max),
                       xlab = x_axis_label, ylab = "Count  (TN above 0 | FP below 0)",
                       main = "True Negatives and False Positives per Time Step",
                       las = 1, xaxt = "n")
      }
      .draw_xaxis(x_tck, x_labels_attr)
      graphics::abline(h = 0, col = "gray40", lwd = 0.8)

      t_uniq <- sort(unique(t_sum))
      bw <- if (length(t_uniq) > 1) diff(range(t_uniq)) / (length(t_uniq) * 2.2) else 0.4

      for (ti in seq_along(t_sum)) {
        tx <- t_sum[ti]
        if (!is.na(tn_sum[ti])) graphics::rect(tx - bw, 0, tx + bw,  tn_sum[ti],
                                               col = grDevices::adjustcolor(col_tn, 0.7), border = "gray30", lwd = 0.4)
        if (!is.na(fp_sum[ti])) graphics::rect(tx - bw, 0, tx + bw, -fp_sum[ti],
                                               col = grDevices::adjustcolor(col_fp, 0.7), border = "gray30", lwd = 0.4)
      }

      pt_tn <- grDevices::adjustcolor(col_tn, 0.7)
      pt_fp <- grDevices::adjustcolor(col_fp, 0.7)
      graphics::points(t_all[valid],  tn_all[valid], pch = 21, bg = pt_tn, cex = 0.9, col = "gray30")
      graphics::points(t_all[valid], -fp_all[valid], pch = 21, bg = pt_fp, cex = 0.9, col = "gray30")

      .bar_legend(
        bar_labs = c("Mean TN", "Mean FP"),
        bar_cols = c(pt_tn, pt_fp),
        pt_lab   = "Individual folds",
        pt_col   = grDevices::adjustcolor("gray30", 0.6)
      )

      plot_list$tn_fp <- grDevices::recordPlot()
    }
  }

  summary_rows <- list()
  if ("Pct_Suitable" %in% names(an)) {
    summary_rows[["Pct_Suitable"]] <- data.frame(
      Metric = "Pct_Suitable",
      Mean   = round(mean(an$Pct_Suitable, na.rm = TRUE), 4),
      SD     = round(stats::sd(an$Pct_Suitable, na.rm = TRUE), 4),
      stringsAsFactors = FALSE
    )
  }
  if ("Sensitivity" %in% names(an)) {
    summary_rows[["Sensitivity"]] <- data.frame(
      Metric = "Sensitivity",
      Mean   = round(mean(an$Sensitivity, na.rm = TRUE), 4),
      SD     = round(stats::sd(an$Sensitivity, na.rm = TRUE), 4),
      stringsAsFactors = FALSE
    )
  }
  if (has_spec) {
    summary_rows[["Specificity"]] <- data.frame(
      Metric = "Specificity",
      Mean   = round(mean(an$Specificity, na.rm = TRUE), 4),
      SD     = round(stats::sd(an$Specificity, na.rm = TRUE), 4),
      stringsAsFactors = FALSE
    )
  }
  if ("CBP" %in% names(an)) {
    pct_good <- round(mean(an$CBP < cbp_threshold, na.rm = TRUE) * 100, 1)
    summary_rows[["CBP"]] <- data.frame(
      Metric = paste0("CBP < ", cbp_threshold, " (%)"),
      Mean   = pct_good,
      SD     = NA_real_,
      stringsAsFactors = FALSE
    )
  }

  if (length(summary_rows) > 0) {
    summary_df <- do.call(rbind, summary_rows)
    rownames(summary_df) <- NULL
    if (verbose) {
      message(paste(utils::capture.output(
        print(summary_df, row.names = FALSE)), collapse = "\n"))
    }
  }

  plot_list$timestep_summary <- an_summary
  if (!is.null(ov_sum)) plot_list$overall_summary <- ov_sum
  invisible(plot_list)
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 5,
  dpi       = 100,
  out.width = "95%"
)

## -----------------------------------------------------------------------------
library(TemporalModelR)
library(terra)
library(sf)

pred_dir <- system.file("extdata/predictions", package = "TemporalModelR")

list.files(pred_dir, pattern = "\\.tif$")[1:5]

## -----------------------------------------------------------------------------
summary_out <- summarize_raster_outputs(
  predictions_dir = pred_dir,
  output_dir      = file.path(tempdir(), "Binary"),
  consensus       = 2,
  overwrite       = TRUE,
  verbose         = FALSE
)

names(summary_out)

## ----fig.width=10, fig.height=12----------------------------------------------
binary_stack   <- summary_out$consensus_stack
frequency_rast <- summary_out$frequency_raster

names(binary_stack) <- paste0("Y", 1:15, "_Spring")

terra::plot(binary_stack, nr = 5, nc = 3,
            mar = c(1.5, 0.5, 1.5, 0.5), legend = FALSE)

## ----fig.width=8, fig.height=4------------------------------------------------
terra::plot(frequency_rast,
            main = "Proportion of years pixel was suitable",
            mar  = c(2.5, 2.5, 2.5, 5.0))

## ----eval=FALSE---------------------------------------------------------------
# install.packages("fastcpd")

## -----------------------------------------------------------------------------
time_steps <- expand.grid(
  year             = 1:15,
  season           = "Spring",
  stringsAsFactors = FALSE
)

patterns <- analyze_temporal_patterns(
  binary_stack            = binary_stack,
  summary_raster          = frequency_rast,
  time_steps              = time_steps,
  output_dir              = file.path(tempdir(), "Patterns"),
  spatial_autocorrelation = TRUE,
  alpha                   = 0.05,
  estimate_time           = FALSE,
  overwrite               = TRUE,
  verbose                 = FALSE
)

names(patterns)

## -----------------------------------------------------------------------------
study_crs <- sf::st_crs(binary_stack)

zones_sf <- rbind(
  sf::st_sf(ZONE = "West",
            geometry = sf::st_sfc(sf::st_polygon(list(
              matrix(c(0, 0, 1500, 1500, 0,
                       0, 1500, 1500, 0, 0), ncol = 2)
            )), crs = study_crs)),
  sf::st_sf(ZONE = "East",
            geometry = sf::st_sfc(sf::st_polygon(list(
              matrix(c(1500, 1500, 3000, 3000, 1500,
                       0,    1500, 1500, 0,    0),    ncol = 2)
            )), crs = study_crs))
)

## -----------------------------------------------------------------------------
zone_summary <- analyze_trends_by_spatial_unit(
  shapefile_path       = zones_sf,
  name_field           = "ZONE",
  binary_stack         = binary_stack,
  pattern_raster       = patterns$pattern,
  time_decrease_raster = patterns$time_decrease,
  time_increase_raster = patterns$time_increase,
  time_steps           = time_steps,
  output_dir           = file.path(tempdir(), "ZoneSummary"),
  create_plot          = FALSE,
  verbose              = FALSE
)

names(zone_summary)

## -----------------------------------------------------------------------------
zone_summary$overall_summary

## -----------------------------------------------------------------------------
head(zone_summary$timestep_summary)

## -----------------------------------------------------------------------------
head(zone_summary$change_by_timestep)

## -----------------------------------------------------------------------------
zone_plots <- analyze_trends_by_spatial_unit(
  shapefile_path       = zones_sf,
  name_field           = "ZONE",
  binary_stack         = binary_stack,
  pattern_raster       = patterns$pattern,
  time_decrease_raster = patterns$time_decrease,
  time_increase_raster = patterns$time_increase,
  time_steps           = time_steps,
  output_dir           = file.path(tempdir(), "ZoneSummary"),
  create_plot          = TRUE,
  verbose              = FALSE
)

utils::globalVariables(c("x", "y", "radius"))

#' Summarize Temporal Patterns and Trends by Spatial Unit
#'
#' Aggregates temporal pattern classifications and change metrics
#' across user-defined spatial units (e.g., states, counties, watersheds).
#' Returns summary tables and optionally generates simple visualizations.
#'
#' @usage
#' analyze_trends_by_spatial_unit(shapefile_path, name_field, binary_stack = NULL,
#'                                pattern_raster = NULL, time_decrease_raster = NULL,
#'                                time_increase_raster = NULL, time_steps = NULL,
#'                                output_dir = NULL, overwrite = FALSE,
#'                                create_plot = TRUE, pie_scale = 0.15,
#'                                verbose = TRUE)
#'
#' @param shapefile_path Character, sf object, or sfc object. Path to a
#'   shapefile or directory containing one, an sf object, or an sfc geometry.
#'   Shapefiles spatial units will be used as the units for the data summary.
#' @param name_field Character. Attribute field to use as spatial unit labels.
#' @param binary_stack \code{SpatRaster} or character. Optional stack of binary
#'   prediction rasters across time. Required for per-unit habitat summaries
#'   and time series plots.
#' @param pattern_raster \code{SpatRaster} or character. Optional pattern
#'   classification raster from \code{\link[TemporalModelR]{analyze_temporal_patterns}}.
#'   Required for pattern composition summaries and scatterpie map.
#' @param time_decrease_raster \code{SpatRaster} or character. Optional raster
#'   of first decrease time step from \code{\link[TemporalModelR]{analyze_temporal_patterns}}.
#' @param time_increase_raster \code{SpatRaster} or character. Optional raster
#'   of first increase time step from \code{\link[TemporalModelR]{analyze_temporal_patterns}}.
#' @param time_steps Vector. Time labels for layers in \code{binary_stack}.
#'   Required when \code{binary_stack} or change rasters are provided.
#' @param output_dir Character or \code{NULL}. Directory for CSV outputs and
#'   plots. If \code{NULL}, results are returned in memory only and no files
#'   are written. Default is \code{NULL}.
#' @param overwrite Logical. If \code{TRUE}, recomputes and overwrites existing
#'   CSVs. Default is \code{FALSE}.
#' @param create_plot Logical. If \code{TRUE} (default), generates and saves
#'   plots.
#' @param pie_scale Numeric in (0, 1]. Largest pie's radius as a fraction of
#'   the smaller map dimension. Default is \code{0.15}, meaning the biggest
#'   pie spans roughly 30\% of the smaller map dimension (radius = 15\%, so
#'   diameter = 30\%). Smaller pies scale so their area is proportional to
#'   the unit's pixel count. The fraction is converted internally to
#'   coordinate units, so this argument works identically across CRSes
#'   (degrees, meters, etc.) without manual rescaling.
#' @param verbose Logical. If \code{TRUE} (default), prints progress
#'   messages during processing. Includes details on raster extraction and
#'   per-spatial-unit summaries.
#'
#' @return Invisibly returns a list containing:
#' \itemize{
#'   \item \code{overall_summary}: Pattern composition per spatial unit
#'     (present when \code{pattern_raster} is supplied).
#'   \item \code{timestep_summary}: Suitable pixel counts per unit per time step
#'     (present when \code{binary_stack} is supplied).
#'   \item \code{change_by_timestep}: Gain and loss pixel counts per unit per time
#'     step (present when both change rasters are supplied).
#'   \item \code{plots}: Named list of recorded plot objects (present when
#'     \code{create_plot = TRUE}).
#' }
#'
#' @details
#' Summarizes results from modeling and post-processing at the scale of specific
#' spatial blocks to allow for a nuanced look at spatiotemporal patterns.
#'
#' @seealso
#' Post-processing: \code{\link{summarize_raster_outputs}},
#'   \code{\link[TemporalModelR]{analyze_temporal_patterns}}
#'
#' @examples
#' con_file <- system.file("extdata/binary/consensus_stack.tif",
#'                         package = "TemporalModelR")
#'
#' binary_stack <- terra::rast(con_file)
#'
#' study_crs <- sf::st_crs(binary_stack)
#'
#' zones_sf <- rbind(
#'   sf::st_sf(ZONE = "West",
#'             geometry = sf::st_sfc(sf::st_polygon(list(
#'               matrix(c(0, 0, 1500, 1500, 0,
#'                        0, 1500, 1500, 0, 0), ncol = 2)
#'             )), crs = study_crs)),
#'   sf::st_sf(ZONE = "East",
#'             geometry = sf::st_sfc(sf::st_polygon(list(
#'               matrix(c(1500, 1500, 3000, 3000, 1500,
#'                        0,    1500, 1500, 0,    0),    ncol = 2)
#'             )), crs = study_crs))
#' )
#'
#' time_steps <- expand.grid(
#'   year             = 1:15,
#'   season           = "Spring",
#'   stringsAsFactors = FALSE
#' )
#'
#' analyze_trends_by_spatial_unit(
#'   shapefile_path = zones_sf,
#'   name_field     = "ZONE",
#'   binary_stack   = binary_stack,
#'   time_steps     = time_steps,
#'   create_plot    = FALSE,
#'   verbose        = FALSE
#' )

#' @export
#' @importFrom sf st_read st_transform st_coordinates st_point_on_surface st_sf
#' @importFrom exactextractr exact_extract
#' @importFrom terra rast nlyr values crs
#' @importFrom graphics plot barplot lines points abline axis legend mtext par
#' @importFrom grDevices dev.cur dev.off png rainbow recordPlot
#' @importFrom utils head read.csv setTxtProgressBar txtProgressBar write.csv
analyze_trends_by_spatial_unit <- function(shapefile_path,
                                           name_field,
                                           binary_stack         = NULL,
                                           pattern_raster       = NULL,
                                           time_decrease_raster = NULL,
                                           time_increase_raster = NULL,
                                           time_steps           = NULL,
                                           output_dir           = NULL,
                                           overwrite            = FALSE,
                                           create_plot          = TRUE,
                                           pie_scale            = 0.15,
                                           verbose              = TRUE) {

  if (missing(shapefile_path)) {
    stop(paste0("ERROR: 'shapefile_path' is required. ",
                "Please provide a file path, sf object, or sfc geometry."))
  }
  if (missing(name_field)) {
    stop(paste0("ERROR: 'name_field' is required. Please specify the attribute ",
                "column to use as spatial unit labels."))
  }

  for (pkg in c("sf", "exactextractr")) {
    if (!requireNamespace(pkg, quietly = TRUE)) {
      stop(paste0("ERROR: The '", pkg, "' package is required. ",
                  "Install with: install.packages('", pkg, "')"))
    }
  }

  has_binary  <- !is.null(binary_stack)
  has_pattern <- !is.null(pattern_raster)
  has_change  <- !is.null(time_decrease_raster) && !is.null(time_increase_raster)

  if (!has_binary && !has_pattern && !has_change) {
    stop("ERROR: At least one raster input must be provided.")
  }
  if ((has_binary || has_change) && is.null(time_steps)) {
    stop("ERROR: 'time_steps' must be provided when using binary_stack or change rasters.")
  }

  if (!is.null(time_steps)) {
    ts_norm           <- .normalize_time_steps(time_steps, "analyze_trends_by_spatial_unit")
    time_steps        <- ts_norm$time_steps
    secondary_filters <- ts_norm$secondary_filters
  } else {
    secondary_filters <- list()
  }
  if (xor(!is.null(time_decrease_raster), !is.null(time_increase_raster))) {
    stop("ERROR: Both 'time_decrease_raster' and 'time_increase_raster' must be provided together.")
  }

  if (is.character(shapefile_path) && dir.exists(shapefile_path)) {
    shp_files <- list.files(shapefile_path, pattern = "\\.shp$", full.names = TRUE)
    if (length(shp_files) == 0) stop(paste0("ERROR: No .shp file found in: ", shapefile_path))
    if (length(shp_files) > 1) stop(paste0("ERROR: Multiple .shp files in: ", shapefile_path))
    spatial_units <- sf::st_read(shp_files[1], quiet = TRUE)
  } else {
    spatial_units <- .load_shapefile_input(shapefile_path, "shapefile_path")
  }

  if (has_pattern) pattern_raster       <- .load_raster_input(pattern_raster,       "pattern_raster")
  if (has_change) {
    time_decrease_raster <- .load_raster_input(time_decrease_raster, "time_decrease_raster")
    time_increase_raster <- .load_raster_input(time_increase_raster, "time_increase_raster")
  }
  if (has_binary)  binary_stack <- .load_raster_input(binary_stack, "binary_stack")

  if (has_binary) lyr_names <- names(binary_stack)

  if (has_binary && length(secondary_filters) > 0) {
    keep_idx  <- seq_len(terra::nlyr(binary_stack))
    for (sc in names(secondary_filters)) {
      val      <- as.character(secondary_filters[[sc]])
      keep_idx <- keep_idx[grepl(val, lyr_names[keep_idx], ignore.case = TRUE)]
    }
    if (length(keep_idx) == 0) {
      stop(paste0(
        "ERROR: No binary raster layers matched the secondary filter(s): ",
        paste(mapply(function(k, v) paste0(k, "=", v),
                     names(secondary_filters), secondary_filters), collapse = ", "),
        ". Layer names: ",
        paste(head(lyr_names, 5), collapse = ", "),
        if (length(lyr_names) > 5) paste0(" ... (", length(lyr_names), " total)") else ""
      ))
    }
    if (length(keep_idx) != length(time_steps)) {
      stop(paste0(
        "ERROR: After filtering by secondary column(s), found ", length(keep_idx),
        " matching raster layer(s) but time_steps has ", length(time_steps), " value(s). ",
        "Ensure time_steps values match the primary time steps in your predictions."
      ))
    }
    binary_stack <- binary_stack[[keep_idx]]
    if (verbose) message(paste0(
      "Filtered binary_stack to ", terra::nlyr(binary_stack), " layer(s) matching: ",
      paste(mapply(function(k, v) paste0(k, "=", v),
                   names(secondary_filters), secondary_filters), collapse = ", ")
    ))
  } else if (has_binary && !is.null(time_steps)) {

    if (terra::nlyr(binary_stack) > length(time_steps)) {
      match_counts <- vapply(as.character(time_steps), function(ts) {
        sum(grepl(paste0("(^|[^0-9])", ts, "([^0-9]|$)"), lyr_names))
      }, integer(1))
      ambiguous <- time_steps[match_counts > 1]
      if (length(ambiguous) > 0) {
        stop(paste0(
          "ERROR: The binary stack has ", terra::nlyr(binary_stack),
          " layers but time_steps only has ", length(time_steps), " value(s). ",
          "Time step value(s) ", paste(ambiguous, collapse = ", "),
          " each match more than one raster layer, suggesting the stack contains ",
          "predictions for multiple secondary time values (e.g. seasons). ",
          "Supply time_steps as a multi-column data frame specifying the secondary ",
          "column value to use. For example: ",
          "time_steps <- expand.grid(time_step = 1:15, season = \"Spring\", stringsAsFactors = FALSE)"
        ))
      }
    }
  }

  if (!is.null(output_dir) && !dir.exists(output_dir)) {
    dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
  }
  plots_dir <- if (!is.null(output_dir)) file.path(output_dir, "plots") else NULL
  if (create_plot && !is.null(plots_dir) && !dir.exists(plots_dir)) {
    dir.create(plots_dir, recursive = TRUE, showWarnings = FALSE)
  }

  if (!name_field %in% names(spatial_units)) {
    stop(paste0("ERROR: '", name_field, "' not found. Available: ",
                paste(names(spatial_units), collapse = ", ")))
  }

  if (verbose) message(paste("Loaded", nrow(spatial_units), "spatial units."))

  ref_raster  <- if (has_pattern) pattern_raster else if (has_binary) binary_stack else time_decrease_raster
  raster_crs  <- terra::crs(ref_raster, proj = TRUE)
  if (is.na(raster_crs) || is.null(raster_crs)) stop("ERROR: Reference raster has no CRS.")

  if (verbose) message("Transforming spatial units to raster CRS...")
  spatial_units_proj <- sf::st_transform(spatial_units, raster_crs)

  all_units <- trimws(as.character(spatial_units_proj[[name_field]]))
  n_units   <- length(all_units)

  overall_summary_file <- if (!is.null(output_dir)) file.path(output_dir, "pattern_summary.csv")     else NULL
  timestep_summary_file  <- if (!is.null(output_dir)) file.path(output_dir, "timestep_habitat.csv")       else NULL
  change_by_timestep_file  <- if (!is.null(output_dir)) file.path(output_dir, "change_by_timestep.csv")       else NULL

  overall_summary <- NULL
  timestep_summary  <- NULL
  change_by_timestep  <- NULL
  plots_list      <- list()

  if (has_pattern) {
    if (verbose) message("Extracting pattern classifications...")

    load_existing_pattern <- !is.null(overall_summary_file) &&
      file.exists(overall_summary_file) && !overwrite

    if (!load_existing_pattern) {

      pat_ext <- suppressWarnings(suppressMessages(
        exactextractr::exact_extract(pattern_raster, spatial_units_proj,
                                     progress = FALSE)
      ))

      pat_counts <- lapply(pat_ext, function(x) {
        v <- x$value
        vapply(seq_len(7), function(k) sum(v == k, na.rm = TRUE), integer(1))
      })

      pm <- do.call(rbind, pat_counts)

      overall_summary <- data.frame(
        Spatial_Unit   = all_units,
        Always_Absent  = pm[, 1],
        Always_Present = pm[, 2],
        No_Pattern     = pm[, 3],
        Increasing     = pm[, 4],
        Decreasing     = pm[, 5],
        Fluctuating    = pm[, 6],
        Failed         = pm[, 7],
        stringsAsFactors = FALSE
      )
      overall_summary[is.na(overall_summary)] <- 0
      overall_summary$Total_Pixels <- rowSums(overall_summary[, 2:8])

      tp <- overall_summary$Total_Pixels
      overall_summary$Pct_Always_Absent  <- round(100 * overall_summary$Always_Absent  / tp, 2)
      overall_summary$Pct_Always_Present <- round(100 * overall_summary$Always_Present / tp, 2)
      overall_summary$Pct_No_Pattern     <- round(100 * overall_summary$No_Pattern     / tp, 2)
      overall_summary$Pct_Increasing     <- round(100 * overall_summary$Increasing     / tp, 2)
      overall_summary$Pct_Decreasing     <- round(100 * overall_summary$Decreasing     / tp, 2)
      overall_summary$Pct_Fluctuating    <- round(100 * overall_summary$Fluctuating    / tp, 2)

      denom_no_absent  <- tp - overall_summary$Always_Absent
      denom_no_present <- tp - overall_summary$Always_Present
      overall_summary$Prop_Increasing       <- round(100 * overall_summary$Increasing     / denom_no_absent,  2)
      overall_summary$Prop_Stable_Suitable  <- round(100 * overall_summary$Always_Present / denom_no_absent,  2)
      overall_summary$Prop_Decreasing       <- round(100 * overall_summary$Decreasing     / denom_no_present, 2)
      overall_summary$Prop_Stable_Unsuitable <- round(100 * overall_summary$Always_Absent / denom_no_present, 2)
      overall_summary$Prop_Increasing[denom_no_absent  == 0] <- NA
      overall_summary$Prop_Stable_Suitable[denom_no_absent  == 0] <- NA
      overall_summary$Prop_Decreasing[denom_no_present == 0] <- NA
      overall_summary$Prop_Stable_Unsuitable[denom_no_present == 0] <- NA

      if (!is.null(overall_summary_file)) {
        utils::write.csv(overall_summary, overall_summary_file, row.names = FALSE)
        if (verbose) message(paste("Saved:", basename(overall_summary_file)))
      }
      gc(verbose = FALSE)

    } else {
      overall_summary <- utils::read.csv(overall_summary_file,
                                         stringsAsFactors = FALSE)
      overall_summary$Spatial_Unit <- trimws(as.character(overall_summary$Spatial_Unit))
      if (nrow(overall_summary) == 0) {
        warning("Loaded pattern_summary.csv is empty - rerun with overwrite = TRUE.")
        overall_summary <- NULL
      } else {
        count_cols <- c("Always_Absent", "Always_Present", "No_Pattern",
                        "Increasing", "Decreasing", "Fluctuating", "Failed")
        missing_cols <- count_cols[!count_cols %in% names(overall_summary)]
        if (length(missing_cols) > 0) {
          warning(paste0("pattern_summary.csv is missing columns: ",
                         paste(missing_cols, collapse = ", "),
                         " - rerun with overwrite = TRUE."))
          overall_summary <- NULL
        } else {
          if (!"Total_Pixels" %in% names(overall_summary))
            overall_summary$Total_Pixels <- rowSums(
              overall_summary[, count_cols, drop = FALSE], na.rm = TRUE
            )
          if (verbose) message(paste("Loaded existing:", basename(overall_summary_file),
                                     paste0("(", nrow(overall_summary), " units)")))
        }
      }
    }
  }

  if (has_binary) {
    if (verbose) message("Extracting per-unit habitat counts per time step...")

    load_existing_timestep <- !is.null(timestep_summary_file) &&
      file.exists(timestep_summary_file) && !overwrite

    if (!load_existing_timestep) {

      pb <- utils::txtProgressBar(min = 0, max = length(time_steps), style = 3, width = 50)
      timestep_results <- vector("list", length(time_steps))

      for (i in seq_along(time_steps)) {
        ts    <- time_steps[i]
        layer <- binary_stack[[i]]
        hab   <- suppressWarnings(suppressMessages(
          exactextractr::exact_extract(
            layer, spatial_units_proj,
            function(values, coverage_fractions)
              sum((values == 1) * coverage_fractions, na.rm = TRUE),
            progress = FALSE)
        ))
        timestep_results[[i]] <- data.frame(
          Spatial_Unit    = all_units,
          Time_Step       = ts,
          Pixels_Suitable = hab,
          stringsAsFactors = FALSE
        )
        utils::setTxtProgressBar(pb, i)
        if (i %% 5 == 0) gc(verbose = FALSE)
      }
      close(pb)
      if (verbose) message("")

      timestep_summary <- do.call(rbind, timestep_results)
      if (!is.null(timestep_summary_file)) {
        utils::write.csv(timestep_summary, timestep_summary_file, row.names = FALSE)
        if (verbose) message(paste("Saved:", basename(timestep_summary_file)))
      }
      gc(verbose = FALSE)

    } else {
      timestep_summary <- utils::read.csv(timestep_summary_file,
                                          stringsAsFactors = FALSE)
      timestep_summary$Spatial_Unit <- trimws(as.character(timestep_summary$Spatial_Unit))
      ### Support both old (Year) and new (Time_Step) column names
      if ("Year" %in% names(timestep_summary) && !"Time_Step" %in% names(timestep_summary)) {
        names(timestep_summary)[names(timestep_summary) == "Year"] <- "Time_Step"
      }
      timestep_summary$Time_Step       <- as.numeric(timestep_summary$Time_Step)
      timestep_summary$Pixels_Suitable <- as.numeric(timestep_summary$Pixels_Suitable)
      if (nrow(timestep_summary) == 0) {
        warning("Loaded timestep_habitat.csv is empty  -  rerun with overwrite = TRUE.")
        timestep_summary <- NULL
      } else {
        if (verbose) message(paste("Loaded existing:", basename(timestep_summary_file)))
      }
    }
  }

  if (has_change) {
    if (verbose) message("Extracting change events per time step...")

    load_existing_change <- !is.null(change_by_timestep_file) &&
      file.exists(change_by_timestep_file) && !overwrite

    if (!load_existing_change) {

      pb <- utils::txtProgressBar(min = 0, max = length(time_steps), style = 3, width = 50)
      dec_results <- vector("list", length(time_steps))
      inc_results <- vector("list", length(time_steps))

      for (i in seq_along(time_steps)) {
        ts  <- time_steps[i]
        dec <- suppressWarnings(suppressMessages(
          exactextractr::exact_extract(
            time_decrease_raster, spatial_units_proj,
            function(values, coverage_fractions) sum(values == ts, na.rm = TRUE),
            progress = FALSE)
        ))
        inc <- suppressWarnings(suppressMessages(
          exactextractr::exact_extract(
            time_increase_raster, spatial_units_proj,
            function(values, coverage_fractions) sum(values == ts, na.rm = TRUE),
            progress = FALSE)
        ))
        dec_results[[i]] <- data.frame(Spatial_Unit = all_units, Time_Step = ts,
                                       Decrease_Pixels = dec, stringsAsFactors = FALSE)
        inc_results[[i]] <- data.frame(Spatial_Unit = all_units, Time_Step = ts,
                                       Increase_Pixels = inc, stringsAsFactors = FALSE)
        utils::setTxtProgressBar(pb, i)
        if (i %% 5 == 0) gc(verbose = FALSE)
      }
      close(pb)
      if (verbose) message("")

      change_by_timestep <- merge(
        do.call(rbind, dec_results),
        do.call(rbind, inc_results),
        by = c("Spatial_Unit", "Time_Step"), all.x = TRUE
      )
      if (!is.null(change_by_timestep_file)) {
        utils::write.csv(change_by_timestep, change_by_timestep_file, row.names = FALSE)
        if (verbose) message(paste("Saved:", basename(change_by_timestep_file)))
      }
      gc(verbose = FALSE)

    } else {
      change_by_timestep <- utils::read.csv(change_by_timestep_file,
                                            stringsAsFactors = FALSE)
      change_by_timestep$Spatial_Unit <- trimws(as.character(change_by_timestep$Spatial_Unit))
      if ("Year" %in% names(change_by_timestep) && !"Time_Step" %in% names(change_by_timestep)) {
        names(change_by_timestep)[names(change_by_timestep) == "Year"] <- "Time_Step"
      }
      change_by_timestep$Time_Step       <- as.numeric(change_by_timestep$Time_Step)
      change_by_timestep$Decrease_Pixels <- as.numeric(change_by_timestep$Decrease_Pixels)
      change_by_timestep$Increase_Pixels <- as.numeric(change_by_timestep$Increase_Pixels)
      if (nrow(change_by_timestep) == 0) {
        warning("Loaded change_by_timestep.csv is empty  -  rerun with overwrite = TRUE.")
        change_by_timestep <- NULL
      } else {
        if (verbose) message(paste("Loaded existing:", basename(change_by_timestep_file)))
      }
    }
  }

  has_pattern <- !is.null(overall_summary)
  has_binary  <- !is.null(timestep_summary)
  has_change  <- !is.null(change_by_timestep)

  if (verbose) message(paste0("Data available - pattern: ", has_pattern,
                              " | binary: ", has_binary,
                              " | change: ", has_change))

  if (!has_pattern && !has_binary && !has_change) {
    stop(paste0(
      "ERROR: All cached CSVs in '", output_dir, "' are empty. ",
      "Rerun with overwrite = TRUE to recompute from rasters."
    ))
  }

  if (!is.null(overall_summary) && verbose) {
    message("\nPattern composition by spatial unit available in result$overall_summary.")
  }
  if (!is.null(timestep_summary) && verbose) {
    message("\nPer-unit habitat summary available in result$timestep_summary.")
  }
  if (!is.null(change_by_timestep) && verbose) {
    message("\nChange by time step available in result$change_by_timestep.")
  }

  if (!create_plot) {
    if (verbose) message("\nSpatial unit analysis complete (create_plot = FALSE, no plots generated).")
    result <- list()
    if (!is.null(overall_summary)) result$overall_summary <- overall_summary
    if (!is.null(timestep_summary))  result$timestep_summary  <- timestep_summary
    if (!is.null(change_by_timestep))  result$change_by_timestep  <- change_by_timestep
    return(invisible(result))
  }

  if (verbose) message("Generating visualizations...")

  c25 <- c(
    "dodgerblue2", "#E31A1C", "green4", "#6A3D9A", "#FF7F00", "black", "gold1",
    "skyblue2", "#FB9A99", "palegreen2", "#CAB2D6", "#FDBF6F", "gray70", "khaki2",
    "maroon", "orchid1", "deeppink1", "blue1", "steelblue4", "darkturquoise",
    "green1", "yellow4", "yellow3", "darkorange4", "brown"
  )
  unit_colors <- if (n_units <= 25) c25[seq_len(n_units)] else
    grDevices::rainbow(n_units, s = 0.8, v = 0.8)
  names(unit_colors) <- all_units

  pat_cols_hex <- c(
    Always_Absent  = "#730000",
    Always_Present = "#267300",
    No_Pattern     = "#B2B2B2",
    Increasing     = "#A3FF73",
    Decreasing     = "#FF7F7F",
    Fluctuating    = "#A900E6",
    Failed         = "#000000"
  )

  opar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(opar), add = TRUE)

  if (has_pattern) {
    has_gg <- requireNamespace("ggplot2",    quietly = TRUE)
    has_sp <- requireNamespace("scatterpie", quietly = TRUE)

    if (!has_gg || !has_sp) {
      missing_pkgs <- c(if (!has_gg) "ggplot2", if (!has_sp) "scatterpie")
      if (verbose) message(paste0(
        "NOTE: Scatterpie map skipped  -  install missing package(s): ",
        paste(missing_pkgs, collapse = ", "),
        "  ->  install.packages(c(",
        paste(paste0('"', missing_pkgs, '"'), collapse = ", "), "))"
      ))
    } else {
      if (verbose) message("Creating scatterpie map...")

      centroids       <- suppressWarnings(
        sf::st_coordinates(sf::st_point_on_surface(spatial_units_proj))
      )
      pie_data        <- overall_summary
      pie_data$x      <- centroids[match(overall_summary$Spatial_Unit, all_units), 1]
      pie_data$y      <- centroids[match(overall_summary$Spatial_Unit, all_units), 2]
      pie_cols        <- c("Always_Absent", "Always_Present", "No_Pattern",
                           "Increasing", "Decreasing", "Fluctuating", "Failed")
      for (col in pie_cols) if (!col %in% names(pie_data)) pie_data[[col]] <- 0

      ### pie_scale is a fraction of the smaller map dimension - convert to
      ### coordinate units here. This keeps the user-facing argument CRS-
      ### independent: pie_scale = 0.15 looks the same on a degrees map as on
      ### a meters map.
      if (!is.numeric(pie_scale) || length(pie_scale) != 1 ||
          pie_scale <= 0 || pie_scale > 1) {
        stop("ERROR: 'pie_scale' must be a single numeric value in (0, 1]. ",
             "It is interpreted as the largest pie's radius expressed as a ",
             "fraction of the smaller map dimension.")
      }

      bb_map        <- sf::st_bbox(spatial_units_proj)
      map_width     <- as.numeric(bb_map["xmax"] - bb_map["xmin"])
      map_height    <- as.numeric(bb_map["ymax"] - bb_map["ymin"])
      map_min_dim   <- min(map_width, map_height)
      r_max         <- pie_scale * map_min_dim

      ### Radius proportional to sqrt(pixel count) so pie *area* is
      ### proportional to count. Normalised by the largest unit so the
      ### biggest pie has radius = r_max.
      max_unit_px     <- max(overall_summary$Total_Pixels, na.rm = TRUE)
      pie_data$radius <- sqrt(overall_summary$Total_Pixels / max_unit_px) * r_max

      p_map <- ggplot2::ggplot() +
        ggplot2::geom_sf(data = spatial_units_proj, fill = NA,
                         color = "black", linewidth = 0.3) +
        scatterpie::geom_scatterpie(
          ggplot2::aes(x = x, y = y, r = radius),
          data = pie_data, cols = pie_cols, color = NA
        ) +
        ggplot2::scale_fill_manual(
          values = pat_cols_hex,
          breaks = pie_cols,
          labels = c("Never Suitable", "Always Suitable", "No Pattern",
                     "Increasing", "Decreasing", "Fluctuating", "Failed")
        ) +
        ggplot2::guides(fill = ggplot2::guide_legend(override.aes = list(colour = NA))) +
        ggplot2::coord_sf() +
        ggplot2::labs(title = "Pattern Composition by Spatial Unit", fill = "Pattern",
                      x = NULL, y = NULL) +
        ggplot2::theme_classic() +
        ggplot2::theme(
          plot.title      = ggplot2::element_text(face = "bold", size = 14, hjust = 0.5),
          legend.position = "right",
          axis.title      = ggplot2::element_blank(),
          axis.text       = ggplot2::element_blank(),
          axis.ticks      = ggplot2::element_blank(),
          axis.line       = ggplot2::element_blank()
        )

      if (!is.null(plots_dir)) {
        map_png <- file.path(plots_dir, "pattern_map.png")
        tryCatch({
          ggplot2::ggsave(map_png, p_map, width = 12, height = 8, dpi = 300)
          if (verbose) message(paste("Saved:", basename(map_png)))
        }, error = function(e) warning(paste("Could not save pattern map:", e$message)))
      }

      tryCatch(
        print(p_map),
        error = function(e) warning(paste("Could not render pattern map inline:", e$message))
      )
      plots_list$pattern_map <- p_map
    }
  }

  if (has_binary) {
    if (verbose) message("Creating time series plot...")

    ts_vals  <- sort(unique(as.numeric(timestep_summary$Time_Step)))
    ts_y_all <- as.numeric(timestep_summary$Pixels_Suitable)
    ts_y_max <- if (any(is.finite(ts_y_all))) max(ts_y_all, na.rm = TRUE) else 1
    ts_x_rng <- if (length(ts_vals) > 1) range(ts_vals) else c(ts_vals[1] - 1, ts_vals[1] + 1)
    x_ticks  <- pretty(ts_x_rng, n = 8)

    if (!is.null(plots_dir)) {
      ts_png <- file.path(plots_dir, "time_series.png")
      tryCatch({
        grDevices::png(ts_png, width = 1400, height = 700, res = 150)
        .plot_trend_timeseries(timestep_summary, all_units, unit_colors,
                               ts_x_rng, ts_y_max, x_ticks)
        grDevices::dev.off()
        if (verbose) message(paste("Saved:", basename(ts_png)))
      }, error = function(e) {
        warning(paste("Could not save time series:", e$message))
        if (grDevices::dev.cur() > 1) grDevices::dev.off()
      })
    }
    tryCatch({
      .plot_trend_timeseries(timestep_summary, all_units, unit_colors,
                             ts_x_rng, ts_y_max, x_ticks)
      plots_list$time_series <- grDevices::recordPlot()
    }, error = function(e) {
      warning(paste("Could not render time_series plot:", e$message))
    })
  }

  if (has_change) {

    change_has_data <- any(change_by_timestep$Increase_Pixels > 0, na.rm = TRUE) ||
      any(change_by_timestep$Decrease_Pixels > 0, na.rm = TRUE)

    if (!change_has_data) {
      dec_vals <- tryCatch(
        sort(unique(as.integer(terra::values(time_decrease_raster, mat = FALSE)),
                    na.rm = TRUE)),
        error = function(e) NULL
      )
      warning(paste0(
        "change_by_timestep has no non-zero pixel counts  -  skipping change plots.\n",
        "  time_steps range: ", min(time_steps), "-", max(time_steps), "\n",
        if (!is.null(dec_vals))
          paste0("  time_decrease_raster unique values (sample): ",
                 paste(head(dec_vals[!is.na(dec_vals)], 10), collapse = ", "))
        else
          "  Could not read time_decrease_raster values.",
        "\n  If ranges differ, check that time_steps matches the raster time step values."
      ))
    } else {

      yrs_chg <- sort(unique(change_by_timestep$Time_Step))

      if (verbose) message("Creating annual gains/losses plot...")

      gain_by_yr <- tapply(change_by_timestep$Increase_Pixels, change_by_timestep$Time_Step,
                           sum, na.rm = TRUE)
      loss_by_yr <- tapply(change_by_timestep$Decrease_Pixels, change_by_timestep$Time_Step,
                           sum, na.rm = TRUE)
      gain_v <- as.numeric(gain_by_yr[as.character(yrs_chg)])
      loss_v <- as.numeric(loss_by_yr[as.character(yrs_chg)])
      gain_v[!is.finite(gain_v)] <- 0
      loss_v[!is.finite(loss_v)] <- 0
      ac_ymax <- max(c(gain_v, loss_v), na.rm = TRUE)
      if (!is.finite(ac_ymax) || ac_ymax == 0) ac_ymax <- 1
      ac_ylim <- c(-ac_ymax * 1.15, ac_ymax * 1.15)

      if (!is.null(plots_dir)) {
        ac_png <- file.path(plots_dir, "annual_change.png")
        tryCatch({
          grDevices::png(ac_png, width = 1200, height = 700, res = 150)
          .plot_trend_change_per_step(gain_v, loss_v, yrs_chg, ac_ylim)
          grDevices::dev.off()
          if (verbose) message(paste("Saved:", basename(ac_png)))
        }, error = function(e) {
          warning(paste("Could not save annual change plot:", e$message))
          if (grDevices::dev.cur() > 1) grDevices::dev.off()
        })
      }
      tryCatch({
        .plot_trend_change_per_step(gain_v, loss_v, yrs_chg, ac_ylim)
        plots_list$annual_change <- grDevices::recordPlot()
      }, error = function(e) {
        warning(paste("Could not render annual_change plot:", e$message))
      })

      if (verbose) message("Creating total change by unit plot...")

      unit_gain_tot <- tapply(change_by_timestep$Increase_Pixels,
                              change_by_timestep$Spatial_Unit, sum, na.rm = TRUE)
      unit_loss_tot <- tapply(change_by_timestep$Decrease_Pixels,
                              change_by_timestep$Spatial_Unit, sum, na.rm = TRUE)

      shared_units  <- intersect(names(unit_gain_tot), names(unit_loss_tot))
      shared_units  <- shared_units[order(
        as.numeric(unit_gain_tot[shared_units]) -
          as.numeric(unit_loss_tot[shared_units])
      )]
      g_tot <- as.numeric(unit_gain_tot[shared_units])
      l_tot <- as.numeric(unit_loss_tot[shared_units])
      g_tot[!is.finite(g_tot)] <- 0
      l_tot[!is.finite(l_tot)] <- 0
      tcu_ymax <- max(c(g_tot, l_tot), na.rm = TRUE)
      if (!is.finite(tcu_ymax) || tcu_ymax == 0) tcu_ymax <- 1

      if (!is.null(plots_dir)) {
        tcu_png <- file.path(plots_dir, "total_change_by_unit.png")
        tryCatch({
          grDevices::png(tcu_png, width = 1000, height = 800, res = 150)
          .plot_trend_total_unit(g_tot, l_tot, shared_units, time_steps, tcu_ymax)
          grDevices::dev.off()
          if (verbose) message(paste("Saved:", basename(tcu_png)))
        }, error = function(e) {
          warning(paste("Could not save total change plot:", e$message))
          if (grDevices::dev.cur() > 1) grDevices::dev.off()
        })
      }
      tryCatch({
        .plot_trend_total_unit(g_tot, l_tot, shared_units, time_steps, tcu_ymax)
        plots_list$total_change_by_unit <- grDevices::recordPlot()
      }, error = function(e) {
        warning(paste("Could not render total_change_by_unit plot:", e$message))
      })

      if (verbose) message("Creating faceted time-step change plot...")

      n_cols <- min(3, n_units)
      n_rows <- ceiling(n_units / n_cols)

      if (!is.null(plots_dir)) {
        faceted_png <- file.path(plots_dir, "faceted_change.png")
        tryCatch({
          grDevices::png(faceted_png, width = 1400, height = 1000, res = 150)
          .plot_trend_facet_units(change_by_timestep, all_units, n_rows, n_cols)
          grDevices::dev.off()
          if (verbose) message(paste("Saved:", basename(faceted_png)))
        }, error = function(e) {
          warning(paste("Could not save faceted change plot:", e$message))
          if (grDevices::dev.cur() > 1) grDevices::dev.off()
        })
      }
      tryCatch({
        .plot_trend_facet_units(change_by_timestep, all_units, n_rows, n_cols)
        plots_list$faceted_change <- grDevices::recordPlot()
      }, error = function(e) {
        warning(paste("Could not render faceted_change plot:", e$message))
      })

    }
  }

  if (verbose) message("\nSpatial unit analysis complete.")
  if (!is.null(time_steps))
    if (verbose) message(paste0("Period: ", min(time_steps), "-", max(time_steps),
                                " | Spatial units: ", n_units))

  result <- list()
  if (!is.null(overall_summary)) result$overall_summary <- overall_summary
  if (!is.null(timestep_summary))  result$timestep_summary  <- timestep_summary
  if (!is.null(change_by_timestep))  result$change_by_timestep  <- change_by_timestep
  if (length(plots_list) > 0)    result$plots           <- plots_list

  invisible(result)
}

#' Spatiotemporal Rarefaction of Species Occurrence Data
#'
#' Preprocessing function that rarefies species occurrence data to one point per
#' raster pixel, optionally accounting for temporal components. Reduces sampling
#' bias and spatial autocorrelation in occurrence datasets.
#'
#' @usage
#' spatiotemporal_rarefaction(points_sp, output_dir, reference_raster,
#'                            time_cols = NULL, xcol = NULL, ycol = NULL,
#'                            points_crs = NULL, output_prefix = "Pts_Database",
#'                            verbose = TRUE)
#'
#' @param points_sp Input point data. Accepts an sf object, data frame,
#'   SpatialPointsDataFrame, or file path to a \code{.csv}, \code{.shp},
#'   \code{.geojson}, or \code{.gpkg} file.
#' @param output_dir Character. Directory where output CSV files will be saved.
#' @param reference_raster Character, SpatRaster, or RasterLayer. Raster used
#'   to define pixel boundaries for rarefaction. Accepts a file path,
#'   RasterLayer, or SpatRaster.
#' @param time_cols Character vector. Column names in \code{points_sp} defining
#'   temporal grouping for spatiotemporal rarefaction. When \code{NULL}
#'   (default), only spatial rarefaction is performed.
#' @param xcol Character. Name of the x-coordinate column. Required when
#'   \code{points_sp} is a CSV file or data frame.
#' @param ycol Character. Name of the y-coordinate column. Required when
#'   \code{points_sp} is a CSV file or data frame.
#' @param points_crs Character or CRS object. CRS of the input points.
#'   Required when \code{points_sp} is a CSV file or data frame.
#' @param output_prefix Character. Prefix for output file names. Default is
#'   \code{"Pts_Database"}. Output files are named
#'   \code{<output_prefix>_OnePerPix.csv} and, when \code{time_cols} are
#'   provided, \code{<output_prefix>_OnePerPixPerTimeStep.csv}.
#' @param verbose Logical. If \code{TRUE} (default), prints progress
#'   messages during processing. Includes file loading, missing-data removal
#'   counts, rarefaction summaries, and a final comparison of spatial
#'   vs spatiotemporal point counts.
#'
#' @return Invisibly returns a list containing:
#' \itemize{
#'   \item \code{input_points}: Integer. Number of input points after CRS
#'     alignment and removal of rows with missing time column values.
#'   \item \code{spatial_points}: Integer. Number of points retained after
#'     spatial-only rarefaction (one per pixel).
#'   \item \code{spatiotemporal_points}: Integer. Number of points retained
#'     after spatiotemporal rarefaction (one per pixel per time combination).
#'     \code{NA} when \code{time_cols} is not provided.
#'   \item \code{time_cols_used}: Character vector of time columns used.
#'     \code{NULL} when \code{time_cols} is not provided.
#'   \item \code{spatial_table}: Data frame of spatially rarefied points with
#'     columns \code{pixel_id}, \code{X}, \code{Y}, and any \code{time_cols}.
#'   \item \code{spatiotemporal_table}: Data frame of spatiotemporally rarefied
#'     points with columns \code{pixel_id}, \code{X}, \code{Y}, and
#'     \code{time_cols}. \code{NULL} when \code{time_cols} is not provided.
#'   \item \code{files_created}: Named list of file paths written. Always
#'     contains \code{$spatial}; additionally contains
#'     \code{$spatiotemporal} when \code{time_cols} are provided.
#' }
#'
#' @details
#' The function assigns each point to a raster pixel using the resolution and
#' extent of \code{reference_raster}, then performs:
#' \itemize{
#'   \item Spatial rarefaction: retains one point per pixel, written to
#'     \code{<output_prefix>_OnePerPix.csv}.
#'   \item Spatiotemporal rarefaction (when \code{time_cols} are provided):
#'     retains one point per pixel per unique combination of time column values,
#'     written to \code{<output_prefix>_OnePerPixPerTimeStep.csv}.
#' }
#'
#' Output CSV files are suitable as direct input to
#' \code{\link{temporally_explicit_extraction}}.
#'
#' @seealso
#' Preprocessing: \code{\link{temporally_explicit_extraction}},
#'   \code{\link{spatiotemporal_partition}}
#'
#' @examples
#' pts_file <- system.file(
#'   "extdata/points/synthetic_occurrence_points.csv",
#'   package = "TemporalModelR"
#' )
#'
#' ref_file <- system.file("extdata/rasters_raw/elevation.tif",
#'                         package = "TemporalModelR")
#'
#' out_dir  <- file.path(tempdir(), "rarefied")
#'
#' spatiotemporal_rarefaction(
#'   points_sp        = pts_file,
#'   output_dir       = out_dir,
#'   reference_raster = ref_file,
#'   time_cols        = c("year", "season"),
#'   xcol             = "x",
#'   ycol             = "y",
#'   points_crs       = terra::crs(terra::rast(ref_file)),
#'   verbose          = FALSE
#' )

#' @export
#' @importFrom terra rast res ext ncell crs extract vect
#' @importFrom sf st_as_sf st_transform st_coordinates st_read st_drop_geometry st_crs
#' @importFrom tools file_ext
#' @importFrom utils read.csv write.csv
#' @importFrom stats complete.cases
spatiotemporal_rarefaction <- function(points_sp,
                                       output_dir,
                                       reference_raster,
                                       time_cols     = NULL,
                                       xcol          = NULL,
                                       ycol          = NULL,
                                       points_crs    = NULL,
                                       output_prefix = "Pts_Database",
                                       verbose       = TRUE) {


  if (missing(points_sp)) {
    stop(paste0("ERROR: 'points_sp' is required but was not provided. Please provide ",
                "an sf object, data frame, SpatialPointsDataFrame, or file path."))
  }
  if (missing(output_dir) || is.null(output_dir) || output_dir == "") {
    stop(paste0("ERROR: 'output_dir' is required but was not provided. ",
                "Please specify the directory where output files should be saved."))
  }
  if (missing(reference_raster)) {
    stop(paste0("ERROR: 'reference_raster' is required but was not provided. Please specify ",
                "a raster file path or raster object used to define pixel boundaries."))
  }

  if (is.character(points_sp)) {
    if (!file.exists(points_sp)) {
      stop(paste0("ERROR: File does not exist: ", points_sp))
    }

    file_ext_lower <- tolower(tools::file_ext(points_sp))

    if (file_ext_lower == "csv") {
      if (is.null(xcol))       stop("ERROR: 'xcol' is required when reading CSV files.")
      if (is.null(ycol))       stop("ERROR: 'ycol' is required when reading CSV files.")
      if (is.null(points_crs)) stop("ERROR: 'points_crs' is required when reading CSV files.")

      if (verbose) message(paste("Reading CSV file:", basename(points_sp)))
      points_data <- utils::read.csv(points_sp, stringsAsFactors = FALSE)

      if (!xcol %in% names(points_data)) stop(paste0("ERROR: Column '", xcol, "' not found in CSV."))
      if (!ycol %in% names(points_data)) stop(paste0("ERROR: Column '", ycol, "' not found in CSV."))

      points_sp <- sf::st_as_sf(points_data, coords = c(xcol, ycol), crs = points_crs)

    } else if (file_ext_lower %in% c("shp", "geojson", "gpkg")) {
      if (verbose) message(paste("Reading spatial file:", basename(points_sp)))
      points_sp <- sf::st_read(points_sp, quiet = TRUE)
    } else {
      stop(paste0("ERROR: Unsupported file format: '", file_ext_lower,
                  "'. Supported formats: .csv, .shp, .geojson, .gpkg"))
    }

  } else if (inherits(points_sp, "sf")) {
    if (verbose) message("Using provided sf object...")

  } else if (inherits(points_sp, "SpatialPointsDataFrame")) {
    warning("Converting 'SpatialPointsDataFrame' to sf object for terra compatibility.")
    points_sp <- sf::st_as_sf(points_sp)

  } else if (is.data.frame(points_sp)) {
    if (is.null(xcol))       stop("ERROR: 'xcol' is required when providing a data frame.")
    if (is.null(ycol))       stop("ERROR: 'ycol' is required when providing a data frame.")
    if (is.null(points_crs)) stop("ERROR: 'points_crs' is required when providing a data frame.")

    if (verbose) message("Converting data frame to sf object...")
    points_sp <- sf::st_as_sf(points_sp, coords = c(xcol, ycol), crs = points_crs)

  } else {
    stop("ERROR: 'points_sp' must be an sf object, data frame, SpatialPointsDataFrame, or file path.")
  }

  reference_raster <- .load_raster_input(reference_raster, "reference_raster")
  if (is.na(terra::crs(reference_raster)) || nchar(terra::crs(reference_raster)) == 0) {
    stop("ERROR: Reference raster has no defined CRS. Please assign a coordinate reference system before processing.")
  }

  if (sf::st_crs(points_sp) != sf::st_crs(reference_raster)) {
    if (verbose) message("Reprojecting points to match reference raster CRS...")
    points_sp <- sf::st_transform(points_sp, terra::crs(reference_raster))
  }

  dir.create(output_dir, showWarnings = FALSE, recursive = TRUE)

  if (is.null(time_cols) || length(time_cols) == 0) {
    if (verbose) message("No time columns provided. Performing spatial-only rarefaction.")
    perform_spatiotemporal <- FALSE
  } else {
    if (!is.character(time_cols)) {
      stop("ERROR: 'time_cols' must be a character vector. Please provide column names as a character vector, e.g. c(\"Year\", \"Month\").")
    }

    missing_cols <- setdiff(time_cols, names(points_sp))
    if (length(missing_cols) > 0) {
      stop(paste0("ERROR: The following time_cols are missing from the input data: ",
                  paste(missing_cols, collapse = ", "),
                  ". Available columns: ",
                  paste(names(points_sp)[names(points_sp) != "geometry"], collapse = ", ")))
    }

    n_original <- nrow(points_sp)
    pts_data      <- sf::st_drop_geometry(points_sp)
    complete_rows <- stats::complete.cases(pts_data[, time_cols, drop = FALSE])
    points_sp     <- points_sp[complete_rows, ]

    n_removed <- n_original - nrow(points_sp)
    if (n_removed > 0) {
      pct_removed <- round(n_removed / n_original * 100, 2)
      if (verbose) message(paste0("Removed ", n_removed, " rows (", pct_removed,
                                  "%) with missing values in time columns"))
    }

    perform_spatiotemporal <- TRUE
  }

  res_x           <- terra::res(reference_raster)[1]
  res_y           <- terra::res(reference_raster)[2]
  pixel_id_raster <- terra::rast(terra::ext(reference_raster), res = c(res_x, res_y))
  terra::crs(pixel_id_raster) <- terra::crs(reference_raster)
  pixel_id_raster[] <- seq_len(terra::ncell(pixel_id_raster))

  points_sp$pixel_id <- terra::extract(pixel_id_raster, terra::vect(points_sp))[, 2]

  coords        <- sf::st_coordinates(points_sp)
  points_sp$X  <- coords[, 1]
  points_sp$Y  <- coords[, 2]

  freq_table           <- as.data.frame(table(points_sp$pixel_id))
  colnames(freq_table) <- c("pixel_id", "Freq")

  if (perform_spatiotemporal) {
    pts_df     <- sf::st_drop_geometry(points_sp)
    group_cols <- c("pixel_id", time_cols)
    pts_df     <- pts_df[!duplicated(pts_df[, group_cols, drop = FALSE]), ]
    points_subset_st <- merge(pts_df, freq_table, by = "pixel_id")

    n_spatiotemporal    <- nrow(points_subset_st)
    cols_to_save        <- c(time_cols, "pixel_id", "X", "Y")
    spatiotemporal_file <- file.path(output_dir,
                                     paste0(output_prefix, "_OnePerPixPerTimeStep.csv"))
    utils::write.csv(points_subset_st[, cols_to_save], spatiotemporal_file, row.names = FALSE)

    time_combinations_n <- nrow(unique(pts_df[, time_cols, drop = FALSE]))
    if (verbose) message(paste("Spatiotemporal file saved:", basename(spatiotemporal_file)))
    if (verbose) message(paste("Retained 1 point per pixel across",
                               time_combinations_n, "unique time combinations"))
  }

  pts_df_spatial        <- sf::st_drop_geometry(points_sp)
  pts_df_spatial        <- pts_df_spatial[!duplicated(pts_df_spatial$pixel_id), ]
  points_subset_spatial <- merge(pts_df_spatial, freq_table, by = "pixel_id")

  n_spatial    <- nrow(points_subset_spatial)
  cols_to_save <- if (perform_spatiotemporal) {
    c(time_cols, "pixel_id", "X", "Y")
  } else {
    c("pixel_id", "X", "Y")
  }

  spatial_file <- file.path(output_dir, paste0(output_prefix, "_OnePerPix.csv"))
  utils::write.csv(points_subset_spatial[, cols_to_save], spatial_file, row.names = FALSE)

  if (verbose) message(paste("Spatial file saved:", basename(spatial_file)))

  if (perform_spatiotemporal) {
    additional_points <- n_spatiotemporal - n_spatial
    pct_additional    <- round((additional_points / n_spatial) * 100, 2)
    if (verbose) message(paste0("Spatial: ", n_spatial, " points | Spatiotemporal: ", n_spatiotemporal,
                                " points | Additional retained: ", additional_points,
                                " (", pct_additional, "% increase)"))
  }

  if (verbose) message("Processing complete.")

  invisible(list(
    input_points          = nrow(points_sp),
    spatial_points        = n_spatial,
    spatiotemporal_points = if (perform_spatiotemporal) n_spatiotemporal else NA,
    time_cols_used        = if (perform_spatiotemporal) time_cols else NULL,
    spatial_table         = points_subset_spatial[, cols_to_save],
    spatiotemporal_table  = if (perform_spatiotemporal) points_subset_st[, cols_to_save] else NULL,
    files_created         = if (perform_spatiotemporal) {
      list(spatiotemporal = spatiotemporal_file, spatial = spatial_file)
    } else {
      list(spatial = spatial_file)
    }
  ))
}

#' Build Temporal Random Forest Models Across Cross-Validation Folds
#'
#' Modeling function that constructs Random Forest classification models for
#' each cross-validation fold using presence and pseudoabsence data. Each model
#' reserves one fold as testing data and uses the remaining folds as training
#' data. The user specifies predictors as a character vector.
#' Predicted probabilities of presence are extracted from the
#' out-of-bag or in-bag vote fractions and thresholded to produce binary
#' suitability classifications. Variable importance is recorded for each fold.
#' The returned object follows the same
#' structure as \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
#' and \code{\link{build_temporal_hv}}, and is accepted directly by
#' \code{\link{generate_spatiotemporal_predictions}}.
#'
#' @usage
#' build_temporal_rf(partition_result, pseudoabsence_result, model_vars,
#'                   rf_params = list(), threshold_method = "tss",
#'                   output_dir = file.path(tempdir(), "RF_Models"),
#'                   create_plot = TRUE, plot_palette = "Dark 2",
#'                   overwrite = FALSE, time_cols = NULL, verbose = TRUE)
#'
#' @param partition_result List or character. Output from
#'   \code{\link{spatiotemporal_partition}} or path to an \code{.rds}
#'   file containing that output.
#' @param pseudoabsence_result List or character. Output from
#'   \code{\link{generate_absences}} or path to an \code{.rds} file
#'   containing that output.
#' @param model_vars Character vector. Names of predictor columns to include in
#'   the Random Forest. All variables must be present as columns in both the
#'   presence and pseudoabsence data.
#' @param rf_params Named list. Additional arguments passed to
#'   \code{\link[randomForest]{randomForest}}, such as \code{ntree} (number of
#'   trees, default 500), \code{mtry} (number of variables tried at each split,
#'   default \code{floor(sqrt(length(model_vars)))}), and \code{nodesize}
#'   (minimum node size, default 1 for classification). Default is an empty
#'   list, which uses \pkg{randomForest} defaults.
#' @param threshold_method Character or numeric. Method used to convert
#'   continuous predicted probabilities to binary suitability. Accepted values:
#'   \itemize{
#'     \item \code{"prevalence"}: Sets threshold equal to the prevalence
#'       (proportion of presences) in the training data for that fold.
#'     \item \code{"tss"}: Selects the threshold that maximizes the True Skill
#'       Statistic (sensitivity + specificity - 1) on the training data.
#'       Default.
#'     \item A numeric value between 0 and 1 (e.g. \code{0.4}): Uses that
#'       value as a fixed threshold for all folds directly.
#'   }
#' @param output_dir Character. Directory to write output files including saved
#'   model objects and plots. Default is \code{file.path(tempdir(), "RF_Models")}.
#' @param create_plot Logical. If \code{TRUE}, generates a per-fold variable
#'   importance plot, partial dependence curves for each predictor, and a
#'   combined ROC curve summary. Default is \code{TRUE}.
#' @param overwrite Logical. If \code{TRUE}, overwrites existing saved model
#'   files. If \code{FALSE}, loads existing files when available. Default is
#'   \code{FALSE}.
#' @param plot_palette Character. Name of an HCL or RColorBrewer palette used
#'   to color folds in diagnostic plots. Accepts any HCL palette name (see
#'   \code{\link[grDevices]{hcl.pals}}) or, if \pkg{RColorBrewer} is installed,
#'   any Brewer palette name. Default is \code{"Dark 2"}.
#' @param time_cols Character. Name of the column(s) containing year or time
#'   step values in the occurrence data. Must match \code{time_cols} used in
#'   \code{\link{spatiotemporal_partition}}. Default is \code{NULL}.
#' @param verbose Logical. If \code{TRUE} (default), prints progress
#'   messages during processing. Includes per-fold training summaries and
#'   file-saved messages.
#'
#' @return A list with class \code{"TemporalRF"} containing:
#' \itemize{
#'   \item \code{models}: Named list of fitted \code{randomForest} objects,
#'     one per fold.
#'   \item \code{thresholds}: Named numeric vector of probability thresholds
#'     used for binary classification, one per fold.
#'   \item \code{threshold_method}: Character string recording the thresholding
#'     method used.
#'   \item \code{model_vars}: Character vector of predictor names used.
#'   \item \code{variable_importance}: Named list of importance data frames,
#'     one per fold, with mean decrease in accuracy for each predictor.
#'   \item \code{fold_training_data}: Named list of training data frames used
#'     to fit each fold model, retained for downstream prediction.
#'   \item \code{fold_test_metrics}: Data frame of held-out test fold metrics
#'     per fold: \code{Threshold}, \code{AUC}, \code{TSS}, \code{Kappa},
#'     \code{Sensitivity}, and \code{Specificity}. Also written to
#'     \code{Fold_Test_Metrics.csv} in \code{output_dir}.
#'   \item \code{output_dir}: Path to the output directory.
#'   \item \code{model_type}: Character string \code{"rf"}, used by
#'     \code{\link{generate_spatiotemporal_predictions}}.
#'   \item \code{plots}: Named list of recorded plot objects when
#'     \code{create_plot = TRUE}. Plots can be replayed with
#'     \code{grDevices::replayPlot()}.
#' }
#'
#' @details
#'
#' Random Forests are fit using \code{\link[randomForest]{randomForest}} from
#' the \pkg{randomForest} package. The response is treated as a factor
#' (\code{0}/\code{1}) so the model runs in classification mode, which produces
#' class vote fractions used as predicted probabilities. Importance is computed
#' with \code{importance = TRUE} and \code{type = 1} (mean decrease in
#' accuracy).
#'
#' Predicted probabilities are the vote fraction for class \code{1} from
#' \code{predict(..., type = "prob")[, "1"]}. These are used for threshold
#' selection and ROC curve construction.
#'
#' Diagnostic plots include: a variable importance bar chart (mean decrease in
#' accuracy across folds), partial dependence curves for each predictor showing
#' the marginal effect of each variable while averaging over all others (with
#' rug marks for presences and pseudoabsences), and a combined ROC curve panel.
#'
#' The returned object is recognized by
#' \code{\link{generate_spatiotemporal_predictions}}, which uses the
#' \code{model_type} field to use the correct prediction and evaluation
#' logic.
#'
#' @seealso
#' Preprocessing: \code{\link{spatiotemporal_partition}},
#'   \code{\link{generate_absences}}
#'
#' Modeling: \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
#'   \code{\link{build_temporal_hv}},
#'   \code{\link{generate_spatiotemporal_predictions}}
#'
#' External: \code{\link[randomForest]{randomForest}}
#'
#' @examples
#' data(tmr_partition, package = "TemporalModelR")
#'
#' data(tmr_absences,  package = "TemporalModelR")
#'
#' build_temporal_rf(
#'   partition_result     = tmr_partition,
#'   pseudoabsence_result = tmr_absences,
#'   model_vars           = c("elevation", "forest_cover", "prseas"),
#'   rf_params            = list(ntree = 100),
#'   threshold_method     = "tss",
#'   output_dir           = tempdir(),
#'   create_plot          = FALSE,
#'   time_cols            = c("year", "season"),
#'   verbose              = FALSE
#' )

#' @export
#' @importFrom sf st_drop_geometry
#' @importFrom stats predict complete.cases
#' @importFrom grDevices adjustcolor dev.cur dev.off png recordPlot
#' @importFrom graphics plot barplot abline legend mtext par rug
#' @importFrom tools file_ext
#' @importFrom utils capture.output write.csv
build_temporal_rf <- function(partition_result,
                              pseudoabsence_result,
                              model_vars,
                              rf_params         = list(),
                              threshold_method  = "tss",
                              output_dir        = file.path(tempdir(), "RF_Models"),
                              create_plot       = TRUE,
                              plot_palette      = "Dark 2",
                              overwrite         = FALSE,
                              time_cols         = NULL,
                              verbose           = TRUE) {


  if (!requireNamespace("randomForest", quietly = TRUE)) {
    stop(paste0("ERROR: The 'randomForest' package is required for build_temporal_rf(). ",
                "Install it with install.packages('randomForest')."))
  }

  if (missing(partition_result)) {
    stop(paste0("ERROR: 'partition_result' is required. Provide output from ",
                "spatiotemporal_partition() or a path to an .rds file."))
  }

  if (missing(pseudoabsence_result)) {
    stop(paste0("ERROR: 'pseudoabsence_result' is required. Provide output from ",
                "generate_absences() or a path to an .rds file."))
  }

  if (missing(model_vars) || !is.character(model_vars) || length(model_vars) == 0) {
    stop("ERROR: 'model_vars' must be a non-empty character vector of predictor names.")
  }

  thr_resolved     <- .resolve_threshold_method(threshold_method)
  threshold_method <- thr_resolved$method
  manual_threshold <- thr_resolved$manual_value

  partition_result     <- .load_partition_result(partition_result, verbose = verbose)
  pseudoabsence_result <- .load_pseudoabsence_result(pseudoabsence_result, verbose = verbose)

  prep             <- .prepare_combined_modeling_data(partition_result, pseudoabsence_result,
                                                      model_vars = model_vars, time_cols = time_cols,
                                                      verbose = verbose)
  combined_df      <- prep$combined_df
  all_folds        <- prep$all_folds
  single_fold_mode <- prep$single_fold_mode

  if (verbose) message(paste0("Building Random Forest for ", length(all_folds), " fold(s)."))
  if (verbose) message(paste("Predictors:", paste(model_vars, collapse = ", ")))

  if (!dir.exists(output_dir)) {
    dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
  }

  combined_file <- file.path(output_dir, "all_rf_models.rds")

  model_list      <- list()
  thresholds <- numeric(length(all_folds))
  names(thresholds) <- paste0("fold", all_folds)
  train_data_list <- list()
  importance_list <- list()
  metric_rows     <- list()

  if (file.exists(combined_file) && !overwrite) {

    if (verbose) message(paste("Loading existing RF model file:", basename(combined_file)))
    saved <- tryCatch(
      readRDS(combined_file),
      error = function(e) stop(paste0("ERROR loading saved models. Try overwrite = TRUE: ", e$message))
    )
    model_list      <- saved$models
    thresholds      <- saved$thresholds
    train_data_list <- saved$fold_training_data
    importance_list <- saved$variable_importance
    metric_rows     <- saved$fold_test_metrics_raw

  } else {

    for (fold in all_folds) {

      fold_key   <- paste0("fold", fold)
      train_data <- if (single_fold_mode) combined_df else combined_df[combined_df$fold != fold, , drop = FALSE]
      test_data  <- combined_df[combined_df$fold == fold, , drop = FALSE]

      n_pr_train <- sum(train_data$presence == 1, na.rm = TRUE)
      n_pa_train <- sum(train_data$presence == 0, na.rm = TRUE)
      n_pr_test  <- sum(test_data$presence == 1, na.rm = TRUE)
      n_pa_test  <- sum(test_data$presence == 0, na.rm = TRUE)

      if (verbose) message(paste0(
        "Fold ", fold, ": training on ", nrow(train_data), " points (",
        n_pr_train, " presences, ", n_pa_train, " pseudoabsences), testing on ",
        nrow(test_data), " (", n_pr_test, " presences, ", n_pa_test, " pseudoabsences)."
      ))

      if (n_pr_train < 5) {
        stop(paste0("ERROR: Fold ", fold, " has only ", n_pr_train, " training presences. Minimum 5 required."))
      }

      if (n_pa_train < 1) {
        stop(paste0("ERROR: Fold ", fold, " has no training pseudoabsences. Check generate_absences() output."))
      }

      train_rf         <- train_data[, model_vars, drop = FALSE]
      train_rf$presence <- factor(train_data$presence, levels = c(0, 1))

      rf_args <- c(
        list(
          formula    = presence ~ .,
          data       = train_rf,
          importance = TRUE
        ),
        rf_params
      )

      model_fit <- tryCatch(
        do.call(randomForest::randomForest, rf_args),
        error = function(e) stop(paste0("ERROR: randomForest() failed for fold ", fold, ": ", e$message))
      )

      train_pred_mat  <- stats::predict(model_fit, newdata = train_rf, type = "prob")
      train_pred_prob <- train_pred_mat[, "1"]

      thr <- {
        .select_threshold_prob(
          pred_prob    = train_pred_prob,
          observed     = train_data$presence,
          method       = threshold_method,
          manual_value = manual_threshold
        )
      }
      thresholds[fold_key] <- thr

      model_list[[fold_key]]      <- model_fit
      train_data_list[[fold_key]] <- train_data

      imp_raw  <- randomForest::importance(model_fit, type = 1)
      imp_gini <- randomForest::importance(model_fit, type = 2)
      importance_list[[fold_key]] <- data.frame(
        variable         = rownames(imp_raw),
        mean_decr_acc    = imp_raw[, 1],
        mean_decr_gini   = imp_gini[, 1],
        row.names        = NULL,
        stringsAsFactors = FALSE
      )

      if (nrow(test_data) > 0) {
        test_rf <- test_data
        test_rf$presence <- factor(test_rf$presence, levels = c(0, 1))
        test_pred_mat  <- stats::predict(model_fit, newdata = test_rf, type = "prob")
        test_pred_prob <- test_pred_mat[, "1"]
        test_pred_bin  <- as.integer(test_pred_prob >= thr)
        m <- .compute_confusion_metrics(
          observed  = test_data$presence,
          predicted = test_pred_bin,
          pred_prob = test_pred_prob
        )
      } else {
        m <- list(auc = NA, tss = NA, sensitivity = NA, specificity = NA,
                  kappa = NA, tp = NA, fn = NA, tn = NA, fp = NA)
      }

      metric_rows[[fold_key]] <- .build_fold_metric_row(fold, thr, m)
    }

    tryCatch({
      saveRDS(
        list(models = model_list, thresholds = thresholds,
             fold_training_data = train_data_list,
             variable_importance = importance_list,
             fold_test_metrics_raw = metric_rows),
        combined_file
      )
      if (verbose) message(paste("Saved RF models to:", basename(combined_file)))
    }, error = function(e) {
      warning(paste0("Could not save RF model file: ", e$message))
    })
  }

  fold_metrics_df           <- do.call(rbind, metric_rows)
  rownames(fold_metrics_df) <- NULL

  metrics_file <- file.path(output_dir, "Fold_Test_Metrics.csv")
  tryCatch(
    utils::write.csv(fold_metrics_df, metrics_file, row.names = FALSE),
    error = function(e) warning(paste0("Could not save fold test metrics: ", e$message))
  )

  plot_list <- list()

  if (create_plot) {
    if (verbose) message("Generating diagnostic plots...")

    fold_colors <- .resolve_palette(length(all_folds), plot_palette)
    names(fold_colors) <- paste0("fold", all_folds)

    imp_file <- file.path(output_dir, "VariableImportance.png")

    tryCatch({
      n_vars_imp  <- length(model_vars)
      n_folds_imp <- length(all_folds)
      n_cols_imp  <- min(n_vars_imp, 3)
      n_rows_imp  <- ceiling(n_vars_imp / n_cols_imp)
      png_w <- 280 * n_cols_imp + 60
      png_h <- (220 + n_folds_imp * 25) * n_rows_imp + 80
      grDevices::png(imp_file, width = png_w, height = png_h, res = 120)
      .plot_rf_importance(model_vars, all_folds, importance_list, fold_colors)
      grDevices::dev.off()
      if (verbose) message(paste("Saved variable importance plot:", basename(imp_file)))
    }, error = function(e) {
      warning(paste0("Could not save variable importance plot: ", e$message))
      if (grDevices::dev.cur() > 1) grDevices::dev.off()
    })

    tryCatch({
      .plot_rf_importance(model_vars, all_folds, importance_list, fold_colors)
      plot_list[["variable_importance"]] <- grDevices::recordPlot()
    }, error = function(e) {
      warning(paste0("Could not record variable importance plot: ", e$message))
    })

    for (fold in all_folds) {
      fold_key   <- paste0("fold", fold)
      model_fit  <- model_list[[fold_key]]
      train_data <- train_data_list[[fold_key]]
      thr <- thresholds[fold_key]

      plot_file <- file.path(output_dir, paste0("MarginalCurves_Fold", fold, ".png"))
      n_vars    <- length(model_vars)
      n_cols    <- min(3, n_vars)
      n_rows    <- ceiling(n_vars / n_cols)

      var_means <- colMeans(train_data[, model_vars, drop = FALSE], na.rm = TRUE)

      tryCatch({
        grDevices::png(plot_file, width = 350 * n_cols, height = 320 * n_rows, res = 120)
        .plot_rf_marginal_curves(model_fit, train_data, model_vars,
                                 var_means, fold_colors[fold_key], thr,
                                 fold, n_rows, n_cols)
        grDevices::dev.off()
        if (verbose) message(paste("Saved marginal curve plot:", basename(plot_file)))
      }, error = function(e) {
        warning(paste0("Could not save marginal curve plot for Fold ", fold, ": ", e$message))
        if (grDevices::dev.cur() > 1) grDevices::dev.off()
      })

      tryCatch({
        .plot_rf_marginal_curves(model_fit, train_data, model_vars,
                                 var_means, fold_colors[fold_key], thr,
                                 fold, n_rows, n_cols)
        plot_list[[paste0("marginal_curves_fold", fold)]] <- grDevices::recordPlot()
      }, error = function(e) {
        warning(paste0("Could not record marginal curve plot for Fold ", fold, ": ", e$message))
      })
    }

  }

  if (verbose) message("\nRandom Forest modeling complete.")
  if (verbose) message(paste0("Threshold: ",
                              threshold_method,
                              " | Folds: ", length(all_folds)))
  if (nrow(fold_metrics_df) > 0 && verbose) {
    message(paste("Metrics saved to:", basename(metrics_file)))
    message(paste(utils::capture.output(
      print(fold_metrics_df, row.names = FALSE)), collapse = "\n"))
  }

  if (create_plot && nrow(fold_metrics_df) > 0) {
    plot_rate_cols <- c("Sensitivity", "Specificity", "TSS", "Kappa", "AUC")
    plot_rate_cols <- plot_rate_cols[plot_rate_cols %in% names(fold_metrics_df) &
                                       !vapply(fold_metrics_df[plot_rate_cols],
                                               function(x) all(is.na(x)), logical(1))]
    if (length(plot_rate_cols) > 0) {
      bar_file <- file.path(output_dir, "Fold_Test_Metrics.png")
      tryCatch({
        grDevices::png(bar_file, width = 300 * length(plot_rate_cols) + 200,
                       height = 700, res = 150)
        .plot_fold_metrics_bars(
          metrics_df  = fold_metrics_df,
          rate_cols   = plot_rate_cols,
          fold_colors = fold_colors,
          title       = "Random Forest Time Independent Assessment Metrics"
        )
        grDevices::dev.off()
        if (verbose) message(paste("Saved fold metrics plot:", basename(bar_file)))
      }, error = function(e) {
        warning(paste0("Could not save fold metrics plot: ", e$message))
        if (grDevices::dev.cur() > 1) grDevices::dev.off()
      })
      tryCatch({
        .plot_fold_metrics_bars(
          metrics_df  = fold_metrics_df,
          rate_cols   = plot_rate_cols,
          fold_colors = fold_colors,
          title       = "Random Forest Time Independent Assessment Metrics"
        )
        plot_list[["fold_test_metrics"]] <- grDevices::recordPlot()
      }, error = function(e) {
        warning(paste0("Could not record fold metrics plot: ", e$message))
      })
    }
  }

  result <- list(
    models              = model_list,
    thresholds          = thresholds,
    threshold_method    = threshold_method,
    model_vars          = model_vars,
    variable_importance = importance_list,
    fold_training_data  = train_data_list,
    fold_test_metrics   = fold_metrics_df,
    output_dir          = output_dir,
    model_type          = "rf",
    plots               = plot_list
  )

  class(result) <- c("TemporalRF", "list")
  invisible(result)
}

#' Summarize Prediction Rasters into Consensus Outputs
#'
#' Post-processing function that synthesizes per-time-step fold-vote rasters
#' (output from \code{\link{generate_spatiotemporal_predictions}}) into binary
#' consensus predictions and a temporal frequency summary. Each input raster
#' contains integer vote counts per pixel the number of cross-validation
#' folds that classified that pixel as suitable. The \code{consensus} threshold
#' controls how many folds must agree for a pixel to be classified as suitable
#' in the output binary rasters.
#'
#' @usage
#' summarize_raster_outputs(predictions_dir, output_dir = NULL,
#'                          consensus = 1, file_pattern = "Prediction_.*\\\.tif$",
#'                          overwrite = FALSE, verbose = TRUE)
#'
#' @param predictions_dir Character. Directory containing prediction raster
#'   files, typically the \code{output_dir} used in
#'   \code{\link{generate_spatiotemporal_predictions}}.
#' @param output_dir Character. Directory for output files. Defaults to
#'   \code{predictions_dir} if \code{NULL}.
#' @param consensus Integer. Minimum number of folds that must agree on
#'   suitability for a pixel to be classified as suitable in the binary output.
#'   For example, \code{consensus = 1} marks any pixel suitable if at least one
#'   fold predicts it suitable; \code{consensus = 4} requires at least four folds to
#'   agree. Must be between 1 and the maximum vote count in the rasters (number of
#'   folds). Default is \code{1}.
#' @param file_pattern Character. Regular expression to match prediction raster
#'   files. Default is \code{"Prediction_.*\\.tif$"}.
#' @param overwrite Logical. If \code{TRUE}, overwrites existing output files.
#'   If \code{FALSE} (default), existing files are skipped.
#' @param verbose Logical. If \code{TRUE} (default), prints progress
#'   messages during processing.
#'
#' @return Invisibly returns a list containing:
#' \itemize{
#'   \item \code{consensus_stack}: \code{SpatRaster} stack of per-time-step
#'     binary consensus rasters (one layer per input file).
#'   \item \code{frequency_raster}: \code{SpatRaster} showing the proportion of
#'     time steps during which each pixel met the consensus threshold. Values
#'     range from 0 (never suitable) to 1 (always suitable).
#'   \item \code{consensus}: the consensus threshold used.
#'   \item \code{n_timesteps}: number of time steps processed.
#'   \item \code{consensus_dir}: path to the directory containing per-time-step
#'     binary consensus rasters.
#'   \item \code{frequency_file}: path to the written frequency raster file.
#' }
#'
#' @details
#' Input rasters are fold-vote-count rasters produced by
#' \code{\link{generate_spatiotemporal_predictions}}, where each pixel value is
#' the number of cross-validation fold models that predicted suitability at that
#' location for that time step. The \code{consensus} threshold is applied as:
#' \code{binary = as.integer(vote_count >= consensus)}.
#'
#' The frequency raster (proportion of time steps suitable under the chosen
#' consensus) serves as input to \code{\link[TemporalModelR]{analyze_temporal_patterns}} for
#' identifying long-term trends in suitability.
#'
#' @seealso
#' Upstream: \code{\link{generate_spatiotemporal_predictions}}
#'
#' Downstream: \code{\link[TemporalModelR]{analyze_temporal_patterns}}
#'
#' @examples
#' pred_dir <- system.file("extdata/predictions",
#'                         package = "TemporalModelR")
#'
#' summarize_raster_outputs(
#'   predictions_dir = pred_dir,
#'   output_dir      = tempdir(),
#'   consensus       = 3,
#'   overwrite       = TRUE,
#'   verbose         = FALSE
#' )

#' @export
#' @importFrom terra rast nlyr app writeRaster values global
summarize_raster_outputs <- function(predictions_dir,
                                     output_dir   = NULL,
                                     consensus    = 1,
                                     file_pattern = "Prediction_.*\\.tif$",
                                     overwrite    = FALSE,
                                     verbose      = TRUE) {

  if (missing(predictions_dir) || is.null(predictions_dir) || predictions_dir == "") {
    stop("ERROR: 'predictions_dir' is required and cannot be NULL or empty.")
  }
  if (!dir.exists(predictions_dir)) {
    stop(paste0("ERROR: predictions_dir does not exist: ", predictions_dir))
  }
  if (!is.numeric(consensus) || length(consensus) != 1 || consensus < 1) {
    stop("ERROR: 'consensus' must be a single positive integer >= 1.")
  }
  consensus <- as.integer(consensus)

  prediction_files <- list.files(
    path       = predictions_dir,
    pattern    = file_pattern,
    full.names = TRUE
  )

  if (length(prediction_files) == 0) {
    stop(paste0("ERROR: No files matching '", file_pattern,
                "' found in: ", predictions_dir))
  }

  prediction_files <- .natural_sort(prediction_files)
  if (verbose) message(paste("Found", length(prediction_files), "prediction rasters."))

  if (is.null(output_dir)) output_dir <- predictions_dir
  if (!dir.exists(output_dir)) dir.create(output_dir, recursive = TRUE,
                                          showWarnings = FALSE)

  consensus_dir  <- file.path(output_dir,
                              paste0("Consensus_", consensus, "_Rasters"))
  frequency_file <- file.path(output_dir,
                              paste0("Frequency_Consensus_", consensus, ".tif"))

  if (dir.exists(consensus_dir) && file.exists(frequency_file) && !overwrite) {
    if (verbose) message("Outputs already exist and overwrite = FALSE. Reloading from disk.")
    con_files <- .natural_sort(list.files(consensus_dir, pattern = "\\.tif$",
                                          full.names = TRUE))
    con_stack      <- terra::rast(con_files)
    frequency_rast <- terra::rast(frequency_file)
    return(invisible(list(
      consensus_stack  = con_stack,
      frequency_raster = frequency_rast,
      consensus        = consensus,
      n_timesteps      = terra::nlyr(con_stack),
      consensus_dir    = consensus_dir,
      frequency_file   = frequency_file
    )))
  }

  if (verbose) message("Loading prediction rasters...")
  vote_stack  <- terra::rast(prediction_files)
  clean_names <- gsub("\\.tif$", "", basename(prediction_files))
  names(vote_stack) <- clean_names

  max_votes <- max(vapply(seq_len(terra::nlyr(vote_stack)), function(i) {
    v <- terra::global(vote_stack[[i]], "max", na.rm = TRUE)[1, 1]
    if (is.na(v)) 0 else as.integer(v)
  }, integer(1)), na.rm = TRUE)

  if (verbose) message(paste0("Maximum fold vote count detected: ", max_votes,
                              " | Consensus threshold: ", consensus))

  if (consensus > max_votes) {
    stop(paste0(
      "ERROR: consensus = ", consensus, " exceeds the maximum vote count (",
      max_votes, ") in the prediction rasters. ",
      "Reduce consensus to <= ", max_votes, "."
    ))
  }

  if (!dir.exists(consensus_dir)) dir.create(consensus_dir, recursive = TRUE,
                                             showWarnings = FALSE)

  if (verbose) message(paste0("Applying consensus threshold (>= ", consensus,
                              " folds) to ", terra::nlyr(vote_stack), " time steps..."))

  consensus_files <- character(terra::nlyr(vote_stack))

  for (i in seq_len(terra::nlyr(vote_stack))) {
    layer <- vote_stack[[i]]
    con_layer <- terra::app(layer, function(x) {
      ifelse(is.na(x), NA_integer_, as.integer(x >= consensus))
    })

    con_file  <- file.path(consensus_dir,
                           paste0(clean_names[i], "_consensus", consensus, ".tif"))
    terra::writeRaster(con_layer, con_file, overwrite = TRUE,
                       datatype = "INT1U", NAflag = 255)
    consensus_files[i] <- con_file
    if (verbose) message(paste0("  [", i, "/", terra::nlyr(vote_stack), "] ",
                                clean_names[i], " done."))
  }

  con_stack <- terra::rast(consensus_files)
  names(con_stack) <- clean_names

  if (verbose) message("Calculating frequency raster (proportion of time steps suitable)...")
  frequency_rast <- terra::app(con_stack, mean, na.rm = TRUE)

  terra::writeRaster(frequency_rast, frequency_file, overwrite = TRUE)
  if (verbose) message(paste("Saved frequency raster:", basename(frequency_file)))

  n_suitable_always <- sum(terra::values(frequency_rast) == 1, na.rm = TRUE)
  n_suitable_ever   <- sum(terra::values(frequency_rast) >  0, na.rm = TRUE)
  n_total           <- sum(!is.na(terra::values(frequency_rast)))

  if (verbose) message("--- Summarize Raster Outputs Complete ---")
  if (verbose) message(paste0("  Consensus threshold : ", consensus, " of ", max_votes, " folds"))
  if (verbose) message(paste0("  Time steps processed: ", terra::nlyr(vote_stack)))
  if (verbose) message(paste0("  Ever suitable       : ", n_suitable_ever, " / ", n_total,
                              " pixels (", round(n_suitable_ever / n_total * 100, 1), "%)"))
  if (verbose) message(paste0("  Always suitable     : ", n_suitable_always, " / ", n_total,
                              " pixels (", round(n_suitable_always / n_total * 100, 1), "%)"))
  if (verbose) message(paste0("  Output dir          : ", output_dir))

  invisible(list(
    consensus_stack    = con_stack,
    frequency_raster   = frequency_rast,
    consensus          = consensus,
    n_timesteps        = terra::nlyr(vote_stack),
    consensus_dir      = consensus_dir,
    frequency_file     = frequency_file
  ))
}

#' Build Temporal GAM Models Across Cross-Validation Folds
#'
#' Modeling function that constructs binomial generalized additive models (GAMs)
#' for each cross-validation fold using presence and pseudoabsence data. Each
#' model reserves one fold as testing data and uses the remaining folds as
#' training data. The user supplies the model formula directly using standard
#' \pkg{mgcv} formula syntax, including smooth terms such as \code{s()},
#' \code{te()}, and \code{ti()}. Supports automatic or manual probability
#' thresholding for converting continuous predictions to binary suitability
#' classifications necessary for downstream analyses. The returned object follows the same
#' structure as \code{\link{build_temporal_glm}}, \code{\link{build_temporal_hv}},
#' and \code{\link{build_temporal_rf}}, and is accepted directly by
#' \code{\link{generate_spatiotemporal_predictions}}.
#'
#' @usage
#' build_temporal_gam(partition_result, pseudoabsence_result, model_formula,
#'                    link = "logit", gam_params = list(method = "REML"),
#'                    threshold_method  = "tss",
#'                    output_dir = file.path(tempdir(), "GAM_Models"),
#'                    create_plot = TRUE, plot_palette = "Dark 2",
#'                    overwrite = FALSE, time_cols = NULL, verbose = TRUE)
#'
#' @param partition_result List or character. Output from
#'   \code{\link{spatiotemporal_partition}} or path to an \code{.rds}
#'   file containing that output.
#' @param pseudoabsence_result List or character. Output from
#'   \code{\link{generate_absences}} or path to an \code{.rds} file
#'   containing that output.
#' @param model_formula Formula or character. The right-hand side of the model
#'   formula supplied as either a formula object or a character string. The
#'   response variable (\code{presence}) is always added automatically on the
#'   left-hand side, so only the right-hand side needs to be provided. Both of
#'   the following are accepted and equivalent:
#'   \itemize{
#'     \item \code{~ s(Var1) + s(Var2) + Var3}
#'     \item \code{"~ s(Var1) + s(Var2) + Var3"}
#'   }
#'   Standard \pkg{mgcv} formula syntax applies. Smooth terms are specified
#'   with \code{s()} for univariate smooths, \code{te()} for tensor product
#'   smooths of two or more variables, and \code{ti()} for tensor product
#'   interaction terms. Parametric terms can be included alongside smooth terms
#'   using \code{+}. The basis type and dimension can be controlled via
#'   arguments to \code{s()}, e.g. \code{s(Var1, k = 5, bs = "tp")}. All
#'   predictor names referenced in the formula must be present as columns in
#'   both the presence and pseudoabsence data.
#' @param link Character. The link function for the binomial GAM. One of
#'   \code{"logit"} (default), \code{"probit"}, \code{"cloglog"}, or
#'   \code{"cauchit"}. See \code{\link[stats]{binomial}} for details on each
#'   link function.
#' @param gam_params Named list. Additional arguments passed to
#'   \code{\link[mgcv]{gam}}, such as \code{method} for the smoothing
#'   parameter estimation method (e.g. \code{"REML"}) or
#'   \code{select} for additional shrinkage. Default is
#'   \code{list(method = "REML")}.
#' @param threshold_method Character or numeric. Method used to convert
#'   continuous predicted probabilities to binary suitability. Accepted values:
#'   \itemize{
#'     \item \code{"prevalence"}: Sets threshold equal to the prevalence
#'       (proportion of presences) in the training data for that fold.
#'     \item \code{"tss"}: Selects the threshold that maximizes the True Skill
#'       Statistic (sensitivity + specificity - 1) on the training data.
#'       Default.
#'     \item A numeric value between 0 and 1 (e.g. \code{0.4}): Uses that
#'       value as a fixed threshold for all folds directly.
#'   }
#' @param output_dir Character. Directory to write output files including saved
#'   model objects and plots. Default is \code{file.path(tempdir(), "GAM_Models")}.
#' @param create_plot Logical. If \code{TRUE}, generates per-fold response
#'   curve plots and a combined ROC curve summary. Default is \code{TRUE}.
#' @param overwrite Logical. If \code{TRUE}, overwrites existing saved model
#'   files. If \code{FALSE}, loads existing files when available. Default is
#'   \code{FALSE}.
#' @param plot_palette Character. Name of an HCL or RColorBrewer palette used
#'   to color folds in diagnostic plots. Accepts any HCL palette name (see
#'   \code{\link[grDevices]{hcl.pals}}) or, if \pkg{RColorBrewer} is installed,
#'   any Brewer palette name. Default is \code{"Dark 2"}.
#' @param time_cols Character. Name of the column(s) containing year or time
#'   step values in the occurrence data. Must match \code{time_cols} used in
#'   \code{\link{spatiotemporal_partition}}. Default is \code{NULL}.
#' @param verbose Logical. If \code{TRUE} (default), prints progress
#'   messages during processing. Includes per-fold training summaries and
#'   file-saved messages. The completion summary and metrics table are
#'   always printed regardless of this setting.
#'
#' @return A list with class \code{"TemporalGAM"} containing:
#' \itemize{
#'   \item \code{models}: Named list of fitted \code{gam} objects, one per
#'     fold.
#'   \item \code{thresholds}: Named numeric vector of probability thresholds
#'     used for binary classification, one per fold.
#'   \item \code{threshold_method}: Character string recording the thresholding
#'     method used.
#'   \item \code{model_formula}: The formula object as passed to the fitting
#'     function.
#'   \item \code{link}: Character string recording the link function used.
#'   \item \code{model_vars}: Character vector of predictor names extracted
#'     from the formula right-hand side.
#'   \item \code{fold_training_data}: Named list of training data frames used
#'     to fit each fold model, retained for downstream prediction.
#'   \item \code{fold_test_metrics}: Data frame of held-out test fold metrics
#'     per fold: \code{Threshold}, \code{AUC}, \code{TSS}, \code{Kappa},
#'     \code{Sensitivity}, and \code{Specificity}. Also written to
#'     \code{Fold_Test_Metrics.csv} in \code{output_dir}.
#'   \item \code{output_dir}: Path to the output directory.
#'   \item \code{model_type}: Character string \code{"gam"}, used by
#'     \code{\link{generate_spatiotemporal_predictions}}.
#'   \item \code{plots}: Named list of recorded plot objects when
#'     \code{create_plot = TRUE}. Plots can be replayed with
#'     \code{grDevices::replayPlot()}.
#' }
#'
#' @details
#'
#' GAMs are fit using \code{\link[mgcv]{gam}} from the \pkg{mgcv} package with
#' \code{family = binomial(link = link)}. Smooth terms default to thin plate
#' regression splines (\code{bs = "tp"}) with the basis dimension \code{k}
#' chosen automatically by \pkg{mgcv} unless specified in the formula. Smoothing
#' parameters are estimated by REML by default.
#'
#' The returned object is recognized by
#' \code{\link{generate_spatiotemporal_predictions}}, which uses the
#' \code{model_type} field to use the correct prediction and evaluation
#' logic.
#'
#'
#' @seealso
#' Preprocessing: \code{\link{spatiotemporal_partition}},
#'   \code{\link{generate_absences}}
#'
#' Modeling: \code{\link{build_temporal_glm}}, \code{\link{build_temporal_rf}},
#'   \code{\link{build_temporal_hv}},
#'   \code{\link{generate_spatiotemporal_predictions}}
#'
#' External: \code{\link[mgcv]{gam}}, \code{\link[mgcv]{s}}
#'
#' @examples
#' data(tmr_partition, package = "TemporalModelR")
#'
#' data(tmr_absences,  package = "TemporalModelR")
#'
#' build_temporal_gam(
#'   partition_result     = tmr_partition,
#'   pseudoabsence_result = tmr_absences,
#'   model_formula        = ~ s(elevation) + s(forest_cover) + s(prseas),
#'   threshold_method     = "tss",
#'   output_dir           = tempdir(),
#'   create_plot          = FALSE,
#'   time_cols            = c("year", "season"),
#'   verbose              = FALSE
#' )

#' @export
#' @importFrom sf st_drop_geometry
#' @importFrom stats binomial predict var complete.cases as.formula
#' @importFrom grDevices adjustcolor dev.cur dev.off png recordPlot
#' @importFrom graphics plot lines abline legend par mtext points rug
#' @importFrom tools file_ext
#' @importFrom utils capture.output write.csv
build_temporal_gam <- function(partition_result,
                               pseudoabsence_result,
                               model_formula,
                               link              = "logit",
                               gam_params        = list(method = "REML"),
                               threshold_method  = "tss",
                               output_dir        = file.path(tempdir(), "GAM_Models"),
                               create_plot       = TRUE,
                               plot_palette      = "Dark 2",
                               overwrite         = FALSE,
                               time_cols         = NULL,
                               verbose           = TRUE) {

  if (!requireNamespace("mgcv", quietly = TRUE)) {
    stop(paste0("ERROR: The 'mgcv' package is required for build_temporal_gam(). ",
                "Install it with install.packages('mgcv')."))
  }

  ### Input validation

  if (missing(partition_result)) {
    stop(paste0("ERROR: 'partition_result' is required. Provide output from ",
                "spatiotemporal_partition() or a path to an .rds file."))
  }

  if (missing(pseudoabsence_result)) {
    stop(paste0("ERROR: 'pseudoabsence_result' is required. Provide output from ",
                "generate_absences() or a path to an .rds file."))
  }

  if (missing(model_formula)) {
    stop(paste0(
      "ERROR: 'model_formula' is required. Provide the right-hand side of the model formula as a ",
      "formula object or character string, e.g. ~ s(Var1) + s(Var2)."
    ))
  }

  if (!link %in% c("logit", "probit", "cloglog", "cauchit")) {
    stop("ERROR: 'link' must be one of 'logit', 'probit', 'cloglog', or 'cauchit'.")
  }

  thr_resolved     <- .resolve_threshold_method(threshold_method)
  threshold_method <- thr_resolved$method
  manual_threshold <- thr_resolved$manual_value

  model_formula <- .parse_model_formula(model_formula)
  rhs_vars      <- all.vars(model_formula)

  family_obj   <- stats::binomial(link = link)
  family_label <- paste0("binomial(", link, ")")

  partition_result     <- .load_partition_result(partition_result, verbose = verbose)
  pseudoabsence_result <- .load_pseudoabsence_result(pseudoabsence_result, verbose = verbose)

  prep             <- .prepare_combined_modeling_data(partition_result, pseudoabsence_result,
                                                      model_vars = rhs_vars, time_cols = time_cols,
                                                      verbose = verbose)
  combined_df      <- prep$combined_df
  all_folds        <- prep$all_folds
  single_fold_mode <- prep$single_fold_mode

  rhs_str      <- gsub("\\s+", " ", paste(deparse(model_formula[[2]]), collapse = " "))
  full_formula <- stats::as.formula(paste("presence ~", rhs_str))

  if (verbose) message(paste0("Building GAM (", family_label, ") for ", length(all_folds), " fold(s)."))
  if (verbose) message(paste("Formula:", gsub("\\s+", " ", paste(deparse(full_formula), collapse = " "))))

  if (!dir.exists(output_dir)) {
    dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
  }

  combined_file <- file.path(output_dir, paste0("all_gam_models_", link, ".rds"))

  model_list      <- list()
  thresholds <- numeric(length(all_folds))
  names(thresholds) <- paste0("fold", all_folds)
  train_data_list <- list()
  metric_rows     <- list()

  if (file.exists(combined_file) && !overwrite) {

    if (verbose) message(paste("Loading existing GAM model file:", basename(combined_file)))
    saved <- tryCatch(
      readRDS(combined_file),
      error = function(e) stop(paste0("ERROR loading saved models. Try overwrite = TRUE: ", e$message))
    )
    model_list      <- saved$models
    thresholds      <- saved$thresholds
    train_data_list <- saved$fold_training_data
    metric_rows     <- saved$fold_test_metrics_raw

  } else {

    for (fold in all_folds) {

      fold_key   <- paste0("fold", fold)
      train_data <- if (single_fold_mode) combined_df else combined_df[combined_df$fold != fold, , drop = FALSE]
      test_data  <- combined_df[combined_df$fold == fold, , drop = FALSE]

      n_pr_train <- sum(train_data$presence == 1, na.rm = TRUE)
      n_pa_train <- sum(train_data$presence == 0, na.rm = TRUE)
      n_pr_test  <- sum(test_data$presence == 1, na.rm = TRUE)
      n_pa_test  <- sum(test_data$presence == 0, na.rm = TRUE)

      if (verbose) message(paste0(
        "Fold ", fold, ": training on ", nrow(train_data), " points (",
        n_pr_train, " presences, ", n_pa_train, " pseudoabsences), testing on ",
        nrow(test_data), " (", n_pr_test, " presences, ", n_pa_test, " pseudoabsences)."
      ))

      if (n_pr_train < 5) {
        stop(paste0("ERROR: Fold ", fold, " has only ", n_pr_train, " training presences. Minimum 5 required."))
      }

      if (n_pa_train < 1) {
        stop(paste0("ERROR: Fold ", fold, " has no training pseudoabsences. Check generate_absences() output."))
      }

      var_check <- vapply(train_data[, rhs_vars, drop = FALSE], function(x) stats::var(x, na.rm = TRUE), numeric(1))
      zero_var  <- names(var_check[!is.na(var_check) & var_check == 0])
      if (length(zero_var) > 0) {
        warning(paste0("Zero variance in Fold ", fold, " for: ", paste(zero_var, collapse = ", "),
                       ". Smooth terms on these variables may fail."))
      }

      gam_args <- c(
        list(formula = full_formula, data = train_data, family = family_obj),
        gam_params
      )

      model_fit <- tryCatch(
        do.call(mgcv::gam, gam_args),
        error = function(e) stop(paste0("ERROR: gam() failed for fold ", fold, ": ", e$message))
      )

      thr <- {
        .select_threshold_prob(
          pred_prob    = stats::predict(model_fit, newdata = train_data, type = "response"),
          observed     = train_data$presence,
          method       = threshold_method,
          manual_value = manual_threshold
        )
      }
      thresholds[fold_key] <- thr

      model_list[[fold_key]]      <- model_fit
      train_data_list[[fold_key]] <- train_data

      if (nrow(test_data) > 0) {
        test_pred_prob <- stats::predict(model_fit, newdata = test_data, type = "response")
        test_pred_bin  <- as.integer(test_pred_prob >= thr)
        m <- .compute_confusion_metrics(
          observed  = test_data$presence,
          predicted = test_pred_bin,
          pred_prob = test_pred_prob
        )
      } else {
        m <- list(auc = NA, tss = NA, sensitivity = NA, specificity = NA,
                  kappa = NA, tp = NA, fn = NA, tn = NA, fp = NA)
      }

      metric_rows[[fold_key]] <- .build_fold_metric_row(fold, thr, m)
    }

    tryCatch({
      saveRDS(
        list(models = model_list, thresholds = thresholds,
             fold_training_data = train_data_list, fold_test_metrics_raw = metric_rows),
        combined_file
      )
      if (verbose) message(paste("Saved GAM models to:", basename(combined_file)))
    }, error = function(e) {
      warning(paste0("Could not save GAM model file: ", e$message))
    })
  }

  fold_metrics_df           <- do.call(rbind, metric_rows)
  rownames(fold_metrics_df) <- NULL

  metrics_file <- file.path(output_dir, "Fold_Test_Metrics.csv")
  tryCatch(
    utils::write.csv(fold_metrics_df, metrics_file, row.names = FALSE),
    error = function(e) warning(paste0("Could not save fold test metrics: ", e$message))
  )

  plot_list <- list()

  if (create_plot) {
    if (verbose) message("Generating diagnostic plots...")

    fold_colors <- .resolve_palette(length(all_folds), plot_palette)
    names(fold_colors) <- paste0("fold", all_folds)

    for (fold in all_folds) {
      fold_key   <- paste0("fold", fold)
      model_fit  <- model_list[[fold_key]]
      train_data <- train_data_list[[fold_key]]
      thr <- thresholds[fold_key]

      plot_file <- file.path(output_dir, paste0("ResponseCurves_Fold", fold, ".png"))
      n_vars    <- length(rhs_vars)
      n_cols    <- min(3, n_vars)
      n_rows    <- ceiling(n_vars / n_cols)
      var_means <- colMeans(train_data[, rhs_vars, drop = FALSE], na.rm = TRUE)

      tryCatch({
        grDevices::png(plot_file, width = 350 * n_cols, height = 320 * n_rows, res = 120)
        .plot_response_curves_parametric(model_fit, train_data, rhs_vars,
                                         var_means, fold_colors[fold_key], thr,
                                         fold, n_rows, n_cols, "GAM")
        grDevices::dev.off()
        if (verbose) message(paste("Saved response curve plot:", basename(plot_file)))
      }, error = function(e) {
        warning(paste0("Could not save response curve plot for Fold ", fold, ": ", e$message))
        if (grDevices::dev.cur() > 1) grDevices::dev.off()
      })

      tryCatch({
        .plot_response_curves_parametric(model_fit, train_data, rhs_vars,
                                         var_means, fold_colors[fold_key], thr,
                                         fold, n_rows, n_cols, "GAM")
        plot_list[[paste0("response_curves_fold", fold)]] <- grDevices::recordPlot()
      }, error = function(e) {
        warning(paste0("Could not record response curve plot for Fold ", fold, ": ", e$message))
      })
    }

    roc_file <- file.path(output_dir, "ROC_Summary.png")

    tryCatch({
      grDevices::png(roc_file, width = 600, height = 580, res = 120)
      .plot_roc_summary_parametric(model_list, train_data_list,
                                   thresholds, fold_metrics_df,
                                   all_folds, fold_colors, "GAM")
      grDevices::dev.off()
      if (verbose) message(paste("Saved ROC summary plot:", basename(roc_file)))
    }, error = function(e) {
      warning(paste0("Could not save ROC summary plot: ", e$message))
      if (grDevices::dev.cur() > 1) grDevices::dev.off()
    })

    tryCatch({
      .plot_roc_summary_parametric(model_list, train_data_list,
                                   thresholds, fold_metrics_df,
                                   all_folds, fold_colors, "GAM")
      plot_list[["roc_summary"]] <- grDevices::recordPlot()
    }, error = function(e) {
      warning(paste0("Could not record ROC plot: ", e$message))
    })
  }

  if (verbose) message("\nGAM modeling complete.")
  if (verbose) message(paste0("Link: ", link, " | Threshold: ",
                              threshold_method,
                              " | Folds: ", length(all_folds)))
  if (nrow(fold_metrics_df) > 0 && verbose) {
    message(paste("Metrics saved to:", basename(metrics_file)))
    message(paste(utils::capture.output(
      print(fold_metrics_df, row.names = FALSE)), collapse = "\n"))
  }

  if (create_plot && nrow(fold_metrics_df) > 0) {
    plot_rate_cols <- c("Sensitivity", "Specificity", "TSS", "Kappa", "AUC")
    plot_rate_cols <- plot_rate_cols[plot_rate_cols %in% names(fold_metrics_df) &
                                       !vapply(fold_metrics_df[plot_rate_cols],
                                               function(x) all(is.na(x)), logical(1))]
    if (length(plot_rate_cols) > 0) {
      bar_file <- file.path(output_dir, "Fold_Test_Metrics.png")
      tryCatch({
        grDevices::png(bar_file, width = 300 * length(plot_rate_cols) + 200,
                       height = 700, res = 150)
        .plot_fold_metrics_bars(
          metrics_df  = fold_metrics_df,
          rate_cols   = plot_rate_cols,
          fold_colors = fold_colors,
          title       = "GAM Time Independent Assessment Metrics"
        )
        grDevices::dev.off()
        if (verbose) message(paste("Saved fold metrics plot:", basename(bar_file)))
      }, error = function(e) {
        warning(paste0("Could not save fold metrics plot: ", e$message))
        if (grDevices::dev.cur() > 1) grDevices::dev.off()
      })
      tryCatch({
        .plot_fold_metrics_bars(
          metrics_df  = fold_metrics_df,
          rate_cols   = plot_rate_cols,
          fold_colors = fold_colors,
          title       = "GAM Time Independent Assessment Metrics"
        )
        plot_list[["fold_test_metrics"]] <- grDevices::recordPlot()
      }, error = function(e) {
        warning(paste0("Could not record fold metrics plot: ", e$message))
      })
    }
  }

  result <- list(
    models             = model_list,
    thresholds         = thresholds,
    threshold_method   = threshold_method,
    model_formula      = full_formula,
    link               = link,
    model_vars         = rhs_vars,
    fold_training_data = train_data_list,
    fold_test_metrics  = fold_metrics_df,
    output_dir         = output_dir,
    model_type         = "gam",
    plots              = plot_list
  )

  class(result) <- c("TemporalGAM", "list")
  invisible(result)
}

#' Extract Time-Aligned Environmental Values at Species Occurrences
#'
#' Preprocessing function that extracts raster values to species occurrence
#' records based on temporal components. Matches environmental layers to
#' occurrence timestamps and optionally computes scaling parameters for standardization.
#'
#' @usage
#' temporally_explicit_extraction(points_sp, raster_dir, variable_patterns,
#'                                time_cols, xcol = NULL, ycol = NULL,
#'                                points_crs = NULL, output_dir,
#'                                output_prefix = "temp_explicit_df",
#'                                save_raw = TRUE, save_scaled = TRUE,
#'                                save_scaling_params = TRUE,
#'                                verbose = TRUE)
#'
#' @param points_sp sf object, SpatialPointsDataFrame, file path to
#'   .csv/.shp/.geojson/.gpkg, or data frame with coordinate columns.
#' @param raster_dir Character. Directory containing environmental raster
#'   files (\code{.tif}), typically the output of \code{\link{raster_align}}.
#'   File names must follow the patterns supplied in \code{variable_patterns}, with any
#'   time placeholder substituted for the corresponding value from
#'   \code{time_cols}.
#' @param variable_patterns Named character vector mapping clean variable names
#'   to raster filename patterns. For time-varying variables include the time
#'   placeholder in the pattern (e.g. \code{"forest_cover" = "forest_cover_YEAR"});
#'   for static variables omit it (e.g. \code{"elevation" = "elevation"}). Time
#'   placeholders must match entries in \code{time_cols}.
#' @param time_cols Character vector of time column names present in the point
#'   data (e.g., c("YEAR"), c("YEAR", "MONTH")).
#' @param xcol Character. Name of the x-coordinate column. Required when
#'   \code{points_sp} is a CSV file or data frame.
#' @param ycol Character. Name of the y-coordinate column. Required when
#'   \code{points_sp} is a CSV file or data frame.
#' @param points_crs Character or CRS object. CRS of the input points.
#'   Required when \code{points_sp} is a CSV file or data frame.
#' @param output_dir Character. Directory to write output files.
#' @param output_prefix Character. Prefix for output filenames. Default is
#'   "temp_explicit_df".
#' @param save_raw Logical. If \code{TRUE} (default), writes raw extracted values
#'   CSV. If \code{FALSE}, skips raw values output.
#' @param save_scaled Logical. If \code{TRUE} (default), writes z-scaled values
#'   CSV. If \code{FALSE}, skips scaled values output.
#' @param save_scaling_params Logical. If \code{TRUE} (default), writes CSV of
#'   per-variable means and standard deviations. If \code{FALSE}, skips scaling
#'   parameters output.
#' @param verbose Logical. If \code{TRUE} (default), prints progress
#'   messages during processing. Includes file loading, extraction
#'   progress, and file-save confirmation.
#'
#' @return Invisibly returns a list containing:
#' \itemize{
#'   \item \code{raw_values}: Data frame of raw extracted values at each
#'     occurrence record (when \code{save_raw = TRUE}; \code{NULL} otherwise).
#'   \item \code{scaled_values}: Data frame of z-scaled extracted values
#'     (when \code{save_scaled = TRUE}; \code{NULL} otherwise).
#'   \item \code{scaling_params}: Data frame of per-variable means and standard
#'     deviations used for scaling (when \code{save_scaling_params = TRUE};
#'     \code{NULL} otherwise). Pass this to \code{\link{scale_rasters}}.
#'   \item \code{files_created}: Named list of file paths written, with elements
#'     \code{raw}, \code{scaled}, and \code{scaling_params} (each \code{NULL}
#'     when the corresponding save flag is \code{FALSE}).
#' }
#'
#' @details
#' Extracts raster values to species occurrence records based on matched temporal components.
#' Matches environmental layers to occurrence timestamps and optionally computes scaling
#' parameters for standardization.
#'
#' Output CSV files are written to output_dir containing raw values, scaled
#' values, and scaling parameters.
#'
#' Scaling parameters (mean and standard deviation) are optionally computed across
#' all occurrence records for each variable. These parameters should be used with
#' \code{\link{scale_rasters}} to standardize prediction layers.
#'
#' @seealso
#' Preprocessing: \code{\link{spatiotemporal_rarefaction}},
#'   \code{\link{scale_rasters}}, \code{\link{spatiotemporal_partition}}
#'
#' @examples
#' pts_file <- system.file(
#'   "extdata/points/synthetic_occurrence_points.csv",
#'   package = "TemporalModelR"
#' )
#'
#' aln_dir  <- system.file("extdata/rasters_aligned",
#'                         package = "TemporalModelR")
#'
#' ref_file <- system.file("extdata/rasters_raw/elevation.tif",
#'                         package = "TemporalModelR")
#'
#' out_dir  <- file.path(tempdir(), "extracted")
#'
#' temporally_explicit_extraction(
#'   points_sp           = pts_file,
#'   raster_dir          = aln_dir,
#'   variable_patterns   = c(
#'     "elevation"    = "elevation",
#'     "forest_cover" = "forest_cover_YEAR",
#'     "prseas"       = "prseas_YEAR_SEASON"
#'   ),
#'   time_cols           = c("year", "season"),
#'   xcol                = "x",
#'   ycol                = "y",
#'   points_crs          = terra::crs(terra::rast(ref_file)),
#'   output_dir          = out_dir,
#'   save_raw            = TRUE,
#'   save_scaled         = FALSE,
#'   save_scaling_params = TRUE,
#'   verbose             = FALSE
#' )

#' @export
#' @importFrom terra rast extract vect
#' @importFrom sf st_as_sf st_drop_geometry st_coordinates
#' @importFrom stats sd setNames
#' @importFrom utils write.csv
temporally_explicit_extraction <- function(points_sp,
                                           raster_dir,
                                           variable_patterns,
                                           time_cols,
                                           xcol = NULL,
                                           ycol = NULL,
                                           points_crs = NULL,
                                           output_dir,
                                           output_prefix       = "temp_explicit_df",
                                           save_raw            = TRUE,
                                           save_scaled         = TRUE,
                                           save_scaling_params = TRUE,
                                           verbose             = TRUE) {

  pts_df    <- .load_points_data(points_sp, xcol, ycol, points_crs, verbose = verbose)
  xcol      <- attr(pts_df, "xcol")
  ycol      <- attr(pts_df, "ycol")
  pts_crs   <- attr(pts_df, "crs")
  points_sp <- sf::st_as_sf(pts_df, coords = c(xcol, ycol), crs = pts_crs)

  .validate_variable_patterns(variable_patterns)

  if (missing(time_cols) || !is.character(time_cols) || length(time_cols) == 0) {
    stop("ERROR: 'time_cols' must be a character vector with at least one column name.")
  }
  missing_cols <- setdiff(time_cols, names(points_sp))
  if (length(missing_cols) > 0) {
    stop(paste0("ERROR: The following time_cols are missing from data: ",
                paste(missing_cols, collapse = ", ")))
  }

  if (!dir.exists(raster_dir)) stop(paste0("ERROR: 'raster_dir' does not exist: ", raster_dir))
  if (verbose) message(paste("Found",
                             length(list.files(raster_dir, "\\.tif$", recursive = TRUE)),
                             "raster files"))
  dir.create(output_dir, showWarnings = FALSE, recursive = TRUE)
  if (verbose) message(paste("Processing", nrow(points_sp), "points"))

  var_classes  <- .classify_variable_patterns(variable_patterns, time_cols)
  dynamic_vars <- var_classes$dynamic
  static_vars  <- var_classes$static

  if (verbose) message(paste("Dynamic variables:",
                             if (length(dynamic_vars) > 0) paste(dynamic_vars, collapse = ", ") else "none"))
  if (verbose) message(paste("Static variables:",
                             if (length(static_vars) > 0) paste(static_vars, collapse = ", ") else "none"))

  for (var in names(variable_patterns)) points_sp[[var]] <- NA

  if (length(static_vars) > 0) {
    if (verbose) message("Extracting static variables...")
    static_paths <- .resolve_raster_paths(variable_patterns, character(0), static_vars,
                                          time_cols,
                                          as.data.frame(stats::setNames(
                                            lapply(time_cols, function(tc) NA_character_),
                                            time_cols), stringsAsFactors = FALSE),
                                          raster_dir)
    if (!is.null(static_paths)) {
      r_static <- tryCatch(
        terra::rast(static_paths),
        error = function(e) { warning(paste("Could not load static rasters:", e$message)); NULL }
      )
      if (!is.null(r_static)) {
        env_ex <- terra::extract(r_static, terra::vect(points_sp), ID = FALSE)
        if (ncol(env_ex) == length(static_vars)) names(env_ex) <- static_vars
        for (v in static_vars) {
          if (v %in% names(env_ex)) points_sp[[v]] <- env_ex[[v]]
        }
      }
    }
  }

  if (length(dynamic_vars) > 0) {
    if (verbose) message("Extracting dynamic variables...")

    pts_data <- sf::st_drop_geometry(points_sp)
    time_combinations <- unique(pts_data[, time_cols, drop = FALSE])
    time_combinations <- time_combinations[do.call(order, time_combinations), , drop = FALSE]

    if (verbose) message(paste("Extracting values for", nrow(time_combinations), "time periods"))
    for (i in seq_len(nrow(time_combinations))) {
      time_values <- time_combinations[i, , drop = FALSE]
      time_filter <- Reduce(`&`, lapply(time_cols,
                                        function(tc) points_sp[[tc]] == time_values[[tc]]))
      points_subset <- points_sp[time_filter, ]
      if (nrow(points_subset) == 0) next

      raster_paths <- .resolve_raster_paths(variable_patterns, dynamic_vars, character(0),
                                            time_cols, time_values, raster_dir)
      if (is.null(raster_paths)) next

      r_ts <- tryCatch(
        terra::rast(raster_paths),
        error = function(e) { warning(paste("Could not load rasters:", e$message)); NULL }
      )
      if (is.null(r_ts)) next

      env_ex <- terra::extract(r_ts, terra::vect(points_subset), ID = FALSE)
      if (ncol(env_ex) == length(dynamic_vars)) names(env_ex) <- dynamic_vars
      for (v in dynamic_vars) {
        if (v %in% names(env_ex)) points_sp[[v]][time_filter] <- env_ex[[v]]
      }
    }
  }

  coords_mat         <- sf::st_coordinates(points_sp)
  coords_df          <- data.frame(x = coords_mat[, 1], y = coords_mat[, 2])
  points_with_coords <- cbind(sf::st_drop_geometry(points_sp), coords_df)

  raw_output_file    <- NULL
  params_file        <- NULL
  scaled_output_file <- NULL
  scaled_data        <- NULL

  if (save_raw) {
    raw_output_file <- file.path(output_dir, paste0(output_prefix, "_Raw_Values.csv"))
    utils::write.csv(points_with_coords, raw_output_file, row.names = FALSE)
    if (verbose) message(paste("Raw values saved to:", basename(raw_output_file)))
  }

  if (verbose) message("Calculating scaling parameters...")
  scaling_params <- data.frame(variable = character(), mean = numeric(), sd = numeric(),
                               stringsAsFactors = FALSE)
  for (var_name in names(variable_patterns)) {
    values <- points_sp[[var_name]]
    values <- values[!is.na(values)]
    if (length(values) > 0) {
      scaling_params <- rbind(scaling_params,
                              data.frame(variable = var_name, mean = mean(values),
                                         sd = stats::sd(values), stringsAsFactors = FALSE))
    } else {
      warning("No valid values found for ", var_name)
    }
  }

  if (save_scaling_params) {
    params_file <- file.path(output_dir, paste0(output_prefix, "_Scaling_Parameters.csv"))
    utils::write.csv(scaling_params, params_file, row.names = FALSE)
    if (verbose) message(paste("Scaling parameters saved to:", basename(params_file)))
  }

  if (save_scaled) {
    if (verbose) message("Applying scaling...")
    scaled_data <- sf::st_drop_geometry(points_sp)
    for (var_name in names(variable_patterns)) {
      var_params <- scaling_params[scaling_params$variable == var_name, ]
      if (nrow(var_params) > 0) {
        scaled_data[[var_name]] <- (scaled_data[[var_name]] - var_params$mean) / var_params$sd
      } else {
        warning("No scaling parameters found for ", var_name)
      }
    }
    scaled_data <- cbind(scaled_data, coords_df)
    scaled_output_file <- file.path(output_dir, paste0(output_prefix, "_Scaled_Values.csv"))
    utils::write.csv(scaled_data, scaled_output_file, row.names = FALSE)
    if (verbose) message(paste("Scaled values saved to:", basename(scaled_output_file)))
  }

  if (verbose) message("Processing complete.")

  invisible(list(
    raw_values     = if (save_raw)            points_with_coords else NULL,
    scaled_values  = if (save_scaled)         scaled_data        else NULL,
    scaling_params = if (save_scaling_params) scaling_params     else NULL,
    files_created  = list(
      raw            = if (save_raw)            raw_output_file else NULL,
      scaled         = if (save_scaled)         scaled_output_file else NULL,
      scaling_params = if (save_scaling_params) params_file else NULL
    )
  ))
}

#' Analyze Temporal Patterns in Binary Raster Time Series
#'
#' Postprocessing function that applies changepoint detection methods to
#' identify temporal trends in habitat suitability across consecutive
#' predictions. Classifies pixels as stable, increasing in quality, or
#' decreasing in quality, and identifies time periods of significant change.
#'
#' @param binary_stack RasterStack, RasterBrick, or character. Stack of binary
#'   raster layers across time, or path to directory containing binary rasters.
#'   Typically from \code{\link{summarize_raster_outputs}}.
#' @param summary_raster RasterLayer. Per-pixel proportion of time periods where
#'  a pixel is suitable. From \code{\link{summarize_raster_outputs}}.
#' @param time_steps Integer vector. Time labels corresponding to raster layers
#'   (same length as number of layers).
#' @param fastcpd_params List. Named list of parameters passed to fastcpd
#'   changepoint detection function. Default is empty list. Supports parameterization
#'   from \code{\link[fastcpd]{fastcpd_binomial}}.
#' @param output_dir Character. Optional. Output directory for pattern rasters.
#'   When \code{NULL} (default), rasters are written to temporary files and not
#'   saved persistently. Provide a path to write named output files to disk.
#' @param n_tiles_x Integer. Number of tiles in the x direction for tiled
#'   processing. Default is \code{1}. Increase to reduce peak memory use for
#'   large rasters and prevent crashes.
#' @param n_tiles_y Integer. Number of tiles in the y direction for tiled
#'   processing. Default is \code{1}. Increase to reduce peak memory use for
#'   large rasters and prevent crashes.
#' @param alpha Numeric. Significance level for changepoint detection. Default
#'   is \code{0.05}.
#' @param spatial_autocorrelation Logical. If \code{TRUE} (default), includes a
#'   neighbor variable in the changepoint analysis to account for spatial
#'   autocorrelation.
#' @param verbose Logical. If \code{TRUE} (default), prints progress
#'   messages during processing.
#' @param estimate_time Logical. If \code{TRUE} (default), estimates runtime from
#'   a sample of pixels before full processing begins. If \code{FALSE}, proceeds
#'   directly to processing without a time estimate.
#' @param overwrite Logical. If \code{TRUE}, overwrites existing output files.
#'   If \code{FALSE} (default), existing files are skipped.
#'
#' @return Invisibly returns a list containing:
#' \itemize{
#'   \item \code{pattern}: \code{SpatRaster} classifying pixels as integer
#'     values 1-6, corresponding to "Never Suitable", "Always Suitable",
#'     "No Pattern", "Increasing Suitability", "Decreasing Suitability",
#'     or "Fluctuating".
#'   \item \code{time_decrease}: \code{SpatRaster} showing the time step of
#'     first significant decrease for pixels classified as decreasing.
#'   \item \code{time_increase}: \code{SpatRaster} showing the time step of
#'     first significant increase for pixels classified as increasing.
#' }
#'
#' @details
#' Applies changepoint detection using fastcpd to identify significant temporal
#' shifts in spatial suitability. Accounts for spatial and temporal
#' autocorrelation when \code{spatial_autocorrelation = TRUE}. The
#' \code{fastcpd_params} list allows customization of the changepoint detection
#' algorithm, and is passed through to \code{\link[fastcpd]{fastcpd.binomial}}
#' (e.g. \code{list(cost_adjustment = "BIC", trim = 0.025)}).
#'
#' Pattern classifications enable identification of expanding, contracting, or
#' stable g-space distributions over time or site level assessments of directional
#' change in suitability.
#'
#' Classification assumes consecutive rasters. Time periods
#' shorter than ~15 time steps may be too short to classify increases or decreases.
#'
#' @seealso
#' Postprocessing: \code{\link{summarize_raster_outputs}}
#'
#' External: \code{\link[fastcpd]{fastcpd}}
#'
#' @examples
#' \donttest{
#'   con_file <- system.file("extdata/binary/consensus_stack.tif",
#'                           package = "TemporalModelR")
#'
#'   frq_file <- system.file("extdata/binary/frequency_raster.tif",
#'                           package = "TemporalModelR")
#'
#'   binary_stack   <- terra::rast(con_file)
#'
#'   summary_raster <- terra::rast(frq_file)
#'
#'   time_steps <- expand.grid(
#'     year             = 1:15,
#'     season           = "Spring",
#'     stringsAsFactors = FALSE
#'   )
#'
#'   analyze_temporal_patterns(
#'     binary_stack            = binary_stack,
#'     summary_raster          = summary_raster,
#'     time_steps              = time_steps,
#'     output_dir              = tempdir(),
#'     spatial_autocorrelation = FALSE,
#'     overwrite               = TRUE,
#'     estimate_time           = FALSE,
#'     verbose                 = FALSE
#'   )
#' }

#' @export
#' @importFrom terra rast ext res crop nlyr app focal values ncell freq plot writeRaster mosaic
#' @importFrom graphics par legend
#' @importFrom grDevices heat.colors terrain.colors
#' @importFrom utils head setTxtProgressBar txtProgressBar
#' @importFrom tools file_path_sans_ext
analyze_temporal_patterns <- function(binary_stack,
                                      summary_raster,
                                      time_steps,
                                      fastcpd_params          = list(),
                                      output_dir              = NULL,
                                      n_tiles_x               = 1,
                                      n_tiles_y               = 1,
                                      alpha                   = 0.05,
                                      spatial_autocorrelation = TRUE,
                                      verbose                 = TRUE,
                                      estimate_time           = TRUE,
                                      overwrite               = FALSE) {


  if (!requireNamespace("fastcpd", quietly = TRUE)) {
    stop(paste0(
      "ERROR: The 'fastcpd' package is required for analyze_temporal_patterns(). ",
      "Install with: install.packages('fastcpd')"
    ))
  }

  if (missing(binary_stack)) {
    stop("ERROR: 'binary_stack' is required")
  }
  if (missing(summary_raster)) {
    stop("ERROR: 'summary_raster' is required")
  }
  if (missing(time_steps)) {
    stop("ERROR: 'time_steps' is required")
  }

  ts_norm           <- .normalize_time_steps(time_steps, "analyze_temporal_patterns")
  time_steps        <- ts_norm$time_steps
  secondary_filters <- ts_norm$secondary_filters

  if (is.character(binary_stack) && dir.exists(binary_stack)) {
    if (verbose) message(paste0("Loading binary rasters from directory: ", binary_stack))
    binary_files <- list.files(binary_stack, pattern = "\\.tif$", full.names = TRUE)
    if (length(binary_files) == 0) {
      stop(paste0("ERROR: No .tif files found in provided binary_stack directory: ", binary_stack))
    }
    binary_files <- .natural_sort(binary_files)
    binary_stack <- terra::rast(binary_files)
    names(binary_stack) <- tools::file_path_sans_ext(basename(binary_files))
    if (verbose) message(paste("Loaded", terra::nlyr(binary_stack), "binary raster layers"))
  } else if (is.character(binary_stack) && file.exists(binary_stack)) {
    if (verbose) message(paste0("Loading binary raster from file: ", binary_stack))
    binary_stack <- terra::rast(binary_stack)
  } else if (inherits(binary_stack, "SpatRaster")) {
    if (verbose) message("Using provided raster object")
  } else {
    stop("ERROR: 'binary_stack' must be a directory path, file path, or SpatRaster object.")
  }

  summary_raster <- .load_raster_input(summary_raster, "summary_raster")

  if (length(secondary_filters) > 0) {
    lyr_names <- names(binary_stack)
    keep_idx  <- seq_len(terra::nlyr(binary_stack))
    for (sc in names(secondary_filters)) {
      val      <- as.character(secondary_filters[[sc]])
      keep_idx <- keep_idx[grepl(val, lyr_names[keep_idx], ignore.case = TRUE)]
    }
    if (length(keep_idx) == 0) {
      stop(paste0(
        "ERROR: No binary raster layers matched the secondary filter(s): ",
        paste(mapply(function(k, v) paste0(k, "=", v),
                     names(secondary_filters), secondary_filters), collapse = ", "),
        ". Layer names in binary_stack: ",
        paste(head(lyr_names, 5), collapse = ", "),
        if (length(lyr_names) > 5) paste0(" ... (", length(lyr_names), " total)") else ""
      ))
    }
    if (length(keep_idx) != length(time_steps)) {
      stop(paste0(
        "ERROR: After filtering by secondary column(s), found ", length(keep_idx),
        " matching raster layer(s) but time_steps has ", length(time_steps), " value(s). ",
        "Ensure time_steps values match the primary time steps in your predictions."
      ))
    }
    binary_stack <- binary_stack[[keep_idx]]
    if (verbose) message(paste0(
      "Filtered binary_stack to ", terra::nlyr(binary_stack), " layer(s) matching: ",
      paste(mapply(function(k, v) paste0(k, "=", v),
                   names(secondary_filters), secondary_filters), collapse = ", ")
    ))
  } else {
    lyr_names <- names(binary_stack)
    if (terra::nlyr(binary_stack) > length(time_steps)) {
      match_counts <- vapply(as.character(time_steps), function(ts) {
        sum(grepl(paste0("(^|[^0-9])", ts, "([^0-9]|$)"), lyr_names))
      }, integer(1))
      ambiguous <- time_steps[match_counts > 1]
      if (length(ambiguous) > 0) {
        stop(paste0(
          "ERROR: The binary stack has ", terra::nlyr(binary_stack),
          " layers but time_steps only has ", length(time_steps), " value(s). ",
          "Time step value(s) ", paste(ambiguous, collapse = ", "),
          " each match more than one raster layer, suggesting the stack contains ",
          "predictions for multiple secondary time values (e.g. seasons). ",
          "Supply time_steps as a multi-column data frame specifying the secondary ",
          "column value to use. For example: ",
          "time_steps <- expand.grid(time_step = 1:15, season = \"Spring\", stringsAsFactors = FALSE)"
        ))
      }
    }
  }

  if (!is.list(fastcpd_params)) {
    stop("ERROR: 'fastcpd_params' must be a named list")
  }

  save_output <- !is.null(output_dir)
  if (save_output && !dir.exists(output_dir)) dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)

  ts_range      <- paste0(min(time_steps), "_", max(time_steps))
  pattern_file  <- if (save_output) file.path(output_dir,
                                              paste0("pattern_raster_",     ts_range, ".tif")) else tempfile(fileext = ".tif")
  decrease_file <- if (save_output) file.path(output_dir,
                                              paste0("time_first_decrease_", ts_range, ".tif")) else tempfile(fileext = ".tif")
  increase_file <- if (save_output) file.path(output_dir,
                                              paste0("time_first_increase_", ts_range, ".tif")) else tempfile(fileext = ".tif")

  if (file.exists(pattern_file) && file.exists(decrease_file) && file.exists(increase_file) && !overwrite) {
    if (verbose) message("Output rasters exist. Set overwrite = TRUE to rerun.")
    return(invisible(list(
      pattern = terra::rast(pattern_file),
      time_decrease = terra::rast(decrease_file),
      time_increase = terra::rast(increase_file)
    )))
  }

  if (spatial_autocorrelation) {
    if (verbose) message("Spatial autocorrelation: ENABLED (neighbor variable included)")
  } else {
    if (verbose) message("Spatial autocorrelation: DISABLED (neighbor variable excluded)")
  }

  skip_tiling <- (is.null(n_tiles_x) || n_tiles_x == 1) && (is.null(n_tiles_y) || n_tiles_y == 1)

  if (skip_tiling) {
    if (verbose) message("Processing entire raster without tiling...")

    if (estimate_time) {
      if (verbose) message("Estimating processing time...")

      n_times <- terra::nlyr(binary_stack)
      n_middle <- n_times - 2

      summary_vals <- terra::values(summary_raster, mat = FALSE)
      valid_indices <- which(!is.na(summary_vals))

      if (length(valid_indices) > 0) {
        mean_vals <- summary_vals[valid_indices]
        n_quick <- sum(mean_vals < 0.01 | mean_vals > 0.99)
        n_complex <- sum(mean_vals >= 0.01 & mean_vals <= 0.99)

        if (verbose) message(paste0("Quick pixels (always absent/present): ", format(n_quick, big.mark = ",")))
        if (verbose) message(paste0("Complex pixels (changepoint analysis): ", format(n_complex, big.mark = ",")))

        if (n_complex > 0) {
          if (verbose) message("Timing sample pixels...")

          middle_times <- binary_stack[[2:(n_times - 1)]]
          lag_stack <- binary_stack[[1:(n_times - 2)]]

          if (spatial_autocorrelation) {
            middle_neighbor <- terra::rast(lapply(seq_len(terra::nlyr(middle_times)), function(i) {
              terra::focal(middle_times[[i]], w = matrix(1/9, 3, 3), fun = mean, na.rm = TRUE)
            }))
            predictor_stack <- c(middle_times, lag_stack, middle_neighbor, summary_raster)
          } else {
            predictor_stack <- c(middle_times, lag_stack, summary_raster)
          }

          pred_vals_all <- terra::values(predictor_stack, mat = TRUE)
          valid_pred <- which(!is.na(pred_vals_all[, 1]))

          if (spatial_autocorrelation) {
            mean_pred <- pred_vals_all[valid_pred, (3 * n_middle + 1)]
          } else {
            mean_pred <- pred_vals_all[valid_pred, (2 * n_middle + 1)]
          }

          complex_indices <- valid_pred[mean_pred >= 0.01 & mean_pred <= 0.99]

          sample_size <- min(100, length(complex_indices))
          sample_indices <- sample(complex_indices, size = sample_size, replace = FALSE)

          start_time <- Sys.time()
          for (idx in sample_indices) {
            result <- .classify_pixel_with_times(
              pixel_vals      = pred_vals_all[idx, ],
              n_middle        = n_middle,
              time_steps      = time_steps,
              fastcpd_params  = fastcpd_params,
              alpha           = alpha,
              use_neighbor    = spatial_autocorrelation
            )
          }
          end_time <- Sys.time()

          time_per_pixel <- as.numeric(difftime(end_time, start_time, units = "secs")) / sample_size
          if (verbose) message(paste("Average time per complex pixel:", round(time_per_pixel, 4), "seconds"))

          base_seconds <- time_per_pixel * n_complex
          overhead <- 30

          lower_seconds <- (base_seconds * 0.8) + (overhead * 0.5)
          upper_seconds <- (base_seconds * 1.2) + (overhead * 1.5)

          if (verbose) message(paste("Estimated processing time:",
                                     .format_time_estimate(lower_seconds), "to",
                                     .format_time_estimate(upper_seconds)))

          rm(middle_times, lag_stack, predictor_stack, pred_vals_all)
          if (spatial_autocorrelation) rm(middle_neighbor)
          gc(verbose = FALSE)
        }
      }
    }

    if (verbose) message("Processing raster...")

    n_times <- terra::nlyr(binary_stack)
    n_middle <- n_times - 2

    middle_times <- binary_stack[[2:(n_times - 1)]]
    lag_stack <- binary_stack[[1:(n_times - 2)]]

    if (spatial_autocorrelation) {
      middle_neighbor <- terra::rast(lapply(seq_len(terra::nlyr(middle_times)), function(i) {
        terra::focal(middle_times[[i]], w = matrix(1/9, 3, 3), fun = mean, na.rm = TRUE)
      }))
      predictor_stack <- c(middle_times, lag_stack, middle_neighbor, summary_raster)
    } else {
      predictor_stack <- c(middle_times, lag_stack, summary_raster)
    }

    if (verbose) {
      n_cells <- terra::ncell(predictor_stack)
      pb <- utils::txtProgressBar(min = 0, max = n_cells, style = 3, width = 50)

      pred_vals <- terra::values(predictor_stack, mat = TRUE)
      pattern_vals <- numeric(n_cells)
      decrease_vals <- numeric(n_cells)
      increase_vals <- numeric(n_cells)

      for (cell_i in seq_len(n_cells)) {
        if (!any(is.na(pred_vals[cell_i, ]))) {
          result <- .classify_pixel_with_times(
            pixel_vals      = pred_vals[cell_i, ],
            n_middle        = n_middle,
            time_steps      = time_steps,
            fastcpd_params  = fastcpd_params,
            alpha           = alpha,
            use_neighbor    = spatial_autocorrelation
          )
          pattern_vals[cell_i] <- result[1]
          decrease_vals[cell_i] <- result[2]
          increase_vals[cell_i] <- result[3]
        } else {
          pattern_vals[cell_i] <- NA
          decrease_vals[cell_i] <- NA
          increase_vals[cell_i] <- NA
        }
        if (cell_i %% 10 == 0) utils::setTxtProgressBar(pb, cell_i)
      }
      close(pb)
      if (verbose) message("")

      pattern_raster <- terra::rast(predictor_stack, nlyr = 1)
      decrease_raster <- terra::rast(predictor_stack, nlyr = 1)
      increase_raster <- terra::rast(predictor_stack, nlyr = 1)

      terra::values(pattern_raster) <- pattern_vals
      terra::values(decrease_raster) <- decrease_vals
      terra::values(increase_raster) <- increase_vals

    } else {
      result_matrix <- terra::app(
        predictor_stack,
        fun = function(x) .classify_pixel_with_times(
          pixel_vals      = x,
          n_middle        = n_middle,
          time_steps      = time_steps,
          fastcpd_params  = fastcpd_params,
          alpha           = alpha,
          use_neighbor    = spatial_autocorrelation
        )
      )

      pattern_raster <- result_matrix[[1]]
      decrease_raster <- result_matrix[[2]]
      increase_raster <- result_matrix[[3]]
    }

  } else {

    tiles_dir <- if (save_output) file.path(output_dir, "tiles") else file.path(tempdir(), "tiles")
    if (!dir.exists(tiles_dir)) dir.create(tiles_dir, recursive = TRUE, showWarnings = FALSE)

    if (verbose) message("Calculating tile extents...")

    full_ext <- terra::ext(binary_stack)
    x_min <- full_ext[1]
    x_max <- full_ext[2]
    y_min <- full_ext[3]
    y_max <- full_ext[4]

    res_vals <- terra::res(binary_stack)
    res_x <- res_vals[1]
    res_y <- res_vals[2]

    x_range <- x_max - x_min
    y_range <- y_max - y_min

    tile_width <- x_range / n_tiles_x
    tile_height <- y_range / n_tiles_y

    tile_extents <- list()
    tile_idx <- 1

    for (i in seq_len(n_tiles_y)) {
      for (j in seq_len(n_tiles_x)) {
        tile_x_min <- x_min + (j - 1) * tile_width
        tile_x_max <- x_min + j * tile_width
        tile_y_min <- y_min + (i - 1) * tile_height
        tile_y_max <- y_min + i * tile_height

        tile_extents[[tile_idx]] <- terra::ext(tile_x_min, tile_x_max, tile_y_min, tile_y_max)
        tile_idx <- tile_idx + 1
      }
    }

    n_tiles <- length(tile_extents)
    if (verbose) message(paste("Created", n_tiles, "tiles"))

    if (estimate_time) {
      if (verbose) message("Estimating processing time...")

      n_times <- terra::nlyr(binary_stack)
      n_middle <- n_times - 2

      total_complex <- 0
      total_quick <- 0
      time_per_pixel <- NULL

      for (tile_i in seq_len(n_tiles)) {
        if (verbose) message(paste0("Scanning tile ", tile_i, "/", n_tiles, "..."))

        tile_ext <- tile_extents[[tile_i]]

        tryCatch({
          suppressWarnings({
            tile_binary <- terra::crop(binary_stack, tile_ext)
            tile_summary <- terra::crop(summary_raster, tile_ext)
          })

          summary_vals <- terra::values(tile_summary, mat = FALSE)
          valid_indices <- which(!is.na(summary_vals))

          if (length(valid_indices) > 0) {
            mean_vals <- summary_vals[valid_indices]
            n_quick <- sum(mean_vals < 0.01 | mean_vals > 0.99)
            n_complex <- sum(mean_vals >= 0.01 & mean_vals <= 0.99)

            total_quick <- total_quick + n_quick
            total_complex <- total_complex + n_complex

            if (is.null(time_per_pixel) && n_complex > 0) {
              if (verbose) message("Timing sample pixels...")

              middle_times <- tile_binary[[2:(n_times - 1)]]
              lag_stack <- tile_binary[[1:(n_times - 2)]]

              if (spatial_autocorrelation) {
                middle_neighbor <- terra::rast(lapply(seq_len(terra::nlyr(middle_times)), function(i) {
                  terra::focal(middle_times[[i]], w = matrix(1/9, 3, 3), fun = mean, na.rm = TRUE)
                }))
                predictor_stack <- c(middle_times, lag_stack, middle_neighbor, tile_summary)
              } else {
                predictor_stack <- c(middle_times, lag_stack, tile_summary)
              }

              pred_vals_all <- terra::values(predictor_stack, mat = TRUE)
              valid_pred <- which(!is.na(pred_vals_all[, 1]))

              if (spatial_autocorrelation) {
                mean_pred <- pred_vals_all[valid_pred, (3 * n_middle + 1)]
              } else {
                mean_pred <- pred_vals_all[valid_pred, (2 * n_middle + 1)]
              }

              complex_indices <- valid_pred[mean_pred >= 0.01 & mean_pred <= 0.99]

              sample_size <- min(100, length(complex_indices))
              sample_indices <- sample(complex_indices, size = sample_size, replace = FALSE)

              start_time <- Sys.time()
              for (idx in sample_indices) {
                result <- .classify_pixel_with_times(
                  pixel_vals      = pred_vals_all[idx, ],
                  n_middle        = n_middle,
                  time_steps      = time_steps,
                  fastcpd_params  = fastcpd_params,
                  alpha           = alpha,
                  use_neighbor    = spatial_autocorrelation
                )
              }
              end_time <- Sys.time()

              time_per_pixel <- as.numeric(difftime(end_time, start_time, units = "secs")) / sample_size
              if (verbose) message(paste("Average time per complex pixel:", round(time_per_pixel, 4), "seconds"))
            }
          }
        }, error = function(e) {
          if (grepl("cannot allocate vector", e$message, ignore.case = TRUE)) {
            stop("ERROR: Memory error. Increase n_tiles_x and n_tiles_y to use smaller tiles.")
          } else {
            stop(e)
          }
        })
      }

      if (verbose) message(paste0("Quick pixels (always absent/present): ", format(total_quick, big.mark = ",")))
      if (verbose) message(paste0("Complex pixels (changepoint analysis): ", format(total_complex, big.mark = ",")))

      if (!is.null(time_per_pixel) && total_complex > 0) {
        base_seconds <- time_per_pixel * total_complex
        overhead_per_tile <- 15
        total_overhead <- (overhead_per_tile * n_tiles) + 30

        lower_seconds <- (base_seconds * 0.8) + (total_overhead * 0.5)
        upper_seconds <- (base_seconds * 1.2) + (total_overhead * 1.5)

        if (verbose) message(paste("Estimated processing time:",
                                   .format_time_estimate(lower_seconds), "to",
                                   .format_time_estimate(upper_seconds)))
      }
    }

    if (verbose) message("Processing tiles...")

    tile_files_pattern <- character(n_tiles)
    tile_files_decrease <- character(n_tiles)
    tile_files_increase <- character(n_tiles)

    for (tile_i in seq_len(n_tiles)) {
      if (verbose) message(paste("Processing tile", tile_i, "of", n_tiles))

      tile_file_pattern <- file.path(tiles_dir, paste0("pattern_tile_", tile_i, ".tif"))
      tile_file_decrease <- file.path(tiles_dir, paste0("decrease_tile_", tile_i, ".tif"))
      tile_file_increase <- file.path(tiles_dir, paste0("increase_tile_", tile_i, ".tif"))

      tile_files_pattern[tile_i] <- tile_file_pattern
      tile_files_decrease[tile_i] <- tile_file_decrease
      tile_files_increase[tile_i] <- tile_file_increase

      tile_ext <- tile_extents[[tile_i]]

      tryCatch({
        suppressWarnings({
          tile_binary <- terra::crop(binary_stack, tile_ext)
          tile_summary <- terra::crop(summary_raster, tile_ext)
        })

        n_times <- terra::nlyr(tile_binary)
        n_middle <- n_times - 2

        middle_times <- tile_binary[[2:(n_times - 1)]]
        lag_stack <- tile_binary[[1:(n_times - 2)]]

        if (spatial_autocorrelation) {
          middle_neighbor <- terra::rast(lapply(seq_len(terra::nlyr(middle_times)), function(i) {
            terra::focal(middle_times[[i]], w = matrix(1/9, 3, 3), fun = mean, na.rm = TRUE)
          }))
          predictor_stack <- c(middle_times, lag_stack, middle_neighbor, tile_summary)
        } else {
          predictor_stack <- c(middle_times, lag_stack, tile_summary)
        }

        if (verbose) {
          n_cells <- terra::ncell(predictor_stack)
          pb <- utils::txtProgressBar(min = 0, max = n_cells, style = 3, width = 50)

          pred_vals <- terra::values(predictor_stack, mat = TRUE)
          pattern_vals <- numeric(n_cells)
          decrease_vals <- numeric(n_cells)
          increase_vals <- numeric(n_cells)

          for (cell_i in seq_len(n_cells)) {
            if (!any(is.na(pred_vals[cell_i, ]))) {
              result <- .classify_pixel_with_times(
                pixel_vals      = pred_vals[cell_i, ],
                n_middle        = n_middle,
                time_steps      = time_steps,
                fastcpd_params  = fastcpd_params,
                alpha           = alpha,
                use_neighbor    = spatial_autocorrelation
              )
              pattern_vals[cell_i] <- result[1]
              decrease_vals[cell_i] <- result[2]
              increase_vals[cell_i] <- result[3]
            } else {
              pattern_vals[cell_i] <- NA
              decrease_vals[cell_i] <- NA
              increase_vals[cell_i] <- NA
            }
            if (cell_i %% 10 == 0) utils::setTxtProgressBar(pb, cell_i)
          }
          close(pb)
          if (verbose) message("")

          tile_pattern <- terra::rast(predictor_stack, nlyr = 1)
          tile_decrease <- terra::rast(predictor_stack, nlyr = 1)
          tile_increase <- terra::rast(predictor_stack, nlyr = 1)

          terra::values(tile_pattern) <- pattern_vals
          terra::values(tile_decrease) <- decrease_vals
          terra::values(tile_increase) <- increase_vals

        } else {
          result_matrix <- terra::app(
            predictor_stack,
            fun = function(x) .classify_pixel_with_times(
              pixel_vals      = x,
              n_middle        = n_middle,
              time_steps      = time_steps,
              fastcpd_params  = fastcpd_params,
              alpha           = alpha,
              use_neighbor    = spatial_autocorrelation
            )
          )

          tile_pattern <- result_matrix[[1]]
          tile_decrease <- result_matrix[[2]]
          tile_increase <- result_matrix[[3]]
        }

        terra::writeRaster(tile_pattern, tile_file_pattern, overwrite = TRUE,
                           datatype = "INT1U", gdal = c("COMPRESS=LZW"))
        terra::writeRaster(tile_decrease, tile_file_decrease, overwrite = TRUE,
                           datatype = "INT2S", gdal = c("COMPRESS=LZW"))
        terra::writeRaster(tile_increase, tile_file_increase, overwrite = TRUE,
                           datatype = "INT2S", gdal = c("COMPRESS=LZW"))

        rm(tile_binary, tile_summary, middle_times, lag_stack,
           predictor_stack, tile_pattern, tile_decrease, tile_increase)
        if (spatial_autocorrelation) rm(middle_neighbor)
        if (exists("pred_vals")) rm(pred_vals, pattern_vals, decrease_vals, increase_vals)
        gc(verbose = FALSE)

      }, error = function(e) {
        if (grepl("cannot allocate vector", e$message, ignore.case = TRUE)) {
          stop(paste0("ERROR: Memory error on tile ", tile_i, ". Increase n_tiles_x and n_tiles_y."))
        } else {
          stop(e)
        }
      })
    }

    if (n_tiles > 1) {
      if (verbose) message("Merging tiles...")

      tryCatch({
        tile_rasters_pattern <- lapply(tile_files_pattern, terra::rast)
        pattern_raster <- do.call(terra::mosaic, c(tile_rasters_pattern, fun = "mean"))

        tile_rasters_decrease <- lapply(tile_files_decrease, terra::rast)
        decrease_raster <- do.call(terra::mosaic, c(tile_rasters_decrease, fun = "mean"))

        tile_rasters_increase <- lapply(tile_files_increase, terra::rast)
        increase_raster <- do.call(terra::mosaic, c(tile_rasters_increase, fun = "mean"))

      }, error = function(e) {
        if (grepl("cannot allocate vector", e$message, ignore.case = TRUE)) {
          stop("ERROR: Memory error merging tiles. Increase n_tiles_x and n_tiles_y.")
        } else {
          stop(e)
        }
      })
    } else {
      if (verbose) message("Single tile detected, skipping merge...")

      pattern_raster <- terra::rast(tile_files_pattern[1])
      decrease_raster <- terra::rast(tile_files_decrease[1])
      increase_raster <- terra::rast(tile_files_increase[1])
    }

    if (verbose) message("Cleaning up tile files...")
    unlink(tiles_dir, recursive = TRUE)
    if (verbose) message("Tiles removed")
  }

  if (verbose) message("Saving results...")
  terra::writeRaster(pattern_raster, pattern_file, overwrite = TRUE,
                     datatype = "INT1U", gdal = c("COMPRESS=LZW"))
  terra::writeRaster(decrease_raster, decrease_file, overwrite = TRUE,
                     datatype = "INT2S", gdal = c("COMPRESS=LZW"))
  terra::writeRaster(increase_raster, increase_file, overwrite = TRUE,
                     datatype = "INT2S", gdal = c("COMPRESS=LZW"))

  if (verbose) message("Generating plots...")

  pat_cols   <- c("#730000", "#267300", "#B2B2B2", "#A3FF73", "#FF7F7F", "#A900E6", "#eed202")
  pat_labels <- c("Always Absent", "Always Present", "No Pattern",
                  "Increasing", "Decreasing", "Fluctuating", "Failed")

  opar1 <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(opar1), add = TRUE)
  graphics::par(mar = c(6.5, 4, 3, 2))
  terra::plot(pattern_raster,
              col    = pat_cols,
              main   = paste("Pattern Classification\n", min(time_steps), "-", max(time_steps)),
              breaks = c(0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5),
              legend = FALSE)
  graphics::legend(
    x      = "bottom",
    legend = pat_labels,
    fill   = pat_cols,
    cex    = 0.7,
    bty    = "n",
    horiz  = FALSE,
    ncol   = 4,
    xpd    = TRUE,
    inset  = c(0, -0.42)
  )
  graphics::par(opar1)

  dec_vals <- terra::values(decrease_raster, na.rm = TRUE)
  if (length(dec_vals) > 0 && !all(is.na(dec_vals))) {
    opar2 <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(opar2), add = TRUE)
    graphics::par(mar = c(4, 4, 3, 2))
    terra::plot(decrease_raster,
                main = paste("Time of First Decrease\n", min(time_steps), "-", max(time_steps)),
                col  = rev(heat.colors(50)))
    graphics::par(opar2)
  } else {
    if (verbose) message("Skipping decrease raster plot: no pixels with a significant decrease detected.")
  }

  inc_vals <- terra::values(increase_raster, na.rm = TRUE)
  if (length(inc_vals) > 0 && !all(is.na(inc_vals))) {
    opar3 <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(opar3), add = TRUE)
    graphics::par(mar = c(4, 4, 3, 2))
    terra::plot(increase_raster,
                main = paste("Time of First Increase\n", min(time_steps), "-", max(time_steps)),
                col  = terrain.colors(50))
    graphics::par(opar3)
  } else {
    if (verbose) message("Skipping increase raster plot: no pixels with a significant increase detected.")
  }

  if (verbose) message("Analysis complete")
  if (verbose) message(paste("Period:", min(time_steps), "-", max(time_steps)))
  if (!skip_tiling) {
    if (verbose) message(paste("Tiles:", n_tiles))
  }
  if (verbose) message(paste("Spatial autocorrelation:", ifelse(spatial_autocorrelation, "ENABLED", "DISABLED")))

  pattern_freq <- terra::freq(pattern_raster)
  if (!is.null(pattern_freq) && nrow(pattern_freq) > 0) {
    pattern_freq <- as.data.frame(pattern_freq)
    pattern_freq$proportion <- round(pattern_freq$count / sum(pattern_freq$count), 3)
    pattern_names <- c("Always Absent", "Always Present", "No Pattern",
                       "Increasing", "Decreasing", "Fluctuating", "Failed")
    pattern_freq$pattern <- pattern_names[pattern_freq$value]

    if (verbose) message("Pattern Classifications:")
    if (verbose) message(pattern_freq[, c("value", "pattern", "count", "proportion")])
  }

  dec_freq <- as.data.frame(terra::freq(decrease_raster))
  dec_freq <- dec_freq[!is.na(dec_freq$value), ]
  if (nrow(dec_freq) > 0) {
    n_decreasing <- sum(dec_freq$count)
    if (verbose) message(paste("Decreasing pixels:", format(n_decreasing, big.mark = ",")))
    if (verbose) message(paste("Time range:", min(dec_freq$value), "-", max(dec_freq$value)))
    if (verbose) message(paste("Most common:", dec_freq$value[which.max(dec_freq$count)],
                               paste0("(", format(max(dec_freq$count), big.mark = ","), " pixels)")))
  }

  inc_freq <- as.data.frame(terra::freq(increase_raster))
  inc_freq <- inc_freq[!is.na(inc_freq$value), ]
  if (nrow(inc_freq) > 0) {
    n_increasing <- sum(inc_freq$count)
    if (verbose) message(paste("Increasing pixels:", format(n_increasing, big.mark = ",")))
    if (verbose) message(paste("Time range:", min(inc_freq$value), "-", max(inc_freq$value)))
    if (verbose) message(paste("Most common:", inc_freq$value[which.max(inc_freq$count)],
                               paste0("(", format(max(inc_freq$count), big.mark = ","), " pixels)")))
  }

  invisible(list(
    pattern = pattern_raster,
    time_decrease = decrease_raster,
    time_increase = increase_raster
  ))
}

#' Generate Spatiotemporal Predictions from Temporal Models
#'
#' Generates temporally explicit habitat suitability predictions by projecting
#' fitted models onto environmental raster stacks matching each time period.
#' Accepts output from any of the four TemporalModelR modeling functions.
#' Produces per-time-step assessment metrics showing how predictions and
#' test point coverage vary over time.
#'
#'
#' @usage
#' generate_spatiotemporal_predictions(partition_result, model_result,
#'                                     pseudoabsence_result = NULL, raster_dir,
#'                                     variable_patterns, time_cols, time_steps,
#'                                     output_dir = file.path(tempdir(), "Predictions"),
#'                                     overwrite = FALSE,
#'                                     verbose = TRUE)
#'
#' @param partition_result List or character. Output from
#'   \code{\link{spatiotemporal_partition}} or path to an \code{.rds}
#'   file containing that output.
#' @param model_result List or character. Output from any of
#'   \code{\link{build_temporal_hv}}, \code{\link{build_temporal_glm}},
#'   \code{\link{build_temporal_gam}}, or \code{\link{build_temporal_rf}}, or
#'   a path to an \code{.rds} file. Model type is detected automatically from
#'   the \code{model_type} field.
#' @param pseudoabsence_result List, character, or \code{NULL}. Optional.
#'   Output from \code{\link{generate_absences}} or path to an
#'   \code{.rds} file. When supplied for presence/absence models (GLM, GAM, RF),
#'   the held-out pseudoabsence test points for each fold are filtered to the
#'   current time step and used alongside presence test points to compute
#'   per-timestep TN, FP, Specificity, and TSS. These columns are added to the
#'   timestep metrics table when pseudoabsences are available. Ignored for
#'   hypervolume models. Default is \code{NULL}.
#' @param raster_dir Character. Directory containing environmental raster
#'   files (\code{.tif}), typically the output of
#'   \code{\link{raster_align}} or \code{\link{scale_rasters}}. File names
#'   must follow the patterns supplied in \code{variable_patterns}, with any
#'   time placeholder substituted for the corresponding value from
#'   \code{time_cols}.
#' @param variable_patterns Named character vector mapping clean variable names
#'   to raster filename patterns. For time-varying variables include the time
#'   placeholder in the pattern (e.g. \code{"forest_cover" = "forest_cover_YEAR"});
#'   for static variables omit it (e.g. \code{"elevation" = "elevation"}). Time
#'   placeholders must match entries in \code{time_cols}.
#' @param time_cols Character. Name of the column(s) containing year or time
#'   step values in the occurrence data. Must match \code{time_cols} used in
#'   \code{\link{spatiotemporal_partition}} and the time placeholders used in
#'   \code{variable_patterns}.
#' @param time_steps Vector, data frame, or matrix of time periods for which
#'   to generate predictions.
#' @param output_dir Character. Directory to write prediction rasters and the
#'   assessment metrics CSV. Default is \code{file.path(tempdir(), "Predictions")}.
#' @param overwrite Logical. If \code{TRUE}, overwrites existing output files.
#'   If \code{FALSE} (default), existing files are skipped.
#' @param verbose Logical. If \code{TRUE} (default), prints progress
#'   messages during processing. Includes per-time-step prediction progress.
#'
#' @return Invisibly returns a list containing:
#' \itemize{
#'   \item \code{timestep_metrics}: data frame of per-fold per-time-step
#'     metrics. Columns always present: \code{Fold}, the time column(s),
#'     \code{Pct_Suitable}, \code{N_Pres}, \code{TP}, \code{FN},
#'     \code{Sensitivity}, \code{CBP}. When \code{pseudoabsence_result} is
#'     supplied for parametric models, additionally: \code{N_Abs}, \code{TN},
#'     \code{FP}, \code{Specificity}, \code{TSS}. Saved as
#'     \code{Timestep_Assessment_Metrics.csv}.
#'   \item \code{overall_summary}: data frame of pooled metrics per fold.
#'     Columns: \code{Fold}, \code{N_Timesteps}, \code{Mean_Pct_Suitable},
#'     \code{Total_TP}, \code{Total_FN}, \code{Overall_Sensitivity},
#'     \code{Overall_CBP}. When pseudoabsences available: additionally
#'     \code{Total_TN}, \code{Total_FP}, \code{Overall_Specificity},
#'     \code{Overall_TSS}.
#'   \item \code{prediction_files}: character vector of paths to saved
#'     prediction rasters.
#'   \item \code{model_type}: character string recording the model type used.
#' }
#'
#' @details
#' G-space predictions are produced as rasters for each time-step and fold of an
#' input model.
#'
#' Per-timestep metrics are computed per fold per time step by extracting
#' raster predictions at the held-out presence points that fall within that
#' time step. \code{Pct_Suitable} records the proportion of the study area
#' predicted suitable. \code{Sensitivity} shows how consistently test points are
#' captured. \code{CBP} (cumulative binomial probability) tests whether the
#' observed number of correctly predicted test points is better than expected
#' by random placement: under the null, each point has \code{Pct_Suitable}
#' probability of falling in suitable area, so
#' \code{dbinom(TP, N_Test_Pts, Pct_Suitable)} gives the probability of
#' observing exactly \code{TP} correct by chance. Small values indicate
#' predictions are better than random.
#'
#' @seealso
#' Preprocessing: \code{\link{scale_rasters}},
#'   \code{\link{spatiotemporal_partition}}
#'
#' Modeling: \code{\link{build_temporal_hv}},
#'   \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
#'   \code{\link{build_temporal_rf}}
#'
#' Post-processing: \code{\link{summarize_raster_outputs}},
#'   \code{\link{plot_model_assessment}}
#'
#' @examples
#' data(tmr_partition, package = "TemporalModelR")
#'
#' data(tmr_glm,       package = "TemporalModelR")
#'
#' data(tmr_absences,  package = "TemporalModelR")
#'
#' scl_dir    <- system.file("extdata/rasters_scaled",
#'                           package = "TemporalModelR")
#'
#' time_steps <- expand.grid(
#'   year             = 1:15,
#'   season           = "Spring",
#'   stringsAsFactors = FALSE
#' )
#'
#' generate_spatiotemporal_predictions(
#'   partition_result     = tmr_partition,
#'   model_result         = tmr_glm,
#'   pseudoabsence_result = tmr_absences,
#'   raster_dir           = scl_dir,
#'   variable_patterns    = c(
#'     "elevation"    = "elevation",
#'     "forest_cover" = "forest_cover_YEAR",
#'     "prseas"       = "prseas_YEAR_SEASON"
#'   ),
#'   time_cols            = c("year", "season"),
#'   time_steps           = time_steps,
#'   output_dir           = tempdir(),
#'   overwrite            = TRUE,
#'   verbose              = FALSE
#' )

#' @export
#' @importFrom terra rast extract values vect ncell nlyr writeRaster
#' @importFrom sf st_drop_geometry
#' @importFrom stats complete.cases predict dbinom setNames
#' @importFrom tools file_ext file_path_sans_ext
#' @importFrom utils read.csv write.csv
generate_spatiotemporal_predictions <- function(partition_result,
                                                model_result,
                                                pseudoabsence_result = NULL,
                                                raster_dir,
                                                variable_patterns,
                                                time_cols,
                                                time_steps,
                                                output_dir = file.path(tempdir(), "Predictions"),
                                                overwrite  = FALSE,
                                                verbose    = TRUE) {

  if (missing(model_result)) {
    stop("ERROR: 'model_result' is required.")
  }

  partition_result  <- .load_partition_result(partition_result)
  occurrence_points <- partition_result$points_sf

  if (is.null(occurrence_points) || nrow(occurrence_points) == 0) {
    stop("ERROR: partition_result$points_sf is empty.")
  }
  if (!"fold" %in% names(occurrence_points)) {
    stop("ERROR: partition_result$points_sf is missing 'fold' column.")
  }

  occ_df    <- sf::st_drop_geometry(occurrence_points)
  all_folds <- sort(unique(occ_df$fold[!is.na(occ_df$fold)]))
  points_sf <- occurrence_points
  if (verbose) message(paste("Loaded", nrow(occ_df), "points across", length(all_folds), "folds."))

  pseudoabs_sf <- NULL
  if (!is.null(pseudoabsence_result)) {
    pseudoabsence_result <- .load_pseudoabsence_result(pseudoabsence_result)
    pseudoabs_sf <- pseudoabsence_result$pseudoabsences
    if (is.null(pseudoabs_sf) || nrow(pseudoabs_sf) == 0) {
      warning("pseudoabsence_result$pseudoabsences is empty  -  per-timestep TN/FP/Specificity/TSS will not be computed.")
      pseudoabs_sf <- NULL
    } else {
      if (verbose) message(paste("Loaded", nrow(pseudoabs_sf), "pseudoabsence points for per-timestep evaluation."))
    }
  }

  if (is.character(model_result)) {
    if (!file.exists(model_result)) stop(paste0("ERROR: 'model_result' file not found: ", model_result))
    if (tolower(tools::file_ext(model_result)) != "rds") stop("ERROR: 'model_result' must be .rds format.")
    if (verbose) message(paste("Reading model results from:", basename(model_result)))
    model_result <- tryCatch(readRDS(model_result),
                             error = function(e) stop(paste0("ERROR reading model_result: ", e$message)))
  }

  if (!is.list(model_result) || is.null(model_result$model_type)) {
    stop("ERROR: 'model_result' must be output from a build_temporal_*() function.")
  }

  model_type <- model_result$model_type
  if (!model_type %in% c("hypervolume", "glm", "gam", "rf")) {
    stop(paste0("ERROR: Unrecognized model_type '", model_type, "'."))
  }
  if (verbose) message(paste("Model type:", model_type))

  if (model_type == "hypervolume" && !requireNamespace("hypervolume", quietly = TRUE)) {
    stop("ERROR: The 'hypervolume' package is required. Install with: install.packages('hypervolume')")
  }

  model_list    <- model_result$models
  model_vars    <- model_result$model_vars
  is_parametric <- model_type != "hypervolume"
  thresholds    <- NULL

  if (is.null(model_list) || length(model_list) == 0) stop("ERROR: model_result$models is empty.")
  if (is.null(model_vars) || length(model_vars) == 0) stop("ERROR: model_result$model_vars is missing.")

  if (is_parametric) {
    thresholds <- model_result$thresholds
    if (is.null(thresholds) || all(is.na(thresholds))) {
      stop("ERROR: model_result$thresholds is missing or all NA. Re-run the model with a valid threshold_method.")
    }
  }

  if (verbose) message(paste("Loaded", length(model_list), "model(s)."))

  if (missing(variable_patterns)) {
    stop("ERROR: 'variable_patterns' is required.")
  }
  .validate_variable_patterns(variable_patterns)
  if (missing(time_cols) || is.null(time_cols)) stop("ERROR: 'time_cols' is required.")
  time_cols <- as.character(time_cols)
  if (any(!time_cols %in% names(occ_df))) {
    stop(paste0("ERROR: 'time_cols' '", paste(setdiff(time_cols, names(occ_df)), collapse = ", "),
                "' not found in occurrence data."))
  }
  if (missing(time_steps)) stop("ERROR: 'time_steps' is required.")
  if (is.data.frame(time_steps) || is.matrix(time_steps)) {
    time_steps_df <- as.data.frame(time_steps, stringsAsFactors = FALSE)
    missing_cols  <- setdiff(time_cols, names(time_steps_df))
    if (length(missing_cols) > 0) stop(paste0("ERROR: 'time_steps' missing columns: ", paste(missing_cols, collapse = ", ")))
  } else if (is.vector(time_steps) && !is.list(time_steps)) {
    if (length(time_cols) == 1) {
      time_steps_df <- stats::setNames(data.frame(time_steps, stringsAsFactors = FALSE), time_cols)
    } else {
      other_cols   <- time_cols[-1]
      other_combos <- unique(occ_df[, other_cols, drop = FALSE])
      cross <- expand.grid(
        c(list(time_steps), as.list(other_combos)),
        stringsAsFactors = FALSE
      )
      names(cross)  <- time_cols
      time_steps_df <- cross
    }
  } else {
    stop("ERROR: 'time_steps' must be a plain vector of time step values, or a data frame with one column per time_cols.")
  }
  if (missing(raster_dir) || !dir.exists(raster_dir)) {
    stop(paste0("ERROR: 'raster_dir' does not exist: ", raster_dir))
  }

  if (verbose) message(paste("Processing", nrow(time_steps_df), "time steps."))

  var_classes  <- .classify_variable_patterns(variable_patterns, time_cols)
  dynamic_vars <- var_classes$dynamic
  static_vars  <- var_classes$static

  if (verbose) message(paste("Dynamic variables:", if (length(dynamic_vars) > 0) paste(dynamic_vars, collapse = ", ") else "none"))
  if (verbose) message(paste("Static variables:",  if (length(static_vars)  > 0) paste(static_vars,  collapse = ", ") else "none"))

  if (!dir.exists(output_dir)) dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)

  fold_keys <- paste0("fold", all_folds)

  timestep_metrics_file <- file.path(output_dir, "Timestep_Assessment_Metrics.csv")
  timestep_metrics <- if (!overwrite && file.exists(timestep_metrics_file)) {
    tryCatch(utils::read.csv(timestep_metrics_file, stringsAsFactors = FALSE),
             error = function(e) data.frame())
  } else data.frame()

  prediction_files <- character(0)

  for (i in seq_len(nrow(time_steps_df))) {

    time_values <- time_steps_df[i, , drop = FALSE]
    time_label  <- paste(vapply(time_cols, function(tc) as.character(time_values[[tc]]),
                                character(1)), collapse = "_")
    output_path <- file.path(output_dir, paste0("Prediction_", time_label, ".tif"))

    if (file.exists(output_path) && !overwrite) {
      if (verbose) message(paste("Skipping", time_label, " - output already exists."))
      prediction_files <- c(prediction_files, output_path)
      next
    }

    if (verbose) message(paste0("--- Time step: ", time_label, " ---"))

    raster_paths <- .resolve_raster_paths(variable_patterns, dynamic_vars, static_vars,
                                          time_cols, time_values, raster_dir)
    if (is.null(raster_paths)) {
      warning(paste("Missing rasters for", time_label, " - skipping."))
      next
    }

    raster_stack <- tryCatch(terra::rast(raster_paths),
                             error = function(e) { warning(paste("Could not load rasters:", e$message)); NULL })
    if (is.null(raster_stack)) next

    expected_names <- vapply(raster_paths, function(p) {
      tools::file_path_sans_ext(basename(p))
    }, character(1))
    if (terra::nlyr(raster_stack) == length(expected_names)) {
      names(raster_stack) <- expected_names
    }

    time_filter <- rep(TRUE, nrow(points_sf))
    for (tc in time_cols) {
      time_filter <- time_filter & !is.na(points_sf[[tc]]) & (points_sf[[tc]] == time_values[[tc]])
    }

    fold_projections <- vector("list", length(all_folds))
    names(fold_projections) <- fold_keys

    for (fi in seq_along(all_folds)) {

      fold     <- all_folds[fi]
      fold_key <- fold_keys[fi]

      if (verbose) message(paste("  Fold", fold, "..."))

      test_pts_time <- points_sf[!is.na(points_sf$fold) & points_sf$fold == fold & time_filter, ]

      pseudo_pts_time <- if (!is.null(pseudoabs_sf) && is_parametric) {
        pa_fold_mask <- !is.na(pseudoabs_sf$fold) & pseudoabs_sf$fold == fold
        pa_time_mask <- rep(TRUE, nrow(pseudoabs_sf))
        for (tc in time_cols) {
          if (tc %in% names(pseudoabs_sf)) {
            pa_time_mask <- pa_time_mask &
              !is.na(pseudoabs_sf[[tc]]) & (pseudoabs_sf[[tc]] == time_values[[tc]])
          }
        }
        pseudoabs_sf[pa_fold_mask & pa_time_mask, ]
      } else NULL

      binary_raster <- if (model_type == "hypervolume") {
        .predict_hypervolume_raster(model_list[[fold_key]], raster_stack, fold, time_label)
      } else {
        thr <- thresholds[fold_key]
        if (is.null(thr) || is.na(thr)) { warning(paste("No threshold for", fold_key)); NULL }
        else .predict_parametric_raster(model_list[[fold_key]], raster_stack, model_vars,
                                        thr, model_type, variable_patterns, time_cols, time_values)
      }

      fold_projections[[fold_key]] <- binary_raster
      if (is.null(binary_raster)) next

      timestep_row <- .compute_timestep_metrics(
        binary_raster   = binary_raster,
        test_pts_time   = test_pts_time,
        pseudo_pts_time = pseudo_pts_time,
        model           = if (is_parametric) model_list[[fold_key]] else NULL,
        model_vars      = model_vars,
        threshold       = if (is_parametric) thresholds[fold_key] else NULL,
        model_type      = model_type,
        time_values     = time_values,
        time_cols       = time_cols,
        fold            = fold
      )
      timestep_metrics <- rbind(timestep_metrics, timestep_row)
    }

    valid_proj <- fold_projections[!vapply(fold_projections, is.null, logical(1))]
    if (length(valid_proj) == 0) { warning(paste("No valid projections for", time_label)); next }

    combined_raster <- Reduce("+", valid_proj)

    tryCatch({
      terra::writeRaster(combined_raster, output_path, overwrite = TRUE)
      if (verbose) message(paste("  Saved:", basename(output_path)))
      prediction_files <- c(prediction_files, output_path)
    }, error = function(e) warning(paste("Could not save raster for", time_label, ":", e$message)))

    tryCatch(utils::write.csv(timestep_metrics, timestep_metrics_file, row.names = FALSE),
             error = function(e) warning(paste("Could not save timestep metrics:", e$message)))
  }

  if (verbose) message("--- Predictions complete ---")
  if (verbose) message(paste("Timestep metrics:", basename(timestep_metrics_file)))

  overall_summary <- data.frame()

  if (nrow(timestep_metrics) > 0) {
    has_pa_metrics <- all(c("TN", "FP") %in% names(timestep_metrics))

    for (fold in all_folds) {
      fm <- timestep_metrics[!is.na(timestep_metrics$Fold) & timestep_metrics$Fold == fold, ]
      if (nrow(fm) == 0) next

      sum_tp   <- sum(fm$TP,           na.rm = TRUE)
      sum_fn   <- sum(fm$FN,           na.rm = TRUE)
      tot_p    <- sum_tp + sum_fn
      mean_pct <- mean(fm$Pct_Suitable, na.rm = TRUE)
      sens     <- if (tot_p > 0) sum_tp / tot_p else NA_real_
      cbp_ov   <- if (tot_p > 0 && !is.na(mean_pct)) {
        stats::dbinom(sum_tp, size = tot_p, prob = mean_pct)
      } else NA_real_

      row <- data.frame(
        Fold               = fold,
        N_Timesteps        = nrow(fm),
        Mean_Pct_Suitable  = round(mean_pct, 4),
        Total_TP           = sum_tp,
        Total_FN           = sum_fn,
        Overall_Sensitivity = round(sens,    4),
        Overall_CBP        = cbp_ov,
        stringsAsFactors   = FALSE
      )

      if (has_pa_metrics) {
        sum_tn  <- sum(fm$TN, na.rm = TRUE)
        sum_fp  <- sum(fm$FP, na.rm = TRUE)
        tot_n   <- sum_tn + sum_fp
        spec    <- if (tot_n > 0) sum_tn / tot_n else NA_real_
        tss     <- if (!is.na(sens) && !is.na(spec)) sens + spec - 1 else NA_real_
        row$Total_TN            <- sum_tn
        row$Total_FP            <- sum_fp
        row$Overall_Specificity <- round(spec, 4)
        row$Overall_TSS         <- round(tss,  4)
      }

      overall_summary <- rbind(overall_summary, row)
    }

    if (nrow(overall_summary) > 0) {
      rate_cols <- c("Overall_Sensitivity", if (has_pa_metrics) c("Overall_Specificity", "Overall_TSS") else NULL)
      hdr_cols  <- c("Fold", "N_Steps", "MeanPct", "Tot_TP", "Tot_FN",
                     if (has_pa_metrics) c("Tot_TN", "Tot_FP") else NULL,
                     "Sens", if (has_pa_metrics) c("Spec", "TSS") else NULL, "CBP")
      if (verbose) message("  Overall pooled metrics per fold:")
      if (verbose) message(paste0("  ", paste(sprintf("%-8s", hdr_cols), collapse = " ")))

      for (i in seq_len(nrow(overall_summary))) {
        r <- overall_summary[i, ]
        vals <- c(r$Fold, r$N_Timesteps,
                  round(r$Mean_Pct_Suitable, 3),
                  r$Total_TP, r$Total_FN,
                  if (has_pa_metrics) c(r$Total_TN, r$Total_FP) else NULL,
                  round(r$Overall_Sensitivity, 3),
                  if (has_pa_metrics) c(round(r$Overall_Specificity, 3), round(r$Overall_TSS, 3)) else NULL,
                  formatC(r$Overall_CBP, format = "e", digits = 2))
        if (verbose) message(paste0("  ", paste(sprintf("%-8s", vals), collapse = " ")))
      }

      means <- c("",
                 round(mean(overall_summary$N_Timesteps, na.rm = TRUE), 0),
                 round(mean(overall_summary$Mean_Pct_Suitable, na.rm = TRUE), 3),
                 round(mean(overall_summary$Total_TP, na.rm = TRUE), 1),
                 round(mean(overall_summary$Total_FN, na.rm = TRUE), 1),
                 if (has_pa_metrics) c(
                   round(mean(overall_summary$Total_TN, na.rm = TRUE), 1),
                   round(mean(overall_summary$Total_FP, na.rm = TRUE), 1)
                 ) else NULL,
                 round(mean(overall_summary$Overall_Sensitivity, na.rm = TRUE), 3),
                 if (has_pa_metrics) c(
                   round(mean(overall_summary$Overall_Specificity, na.rm = TRUE), 3),
                   round(mean(overall_summary$Overall_TSS,         na.rm = TRUE), 3)
                 ) else NULL,
                 formatC(mean(overall_summary$Overall_CBP, na.rm = TRUE), format = "e", digits = 2))
      if (verbose) message(paste0("  Means    ", paste(sprintf("%-8s", means), collapse = " ")))
    }
  }

  invisible(list(
    timestep_metrics = timestep_metrics,
    overall_summary  = overall_summary,
    prediction_files = prediction_files,
    model_type       = model_type
  ))
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 5,
  dpi       = 100,
  out.width = "95%"
)

## -----------------------------------------------------------------------------
library(TemporalModelR)
library(terra)
library(sf)

raw_dir  <- system.file("extdata/rasters_raw", package = "TemporalModelR")

pts_file <- system.file("extdata/points/synthetic_occurrence_points.csv",
                        package = "TemporalModelR")

ref_file <- file.path(raw_dir, "elevation.tif")
study_crs <- sf::st_crs(terra::rast(ref_file))

## ----eval=FALSE---------------------------------------------------------------
# aligned_dir <- file.path(tempdir(), "rasters_aligned")
# 
# raster_align(
#   input_dir        = raw_dir,
#   output_dir       = aligned_dir,
#   reference_raster = ref_file,
#   resample_method  = "bilinear",
#   overwrite        = TRUE
# )

## -----------------------------------------------------------------------------
aligned_dir <- system.file("extdata/rasters_aligned", package = "TemporalModelR")

list.files(aligned_dir, pattern = "\\.tif$")[1:6]

## -----------------------------------------------------------------------------
rare_dir <- file.path(tempdir(), "rarefied")

rare_out <- spatiotemporal_rarefaction(
  points_sp        = pts_file,
  output_dir       = rare_dir,
  reference_raster = ref_file,
  time_cols        = c("year", "season"),
  xcol             = "x",
  ycol             = "y",
  points_crs       = study_crs,
  output_prefix    = "Pts_seasonal",
  verbose          = FALSE
)

rare_out$input_points

rare_out$spatial_points

rare_out$spatiotemporal_points

## -----------------------------------------------------------------------------
rare_out$files_created

## -----------------------------------------------------------------------------
rare_dir <- file.path(tempdir(), "rarefied")

rare_out_annual <- spatiotemporal_rarefaction(
  points_sp        = pts_file,
  output_dir       = rare_dir,
  reference_raster = ref_file,
  time_cols        = "year",
  xcol             = "x",
  ycol             = "y",
  points_crs       = study_crs,
  output_prefix    = "Pts_ann",
  verbose          = FALSE
)

rare_out_annual$input_points

rare_out_annual$spatial_points

rare_out_annual$spatiotemporal_points

## -----------------------------------------------------------------------------
ext_dir <- file.path(tempdir(), "extracted")

ext_out <- temporally_explicit_extraction(
  points_sp           = rare_out$files_created$spatiotemporal,
  raster_dir          = aligned_dir,
  variable_patterns   = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR",
    "prseas"       = "prseas_YEAR_SEASON"
  ),
  time_cols           = c("year", "season"),
  xcol                = "X",
  ycol                = "Y",
  points_crs          = study_crs,
  output_dir          = ext_dir,
  output_prefix       = "extracted_seasonal",
  save_raw            = TRUE,
  save_scaled         = TRUE,
  save_scaling_params = TRUE,
  verbose             = FALSE
)

head(ext_out$raw_values)

## -----------------------------------------------------------------------------
ext_out$files_created

## -----------------------------------------------------------------------------
ext_out$scaling_params

## ----eval=FALSE---------------------------------------------------------------
# scaled_dir <- file.path(tempdir(), "rasters_scaled")
# 
# scale_rasters(
#   input_dir           = aligned_dir,
#   output_dir          = scaled_dir,
#   scaling_params_file = ext_out$files_created$scaling_params,
#   variable_patterns   = c(
#     "elevation"    = "elevation",
#     "forest_cover" = "forest_cover_YEAR",
#     "prseas"       = "prseas_YEAR_SEASON"
#   ),
#   time_cols           = c("year", "season"),
#   overwrite           = TRUE,
#   verbose             = FALSE
# )

## -----------------------------------------------------------------------------
scaled_dir <- system.file("extdata/rasters_scaled", package = "TemporalModelR")

list.files(scaled_dir, pattern = "\\.tif$")[1:6]

## -----------------------------------------------------------------------------
ext_scaled_file <- system.file(
  "extdata/points/extracted_seasonal_Scaled_Values.csv",
  package = "TemporalModelR"
)

study_area_sf <- sf::st_as_sf(sf::st_as_sfc(
  sf::st_bbox(c(xmin = 0, xmax = 3000, ymin = 0, ymax = 1500),
              crs = study_crs)
))

partition <- spatiotemporal_partition(
  reference_shapefile_path = study_area_sf,
  points_file_path         = ext_scaled_file,
  xcol                     = "x",
  ycol                     = "y",
  points_crs               = study_crs,
  time_cols                = "year",
  n_spatial_folds          = 2,
  n_temporal_folds         = 2,
  max_attempts = 10,
  max_imbalance = 0.15,
  create_plot              = TRUE,
  verbose                  = FALSE
)

partition$summary

## -----------------------------------------------------------------------------
absences <- generate_absences(
  partition_result         = partition,
  reference_shapefile_path = study_area_sf,
  raster_dir               = scaled_dir,
  variable_patterns        = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR",
    "prseas"       = "prseas_YEAR_SEASON"
  ),
  method                   = "buffer",
  buffer_distance          = 300,
  ratio                    = 2,
  time_cols                = c("year", "season"),
  create_plot              = TRUE,
  plot_by_fold             = TRUE,
  verbose                  = FALSE
)

absences$summary

## -----------------------------------------------------------------------------
user_pts_file <- system.file(
  "extdata/points/synthetic_user_presences.csv",
  package = "TemporalModelR"
)

user_pts <- utils::read.csv(user_pts_file)

head(user_pts)

nrow(user_pts)

## -----------------------------------------------------------------------------
user_rare_dir <- file.path(tempdir(), "rarefied_user")

user_rare_out <- spatiotemporal_rarefaction(
  points_sp        = user_pts_file,
  output_dir       = user_rare_dir,
  reference_raster = ref_file,
  time_cols        = c("year", "season"),
  xcol             = "x",
  ycol             = "y",
  points_crs       = study_crs,
  output_prefix    = "Pts_user",
  verbose          = FALSE
)

user_rare_out$input_points

user_rare_out$spatiotemporal_points

## -----------------------------------------------------------------------------
absences_user <- generate_absences(
  partition_result         = partition,
  reference_shapefile_path = study_area_sf,
  raster_dir               = scaled_dir,
  variable_patterns        = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR",
    "prseas"       = "prseas_YEAR_SEASON"
  ),
  method                   = "user_data",
  user_absence_data        = user_rare_out$files_created$spatiotemporal,
  xcol                     = "X",
  ycol                     = "Y",
  points_crs               = study_crs,
  time_cols                = c("year", "season"),
  create_plot              = TRUE,
  plot_by_fold             = TRUE,
  verbose                  = FALSE
)

absences_user$summary

#' Generate Temporally Explicit Pseudoabsence Points
#'
#' Generates pseudoabsence or background points for each fold produced by
#' \code{\link{spatiotemporal_partition}}, distributed across time steps
#' proportionally to the number of presence points in each time step within each
#' fold. Three generation methods are supported: random sampling within the
#' study area, buffer-constrained sampling around presence points, and
#' environmentally biased sampling that targets areas outside the known
#' environmental tolerance of the species. Additionally, this function can be used
#' to process user defined absences to work with downflow operations. This may be
#' any list of points for which a user wants to count as 'absences' in downflow
#' operations, including negative occupancy surveys, or alternatively surveys for
#' similar species which may act as pseudoabsences for the presence of the species
#' of interest.
#'
#' @usage
#' generate_absences(partition_result, reference_shapefile_path, raster_dir,
#'                   variable_patterns, method = "random", ratio = 1,
#'                   buffer_distance = NULL, env_percentile = 0.05,
#'                   time_cols = NULL, pseudoabsence_times = NULL,
#'                   min_points_per_timestep = 1, user_absence_data = NULL,
#'                   xcol = NULL, ycol = NULL, points_crs = NULL,
#'                   create_plot = TRUE, plot_by_fold = FALSE,
#'                   plot_palette = "Dark 2", output_file = NULL,
#'                   verbose = TRUE)
#'
#' @param partition_result List or character. Output from
#'   \code{\link{spatiotemporal_partition}} or path to an \code{.rds}
#'   file containing that output.
#' @param reference_shapefile_path Character or sf object. Path to a polygon
#'   file or an \code{sf} polygon object defining the study area.
#' @param raster_dir Character. Directory containing environmental raster
#'   files (\code{.tif}), typically the output of
#'   \code{\link{raster_align}} or \code{\link{scale_rasters}}. File names
#'   must follow the patterns supplied in \code{variable_patterns}, with any
#'   time placeholder substituted for the corresponding value from
#'   \code{time_cols}. Required for all methods.
#' @param variable_patterns Named character vector mapping clean variable names
#'   to raster filename patterns. For time-varying variables include the time
#'   placeholder in the pattern (e.g. \code{"forest_cover" = "forest_cover_YEAR"});
#'   for static variables omit it (e.g. \code{"elevation" = "elevation"}). Time
#'   placeholders must match entries in \code{time_cols}.
#' @param method Character. Pseudoabsence generation method. One of
#'   \code{"random"}, \code{"buffer"}, \code{"environmental"}, or
#'   \code{"user_data"}. Default is \code{"random"}. When \code{"user_data"}
#'   is specified, \code{user_absence_data} is used as the source of absence
#'   locations instead of generating them. \code{ratio}, \code{buffer_distance},
#'   \code{env_percentile}, and \code{pseudoabsence_times} are ignored.
#'   \code{raster_dir} and \code{variable_patterns} are required for all
#'   methods, including \code{"user_data"}, for temporally-matched environmental
#'   extraction at the absence points.
#' @param ratio Numeric. Number of pseudoabsence points to generate per
#'   presence point. Default is \code{1}. Values of 2, 10, 50, etc. are
#'   accepted. Points are always distributed proportionally across time steps
#'   within each fold. Set to \code{0} to disable proportional allocation and
#'   use a fixed number of points per time step instead, in which case
#'   \code{min_points_per_timestep} must be greater than 0. \code{ratio} and
#'   \code{min_points_per_timestep} cannot both be 0.
#' @param buffer_distance Numeric. Distance in the units of the CRS (typically
#'   meters for projected CRS) within which pseudoabsence points are sampled.
#'   Required when \code{method = "buffer"}. When \code{method =
#'   "environmental"}, supplying a value automatically applies a spatial buffer
#'   constraint before environmental profiling, following the three-step approach
#'   of Senay et al. (2013). If \code{NULL} for the environmental method, no
#'   spatial constraint is applied. Default is \code{NULL}.
#' @param env_percentile Numeric between 0 and 1. Quantile threshold used to
#'   define the boundary of the known environmental tolerance when
#'   \code{method = "environmental"}. Environmental cells within this quantile
#'   range across all variables are excluded from pseudoabsence sampling.
#'   Default is \code{0.05} (5th to 95th percentile envelope).
#' @param time_cols Character or character vector. Name of the column(s)
#'   containing the time step values. Must match \code{time_cols} used in
#'   \code{\link{spatiotemporal_partition}} and the time placeholders used
#'   in \code{variable_patterns}. Default is \code{NULL}.
#' @param pseudoabsence_times Vector. Optional vector of specific time step
#'   values (for the first time column) at which to generate pseudoabsences.
#'   When \code{NULL} (default), all time steps present in the occurrence data
#'   are used.
#' @param min_points_per_timestep Integer. Minimum number of pseudoabsence
#'   points to generate per time step per fold. Default is \code{1}. When
#'   \code{ratio = 0}, this value sets the exact (fixed) number of points
#'   generated per time step per fold, independent of the number of
#'   presence points. \code{ratio} and \code{min_points_per_timestep} cannot
#'   both be 0.
#' @param user_absence_data  Character, sf object, sfc object, Spatial object,
#'   or data frame. Path to occurrence data (\code{.csv}, \code{.shp},
#'   \code{.geojson}, \code{.gpkg}) or a spatial object to be processed as absence
#'   data for downstream operations. Required when \code{method = "user_data"}.
#'   Should be preprocessed with \code{\link{spatiotemporal_rarefaction}} in the
#'   same formatting as presence data if not already thinned.
#' @param xcol Character. Name of the x-coordinate column in \code{user_absence_data}.
#'   Required when when \code{method = "environmental"} and \code{user_absence_data}
#'   is a CSV file or data frame.
#' @param ycol Character. Name of the y-coordinate column in \code{user_absence_data}.
#'   Required when when \code{method = "environmental"} and \code{user_absence_data}
#'   is a CSV file or data frame.
#' @param points_crs Character or CRS object. CRS of the \code{user_absence_data}.
#'   Required when when \code{method = "environmental"} and
#'   \code{user_absence_data} is a CSV file or data frame.
#' @param create_plot Logical. If \code{TRUE} (default), generates diagnostic
#'   plots showing the spatial and temporal distribution of generated
#'   pseudoabsence points alongside presence points.
#' @param plot_by_fold Logical. If \code{TRUE}, generates one map per fold. If
#'   \code{FALSE} (default), generates a single combined map.
#' @param plot_palette Character. Name of an HCL or RColorBrewer palette used
#'   to color folds in diagnostic plots. Accepts any HCL palette name (see
#'   \code{\link[grDevices]{hcl.pals}}) or, if \pkg{RColorBrewer} is installed,
#'   any Brewer palette name. Default is \code{"Dark 2"}.
#' @param output_file Character. Optional path to save the result as an
#'   \code{.rds} file. The parent directory will be created if it does not
#'   exist. Default is \code{NULL}.
#' @param verbose Logical. If \code{TRUE} (default), prints progress
#'   messages during processing. Includes per-fold and per-time-step
#'   pseudoabsence counts.
#'
#' @return Invisibly returns a list containing:
#' \itemize{
#'   \item \code{pseudoabsences}: An sf object of all generated pseudoabsence
#'     points with columns \code{fold}, \code{temporal_block}, \code{presence}
#'     (always 0), the time column(s) if provided, and extracted environmental
#'     variable values matched to each point's time step.
#'   \item \code{plots}: A named list of recorded plot objects when
#'     \code{create_plot = TRUE}. Contains \code{temporal_distribution} and
#'     either \code{spatial_combined} or one \code{spatial_fold_N} entry per
#'     fold. Plots can be replayed with \code{grDevices::replayPlot()}.
#'   \item \code{summary}: A data frame summarising points generated per fold
#'     with columns \code{fold}, \code{n_presences}, \code{n_pseudoabsences},
#'     and \code{ratio_achieved}.
#' }
#'
#' @details
#' Generates sets of background data based on user-specified methodology that
#' can be used as pseudoabsence data for the purposes of training
#' presence/absence models.
#'
#' The four generation methods differ in how the absence locations are obtained:
#' \itemize{
#'   \item \strong{Random}: Points are sampled uniformly at random from the
#'     full study area, excluding a negligible buffer around presence locations
#'     to prevent exact overlap.
#'   \item \strong{Buffer}: Points are sampled within buffers of radius
#'     \code{buffer_distance} drawn around all fold presences, clipped to the
#'     reference shapefile boundary.
#'   \item \strong{Environmental}: Raster cells whose values fall outside the
#'     species tolerance envelope in at least one variable are identified as
#'     candidates. K-means clustering then selects a spatially
#'     representative subset. If \code{buffer_distance} is supplied the
#'     environmental filtering is applied only within that buffered region,
#'     implementing the full three-step approach of Senay et al. (2013).
#'   \item \strong{User data}: Absence locations are taken directly from
#'     \code{user_absence_data} to be used when user has a predefined set of
#'     absense points. Points are assigned to folds by spatial join against the
#'     partition fold boundaries, with unmatched points routed to
#'     temporal folds by time value. Environmental values are then extracted
#'     at the supplied locations using the same time-matched logic as the
#'     generated methods.
#' }
#'
#' @references
#' Senay SD, Worner SP, Ikeda T (2013) Novel Three-Step Pseudo-Absence
#' Selection Technique for Improved Species Distribution Modeling.
#' PLoS ONE 8(8): e71218.
#'
#' @seealso
#' Preprocessing: \code{\link{spatiotemporal_partition}},
#'   \code{\link{temporally_explicit_extraction}}
#'
#' Modeling: \code{\link{build_temporal_hv}},
#'   \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
#'   \code{\link{build_temporal_rf}}
#'
#' @examples
#' data(tmr_partition_small, package = "TemporalModelR")
#'
#' scl_dir   <- system.file("extdata/rasters_scaled",
#'                          package = "TemporalModelR")
#'
#' ref_file  <- system.file("extdata/rasters_raw/elevation.tif",
#'                          package = "TemporalModelR")
#'
#' study_crs <- sf::st_crs(terra::rast(ref_file))
#'
#' study_area_sf <- sf::st_as_sf(sf::st_as_sfc(
#'   sf::st_bbox(c(xmin = 0, xmax = 3000, ymin = 0, ymax = 1500),
#'               crs = study_crs)
#' ))
#'
#' generate_absences(
#'   partition_result         = tmr_partition_small,
#'   reference_shapefile_path = study_area_sf,
#'   raster_dir               = scl_dir,
#'   variable_patterns        = c(
#'     "elevation"    = "elevation",
#'     "forest_cover" = "forest_cover_YEAR"
#'   ),
#'   method                   = "random",
#'   ratio                    = 1,
#'   time_cols                = c("year"),
#'   create_plot              = FALSE,
#'   verbose                  = FALSE
#' )

#' @export
#' @importFrom sf st_as_sf st_as_sfc st_buffer st_coordinates st_crs
#' @importFrom sf st_difference st_drop_geometry st_geometry st_intersection
#' @importFrom sf st_intersects st_is_empty st_point st_read st_sample st_sf
#' @importFrom sf st_transform st_union
#' @importFrom terra rast extract values vect ncell crs xyFromCell is.related
#' @importFrom graphics barplot legend par plot points
#' @importFrom grDevices adjustcolor recordPlot
#' @importFrom stats complete.cases kmeans quantile setNames
#' @importFrom tools file_ext
#' @importFrom utils read.csv
generate_absences <- function(partition_result,
                              reference_shapefile_path,
                              raster_dir,
                              variable_patterns,
                              method                   = "random",
                              ratio                    = 1,
                              buffer_distance          = NULL,
                              env_percentile           = 0.05,
                              time_cols                = NULL,
                              pseudoabsence_times      = NULL,
                              min_points_per_timestep  = 1,
                              user_absence_data        = NULL,
                              xcol                     = NULL,
                              ycol                     = NULL,
                              points_crs               = NULL,
                              create_plot              = TRUE,
                              plot_by_fold             = FALSE,
                              plot_palette             = "Dark 2",
                              output_file              = NULL,
                              verbose                  = TRUE) {

  if (missing(partition_result)) {
    stop(paste0("ERROR: 'partition_result' is required. Provide output from ",
                "spatiotemporal_partition() or a path to an .rds file."))
  }

  if (missing(reference_shapefile_path)) {
    stop("ERROR: 'reference_shapefile_path' is required.")
  }

  if (missing(raster_dir)) {
    stop("ERROR: 'raster_dir' is required for all methods.")
  }

  if (missing(variable_patterns)) {
    stop(paste0("ERROR: 'variable_patterns' is required. Provide a named character vector ",
                "mapping variable names to filename patterns, ",
                "e.g. c(\"Var1\" = \"Var1_YEAR\", \"StaticVar\" = \"StaticVar\")."))
  }

  .validate_variable_patterns(variable_patterns)

  if (!dir.exists(raster_dir)) {
    stop(paste0("ERROR: 'raster_dir' does not exist: ", raster_dir))
  }

  if (!is.null(output_file) && tolower(tools::file_ext(output_file)) != "rds") {
    stop("ERROR: 'output_file' must have a .rds extension.")
  }

  method <- match.arg(method, c("random", "buffer", "environmental", "user_data"))

  if (method == "user_data" && is.null(user_absence_data)) {
    stop("ERROR: 'user_absence_data' is required when method = 'user_data'.")
  }

  if (!is.numeric(ratio) || ratio < 0) {
    stop("ERROR: 'ratio' must be a non-negative number.")
  }

  if (!is.numeric(min_points_per_timestep) || min_points_per_timestep < 0) {
    stop("ERROR: 'min_points_per_timestep' must be a non-negative number.")
  }

  if (ratio == 0 && min_points_per_timestep == 0) {
    stop(paste0("ERROR: 'ratio' and 'min_points_per_timestep' cannot both be 0. ",
                "Set 'ratio' > 0 for proportional allocation, or ",
                "'min_points_per_timestep' > 0 for a fixed number of points per time step."))
  }

  if (method == "buffer" && is.null(buffer_distance)) {
    stop("ERROR: 'buffer_distance' is required when method = 'buffer'.")
  }

  if (!is.null(pseudoabsence_times)) {
    if (!is.vector(pseudoabsence_times) || is.list(pseudoabsence_times)) {
      stop("ERROR: 'pseudoabsence_times' must be a plain vector of time step values.")
    }
  }

  partition_result <- .load_partition_result(partition_result, verbose = verbose)

  use_buffer_with_env <- method == "environmental" && !is.null(buffer_distance)

  time_cols    <- if (is.null(time_cols)) character(0) else as.character(time_cols)
  var_classes  <- .classify_variable_patterns(variable_patterns, time_cols)
  dynamic_vars <- var_classes$dynamic
  static_vars  <- var_classes$static

  if (verbose) message(paste("Dynamic variables:", if (length(dynamic_vars) > 0) paste(dynamic_vars, collapse = ", ") else "none"))
  if (verbose) message(paste("Static variables:",  if (length(static_vars)  > 0) paste(static_vars,  collapse = ", ") else "none"))

  reference_shapefile <- .load_shapefile_input(reference_shapefile_path, "reference_shapefile_path")

  pts_sf <- partition_result$points_sf
  if (!"fold" %in% names(pts_sf)) {
    stop("ERROR: 'partition_result$points_sf' must contain a 'fold' column.")
  }

  reference_shapefile <- sf::st_transform(reference_shapefile, crs = sf::st_crs(pts_sf))
  ref_union           <- sf::st_union(reference_shapefile)
  all_folds           <- sort(unique(pts_sf$fold[!is.na(pts_sf$fold)]))

  has_time <- length(time_cols) > 0 && all(time_cols %in% names(pts_sf))
  if (length(time_cols) > 0 && !has_time) {
    missing_tc <- time_cols[!time_cols %in% names(pts_sf)]
    warning(paste0("time_cols column(s) not found in points_sf: ",
                   paste(missing_tc, collapse = ", "),
                   ". Pseudoabsences will not be matched to time steps."))
  }

  if (!is.null(pseudoabsence_times) && !has_time) {
    stop("ERROR: 'pseudoabsence_times' requires 'time_cols' to be specified and present in the data.")
  }

  if (!is.null(pseudoabsence_times)) {
    pseudoabsence_times <- sort(unique(pseudoabsence_times))
  }

  pts_sf$presence <- 1

  env_candidate_coords_by_year <- list()
  env_var_names                <- NULL

  if (method == "environmental") {

    if (verbose) message("Profiling environmental space from presence points...")

    unique_time_steps <- if (has_time) {
      {
        pts_df_all <- sf::st_drop_geometry(pts_sf)
        sub        <- unique(pts_df_all[, time_cols, drop = FALSE])
        sub        <- sub[do.call(order, sub), , drop = FALSE]
        lapply(seq_len(nrow(sub)), function(i) as.list(sub[i, , drop = FALSE]))
      }
    } else list(list())

    lower_bounds_by_timestep <- list()
    upper_bounds_by_timestep <- list()

    for (combo in unique_time_steps) {
      ts_key      <- if (has_time) paste(vapply(time_cols, function(tc) as.character(combo[[tc]]), character(1)), collapse = "_") else "static"
      time_values <- if (has_time) as.data.frame(combo, stringsAsFactors = FALSE) else
        as.data.frame(stats::setNames(lapply(time_cols, function(tc) NA_character_), time_cols),
                      stringsAsFactors = FALSE)
      raster_paths <- .resolve_raster_paths(variable_patterns, dynamic_vars, static_vars,
                                            time_cols, time_values, raster_dir)
      r_ts <- if (!is.null(raster_paths)) tryCatch(
        terra::rast(raster_paths),
        error = function(e) { warning(paste("Could not load rasters:", e$message)); NULL }
      ) else NULL
      if (is.null(r_ts)) {
        warning(paste0("Could not load rasters for time step ", ts_key,
                       ". Skipping environmental profiling for this time step."))
        next
      }

      ts_pts <- if (has_time) {
        pts_df <- sf::st_drop_geometry(pts_sf)
        pts_sf[Reduce(`&`, lapply(time_cols, function(tc) !is.na(pts_df[[tc]]) & pts_df[[tc]] == combo[[tc]])), ]
      } else pts_sf

      if (nrow(ts_pts) == 0) next

      presence_vect    <- suppressWarnings(terra::vect(ts_pts))
      env_at_presences <- terra::extract(r_ts, presence_vect, ID = FALSE)
      env_at_presences <- env_at_presences[stats::complete.cases(env_at_presences), , drop = FALSE]

      if (nrow(env_at_presences) == 0) next
      if (is.null(env_var_names)) env_var_names <- names(env_at_presences)

      lower_bounds_by_timestep[[ts_key]] <- vapply(env_var_names, function(v) {
        stats::quantile(env_at_presences[[v]], probs = env_percentile, na.rm = TRUE)
      }, numeric(1))
      upper_bounds_by_timestep[[ts_key]] <- vapply(env_var_names, function(v) {
        stats::quantile(env_at_presences[[v]], probs = 1 - env_percentile, na.rm = TRUE)
      }, numeric(1))
    }

    if (is.null(env_var_names)) {
      stop("ERROR: No environmental values could be extracted at presence points. ",
           "Check that raster_dir contains files matching variable_patterns and ",
           "that the rasters overlap with the presence locations.")
    }

    if (verbose) message(paste0("  Tolerance envelope: ", env_percentile * 100, "th to ",
                                (1 - env_percentile) * 100, "th percentiles across ",
                                length(env_var_names), " variables."))
    if (use_buffer_with_env) {
      if (verbose) message(paste0("  Spatial buffer constraint: ", buffer_distance, " units."))
    }

    for (combo in unique_time_steps) {
      ts_key       <- if (has_time) paste(vapply(time_cols, function(tc) as.character(combo[[tc]]), character(1)), collapse = "_") else "static"
      lower_bounds <- lower_bounds_by_timestep[[ts_key]]
      upper_bounds <- upper_bounds_by_timestep[[ts_key]]
      if (is.null(lower_bounds)) next

      time_values  <- if (has_time) as.data.frame(combo, stringsAsFactors = FALSE) else
        as.data.frame(stats::setNames(lapply(time_cols, function(tc) NA_character_), time_cols),
                      stringsAsFactors = FALSE)
      raster_paths <- .resolve_raster_paths(variable_patterns, dynamic_vars, static_vars,
                                            time_cols, time_values, raster_dir)
      r_ts <- if (!is.null(raster_paths)) tryCatch(
        terra::rast(raster_paths),
        error = function(e) { warning(paste("Could not load rasters:", e$message)); NULL }
      ) else NULL
      if (is.null(r_ts)) next

      all_cell_idx <- seq_len(terra::ncell(r_ts))
      all_env_vals <- as.data.frame(terra::values(r_ts))
      names(all_env_vals) <- env_var_names
      cell_xy      <- terra::xyFromCell(r_ts, all_cell_idx)
      ref_vect     <- terra::vect(ref_union)
      cell_vect    <- suppressWarnings(
        terra::vect(as.data.frame(cell_xy), geom = c("x", "y"),
                    crs = terra::crs(r_ts))
      )
      inside_ref <- terra::is.related(cell_vect, ref_vect, "intersects")

      outside_envelope <- rep(FALSE, nrow(all_env_vals))
      for (v in env_var_names) {
        v_vals           <- all_env_vals[, v]
        outside_envelope <- outside_envelope |
          (!is.na(v_vals) & (v_vals < lower_bounds[v] | v_vals > upper_bounds[v]))
      }

      candidate_mask   <- inside_ref & outside_envelope &
        !apply(is.na(all_env_vals), 1, any)
      coords_candidate <- cell_xy[candidate_mask, , drop = FALSE]
      valid_coords     <- !apply(is.na(coords_candidate), 1, any)
      env_candidate_coords_by_year[[ts_key]] <- coords_candidate[valid_coords, , drop = FALSE]

      if (verbose) message(paste0("  ", ts_key, ": ",
                                  nrow(env_candidate_coords_by_year[[ts_key]]),
                                  " candidate cells outside tolerance envelope."))
    }
  }

  if (method == "user_data") {

    if (verbose) message("Processing user-supplied absence data...")

    ud_df <- if (inherits(user_absence_data, "sf")) {
      sf::st_drop_geometry(user_absence_data)
    } else if (is.character(user_absence_data) && file.exists(user_absence_data)) {
      file_ext_lower <- tolower(tools::file_ext(user_absence_data))
      if (file_ext_lower == "csv") {
        if (is.null(xcol))       stop("ERROR: 'xcol' is required when 'user_absence_data' is a .csv file.")
        if (is.null(ycol))       stop("ERROR: 'ycol' is required when 'user_absence_data' is a .csv file.")
        if (is.null(points_crs)) stop("ERROR: 'points_crs' is required when 'user_absence_data' is a .csv file.")
        utils::read.csv(user_absence_data, stringsAsFactors = FALSE)
      } else if (file_ext_lower %in% c("shp", "geojson", "gpkg")) {
        tmp_sf <- sf::st_read(user_absence_data, quiet = TRUE)
        sf::st_drop_geometry(tmp_sf)
      } else {
        stop(paste0("ERROR: Unsupported file format '.", file_ext_lower,
                    "'. 'user_absence_data' must be a .csv, .shp, .geojson, or .gpkg file."))
      }
    } else if (inherits(user_absence_data, "SpatialPointsDataFrame")) {
      sf::st_drop_geometry(sf::st_as_sf(user_absence_data))
    } else if (is.data.frame(user_absence_data)) {
      if (is.null(xcol))       stop("ERROR: 'xcol' is required when 'user_absence_data' is a data frame.")
      if (is.null(ycol))       stop("ERROR: 'ycol' is required when 'user_absence_data' is a data frame.")
      if (is.null(points_crs)) stop("ERROR: 'points_crs' is required when 'user_absence_data' is a data frame.")
      user_absence_data
    } else {
      stop("ERROR: 'user_absence_data' must be an sf object, SpatialPointsDataFrame, data frame, or file path to a .csv, .shp, .geojson, or .gpkg file.")
    }

    if (inherits(user_absence_data, "sf")) {
      xcol_ud <- if (!is.null(xcol)) xcol else if ("X" %in% names(ud_df)) "X" else if ("x" %in% names(ud_df)) "x" else NULL
      ycol_ud <- if (!is.null(ycol)) ycol else if ("Y" %in% names(ud_df)) "Y" else if ("y" %in% names(ud_df)) "y" else NULL
      ud_crs  <- sf::st_crs(user_absence_data)
    } else if (is.character(user_absence_data) &&
               tolower(tools::file_ext(user_absence_data)) %in% c("shp", "geojson", "gpkg")) {
      xcol_ud <- if (!is.null(xcol)) xcol else if ("X" %in% names(ud_df)) "X" else if ("x" %in% names(ud_df)) "x" else NULL
      ycol_ud <- if (!is.null(ycol)) ycol else if ("Y" %in% names(ud_df)) "Y" else if ("y" %in% names(ud_df)) "y" else NULL
      ud_crs  <- sf::st_crs(sf::st_read(user_absence_data, quiet = TRUE))
    } else if (inherits(user_absence_data, "SpatialPointsDataFrame")) {
      xcol_ud <- if (!is.null(xcol)) xcol else if ("X" %in% names(ud_df)) "X" else if ("x" %in% names(ud_df)) "x" else NULL
      ycol_ud <- if (!is.null(ycol)) ycol else if ("Y" %in% names(ud_df)) "Y" else if ("y" %in% names(ud_df)) "y" else NULL
      ud_crs  <- sf::st_crs(sf::st_as_sf(user_absence_data))
    } else {
      xcol_ud <- xcol
      ycol_ud <- ycol
      ud_crs  <- sf::st_crs(points_crs)
    }

    if (is.null(xcol_ud) || is.null(ycol_ud)) {
      stop("ERROR: Could not determine coordinate columns for 'user_absence_data'. ",
           "Supply 'xcol' and 'ycol' explicitly.")
    }
    if (!xcol_ud %in% names(ud_df)) stop(paste0("ERROR: Column '", xcol_ud, "' not found in 'user_absence_data'."))
    if (!ycol_ud %in% names(ud_df)) stop(paste0("ERROR: Column '", ycol_ud, "' not found in 'user_absence_data'."))

    if (has_time && !all(time_cols %in% names(ud_df))) {
      missing_tc <- time_cols[!time_cols %in% names(ud_df)]
      stop(paste0("ERROR: 'user_absence_data' is missing time column(s): ",
                  paste(missing_tc, collapse = ", "), "."))
    }

    ud_sf <- sf::st_as_sf(ud_df, coords = c(xcol_ud, ycol_ud),
                          crs = ud_crs, remove = FALSE)

    if (!identical(sf::st_crs(ud_sf), sf::st_crs(pts_sf))) {
      ud_sf <- sf::st_transform(ud_sf, crs = sf::st_crs(pts_sf))
    }

    voronoi_sf    <- partition_result$voronoi_folds
    pts_df_full   <- sf::st_drop_geometry(pts_sf)

    ud_sf$fold           <- NA_integer_
    ud_sf$presence       <- 0L
    ud_sf$temporal_block <- NA_integer_

    if (!is.null(voronoi_sf) && "fold" %in% names(voronoi_sf)) {

      sentinel   <- max(all_folds) + 1L
      voronoi_t  <- sf::st_transform(voronoi_sf, crs = sf::st_crs(ud_sf))
      spatial_polys <- voronoi_t[!is.na(voronoi_t$fold) & voronoi_t$fold != sentinel, ]

      if (nrow(spatial_polys) > 0) {
        spatial_polys$voronoi_fold <- spatial_polys$fold
        spatial_polys <- spatial_polys[, "voronoi_fold"]

        ud_sf$.row_id <- seq_len(nrow(ud_sf))
        ud_joined     <- suppressMessages(suppressWarnings(
          sf::st_join(ud_sf, spatial_polys, left = TRUE)
        ))
        ud_joined     <- ud_joined[!duplicated(ud_joined$.row_id), ]
        joined_folds  <- ud_joined$voronoi_fold[match(seq_len(nrow(ud_sf)), ud_joined$.row_id)]
        ud_sf$fold    <- joined_folds
        ud_sf$.row_id <- NULL
      }
    }

    unassigned_idx <- which(is.na(ud_sf$fold))

    if (length(unassigned_idx) > 0 && has_time) {
      spatial_fold_ids  <- sort(unique(ud_sf$fold[!is.na(ud_sf$fold)]))
      temporal_fold_ids <- sort(setdiff(all_folds, spatial_fold_ids))

      if (length(temporal_fold_ids) > 0) {
        time_col1 <- time_cols[1]
        tf_ranges <- lapply(temporal_fold_ids, function(f) {
          tv <- pts_df_full[[time_col1]][!is.na(pts_df_full$fold) & pts_df_full$fold == f]
          tv <- tv[!is.na(tv)]
          if (length(tv) == 0) return(NULL)
          list(fold = f, min = min(tv), max = max(tv))
        })
        tf_ranges <- Filter(Negate(is.null), tf_ranges)

        if (length(tf_ranges) > 0) {
          ud_times    <- ud_df[[time_col1]][unassigned_idx]
          assigned_tf <- vapply(ud_times, function(t) {
            if (is.na(t)) return(NA_integer_)
            for (r in tf_ranges) if (t >= r$min && t <= r$max) return(r$fold)
            dists <- vapply(tf_ranges, function(r) min(abs(t - r$min), abs(t - r$max)), numeric(1))
            tf_ranges[[which.min(dists)]]$fold
          }, integer(1))
          ud_sf$fold[unassigned_idx] <- assigned_tf
        }
      }
    }

    n_unassigned <- sum(is.na(ud_sf$fold))
    if (n_unassigned > 0) {
      warning(paste0(n_unassigned, " point(s) in 'user_absence_data' could not be ",
                     "assigned to any fold and were dropped."))
      ud_sf <- ud_sf[!is.na(ud_sf$fold), ]
    }
    if (nrow(ud_sf) == 0) {
      stop("ERROR: No 'user_absence_data' points could be assigned to any fold. ",
           "Check that CRS, spatial extent, and time values match the partition.")
    }

    pseudoabs_sf <- ud_sf

    if (verbose) {
      message(paste0("  ", nrow(pseudoabs_sf),
                     " user-supplied absence points assigned across folds."))
      for (f in all_folds) {
        n_f <- sum(!is.na(pseudoabs_sf$fold) & pseudoabs_sf$fold == f)
        message(paste0("  Fold ", f, ": ", n_f, " absence points"))
      }
    }

  } else {

    if (verbose) message(paste0("Generating pseudoabsence points across ",
                                length(all_folds), " fold",
                                if (length(all_folds) > 1) "s" else "", "..."))

    result_list <- list()

    for (fold_id in all_folds) {

      fold_pts <- pts_sf[!is.na(pts_sf$fold) & pts_sf$fold == fold_id, ]
      if (nrow(fold_pts) == 0) next

      n_presences_fold <- nrow(fold_pts)
      if (ratio == 0) {
        n_pseudoabs_fold <- NA_integer_
      } else {
        n_pseudoabs_fold <- max(1, round(n_presences_fold * ratio))
      }

      if (has_time) {
        fold_df        <- sf::st_drop_geometry(fold_pts)
        sub_ft         <- unique(fold_df[, time_cols, drop = FALSE])
        sub_ft         <- sub_ft[do.call(order, sub_ft), , drop = FALSE]
        presence_times <- lapply(seq_len(nrow(sub_ft)),
                                 function(i) as.list(sub_ft[i, , drop = FALSE]))
      } else {
        presence_times <- list(list())
      }

      unique_times <- if (!is.null(pseudoabsence_times) && has_time) {
        other_cols <- time_cols[-1]
        if (length(other_cols) > 0) {
          all_df <- sf::st_drop_geometry(pts_sf)
          other_combos <- unique(all_df[, other_cols, drop = FALSE])
          cross <- expand.grid(
            c(list(pseudoabsence_times), as.list(other_combos)),
            stringsAsFactors = FALSE
          )
          names(cross) <- time_cols
          lapply(seq_len(nrow(cross)), function(i) as.list(cross[i, , drop = FALSE]))
        } else {
          lapply(pseudoabsence_times, function(v) stats::setNames(list(v), time_cols[1]))
        }
      } else {
        presence_times
      }

      if (length(unique_times) == 0 || (length(unique_times) == 1 && length(unique_times[[1]]) == 0)) {
        if (ratio == 0) {
          time_alloc <- as.integer(min_points_per_timestep)
        } else {
          time_alloc <- n_pseudoabs_fold
        }
        names(time_alloc) <- "all"
      } else {
        ts_keys <- vapply(unique_times, function(combo) {
          if (has_time) paste(vapply(time_cols, function(tc) as.character(combo[[tc]]), character(1)), collapse = "_") else "all"
        }, character(1))

        time_counts <- vapply(unique_times, function(combo) {
          if (!has_time || length(presence_times) == 0) return(0)
          fold_df <- sf::st_drop_geometry(fold_pts)
          sum(Reduce(`&`, lapply(time_cols, function(tc) !is.na(fold_df[[tc]]) & fold_df[[tc]] == combo[[tc]])))
        }, integer(1))
        names(time_counts) <- ts_keys

        if (ratio == 0) {
          time_alloc <- rep(as.integer(min_points_per_timestep), length(unique_times))
        } else {
          raw_alloc  <- (time_counts / max(sum(time_counts), 1)) * n_pseudoabs_fold
          time_alloc <- pmax(as.integer(round(raw_alloc)), as.integer(min_points_per_timestep))

          diff_pts <- sum(time_alloc) - n_pseudoabs_fold
          if (diff_pts > 0) {
            ord <- order(time_alloc, decreasing = TRUE)
            for (i in seq_len(min(diff_pts, length(ord)))) {
              if (time_alloc[ord[i]] > min_points_per_timestep) {
                time_alloc[ord[i]] <- time_alloc[ord[i]] - 1
              }
            }
          } else if (diff_pts < 0) {
            ord <- order(time_alloc, decreasing = FALSE)
            for (i in seq_len(min(abs(diff_pts), length(ord)))) {
              time_alloc[ord[i]] <- time_alloc[ord[i]] + 1
            }
          }
        }
        names(time_alloc) <- ts_keys
      }

      fold_buffer_region <- NULL
      if (method == "buffer" || use_buffer_with_env) {
        fold_buffer_region <- suppressWarnings(
          sf::st_intersection(
            sf::st_buffer(sf::st_union(fold_pts), dist = buffer_distance),
            ref_union
          )
        )
      }

      for (ti in seq_along(unique_times)) {

        combo  <- unique_times[[ti]]
        ts_key <- names(time_alloc)[ti]
        n_pts  <- time_alloc[ti]
        if (n_pts <= 0) next

        ts_pts <- if (has_time && length(combo) > 0) {
          fold_df <- sf::st_drop_geometry(fold_pts)
          fold_pts[Reduce(`&`, lapply(time_cols, function(tc) !is.na(fold_df[[tc]]) & fold_df[[tc]] == combo[[tc]])), ]
        } else fold_pts

        excl_buffer <- if (nrow(ts_pts) > 0) {
          suppressWarnings(sf::st_buffer(sf::st_union(ts_pts), dist = 1e-6))
        } else NULL

        if (method == "random") {
          sampling_region <- if (!is.null(excl_buffer)) {
            suppressWarnings(sf::st_difference(ref_union, excl_buffer))
          } else ref_union

          sampled <- tryCatch({
            s <- sf::st_sample(sampling_region, size = n_pts * 5, type = "random")
            s <- s[!sf::st_is_empty(s)]
            if (length(s) > n_pts) s[seq_len(n_pts)] else s
          }, error = function(e) {
            warning(paste0("Fold ", fold_id, " / time ", ts_key,
                           ": random sampling failed -- ", e$message))
            NULL
          })
        }

        if (method == "buffer") {
          sampling_region <- if (!is.null(fold_buffer_region) &&
                                 !sf::st_is_empty(fold_buffer_region)) {
            fold_buffer_region
          } else {
            warning(paste0("Fold ", fold_id, " / time ", ts_key,
                           ": buffer region is empty -- falling back to full study area."))
            ref_union
          }
          if (!is.null(excl_buffer)) {
            sampling_region <- suppressWarnings(sf::st_difference(sampling_region, excl_buffer))
          }
          sampled <- tryCatch({
            s <- sf::st_sample(sampling_region, size = n_pts * 5, type = "random")
            s <- s[!sf::st_is_empty(s)]
            if (length(s) > n_pts) s[seq_len(n_pts)] else s
          }, error = function(e) {
            warning(paste0("Fold ", fold_id, " / time ", ts_key,
                           ": buffer sampling failed -- ", e$message))
            NULL
          })
        }

        if (method == "environmental") {
          env_ts_key <- if (has_time && length(env_candidate_coords_by_year) > 0) {
            if (ts_key %in% names(env_candidate_coords_by_year)) {
              ts_key
            } else if ("static" %in% names(env_candidate_coords_by_year)) {
              "static"
            } else names(env_candidate_coords_by_year)[1]
          } else names(env_candidate_coords_by_year)[1]

          candidate_coords <- env_candidate_coords_by_year[[env_ts_key]]

          if (is.null(candidate_coords) || nrow(candidate_coords) == 0) {
            warning(paste0("Fold ", fold_id, " / time ", ts_key,
                           ": no candidate cells outside tolerance envelope. Skipping."))
            sampled <- NULL
          } else {
            if (use_buffer_with_env && !is.null(fold_buffer_region) &&
                !sf::st_is_empty(fold_buffer_region)) {
              cand_sf    <- sf::st_as_sf(as.data.frame(candidate_coords),
                                         coords = c("x", "y"), crs = sf::st_crs(pts_sf))
              inside_buf <- suppressMessages(suppressWarnings(
                lengths(sf::st_intersects(cand_sf, fold_buffer_region)) > 0
              ))
              candidate_coords <- candidate_coords[inside_buf, , drop = FALSE]
            }

            if (nrow(candidate_coords) == 0) {
              warning(paste0("Fold ", fold_id, " / time ", ts_key,
                             ": no candidate cells within buffer region. Skipping."))
              sampled <- NULL
            } else {
              n_clusters <- min(n_pts, nrow(candidate_coords))
              km <- if (n_clusters < nrow(candidate_coords)) {
                suppressWarnings(tryCatch(
                  stats::kmeans(candidate_coords, centers = n_clusters,
                                nstart = 5, iter.max = 300, algorithm = "Lloyd"),
                  error = function(e) {
                    warning(paste0("Fold ", fold_id, " / time ", ts_key,
                                   ": k-means failed -- ", e$message,
                                   ". Falling back to random sample."))
                    NULL
                  }
                ))
              } else NULL

              selected_coords <- if (!is.null(km)) {
                km$centers
              } else {
                idx <- if (nrow(candidate_coords) > n_clusters) {
                  sample(nrow(candidate_coords), n_clusters)
                } else seq_len(nrow(candidate_coords))
                candidate_coords[idx, , drop = FALSE]
              }

              sampled <- tryCatch({
                sf::st_as_sfc(
                  lapply(seq_len(nrow(selected_coords)), function(i) {
                    sf::st_point(selected_coords[i, ])
                  }),
                  crs = sf::st_crs(pts_sf)
                )
              }, error = function(e) {
                warning(paste0("Fold ", fold_id, " / time ", ts_key,
                               ": could not build geometry from cluster centroids -- ",
                               e$message))
                NULL
              })
            }
          }
        }

        if (is.null(sampled) || length(sampled) == 0) next

        sampled_sf           <- sf::st_as_sf(sampled)
        names(sampled_sf)[1] <- "geometry"
        sf::st_geometry(sampled_sf) <- "geometry"
        sampled_sf$fold      <- fold_id
        sampled_sf$presence  <- 0

        sampled_sf$temporal_block <- if ("temporal_block" %in% names(fold_pts)) {
          if (has_time && length(combo) > 0) {
            fold_df <- sf::st_drop_geometry(fold_pts)
            tb_v    <- fold_df$temporal_block[Reduce(`&`, lapply(time_cols, function(tc) !is.na(fold_df[[tc]]) & fold_df[[tc]] == combo[[tc]]))]
            if (length(tb_v) > 0) tb_v[1] else NA_integer_
          } else NA_integer_
        } else NA_integer_

        if (has_time && length(combo) > 0) {
          for (tc in time_cols) {
            sampled_sf[[tc]] <- combo[[tc]]
          }
        }

        result_list[[length(result_list) + 1]] <- sampled_sf
      }
    }

    if (length(result_list) == 0) {
      stop("ERROR: No pseudoabsence points could be generated. ",
           "Check method parameters, buffer distance, and raster extent.")
    }

    all_cols    <- unique(unlist(lapply(result_list, names)))
    result_list <- lapply(result_list, function(x) {
      for (col in setdiff(all_cols, names(x))) x[[col]] <- NA
      x[, all_cols, drop = FALSE]
    })
    pseudoabs_sf <- do.call(rbind, result_list)

  }

  if (verbose) message("Extracting time-matched environmental values at pseudoabsence points...")

  if (has_time && all(time_cols %in% names(pseudoabs_sf))) {
    pa_df       <- sf::st_drop_geometry(pseudoabs_sf)
    sub_pa    <- unique(pa_df[, time_cols, drop = FALSE])
    sub_pa    <- sub_pa[do.call(order, sub_pa), , drop = FALSE]
    pa_combos <- lapply(seq_len(nrow(sub_pa)),
                        function(i) as.list(sub_pa[i, , drop = FALSE]))

    for (combo in pa_combos) {
      pa_idx  <- which(Reduce(`&`, lapply(time_cols, function(tc) !is.na(pa_df[[tc]]) & pa_df[[tc]] == combo[[tc]])))
      if (length(pa_idx) == 0) next
      raster_paths <- .resolve_raster_paths(variable_patterns, dynamic_vars, static_vars,
                                            time_cols,
                                            as.data.frame(combo, stringsAsFactors = FALSE),
                                            raster_dir)
      r_ts <- if (!is.null(raster_paths)) tryCatch(
        terra::rast(raster_paths),
        error = function(e) { warning(paste("Could not load rasters:", e$message)); NULL }
      ) else NULL
      if (is.null(r_ts)) next
      pa_sub <- suppressWarnings(terra::vect(pseudoabs_sf[pa_idx, ]))
      env_ex <- terra::extract(r_ts, pa_sub, ID = FALSE)
      var_order <- c(dynamic_vars, static_vars)
      if (ncol(env_ex) == length(var_order)) names(env_ex) <- var_order
      for (v in var_order) {
        if (!v %in% names(env_ex)) next
        if (!v %in% names(pseudoabs_sf)) pseudoabs_sf[[v]] <- NA_real_
        pseudoabs_sf[[v]][pa_idx] <- env_ex[[v]]
      }
    }
  } else {
    static_paths <- .resolve_raster_paths(variable_patterns, dynamic_vars, static_vars,
                                          time_cols,
                                          as.data.frame(stats::setNames(
                                            lapply(time_cols, function(tc) NA_character_),
                                            time_cols), stringsAsFactors = FALSE),
                                          raster_dir)
    r_static <- if (!is.null(static_paths)) tryCatch(
      terra::rast(static_paths),
      error = function(e) { warning(paste("Could not load rasters:", e$message)); NULL }
    ) else NULL
    if (!is.null(r_static)) {
      pa_vect <- suppressWarnings(terra::vect(pseudoabs_sf))
      env_ex  <- terra::extract(r_static, pa_vect, ID = FALSE)
      var_order <- c(dynamic_vars, static_vars)
      if (ncol(env_ex) == length(var_order)) names(env_ex) <- var_order
      for (v in var_order) {
        if (!v %in% names(env_ex)) next
        pseudoabs_sf[[v]] <- env_ex[[v]]
      }
    }
  }

  if (verbose) message(paste0("  Done. ", nrow(pseudoabs_sf), " pseudoabsence points generated."))

  fold_summary <- do.call(rbind, lapply(all_folds, function(f) {
    n_pr <- sum(!is.na(pts_sf$fold) & pts_sf$fold == f)
    n_pa <- sum(!is.na(pseudoabs_sf$fold) & pseudoabs_sf$fold == f)
    data.frame(fold             = f,
               n_presences      = n_pr,
               n_pseudoabsences = n_pa,
               ratio_achieved   = round(n_pa / max(n_pr, 1), 3),
               stringsAsFactors = FALSE)
  }))

  method_label <- switch(method,
                         random        = "Random",
                         buffer        = paste0("Buffer (", buffer_distance, " units)"),
                         environmental = if (use_buffer_with_env) {
                           paste0("Environmental + Buffer (", buffer_distance, " units)")
                         } else "Environmental",
                         user_data     = "User-supplied data"
  )

  plot_list    <- list()
  presence_col <- "steelblue"
  absence_col  <- "tomato"

  if (create_plot) {

    fold_colors <- .resolve_palette(length(all_folds), plot_palette)
    names(fold_colors) <- as.character(all_folds)

    if (has_time && all(time_cols %in% names(pseudoabs_sf))) {
      pa_df    <- sf::st_drop_geometry(pseudoabs_sf)
      pr_df    <- sf::st_drop_geometry(pts_sf)
      pa_time  <- apply(pa_df[, time_cols, drop = FALSE], 1,
                        function(r) paste(r, collapse = "_"))
      pr_time  <- apply(pr_df[, time_cols, drop = FALSE], 1,
                        function(r) paste(r, collapse = "_"))
      all_time <- sort(unique(c(pa_time, pr_time)))
      all_time <- all_time[!is.na(all_time) & all_time != "NA"]
      time_label <- paste(time_cols, collapse = " / ")

      count_matrix <- rbind(
        Presence      = vapply(all_time, function(t)
          sum(!is.na(pr_time) & pr_time == t), integer(1)),
        Pseudoabsence = vapply(all_time, function(t)
          sum(!is.na(pa_time) & pa_time == t), integer(1))
      )
      colnames(count_matrix) <- as.character(all_time)

      label_interval <- max(1, round(length(all_time) / 15))
      display_labels <- rep("", length(all_time))
      display_labels[seq(1, length(all_time), by = label_interval)] <-
        as.character(all_time)[seq(1, length(all_time), by = label_interval)]

      opar <- graphics::par(no.readonly = TRUE)
      on.exit(graphics::par(opar), add = TRUE)
      tryCatch({
        graphics::barplot(count_matrix, beside = TRUE,
                          col = c(presence_col, absence_col),
                          names.arg = display_labels, las = 2, border = NA,
                          main = "Temporal Distribution",
                          xlab = time_label, ylab = "Count")
        graphics::legend("topright",
                         legend = c("Presence", "Pseudoabsence"),
                         fill = c(presence_col, absence_col),
                         border = NA, cex = 0.8, bty = "n")
        plot_list$temporal_distribution <- grDevices::recordPlot()
      }, error = function(e) {
        warning(paste0("Could not generate temporal distribution plot: ", e$message))
      })
    }

    if (plot_by_fold) {
      for (fi in seq_along(all_folds)) {
        fold_id <- all_folds[fi]
        fcol    <- fold_colors[as.character(fold_id)]
        fold_pa <- pseudoabs_sf[!is.na(pseudoabs_sf$fold) & pseudoabs_sf$fold == fold_id, ]
        fold_pr <- pts_sf[!is.na(pts_sf$fold) & pts_sf$fold == fold_id, ]

        opar <- graphics::par(no.readonly = TRUE)
        on.exit(graphics::par(opar), add = TRUE)
        tryCatch({
          graphics::plot(sf::st_geometry(reference_shapefile),
                         col = "gray97", border = "gray50",
                         main = paste0("Fold ", fold_id, " Pseudoabsences"))
          if (nrow(fold_pa) > 0) {
            pa_coords <- sf::st_coordinates(fold_pa)
            graphics::points(pa_coords[, 1], pa_coords[, 2],
                             col = grDevices::adjustcolor(absence_col, alpha.f = 0.6),
                             pch = 4, cex = 0.7)
          }
          if (nrow(fold_pr) > 0) {
            pr_coords <- sf::st_coordinates(fold_pr)
            graphics::points(pr_coords[, 1], pr_coords[, 2],
                             col = fcol, pch = 16, cex = 0.8)
          }
          graphics::legend("topright",
                           legend = c(paste0("Presence (n = ", nrow(fold_pr), ")"),
                                      paste0("Pseudoabsence (n = ", nrow(fold_pa), ")")),
                           col = c(fcol, grDevices::adjustcolor(absence_col, alpha.f = 0.6)),
                           pch = c(16, 4), cex = 0.8, bty = "n")
          plot_list[[paste0("spatial_fold_", fold_id)]] <- grDevices::recordPlot()
        }, error = function(e) {
          warning(paste0("Could not generate plot for fold ", fold_id, ": ", e$message))
        })
      }
    } else {
      opar <- graphics::par(no.readonly = TRUE)
      on.exit(graphics::par(opar), add = TRUE)
      tryCatch({
        graphics::plot(sf::st_geometry(reference_shapefile),
                       col = "gray97", border = "gray50",
                       main = "Pseudoabsences")
        if (nrow(pseudoabs_sf) > 0) {
          pa_coords <- sf::st_coordinates(pseudoabs_sf)
          graphics::points(pa_coords[, 1], pa_coords[, 2],
                           col = grDevices::adjustcolor(absence_col, alpha.f = 0.5),
                           pch = 4, cex = 0.6)
        }
        for (fi in seq_along(all_folds)) {
          fold_id <- all_folds[fi]
          fold_pr <- pts_sf[!is.na(pts_sf$fold) & pts_sf$fold == fold_id, ]
          if (nrow(fold_pr) == 0) next
          pr_coords <- sf::st_coordinates(fold_pr)
          graphics::points(pr_coords[, 1], pr_coords[, 2],
                           col = fold_colors[as.character(fold_id)], pch = 16, cex = 0.8)
        }
        graphics::legend("topright",
                         legend = c(paste0("Presence - Fold ", all_folds),
                                    paste0("Pseudoabsence (n = ", nrow(pseudoabs_sf), ")")),
                         col = c(fold_colors,
                                 grDevices::adjustcolor(absence_col, alpha.f = 0.5)),
                         pch = c(rep(16, length(all_folds)), 4),
                         cex = 0.75, bty = "n")
        plot_list$spatial_combined <- grDevices::recordPlot()
      }, error = function(e) {
        warning(paste0("Could not generate spatial plot: ", e$message))
      })
    }
  }

  if (verbose) message("\nPseudoabsence generation complete.")
  if (verbose) message(paste0("  Method: ", method_label))
  if (method != "user_data") {
    if (verbose) message(paste0("  Ratio: ", ratio,
                                " | Presences: ", nrow(pts_sf),
                                " | Pseudoabsences: ", nrow(pseudoabs_sf)))
  } else {
    if (verbose) message(paste0("  Presences: ", nrow(pts_sf),
                                " | User absences: ", nrow(pseudoabs_sf)))
  }
  for (f in all_folds) {
    row <- fold_summary[fold_summary$fold == f, ]
    if (verbose) message(paste0("  Fold ", f, ": ",
                                row$n_presences, " presences, ",
                                row$n_pseudoabsences, " pseudoabsences"))
  }

  if (!is.null(output_file)) {
    out_dir <- dirname(output_file)
    if (!dir.exists(out_dir)) dir.create(out_dir, recursive = TRUE, showWarnings = FALSE)
    saveRDS(list(pseudoabsences = pseudoabs_sf,
                 plots          = plot_list,
                 summary        = fold_summary),
            output_file)
    if (verbose) message(paste0("Results saved to: ", output_file))
  }

  invisible(list(pseudoabsences = pseudoabs_sf,
                 plots          = plot_list,
                 summary        = fold_summary))
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 5,
  dpi       = 100,
  out.width = "95%"
)

## -----------------------------------------------------------------------------
library(TemporalModelR)
library(terra)

data(tmr_partition, package = "TemporalModelR")

data(tmr_absences,  package = "TemporalModelR")

## -----------------------------------------------------------------------------
glm_out <- build_temporal_glm(
  partition_result     = tmr_partition,
  pseudoabsence_result = tmr_absences,
  model_formula        = ~ elevation + forest_cover + prseas,
  link                 = "logit",
  threshold_method     = "tss",
  output_dir           = file.path(tempdir(), "GLM_Models"),
  create_plot          = TRUE,
  time_cols            = c("year", "season"),
  verbose              = FALSE
)

## -----------------------------------------------------------------------------
class(glm_out)

names(glm_out)

## ----eval=FALSE---------------------------------------------------------------
# ### Additive
# ~ elevation + forest_cover + prseas
# 
# ### Quadratic term on elevation
# ~ elevation + I(elevation^2) + forest_cover + prseas
# 
# ### Orthogonal polynomial of degree 2 on elevation
# ~ poly(elevation, 2) + forest_cover + prseas
# 
# ### Forest cover by precipitation interaction
# ~ elevation + forest_cover * prseas

## -----------------------------------------------------------------------------
glm_out$thresholds

## -----------------------------------------------------------------------------
glm_out$fold_test_metrics

## ----eval=FALSE---------------------------------------------------------------
# scaled_dir <- system.file("extdata/rasters_scaled", package = "TemporalModelR")
# 
# time_steps <- expand.grid(
#   year             = 1:15,
#   season           = c("Spring", "Summer", "Autumn", "Winter"),
#   stringsAsFactors = FALSE
# )
# 
# preds <- generate_spatiotemporal_predictions(
#   partition_result     = tmr_partition,
#   model_result         = glm_out,
#   pseudoabsence_result = tmr_absences,
#   raster_dir           = scaled_dir,
#   variable_patterns    = c(
#     "elevation"    = "elevation",
#     "forest_cover" = "forest_cover_YEAR",
#     "prseas"       = "prseas_YEAR_SEASON"
#   ),
#   time_cols            = c("year", "season"),
#   time_steps           = time_steps,
#   output_dir           = file.path(tempdir(), "GLM_Predictions"),
#   overwrite            = TRUE,
#   verbose              = FALSE
# )

## ----echo=FALSE---------------------------------------------------------------
preds <- readRDS(system.file("extdata/precomputed/glm/preds.rds",
                             package = "TemporalModelR"))
preds$prediction_files <- list.files(
  system.file("extdata/precomputed/glm/pred_tifs", package = "TemporalModelR"),
  pattern = "\\.tif$", full.names = TRUE
)

## ----fig.width=10, fig.height=5-----------------------------------------------
pred_stack <- terra::rast(preds$prediction_files)

pred_names    <- basename(preds$prediction_files)
pred_seasons  <- sub(".*_(Spring|Summer|Autumn|Winter)\\.tif$", "\\1", pred_names)
pred_years    <- as.numeric(sub(".*_(\\d+)_(Spring|Summer|Autumn|Winter)\\.tif$",
                                "\\1", pred_names))
season_levels <- c("Spring", "Summer", "Autumn", "Winter")
stack_order   <- order(pred_years, match(pred_seasons, season_levels))

pred_stack <- pred_stack[[stack_order]]
ordered_years   <- pred_years[stack_order]
ordered_seasons <- pred_seasons[stack_order]
names(pred_stack) <- paste0("Y", ordered_years, "_", ordered_seasons)

block1 <- which(ordered_years %in%  1:4)
block2 <- which(ordered_years %in%  5:8)
block3 <- which(ordered_years %in%  9:12)
block4 <- which(ordered_years %in% 13:15)

## ----fig.width=10, fig.height=5-----------------------------------------------
terra::plot(pred_stack[[block1]], nr = 4, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0),  legend = FALSE)

## ----fig.width=10, fig.height=5-----------------------------------------------
terra::plot(pred_stack[[block2]], nr = 4, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0),  legend = FALSE)

## ----fig.width=10, fig.height=5-----------------------------------------------
terra::plot(pred_stack[[block3]], nr = 4, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0),  legend = FALSE)

## ----fig.width=10, fig.height=4-----------------------------------------------
terra::plot(pred_stack[[block4]], nr = 3, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0),  legend = FALSE)

## -----------------------------------------------------------------------------
head(preds$timestep_metrics)

## -----------------------------------------------------------------------------
preds$overall_summary

## -----------------------------------------------------------------------------
plot_model_assessment(
  predictions         = preds,
  time_column         = c("year", "season"),
  secondary_time_mode = "combine",
  model_result        = glm_out
)

## ----fig.height=20------------------------------------------------------------
plot_model_assessment(
  predictions         = preds,
  time_column         = c("year", "season"),
  secondary_time_mode = "facet",
  model_result        = glm_out,
  cbp_threshold       = 0.001
)

#' Align and Standardize Raster Files to a Reference Raster
#'
#' Preprocessing function that aligns a batch of raster files to a specified
#' reference raster by performing reprojection, resampling, and masking
#' operations. Ensures all rasters share identical CRS, resolution, and spatial
#' extent for later analyses.
#'
#' @usage
#' raster_align(input_dir, output_dir, reference_raster,
#'              output_suffix = "_Masked_Updated", pattern = ".*\\\\.tif$",
#'              resample_method = "bilinear",
#'              overwrite = FALSE, verbose = TRUE)
#'
#' @param input_dir Character. Directory containing the input raster files.
#' @param output_dir Character. Directory where processed rasters will be saved.
#' @param reference_raster Character, SpatRaster, or RasterLayer. File path or
#'   raster object used as the alignment reference.
#' @param output_suffix Character. Suffix appended to output filenames. Default
#'   is \code{"_Masked_Updated"}. Output files are named
#'   \code{<original_name><output_suffix>.tif}.
#' @param pattern Character. Regular expression used to match raster files
#'   within \code{input_dir}. Default is \code{".*\\.tif$"}.
#' @param resample_method Character. Resampling method passed to
#'   \code{\link[terra]{resample}}. Default is \code{"bilinear"}, which is
#'   appropriate for continuous variables. Use \code{"near"} for categorical
#'   rasters (e.g. land cover classes) to avoid interpolating between class
#'   codes. Other accepted values include \code{"cubic"} and \code{"lanczos"}.
#' @param overwrite Logical. If \code{TRUE}, overwrites existing output files.
#'   If \code{FALSE} (default), existing files are skipped.
#' @param verbose Logical. If \code{TRUE} (default), prints progress
#'   messages during processing. Includes file counts, overwrite mode,
#'   per-file progress, and a completion summary.
#'
#' @return Invisibly returns a list containing:
#' \itemize{
#'   \item \code{output_files}: Character vector of file paths written to
#'     \code{output_dir}.
#'   \item \code{n_processed}: Integer. Number of rasters successfully processed
#'     in this call (excludes skipped files when \code{overwrite = FALSE}).
#' }
#'
#' @details
#' For each raster in \code{input_dir} matching \code{pattern}, the function:
#' \enumerate{
#'   \item Reprojects to the CRS of the reference raster
#'   \item Resamples to match the reference resolution using \code{resample_method}
#'   \item Masks values outside the reference raster's non-NA extent
#'   \item Saves the result as a GeoTIFF to \code{output_dir}
#' }
#'
#' This preprocessing step ensures spatial alignment before applying other downstream
#' analyses such as \code{\link{temporally_explicit_extraction}} or \code{\link{scale_rasters}}.
#'
#' @seealso
#' Preprocessing: \code{\link{scale_rasters}},
#'   \code{\link{temporally_explicit_extraction}}
#'
#' @examples
#' raw_dir <- system.file("extdata/rasters_raw", package = "TemporalModelR")
#'
#' ref     <- file.path(raw_dir, "elevation.tif")
#'
#' out_dir <- file.path(tempdir(), "aligned")
#'
#' raster_align(
#'   input_dir        = raw_dir,
#'   output_dir       = out_dir,
#'   reference_raster = ref,
#'   overwrite        = TRUE,
#'   verbose          = FALSE
#' )

#' @export
#' @importFrom terra rast project resample mask writeRaster crs classify
raster_align <- function(input_dir,
                         output_dir,
                         reference_raster,
                         output_suffix   = "_Masked_Updated",
                         pattern         = ".*\\.tif$",
                         resample_method = "bilinear",
                         overwrite       = FALSE,
                         verbose         = TRUE) {

  if (missing(input_dir)) {
    stop(paste0("ERROR: 'input_dir' is required but was not provided. ",
                "Please specify the directory containing input raster files."))
  }
  if (missing(output_dir)) {
    stop(paste0("ERROR: 'output_dir' is required but was not provided. ",
                "Please specify the directory where processed rasters should be saved."))
  }
  if (missing(reference_raster)) {
    stop(paste0("ERROR: 'reference_raster' is required but was not provided. Please specify ",
                "the path to the reference raster file or a raster object used for alignment."))
  }
  if (!dir.exists(input_dir)) {
    stop(paste0("ERROR: Input directory does not exist: '", input_dir,
                "'. Please check the path and try again."))
  }
  if (!resample_method %in% c("bilinear", "near", "cubic", "lanczos",
                              "cubicspline", "average", "mode")) {
    stop(paste0("ERROR: 'resample_method' must be one of 'bilinear', 'near', 'cubic', 'lanczos', ",
                "'cubicspline', 'average', or 'mode'. Please adjust resample_method and try again."))
  }

  reference_raster <- .load_raster_input(reference_raster, "reference_raster")
  if (is.na(terra::crs(reference_raster)) || nchar(terra::crs(reference_raster)) == 0) {
    stop("ERROR: Reference raster has no defined CRS. Please assign a coordinate reference system before processing.")
  }

  reference_raster <- terra::classify(reference_raster, cbind(NA, 0))

  dir.create(output_dir, showWarnings = FALSE, recursive = TRUE)

  all_tif_files <- list.files(input_dir, pattern = pattern, full.names = TRUE)
  total_files   <- length(all_tif_files)

  if (total_files == 0) {
    stop(paste0("ERROR: No raster files found in '", input_dir,
                "' matching pattern '", pattern, "'. Please check that the directory contains .tif files or adjust the pattern parameter."))
  }

  if (verbose) message(paste("Found", total_files, "raster files in directory"))

  if (overwrite) {
    if (verbose) message("Overwrite mode: ON - will process and overwrite all files")
    tif_files        <- all_tif_files
    files_to_process <- total_files
  } else {
    if (verbose) message("Overwrite mode: OFF - will skip files that already exist")

    output_exists <- vapply(all_tif_files, function(f) {
      out_name <- sub("\\.tif$", paste0(output_suffix, ".tif"), basename(f))
      file.exists(file.path(output_dir, out_name))
    }, logical(1))

    already_processed <- sum(output_exists)
    percent_processed <- round((already_processed / total_files) * 100, 1)
    if (verbose) message(paste0(already_processed, " of ", total_files,
                                " files already exist in output directory (",
                                percent_processed, "%)"))

    tif_files        <- all_tif_files[!output_exists]
    files_to_process <- length(tif_files)

    if (files_to_process == 0) {
      if (verbose) message("All rasters already processed. No new files to process.")
      return(invisible(list(output_files = character(0), n_processed = 0)))
    }
  }

  if (verbose) message(paste("Processing", files_to_process, "files"))

  output_paths <- character(files_to_process)

  for (i in seq_along(tif_files)) {
    original_name <- basename(tif_files[i])
    output_name   <- sub("\\.tif$", paste0(output_suffix, ".tif"), original_name)
    output_path   <- file.path(output_dir, output_name)

    if (verbose) message(paste0("  [", i, "/", files_to_process, "] ", original_name))

    r <- terra::rast(tif_files[i])

    if (is.na(terra::crs(r))) {
      stop(paste0("ERROR: Raster '", original_name, "' has no defined CRS. Please assign a coordinate reference system to this raster before processing."))
    }

    r <- terra::project(r, terra::crs(reference_raster))
    r <- terra::resample(r, reference_raster, method = resample_method)
    r <- terra::mask(r, reference_raster, maskvalue = 0)

    terra::writeRaster(r, output_path, overwrite = TRUE)
    output_paths[i] <- output_path
  }

  if (verbose) message("Processing complete.")
  invisible(list(
    output_files = output_paths,
    n_processed  = files_to_process
  ))
}

#' Scale Environmental Rasters Using Species Occurrence Data
#'
#' Preprocessing function that standardizes raster files using precomputed mean
#' and standard deviation values from species occurrence data. Supports both
#' temporally static and dynamic rasters.
#'
#' @usage
#' scale_rasters(input_dir, output_dir, scaling_params_file,
#'               variable_patterns, time_cols = NULL, output_suffix = "_Scaled",
#'               overwrite = FALSE, verbose = TRUE)
#'
#' @param input_dir Character. Directory containing aligned input .tif raster files,
#' typically the output of \code{\link{raster_align}}.
#' @param output_dir Character. Directory where scaled rasters will be saved.
#' @param scaling_params_file Character. Path to CSV file with columns:
#'   variable, mean, sd. Typically generated by
#'   \code{\link{temporally_explicit_extraction}}.
#' @param variable_patterns Named character vector mapping clean variable names
#'   to raster filename patterns. For time-varying variables include the time
#'   placeholder in the pattern (e.g. \code{"forest_cover" = "forest_cover_YEAR"});
#'   for static variables omit it (e.g. \code{"elevation" = "elevation"}). Time
#'   placeholders must match entries in \code{time_cols}.
#' @param time_cols Character vector of time placeholders for dynamic variables.
#'   Default is \code{NULL}.
#' @param output_suffix Character. Suffix to append to output raster filenames.
#'   Default is \code{"_Scaled"}.
#' @param overwrite Logical. If \code{TRUE}, overwrites existing output files.
#'   If \code{FALSE} (default), existing files are skipped.
#' @param verbose Logical. If \code{TRUE} (default), prints progress
#'   messages during processing. Includes per-variable scaling progress.
#'
#' @return Invisibly returns a list containing:
#' \itemize{
#'   \item \code{n_scaled}: Integer. Total number of variables successfully
#'     scaled (static plus dynamic).
#'   \item \code{output_dir}: Character. Path to the directory containing scaled
#'     rasters.
#' }
#'
#' @details
#' Applies scaling to rasters via z-score transformation: (value - mean) / sd
#' as is calculated during \code{\link{temporally_explicit_extraction}}. Scaled
#' rasters are written to output_dir.
#'
#' Scaled rasters should use the same scaling parameters derived from species
#' occurrence data to ensure consistent standardization between training data
#' and prediction layers.
#'
#' @seealso
#' Preprocessing: \code{\link{temporally_explicit_extraction}},
#'   \code{\link{raster_align}}
#'
#' @examples
#' aln_dir     <- system.file("extdata/rasters_aligned",
#'                            package = "TemporalModelR")
#'
#' params_file <- system.file(
#'   "extdata/points/extracted_seasonal_Scaling_Parameters.csv",
#'   package = "TemporalModelR"
#' )
#'
#' out_dir <- file.path(tempdir(), "scaled")
#'
#' scale_rasters(
#'   input_dir           = aln_dir,
#'   output_dir          = out_dir,
#'   scaling_params_file = params_file,
#'   variable_patterns   = c(
#'     "elevation"    = "elevation",
#'     "forest_cover" = "forest_cover_YEAR",
#'     "prseas"       = "prseas_YEAR_SEASON"
#'   ),
#'   time_cols           = c("year", "season"),
#'   overwrite           = TRUE,
#'   verbose             = FALSE
#' )

#' @export
#' @importFrom terra rast writeRaster
#' @importFrom tools file_path_sans_ext file_ext
#' @importFrom utils read.csv
#' @importFrom stats setNames
scale_rasters <- function(input_dir,
                          output_dir,
                          scaling_params_file,
                          variable_patterns,
                          time_cols     = NULL,
                          output_suffix = "_Scaled",
                          overwrite     = FALSE,
                          verbose       = TRUE) {

  .validate_variable_patterns(variable_patterns)

  if (!dir.exists(input_dir)) {
    stop(paste0("ERROR: Input directory does not exist: ", input_dir))
  }

  if (!file.exists(scaling_params_file)) {
    stop(paste0("ERROR: Scaling parameters file does not exist: ", scaling_params_file))
  }

  dir.create(output_dir, showWarnings = FALSE, recursive = TRUE)

  scaling_params <- utils::read.csv(scaling_params_file, stringsAsFactors = FALSE)
  if (verbose) message(paste("Loaded scaling parameters for", nrow(scaling_params), "variables"))

  missing_params <- setdiff(names(variable_patterns), scaling_params$variable)
  if (length(missing_params) > 0) {
    warning(paste0("The following variables lack scaling parameters: ",
                   paste(missing_params, collapse = ", ")))
  }

  all_files <- list.files(path = input_dir, pattern = "\\.tif$",
                          recursive = TRUE, full.names = TRUE)
  if (verbose) message(paste("Found", length(all_files), "raster files"))

  if (length(all_files) == 0) {
    stop(paste0("ERROR: No .tif files found in input directory: ", input_dir))
  }

  if (overwrite) {
    if (verbose) message("Overwrite mode: ON - Will process and overwrite all files")
  } else {
    if (verbose) message("Overwrite mode: OFF - Will skip files that already exist")
  }

  var_classes  <- .classify_variable_patterns(variable_patterns, time_cols)
  dynamic_vars <- var_classes$dynamic
  static_vars  <- var_classes$static

  if (verbose) message(paste("Dynamic variables:",
                             if (length(dynamic_vars) > 0) paste(dynamic_vars, collapse = ", ") else "none"))
  if (verbose) message(paste("Static variables:",
                             if (length(static_vars) > 0) paste(static_vars, collapse = ", ") else "none"))

  static_vars_scaled  <- 0
  static_vars_skipped <- 0
  if (length(static_vars) > 0) {
    if (verbose) message("Processing static variables...")
    dummy_time_values <- as.data.frame(stats::setNames(
      lapply(time_cols, function(tc) NA_character_), time_cols
    ), stringsAsFactors = FALSE)
    for (var_name in static_vars) {
      if (verbose) message(paste("Scaling static variable:", var_name))
      var_params <- scaling_params[scaling_params$variable == var_name, ]
      if (nrow(var_params) == 0) {
        if (verbose) message(paste0("  Skipping ", var_name, " - no scaling parameters found"))
        next
      }
      var_mean <- var_params$mean
      var_sd   <- var_params$sd
      resolved <- .resolve_raster_paths(variable_patterns, character(0), var_name,
                                        time_cols, dummy_time_values, input_dir)
      if (is.null(resolved)) {
        warning(paste0("No files found for static variable ", var_name))
        next
      }
      file_path   <- resolved[1]
      output_file <- file.path(output_dir, paste0(var_name, output_suffix, ".tif"))
      if (file.exists(output_file) && !overwrite) {
        if (verbose) message("  Skipped (already exists)")
        static_vars_skipped <- static_vars_skipped + 1
      } else {
        raster_layer  <- terra::rast(file_path)
        raster_scaled <- (raster_layer - var_mean) / var_sd
        names(raster_scaled) <- tools::file_path_sans_ext(basename(output_file))
        terra::writeRaster(raster_scaled, output_file, overwrite = overwrite)
        static_vars_scaled <- static_vars_scaled + 1
      }
    }
  }

  dynamic_vars_scaled <- 0
  if (length(dynamic_vars) > 0) {
    if (verbose) message("Processing dynamic variables...")

    all_time_combos <- list()

    for (var_name in dynamic_vars) {
      search_pattern <- variable_patterns[var_name]
      relevant_time_cols <- time_cols[vapply(time_cols, function(tc) {
        grepl(tc, search_pattern, ignore.case = TRUE)
      }, logical(1))]
      if (length(relevant_time_cols) == 0) next

      pattern_parts  <- strsplit(search_pattern, "_")[[1]]
      tc_positions   <- vapply(relevant_time_cols, function(tc) {
        pos <- which(toupper(pattern_parts) == toupper(tc))
        if (length(pos) > 0) pos[1] else NA_integer_
      }, integer(1))

      candidate_pattern <- variable_patterns[var_name]
      for (tc in relevant_time_cols) {
        candidate_pattern <- gsub(tc, "[^_]+", candidate_pattern,
                                  ignore.case = TRUE)
      }
      candidate_pattern <- paste0("^", candidate_pattern, "(?=[^0-9]|$)")
      candidate_files <- all_files[vapply(all_files, function(f) {
        grepl(candidate_pattern, basename(f),
              ignore.case = TRUE, perl = TRUE)
      }, logical(1))]

      for (file_path in candidate_files) {
        filename_parts <- strsplit(tools::file_path_sans_ext(basename(file_path)), "_")[[1]]
        time_vals      <- stats::setNames(
          lapply(seq_along(relevant_time_cols), function(i) {
            pos <- tc_positions[i]
            if (!is.na(pos) && pos <= length(filename_parts)) filename_parts[pos] else NA_character_
          }),
          relevant_time_cols
        )
        if (any(is.na(unlist(time_vals)))) next
        ### Build full time_values row with all time_cols (NAs for cols not in this pattern)
        full_row <- stats::setNames(
          lapply(time_cols, function(tc) {
            if (tc %in% relevant_time_cols) time_vals[[tc]] else NA_character_
          }),
          time_cols
        )
        combo_key <- paste(vapply(relevant_time_cols, function(tc) time_vals[[tc]], character(1)), collapse = "_")
        all_time_combos[[combo_key]] <- as.data.frame(full_row, stringsAsFactors = FALSE)
      }
    }

    if (length(all_time_combos) == 0) {
      warning("No valid time information found in filenames for dynamic variables.")
    } else {
      if (verbose) message(paste("  Found", length(all_time_combos), "unique time periods"))

      for (var_name in dynamic_vars) {
        if (verbose) message(paste("Scaling dynamic variable:", var_name))
        var_params <- scaling_params[scaling_params$variable == var_name, ]
        if (nrow(var_params) == 0) {
          if (verbose) message(paste0("  Skipping ", var_name, " - no scaling parameters found"))
          next
        }
        var_mean <- var_params$mean
        var_sd   <- var_params$sd

        search_pattern <- variable_patterns[var_name]
        relevant_time_cols <- time_cols[vapply(time_cols, function(tc) {
          grepl(tc, search_pattern, ignore.case = TRUE)
        }, logical(1))]

        processed_count <- 0
        skipped_count   <- 0

        for (combo_key in names(all_time_combos)) {
          time_values <- all_time_combos[[combo_key]]

          if (any(is.na(unlist(time_values[relevant_time_cols])))) next

          resolved <- .resolve_raster_paths(variable_patterns, var_name, character(0),
                                            time_cols, time_values, input_dir)
          if (is.null(resolved)) next

          file_path   <- resolved[1]
          time_string <- paste(vapply(relevant_time_cols, function(tc) as.character(time_values[[tc]]), character(1)), collapse = "_")
          output_file <- file.path(output_dir, paste0(var_name, "_", time_string, output_suffix, ".tif"))

          if (file.exists(output_file) && !overwrite) {
            skipped_count <- skipped_count + 1
            next
          }
          raster_layer  <- terra::rast(file_path)
          raster_scaled <- (raster_layer - var_mean) / var_sd
          names(raster_scaled) <- tools::file_path_sans_ext(basename(output_file))
          terra::writeRaster(raster_scaled, output_file, overwrite = overwrite)
          processed_count <- processed_count + 1
        }

        if (processed_count > 0) dynamic_vars_scaled <- dynamic_vars_scaled + 1
        if (verbose) message(paste("  Processed:", processed_count, "| Skipped:", skipped_count))
      }
    }
  }

  total_vars_scaled  <- static_vars_scaled + dynamic_vars_scaled
  total_vars_skipped <- static_vars_skipped

  if (total_vars_scaled == 0 && total_vars_skipped == 0) {
    stop(paste0(
      "ERROR: No raster files could be matched and scaled for any variables. ",
      "Check that 'variable_patterns' values match your raster filenames and that ",
      "'time_cols' placeholders correspond to the time values present in those filenames. ",
      "See ?scale_rasters for details."
    ))
  }

  if (verbose) message("All scaling complete.")
  if (verbose) message(paste("  Static:", static_vars_scaled, "new |",
                             static_vars_skipped, "skipped | Dynamic:", dynamic_vars_scaled, "variables"))

  invisible(list(
    n_scaled   = total_vars_scaled,
    output_dir = output_dir
  ))
}

### This file documents the pre-built workflow objects stored in data/ and
### the bundled raster / point files stored in inst/extdata/.
### Documentation is generated from roxygen2 tags below.


### Seasonal workflow .rda objects --------------------------------------------

#' Pre-built spatiotemporal partition (seasonal workflow)
#'
#' Output from \code{\link{spatiotemporal_partition}} run on the synthetic
#' occurrence dataset bundled in \code{inst/extdata/}. Built with
#' \code{n_spatial_folds = 2} and \code{n_temporal_folds = 2}, producing
#' four cross-validation folds (2 spatial + 2 temporal). Points span 15
#' years and the seasons Spring, Summer, and Autumn on a 15 x 30 cell
#' study area (3000 m x 1500 m, 100 m pixels) in a custom synthetic local
#' CRS. Predictors attached to each point are \code{forest_cover},
#' \code{prseas} (seasonal precipitation), and \code{elevation}, all
#' z-scored. Loaded with \code{data(tmr_partition)}.
#'
#' @format A list as returned by \code{\link{spatiotemporal_partition}},
#'   containing \code{$folds}, \code{$points_sf}, \code{$voronoi_blocks},
#'   \code{$voronoi_folds}, \code{$summary}, and \code{$plots}.
"tmr_partition"

#' Pre-built spatiotemporal partition, small version (seasonal workflow)
#'
#' A subsampled version of \code{\link{tmr_partition}} retaining
#' approximately half the points per fold, for use in examples and tests
#' where runtime is a concern. Built by sampling
#' \code{ceiling(n / 2)} rows from each fold of \code{tmr_partition$points_sf}.
#' Loaded with \code{data(tmr_partition_small)}.
#'
#' @format A list with the same structure as \code{\link{tmr_partition}},
#'   containing \code{$folds}, \code{$points_sf}, \code{$voronoi_blocks},
#'   \code{$voronoi_folds}, \code{$summary}, and \code{$plots}.
#' @seealso \code{\link{tmr_partition}}
"tmr_partition_small"

#' Pre-built pseudoabsence result (seasonal workflow)
#'
#' Output from \code{\link{generate_absences}} using the buffer method with
#' a 300 m buffer (3 pixels at the synthetic landscape's 100 m resolution)
#' and a 2:1 pseudoabsence-to-presence ratio. Pseudoabsences are stratified
#' by fold from \code{tmr_partition} and have \code{forest_cover},
#' \code{prseas}, and \code{elevation} extracted at each location's
#' year-season combination. Loaded with \code{data(tmr_absences)}.
#'
#' @format A list as returned by \code{\link{generate_absences}}, containing
#'   \code{$pseudoabsences} (an sf object with attached predictor columns),
#'   \code{$plots}, and \code{$summary}.
"tmr_absences"

#' Pre-built GLM result (seasonal workflow)
#'
#' Output from \code{\link{build_temporal_glm}} fit to \code{tmr_partition}
#' and \code{tmr_absences} with the formula
#' \code{~ forest_cover + prseas + elevation}, a logit link, and TSS-based
#' threshold selection. One model per fold. Loaded with
#' \code{data(tmr_glm)}.
#'
#' @format A list of class \code{"TemporalGLM"} as returned by
#'   \code{\link{build_temporal_glm}}, containing \code{$models},
#'   \code{$thresholds}, \code{$threshold_method}, \code{$model_formula},
#'   \code{$link}, \code{$model_vars}, \code{$fold_training_data},
#'   \code{$fold_test_metrics}, \code{$output_dir}, \code{$model_type},
#'   and \code{$plots}.
"tmr_glm"

#' Pre-built spatiotemporal predictions (seasonal workflow)
#'
#' Output from \code{\link{generate_spatiotemporal_predictions}} projecting
#' \code{tmr_glm} across 15 years for the Spring season only (one
#' prediction layer per year-season combination, 15 total). Loaded with
#' \code{data(tmr_predictions)}.
#'
#' Note that \code{$prediction_files} stores the absolute paths used at
#' build time. To work with the per-timestep rasters on a user machine,
#' regenerate them via \code{\link{generate_spatiotemporal_predictions}}
#' or use the bundled prediction set in \code{inst/extdata/predictions/}.
#'
#' @format A list as returned by
#'   \code{\link{generate_spatiotemporal_predictions}}, containing
#'   \code{$timestep_metrics}, \code{$overall_summary},
#'   \code{$prediction_files}, and \code{$model_type}.
"tmr_predictions"


### Annual workflow .rda objects --------------------------------------------

#' Pre-built spatiotemporal partition (annual workflow)
#'
#' Annual variant of \code{tmr_partition}. Built from points rarefied with
#' \code{time_cols = "year"} only (one point per pixel per year) and
#' extracted against annual predictors. Predictors attached to each point
#' are \code{forest_cover}, \code{pr_ann} (annual precipitation), and
#' \code{elevation}, all z-scored. Uses
#' \code{n_spatial_folds = 2} and \code{n_temporal_folds = 2} for four
#' folds. Loaded with \code{data(tmr_partition_annual)}.
#'
#' @format A list as returned by \code{\link{spatiotemporal_partition}}.
"tmr_partition_annual"

#' Pre-built pseudoabsence result (annual workflow)
#'
#' Annual variant of \code{tmr_absences}. Buffer method with 300 m buffer
#' and 2:1 ratio, generated against \code{tmr_partition_annual} with the
#' annual predictor set (\code{forest_cover}, \code{pr_ann},
#' \code{elevation}). Loaded with \code{data(tmr_absences_annual)}.
#'
#' @format A list as returned by \code{\link{generate_absences}}.
"tmr_absences_annual"

#' Pre-built GLM result (annual workflow)
#'
#' Annual variant of \code{tmr_glm}. Fit with the formula
#' \code{~ forest_cover + pr_ann + elevation}, a logit link, and
#' TSS-based threshold selection. One model per fold. Loaded with
#' \code{data(tmr_glm_annual)}.
#'
#' @format A list of class \code{"TemporalGLM"}.
"tmr_glm_annual"

#' Pre-built spatiotemporal predictions (annual workflow)
#'
#' Annual variant of \code{tmr_predictions}. Projects \code{tmr_glm_annual}
#' across 15 years on a single (year-only) time axis. Loaded with
#' \code{data(tmr_predictions_annual)}.
#'
#' @format A list as returned by
#'   \code{\link{generate_spatiotemporal_predictions}}.
"tmr_predictions_annual"


#' Pre-built user_data pseudoabsences (seasonal workflow)
#'
#' Output of \code{\link{generate_absences}} run with
#' \code{method = "user_data"} against \code{\link{tmr_partition}}, using
#' \code{inst/extdata/points/synthetic_user_presences.csv} as the supplied
#' absence locations. The CSV contains occurrence-formatted points for a
#' 'related speecies' to that data of interest to be used as a proxy for survey
#' effort for the species of interest, serving as an alternative method for
#' generating pseudoabsences.
#'
#' @format A list as returned by \code{\link{generate_absences}}, containing
#'   \code{$pseudoabsences} (an sf object with attached predictor columns),
#'   \code{$plots}, and \code{$summary}.
#'
"tmr_absences_user"


### Bundled extdata --------------------------------------------------------

#' Bundled rasters, point files, and prediction outputs
#'
#' Several non-R objects ship in \code{inst/extdata/} for use in function
#' examples and the package vignette. They cannot be portably serialized
#' as \code{.rda} files, so they are stored as GeoTIFF and CSV files and
#' loaded via \code{\link[base]{system.file}}.
#'
#' \describe{
#'   \item{\code{extdata/rasters_raw/}}{Raw synthetic environmental rasters
#'     before alignment. Includes \code{elevation.tif},
#'     \code{forest_cover_1.tif} through \code{forest_cover_15.tif},
#'     \code{pr_ann_1.tif} through \code{pr_ann_15.tif}, and
#'     \code{prseas_<year>_<season>.tif} for each of 15 years and 4
#'     seasons (60 files).}
#'   \item{\code{extdata/rasters_aligned/}}{Same rasters after
#'     \code{\link{raster_align}} has reprojected and masked them to the
#'     reference grid.}
#'   \item{\code{extdata/rasters_scaled/}}{Aligned rasters z-scored using
#'     the scaling parameters from the seasonal extraction. Used by
#'     examples and modeling functions that work with the seasonal
#'     predictor set.}
#'   \item{\code{extdata/rasters_scaled_annual/}}{Aligned rasters z-scored
#'     using the scaling parameters from the annual extraction. Used by
#'     examples and modeling functions that work with the annual
#'     predictor set (\code{pr_ann} instead of \code{prseas}).}
#'   \item{\code{extdata/points/synthetic_occurrence_points.csv}}{The raw
#'     synthetic presence dataset: 150 points with \code{x}, \code{y},
#'     \code{year}, \code{season}, and \code{pres = 1}.}
#'   \item{\code{extdata/points/synthetic_occurrence_points.shp}}{Same
#'     points as a shapefile.}
#'   \item{\code{extdata/points/synthetic_user_presences.csv}}{A second-species
#'     presence dataset for demonstrating \code{method = "user_data"} in
#'     \code{\link{generate_absences}}. Points were derived from buffer-
#'     constrained environmental pseudoabsences of the primary species and
#'     reformatted as presences (\code{pres = 1}). Same column structure as
#'     \code{synthetic_occurrence_points.csv} (\code{x}, \code{y},
#'     \code{year}, \code{season}, \code{pres}).}
#'   \item{\code{extdata/points/extracted_seasonal_*.csv}}{Outputs from
#'     \code{\link{temporally_explicit_extraction}} using the seasonal
#'     predictor set: \code{_Raw_Values.csv}, \code{_Scaled_Values.csv},
#'     and \code{_Scaling_Parameters.csv}.}
#'   \item{\code{extdata/points/extracted_annual_*.csv}}{Same outputs but
#'     using the annual predictor set.}
#'   \item{\code{extdata/predictions/}}{Per-timestep fold-vote rasters
#'     from the seasonal workflow's
#'     \code{\link{generate_spatiotemporal_predictions}} call. Fifteen
#'     files (\code{Prediction_<year>_Spring.tif}) suitable for direct
#'     input to \code{\link{summarize_raster_outputs}}.}
#'   \item{\code{extdata/binary/consensus_stack.tif}}{Multi-layer
#'     GeoTIFF of binary suitable / unsuitable rasters produced by
#'     \code{\link{summarize_raster_outputs}} with \code{consensus = 3}.
#'     Fifteen layers, one per year, ordered 1 through 15.}
#'   \item{\code{extdata/binary/frequency_raster.tif}}{Companion
#'     single-layer raster giving the proportion of years each pixel was
#'     classified as suitable.}
#'   \item{\code{extdata/precomputed/}}{Precomputed prediction outputs read
#'     directly by the modeling vignettes (V3a-V3d) so that
#'     \code{\link{generate_spatiotemporal_predictions}} does not need to be
#'     rerun at vignette build time. One subdirectory per model type
#'     (\code{glm/}, \code{gam/}, \code{rf/}, \code{hv/}), each containing
#'     \code{preds.rds} and a \code{pred_tifs/} folder. \code{preds.rds} is
#'     the list returned by
#'     \code{\link{generate_spatiotemporal_predictions}} for that model
#'     (\code{$timestep_metrics}, \code{$overall_summary},
#'     \code{$prediction_files}, and \code{$model_type}), with
#'     \code{$prediction_files} reduced to bare file names rather than
#'     absolute build-time paths. \code{pred_tifs/} holds the 60
#'     per-timestep fold-vote rasters
#'     (\code{Prediction_<year>_<season>.tif}, 15 years x 4 seasons)
#'     projected from that model. The vignettes load \code{preds.rds} and
#'     rebuild \code{$prediction_files} from \code{pred_tifs/} via
#'     \code{\link[base]{system.file}}.}
#' }
#'
#' Example load patterns:
#' \preformatted{
#'   ### Aligned raster directory
#'   aln_dir <- system.file("extdata/rasters_aligned",
#'                          package = "TemporalModelR")
#'
#'   ### Consensus stack (multi-layer)
#'   binary_stack <- terra::rast(system.file(
#'     "extdata/binary/consensus_stack.tif", package = "TemporalModelR"
#'   ))
#'
#'   ### Frequency raster
#'   frequency_rast <- terra::rast(system.file(
#'     "extdata/binary/frequency_raster.tif", package = "TemporalModelR"
#'   ))
#'
#'   ### Per-timestep prediction directory
#'   pred_dir <- system.file("extdata/predictions",
#'                           package = "TemporalModelR")
#'
#'   ### Precomputed GLM prediction object and its rasters
#'   glm_preds <- readRDS(system.file(
#'     "extdata/precomputed/glm/preds.rds", package = "TemporalModelR"
#'   ))
#'   glm_pred_files <- list.files(
#'     system.file("extdata/precomputed/glm/pred_tifs",
#'                 package = "TemporalModelR"),
#'     pattern = "\\.tif$", full.names = TRUE
#'   )
#' }
#'
#' @name extdata
#' @docType data
NULL

#' Spatiotemporal Cross-Validation Partitioning
#'
#' Preprocesses species occurrence data by partitioning it into spatially and
#' temporally structured folds for cross-validation. Supports creation of
#' spatial-only folds, temporal-only folds, and random folds.
#'
#' Works better with smaller numbers of folds and may have difficulties
#' creating even folds for large numbers of groups or where sample sizes
#' are very small.
#'
#' @usage
#' spatiotemporal_partition(reference_shapefile_path, points_file_path,
#'                          time_cols = NULL, xcol = NULL, ycol = NULL,
#'                          points_crs = NULL, n_spatial_folds = 0,
#'                          n_temporal_folds = 0, n_balanced_folds = 0,
#'                          n_random_folds = 0, single_fold= FALSE,
#'                          max_imbalance = 0.05, max_attempts = 10,
#'                          create_plot = TRUE, plot_palette = "Dark 2",
#'                          output_file= NULL, verbose    = TRUE)
#'
#' @param reference_shapefile_path Character or sf object. Path to a polygon
#'   file or an \code{sf} polygon object defining the study area.
#' @param points_file_path Character, sf object, sfc object, Spatial object,
#'   or data frame. Path to occurrence data (\code{.csv}, \code{.shp},
#'   \code{.geojson}, \code{.gpkg}) or a spatial object.
#' @param time_cols Character. Name of a single column containing temporal
#'   values (e.g. year). Used to define temporal blocks. Required when using
#'   temporal folds. Must be a single column name; does not support more than one
#'   time column unlike other functions in this package. Compound time representations
#'   (e.g. year + season) should be encoded into a single ordered numeric column
#'   before partitioning, or only one (e.g. year) should be used.
#' @param xcol Character. Name of the x-coordinate column. Required when
#'   \code{points_file_path} is a CSV file or data frame.
#' @param ycol Character. Name of the y-coordinate column. Required when
#'   \code{points_file_path} is a CSV file or data frame.
#' @param points_crs Character or CRS object. CRS of the input points.
#'   Required when \code{points_file_path} is a CSV file or data frame.
#' @param n_spatial_folds Integer. Number of spatially explicit folds. Ignored
#'   when using random folds. Default is \code{0}.
#' @param n_temporal_folds Integer. Number of temporally explicit folds.
#'   When used alone (with \code{n_spatial_folds = 0}), creates temporal-only
#'   folds where each fold spans the full study area but covers a distinct
#'   slice of the time series. When combined with \code{n_spatial_folds},
#'   creates a spatiotemporal design. Ignored when using random folds.
#'   Default is \code{0}.
#' @param n_balanced_folds Integer. Reserved for future use. Default is
#'   \code{0} (disabled).
#' @param n_random_folds Integer. Number of random folds with no spatial or
#'   temporal structure. Overrides all other fold parameters. Default is
#'   \code{0}.
#' @param single_fold Logical. If \code{TRUE}, bypasses all partitioning and
#'   assigns all points to a single fold (fold 1). In this mode all points are
#'   used for both training and testing, producing a single model trained on
#'   the full dataset. All downstream functions accept the result identically
#'   to a standard multi-fold partition. Overrides all fold count parameters.
#'   Default is \code{FALSE}.
#' @param max_imbalance Numeric. Maximum allowed fold size imbalance as a
#'   proportion between 0 and 1. Default is \code{0.05}.
#' @param max_attempts Integer. Maximum number of partitioning attempts for
#'   spatiotemporal and balanced modes. Each attempt re-runs the spatial block
#'   construction; the attempt with the lowest imbalance is returned. Ignored
#'   for random and spatial-only modes. Default is \code{10}.
#' @param create_plot Logical. If \code{TRUE} (default), generates
#'   diagnostic plots showing fold distributions.
#' @param plot_palette Character. Name of an HCL or RColorBrewer palette used
#'   to color folds in diagnostic plots. Accepts any HCL palette name (see
#'   \code{\link[grDevices]{hcl.pals}}) or, if \pkg{RColorBrewer} is installed,
#'   any Brewer palette name. Default is \code{"Dark 2"}.
#' @param output_file Character. Optional path to save the result as an
#'   \code{.rds} file. The parent directory will be created if it does not
#'   exist. Default is \code{NULL}.
#' @param verbose Logical. If \code{TRUE} (default), prints progress
#'   messages during processing. Includes the partition mode, fold
#'   structure, per-fold point counts, and file-save confirmation.
#'
#' @return Invisibly returns a list containing:
#' \itemize{
#'   \item \code{folds}: Data frame of fold assignments with a \code{fold}
#'     column identifying each point's cross-validation fold.
#'   \item \code{points_sf}: sf object of occurrence points with assigned
#'     folds.
#'   \item \code{voronoi_folds}: sf object of Voronoi polygons representing
#'     the spatial fold boundaries. \code{NULL} for random folds, temporal-only
#'     folds, and single-fold mode.
#'   \item \code{summary}: Data frame of partitioning summary statistics.
#'   \item \code{plots}: Named list of recorded plot objects when
#'     \code{create_plot = TRUE}. Empty list in single-fold mode.
#' }
#'
#' @details
#' The function partitions data into folds using one of five modes:
#' \itemize{
#'   \item \strong{Single fold}: All points are assigned to fold 1 and used
#'     for both training and testing. This produces a single model trained on
#'     the full dataset with no held-out validation. Useful when sample sizes
#'     are too small for cross-validation, or as a final production model step
#'     after cross-validation has already established model quality. Set
#'     \code{single_fold = TRUE}. All downstream functions accept the result
#'     identically to standard multi-fold output.
#'
#'   \item \strong{Random}: Points are assigned to folds by random shuffling
#'     with no spatial or temporal structure. Each fold is a simple random
#'     sample of the full dataset, intended as a naive baseline that makes no
#'     attempt to reduce spatial or temporal autocorrelation between training
#'     and test sets. Use \code{n_random_folds}.
#'
#'   \item \strong{Spatial-only}: The study area is divided into \eqn{k}
#'     contiguous spatial regions using a recursive k-d tree bisection
#'     algorithm. At each step the point set is split along its longest
#'     spatial axis, recursively halving until the target number of folds
#'     is reached. A centroid reassignment pass then refines boundaries
#'     to improve balance. Each region becomes one fold, so training
#'     always occurs on data from geographically distinct areas relative
#'     to the test fold. No temporal separation is imposed, meaning that
#'     points from any time period may appear in any fold. Use
#'     \code{n_spatial_folds} alone.
#'
#'   \item \strong{Temporal-only}: Each fold covers the full spatial extent
#'     of the study area but is restricted to a distinct, non-overlapping
#'     slice of the time series. The global time series is divided into
#'     \code{n_temporal_folds} equal intervals using quantile-based breaks,
#'     and all points within each interval form one fold. This design tests
#'     model transferability across time while retaining full spatial coverage
#'     in every fold. Use \code{n_temporal_folds} alone (with
#'     \code{n_spatial_folds = 0}). Requires \code{time_cols}.
#'
#'   \item \strong{Spatiotemporal}: Folds are assigned using the same
#'     recursive k-d tree bisection as spatial-only mode, operating on
#'     the full point set to produce spatially contiguous groups. The
#'     resulting groups are then split into a spatial pool
#'     (\code{n_spatial_folds} folds drawn from geographically distinct
#'     regions) and a temporal pool (\code{n_temporal_folds} folds each
#'     restricted to a distinct slice of the time series but spanning
#'     the full study area). Together the two pools assess both geographic
#'     and temporal transferability in a single cross-validation design.
#'     Use \code{n_spatial_folds} and \code{n_temporal_folds} together.
#'     Requires \code{time_cols}.
#' }
#'
#' Fold assignment uses a recursive k-d tree bisection algorithm that splits
#' points along their longest spatial axis at each step, followed by a
#' centroid reassignment pass to improve boundary regularity and point-count
#' balance. Voronoi tessellation on fold centroids is used only for
#' visualisation of the resulting spatial boundaries. For temporal
#' mode, temporal blocks are defined by dividing the global time series
#' into equal intervals using quantile-based breaks. For spatiotemporal mode,
#' the typical spatial assignment is done, but with one larger spatial block
#' made with enough points to represent all of the temporal folds, then the
#' temporal blocking is applied to those points.
#'
#' Partitioned datasets are suitable for cross-validation in modeling
#' workflows, ensuring spatial and/or temporal independence between folds.
#'
#' @seealso
#' Preprocessing: \code{\link{spatiotemporal_rarefaction}},
#'   \code{\link{temporally_explicit_extraction}},
#'   \code{\link{generate_absences}}
#'
#' Modeling: \code{\link{build_temporal_hv}},
#'   \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
#'   \code{\link{build_temporal_rf}}
#'
#' @examples
#' pts_file <- system.file(
#'   "extdata/points/extracted_seasonal_Scaled_Values.csv",
#'   package = "TemporalModelR"
#' )
#'
#' ref_file <- system.file("extdata/rasters_raw/elevation.tif",
#'                         package = "TemporalModelR")
#'
#' study_crs <- sf::st_crs(terra::rast(ref_file))
#'
#' study_area_sf <- sf::st_as_sf(sf::st_as_sfc(
#'   sf::st_bbox(c(xmin = 0, xmax = 3000, ymin = 0, ymax = 1500),
#'               crs = study_crs)
#' ))
#'
#' spatiotemporal_partition(
#'   reference_shapefile_path = study_area_sf,
#'   points_file_path         = pts_file,
#'   xcol                     = "x",
#'   ycol                     = "y",
#'   points_crs               = study_crs,
#'   time_cols                = "year",
#'   n_spatial_folds          = 2,
#'   n_temporal_folds         = 2,
#'   create_plot              = FALSE,
#'   verbose                  = FALSE
#' )

#' @export
#' @importFrom sf st_read st_as_sf st_transform st_coordinates st_crs st_bbox
#' @importFrom sf st_drop_geometry st_intersection st_polygon st_sf st_sfc st_union
#' @importFrom deldir deldir tile.list
#' @importFrom stats complete.cases median quantile
#' @importFrom tools file_ext
#' @importFrom utils capture.output
spatiotemporal_partition <- function(reference_shapefile_path,
                                     points_file_path,
                                     time_cols         = NULL,
                                     xcol             = NULL,
                                     ycol             = NULL,
                                     points_crs       = NULL,
                                     n_spatial_folds  = 0,
                                     n_temporal_folds = 0,
                                     n_balanced_folds = 0,
                                     n_random_folds   = 0,
                                     single_fold      = FALSE,
                                     max_imbalance    = 0.05,
                                     max_attempts     = 10,
                                     create_plot   = TRUE,
                                     plot_palette     = "Dark 2",
                                     output_file      = NULL,
                                     verbose          = TRUE) {

  if (missing(reference_shapefile_path) || is.null(reference_shapefile_path)) {
    stop(paste0("ERROR: 'reference_shapefile_path' is required but was not provided. ",
                "Please provide a file path or sf object defining the study area."))
  }
  if (missing(points_file_path) || is.null(points_file_path)) {
    stop(paste0("ERROR: 'points_file_path' is required but was not provided. ",
                "Please provide a file path, sf object, or data frame of occurrence points."))
  }

  if (!is.null(output_file)) {
    if (tolower(tools::file_ext(output_file)) != "rds") {
      stop("ERROR: 'output_file' must have a '.rds' extension.")
    }
    out_dir <- dirname(output_file)
    if (!dir.exists(out_dir)) {
      dir.create(out_dir, recursive = TRUE, showWarnings = FALSE)
    }
  }

  if (isTRUE(single_fold)) {
    use_random            <- FALSE
    use_balanced          <- FALSE
    temporal_partitioning <- FALSE
    total_folds           <- 1
    n_spatial             <- 1
    n_temporal            <- 1
    partition_mode        <- "single_fold"
  } else {

    use_random   <- n_random_folds > 0
    use_balanced <- n_balanced_folds > 0

    modes_active <- sum(c(
      use_random,
      use_balanced,
      (n_spatial_folds > 0 || n_temporal_folds > 0)
    ))
    if (modes_active == 0) {
      stop(paste0("ERROR: Must specify one of n_random_folds, n_balanced_folds, ",
                  "n_spatial_folds/n_temporal_folds, or set single_fold = TRUE."))
    }
    if (modes_active > 1) {
      stop(paste0("ERROR: Only one partitioning mode is allowed at a time. ",
                  "Use n_random_folds, n_balanced_folds, or n_spatial_folds/n_temporal_folds."))
    }
    if (!use_random) {
      if (n_spatial_folds == 1) {
        stop("ERROR: n_spatial_folds must be 0 or >= 2.")
      }
      if (n_temporal_folds == 1) {
        stop("ERROR: n_temporal_folds must be 0 or >= 2.")
      }
    }

    if (use_random) {
      total_folds           <- n_random_folds
      n_spatial             <- 1
      n_temporal            <- 1
      temporal_partitioning <- FALSE
      partition_mode        <- "random"

    } else if (use_balanced) {
      if (is.null(time_cols)) {
        stop("ERROR: 'time_cols' is required for balanced folds. Please specify the column name containing time or year values.")
      }
      total_folds           <- n_balanced_folds
      n_spatial             <- n_balanced_folds * ifelse(n_balanced_folds %% 2 == 0, 4, 5)
      n_temporal            <- 2
      temporal_partitioning <- TRUE
      partition_mode        <- "balanced"
      warning("Balanced folds are more difficult to perfectly create. You may need a more flexible max_imbalance threshold.")

    } else {
      total_folds <- n_spatial_folds + n_temporal_folds
      if (n_spatial_folds == 0 && n_temporal_folds > 0) {
        if (is.null(time_cols)) {
          stop("ERROR: 'time_cols' is required for temporal folds. Please specify the column name containing time or year values.")
        }
        warning("Only temporal folds specified. Spatial structure will not be considered.")
        n_spatial             <- 1
        n_temporal            <- n_temporal_folds
        temporal_partitioning <- TRUE
        partition_mode        <- "temporal_only"
      } else if (n_temporal_folds == 0 && n_spatial_folds > 0) {
        warning("Only spatial folds specified. Temporal structure will not be considered.")
        n_spatial             <- NULL
        n_temporal            <- 1
        temporal_partitioning <- FALSE
        partition_mode        <- "spatial_only"
      } else {
        if (is.null(time_cols)) {
          stop("ERROR: 'time_cols' is required when temporal partitioning is enabled. Please specify the column name containing time or year values.")
        }
        n_spatial             <- n_spatial_folds * 2 * n_temporal_folds
        n_temporal            <- n_temporal_folds
        temporal_partitioning <- TRUE
        partition_mode        <- "spatiotemporal"
      }
    }

  }

  if (is.character(reference_shapefile_path) && verbose) {
    message(paste("Reading shapefile:", basename(reference_shapefile_path)))
  }
  reference_shapefile <- .load_shapefile_input(reference_shapefile_path, "reference_shapefile_path")

  pts        <- .load_points_data(points_file_path, xcol, ycol, points_crs,
                                  verbose = verbose)
  xcol       <- attr(pts, "xcol")
  ycol       <- attr(pts, "ycol")
  points_crs <- attr(pts, "crs")

  n_original <- nrow(pts)
  pts        <- pts[stats::complete.cases(pts), ]
  n_removed  <- n_original - nrow(pts)
  if (n_removed > 0) {
    warning(paste0("Removed ", n_removed, " incomplete rows (",
                   round(n_removed / n_original * 100, 2), "%)"))
  }
  if (nrow(pts) == 0) {
    stop("ERROR: No complete rows remaining after removing incomplete data. Please check your input data.")
  }
  if (!xcol %in% names(pts)) {
    stop(paste0("ERROR: Column '", xcol, "' not found in the input data."))
  }
  if (!ycol %in% names(pts)) {
    stop(paste0("ERROR: Column '", ycol, "' not found in the input data."))
  }
  if (temporal_partitioning && !time_cols %in% names(pts)) {
    stop(paste0("ERROR: Column '", time_cols, "' not found in the input data."))
  }

  pts_sf       <- sf::st_as_sf(pts, coords = c(xcol, ycol), crs = points_crs)
  pts_sf       <- sf::st_transform(pts_sf, crs = sf::st_crs(reference_shapefile))
  total_points <- nrow(pts_sf)

  if (isTRUE(single_fold)) {
    if (verbose) message("\nPartition mode: SINGLE FOLD (all points used for training and testing)")

    pts_sf$fold <- 1

    summary_stats <- data.frame(
      parameter = c("total_folds", "n_spatial_folds", "n_temporal_folds",
                    "n_balanced_folds", "n_random_folds", "partition_mode",
                    "total_points", "points_removed", "pct_rows_removed",
                    "final_imbalance_pct"),
      value = c(1, 0, 0, 0, 0, "single_fold",
                total_points, n_removed,
                ifelse(n_removed > 0, round(n_removed / n_original * 100, 2), 0),
                0),
      stringsAsFactors = FALSE
    )
    if (verbose) message(paste0("  1 fold | ", total_points, " points (100%)"))

    result <- list(
      folds        = sf::st_drop_geometry(pts_sf)[, "fold", drop = FALSE],
      points_sf    = pts_sf,
      voronoi_folds = NULL,
      summary      = summary_stats,
      plots        = list()
    )

    if (!is.null(output_file)) {
      out_dir <- dirname(output_file)
      if (!dir.exists(out_dir)) dir.create(out_dir, recursive = TRUE, showWarnings = FALSE)
      saveRDS(result, output_file)
      if (verbose) message(paste0("Results saved to: ", output_file))
    }

    return(invisible(result))
  }

  if (total_points < total_folds) {
    stop(paste0("ERROR: Not enough points (", total_points, ") for ", total_folds,
                " folds. Please reduce the number of folds or provide more occurrence data."))
  }

  if (partition_mode == "spatial_only") {
    n_unique_coords <- nrow(unique(sf::st_coordinates(pts_sf)))
    n_spatial <- min(
      floor(total_points / n_spatial_folds),
      n_unique_coords - 1
    )
    n_spatial <- max(n_spatial, n_spatial_folds * 2)
  }


  if (use_random) {
    if (verbose) message(paste0("\nPartition mode: RANDOM (", n_random_folds, " folds)"))

    shuffled   <- sample(total_points)
    fold_sizes <- diff(round(seq(0, total_points, length.out = n_random_folds + 1)))
    pts_sf$fold          <- NA_integer_
    pts_sf$fold[shuffled] <- rep(seq_len(n_random_folds), times = fold_sizes)
    pts_sf$spatial_block  <- NA_integer_
    pts_sf$temporal_block <- NA_integer_
    pts_sf$block_type     <- "random"

    final_fold_counts <- table(factor(pts_sf$fold, levels = seq_len(n_random_folds)))
    mean_per_fold     <- mean(final_fold_counts)
    imbalance         <- max(abs(final_fold_counts - mean_per_fold)) / mean_per_fold

    plot_result <- if (create_plot) {
      .plot_partitions_base(pts_sf, reference_shapefile, final_fold_counts, mean_per_fold,
                            n_random_folds, "random", time_cols, !is.null(time_cols), 1, NULL,
                            plot_palette)
    } else {
      list(plots = list(), fold_polygons = NULL)
    }

    result <- .build_partition_result(pts_sf, NULL, final_fold_counts, mean_per_fold, imbalance,
                                      n_spatial, n_temporal, total_folds, partition_mode,
                                      temporal_partitioning, FALSE, TRUE,
                                      0, 0, 0, n_random_folds,
                                      n_removed, n_original, plot_result$plots, output_file,
                                      reference_shapefile = reference_shapefile,
                                      verbose = verbose)
    if (verbose) message(paste(utils::capture.output(
      print(result$summary, row.names = FALSE)), collapse = "\n"))
    return(invisible(result))
  }

  if (n_spatial >= total_points) {
    stop(paste0("ERROR: Spatial blocks (", n_spatial, ") >= points (", total_points,
                "). Please reduce the number of folds or provide more occurrence data."))
  }

  if (partition_mode == "spatial_only") {
    if (verbose) message(paste0("\nPartition mode: SPATIAL_ONLY (", n_spatial_folds, " spatial folds)"))

    pts_work               <- pts_sf
    pts_work$spatial_block <- NA_integer_
    pts_work$temporal_block <- 1
    pts_work$fold          <- NA_integer_
    pts_work$block_type    <- NA_character_

    spatial_info           <- .build_spatial_blocks(pts_work, reference_shapefile, n_spatial)
    pts_work$spatial_block <- spatial_info$clusters

    pts_work <- .assign_spatiotemporal_folds(pts_work, n_spatial_folds, n_temporal_folds,
                                             total_folds, n_spatial, n_temporal,
                                             spatial_info$centers, spatial_info$dist_matrix,
                                             spatial_info$adjacency, temporal_partitioning,
                                             max_imbalance, use_balanced)

    pts_work$spatial_block <- pts_work$fold

    fold_coords <- sf::st_coordinates(pts_work)
    fold_cx <- vapply(seq_len(total_folds), function(f) {
      mean(fold_coords[pts_work$fold == f, 1])
    }, numeric(1))
    fold_cy <- vapply(seq_len(total_folds), function(f) {
      mean(fold_coords[pts_work$fold == f, 2])
    }, numeric(1))
    bbox  <- sf::st_bbox(reference_shapefile)
    vor   <- deldir::deldir(fold_cx, fold_cy,
                            rw = c(bbox["xmin"], bbox["xmax"], bbox["ymin"], bbox["ymax"]))
    tiles <- deldir::tile.list(vor)
    polys <- lapply(seq_along(tiles), function(i) {
      tile <- tiles[[i]]
      sf::st_polygon(list(cbind(c(tile$x, tile$x[1]), c(tile$y, tile$y[1]))))
    })
    voronoi_sf <- sf::st_sf(fold_group = seq_len(total_folds),
                            geometry = sf::st_sfc(polys, crs = sf::st_crs(pts_work)))
    voronoi_sf <- suppressWarnings(sf::st_intersection(voronoi_sf, sf::st_union(reference_shapefile)))

    fc  <- table(factor(pts_work$fold, levels = seq_len(total_folds)))
    mpf <- mean(fc)
    imb <- max(abs(fc - mpf)) / mpf

    if (imb > max_imbalance) {
      warning(paste0("Spatial-only folds are imbalanced (", round(imb * 100, 2),
                     "%) due to uneven point density. This is expected for spatial CV."))
    }

    plot_result <- if (create_plot) {
      .plot_partitions_base(pts_work, reference_shapefile, fc, mpf,
                            total_folds, partition_mode, time_cols, temporal_partitioning,
                            n_temporal, voronoi_sf, plot_palette)
    } else {
      list(plots = list(), fold_polygons = NULL)
    }

    result <- .build_partition_result(pts_work, voronoi_sf, fc, mpf, imb,
                                      n_spatial, n_temporal, total_folds, partition_mode,
                                      temporal_partitioning, use_balanced, FALSE,
                                      n_spatial_folds, n_temporal_folds, n_balanced_folds, 0,
                                      n_removed, n_original, plot_result$plots, output_file,
                                      reference_shapefile = reference_shapefile,
                                      verbose = verbose)
    if (verbose) message(paste(utils::capture.output(
      print(result$summary, row.names = FALSE)), collapse = "\n"))
    return(invisible(result))
  }

  best_imbalance       <- Inf
  best_results         <- NULL
  best_voronoi         <- NULL

  for (attempt in seq_len(max_attempts)) {
    pts_work               <- pts_sf
    pts_work$spatial_block  <- NA_integer_
    pts_work$temporal_block <- NA_integer_
    pts_work$fold           <- NA_integer_
    pts_work$block_type     <- NA_character_

    if (attempt == 1) {
      if (partition_mode == "spatiotemporal") {
        if (verbose) message(paste0("\nPartition mode: SPATIOTEMPORAL (",
                                    n_spatial_folds, " spatial folds, ",
                                    n_temporal_folds, " temporal folds, ",
                                    total_folds, " total folds)"))
      } else {
        if (verbose) message(paste0("\nPartition mode: ", toupper(partition_mode), " (",
                                    n_spatial, " spatial blocks, ",
                                    n_temporal, " temporal blocks, ",
                                    total_folds, " folds)"))
      }
    }

    spatial_info            <- .build_spatial_blocks(pts_work, reference_shapefile, n_spatial)
    pts_work$spatial_block  <- spatial_info$clusters
    voronoi_sf              <- spatial_info$voronoi_sf

    if (temporal_partitioning) {
      tv <- pts_work[[time_cols]]

      if (n_temporal == 1) {
        pts_work$temporal_block <- 1
      } else if (n_temporal == 2) {
        global_mid <- stats::median(tv, na.rm = TRUE)
        pts_work$temporal_block <- ifelse(tv <= global_mid, 1, 2)
      } else {
        breaks <- stats::quantile(tv, probs = seq(0, 1, length.out = n_temporal + 1),
                                  na.rm = TRUE)
        if (any(duplicated(breaks))) {
          breaks <- unique(breaks)
          warning("Duplicate temporal breaks detected.")
        }
        pts_work$temporal_block <- as.integer(cut(tv, breaks = breaks,
                                                  labels = FALSE, include.lowest = TRUE))
      }

      if (n_temporal == 2 && attempt > 1) {
        tb_counts <- table(factor(pts_work$temporal_block, levels = 1:2))
        tb_imb    <- max(tb_counts) / sum(tb_counts)
        if (tb_imb > 0.65) {
          pool_idx <- which(!is.na(pts_work$spatial_block))
          tv_pool  <- tv[pool_idx]
          alt_mid  <- stats::median(tv_pool, na.rm = TRUE)
          if (!is.na(alt_mid) && alt_mid != global_mid) {
            pts_work$temporal_block[pool_idx] <- ifelse(tv_pool <= alt_mid, 1, 2)
          }
        }
      }

    } else {
      pts_work$temporal_block <- 1
    }

    if (use_balanced) {
      pts_work <- .assign_balanced_folds(pts_work, n_balanced_folds, n_spatial,
                                         spatial_info$centers, spatial_info$dist_matrix,
                                         spatial_info$adjacency, max_imbalance, total_folds)
    } else {
      pts_work <- .assign_spatiotemporal_folds(pts_work, n_spatial_folds, n_temporal_folds,
                                               total_folds, n_spatial, n_temporal,
                                               spatial_info$centers, spatial_info$dist_matrix,
                                               spatial_info$adjacency, temporal_partitioning,
                                               max_imbalance, use_balanced)
    }

    unassigned <- which(is.na(pts_work$fold))
    for (idx in unassigned) {
      counts <- table(factor(pts_work$fold[!is.na(pts_work$fold)],
                             levels = seq_len(total_folds)))
      pts_work$fold[idx]       <- as.integer(names(which.min(counts))[1])
      pts_work$block_type[idx] <- "remainder"
    }

    fc  <- table(factor(pts_work$fold, levels = seq_len(total_folds)))
    mpf <- mean(fc)
    imb <- max(abs(fc - mpf)) / mpf

    if (imb < best_imbalance) {
      best_imbalance <- imb
      best_results   <- pts_work
      best_voronoi   <- voronoi_sf
      best_counts    <- fc
      best_mean      <- mpf
    }
    if (imb <= max_imbalance) break
  }

  if (best_imbalance > max_imbalance) {
    warning(paste0("Could not achieve target balance within ",
                   max_attempts, " attempts. ",
                   "Final imbalance: ", round(best_imbalance * 100, 2), "%. ",
                   "Returning best result achieved. Try increasing max_imbalance or adjusting the fold configuration."))
  }

  pts_sf            <- best_results
  voronoi_sf        <- best_voronoi
  final_fold_counts <- best_counts
  mean_per_fold     <- best_mean
  imbalance         <- best_imbalance

  plot_result <- if (create_plot) {
    .plot_partitions_base(pts_sf, reference_shapefile, final_fold_counts, mean_per_fold,
                          total_folds, partition_mode, time_cols, temporal_partitioning,
                          n_temporal, voronoi_sf, plot_palette)
  } else {
    list(plots = list(), fold_polygons = NULL)
  }

  result <- .build_partition_result(pts_sf, voronoi_sf, final_fold_counts, mean_per_fold,
                                    imbalance, n_spatial, n_temporal, total_folds, partition_mode,
                                    temporal_partitioning, use_balanced, FALSE,
                                    n_spatial_folds, n_temporal_folds, n_balanced_folds, 0,
                                    n_removed, n_original, plot_result$plots, output_file,
                                    reference_shapefile = reference_shapefile,
                                    verbose = verbose)
  if (verbose) message(paste(utils::capture.output(
    print(result$summary, row.names = FALSE)), collapse = "\n"))
  invisible(result)
}

#' Build Temporal Hypervolume Models Across Cross-Validation Folds
#'
#' Modeling function that constructs hypervolume models for each
#' cross-validation fold using either Gaussian kernel density estimation or
#' one-class SVM. Each hypervolume reserves one fold as testing data and uses
#' the remaining folds as training data. The returned object follows the same
#' structure as \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
#' and \code{\link{build_temporal_rf}}, and is accepted directly by
#' \code{\link{generate_spatiotemporal_predictions}}.
#'
#' @usage
#' build_temporal_hv(partition_result, model_vars, method,
#'                   hypervolume_params = list(),
#'                   output_dir = file.path(tempdir(), "Hypervolume_Models"),
#'                   create_plot = TRUE, overwrite = FALSE,
#'                   plot_palette = "Dark 2", verbose = TRUE)
#'
#' @param partition_result List or character. Output from
#'   \code{\link{spatiotemporal_partition}} or path to an \code{.rds}
#'   file containing that output.
#' @param model_vars Character vector. Names of predictor columns to use in
#'   hypervolume construction. All variables must be present as columns in
#'   the occurrence data produced by \code{\link{temporally_explicit_extraction}}.
#' @param method Character. Hypervolume method. One of \code{"gaussian"} for
#'   Gaussian kernel density estimation or \code{"svm"} for one-class support
#'   vector machine. Required.
#' @param hypervolume_params Named list. Additional parameters passed to
#'   \code{\link[hypervolume]{hypervolume_gaussian}} or
#'   \code{\link[hypervolume]{hypervolume_svm}}. For Gaussian models, valid
#'   keys are \code{kde.bandwidth}, \code{quantile.requested},
#'   \code{quantile.requested.type}, \code{chunk.size}, \code{verbose}, and
#'   \code{samples.per.point}. For SVM models, valid keys are \code{svm.nu},
#'   \code{svm.gamma}, \code{chunk.size}, \code{verbose}, and
#'   \code{samples.per.point}. Default is an empty list, which uses the
#'   built-in defaults.
#' @param output_dir Character. Directory to write output files including saved
#'   hypervolume objects and plots. Default is \code{file.path(tempdir(), "Hypervolume_Models")}.
#' @param create_plot Logical. If \code{TRUE}, generates pairplot visualisations
#'   of each fold's hypervolume and a combined comparison plot. Default is
#'   \code{TRUE}.
#' @param overwrite Logical. If \code{TRUE}, overwrites existing saved
#'   hypervolume files. If \code{FALSE}, loads existing files when available.
#'   Default is \code{FALSE}.
#' @param plot_palette Character. Name of an HCL or RColorBrewer palette used
#'   to color folds in diagnostic plots. Accepts any HCL palette name (see
#'   \code{\link[grDevices]{hcl.pals}}) or, if \pkg{RColorBrewer} is installed,
#'   any Brewer palette name. Default is \code{"Dark 2"}.
#' @param verbose Logical. If \code{TRUE} (default), prints progress
#'   messages during processing. Includes per-fold hypervolume construction
#'   progress, volume summaries, and overlap statistics.
#'
#' @return A list with class \code{"TemporalHypervolume"} containing:
#' \itemize{
#'   \item \code{models}: Named list of fitted \code{Hypervolume} objects, one
#'     per fold, named \code{fold1}, \code{fold2}, etc. This naming convention
#'     matches the other \code{build_temporal_*()} functions and is required by
#'     \code{\link{generate_spatiotemporal_predictions}}.
#'   \item \code{volumes}: Named numeric vector of hypervolume sizes (units
#'     depend on the number of dimensions and bandwidth).
#'   \item \code{overlaps}: Named list of pairwise percent volume overlaps
#'     between all fold combinations.
#'   \item \code{method}: Character string recording the method used.
#'   \item \code{model_vars}: Character vector of predictor names used.
#'   \item \code{fold_training_data}: Named list of training data frames used
#'     to fit each fold model, retained for consistency with the other model
#'     types and for downstream use.
#'   \item \code{fold_test_metrics}: Data frame of E-space inclusion metrics
#'     computed on the held-out test points for each fold. Columns: \code{fold},
#'     \code{n_test}, \code{volume}, \code{tp}, \code{fn}, \code{sensitivity},
#'     \code{Sensitivity}. Printed as a summary table at the end of
#'     model building.
#'   \item \code{output_dir}: Path to the output directory.
#'   \item \code{model_type}: Character string \code{"hypervolume"}, used by
#'     \code{\link{generate_spatiotemporal_predictions}}.
#'   \item \code{plots}: List of recorded plot objects (if
#'     \code{create_plot = TRUE}). Plots can be replayed with
#'     \code{grDevices::replayPlot()}.
#' }
#'
#' @details
#' For N folds, constructs N hypervolumes where each fold reserves one group
#' of points for testing and uses the remaining N-1 groups for training.
#' Pairwise overlap statistics quantify similarity between hypervolumes across
#' folds, with low overlap indicating that the environmental space sampled
#' differs substantially across folds.
#'
#' Hypervolume objects are saved as a combined \code{.rds} file in
#' \code{output_dir}. If the file already exists and \code{overwrite = FALSE},
#' the saved file is loaded rather than re-fitting.
#'
#' The returned object is accepted directly by
#' \code{\link{generate_spatiotemporal_predictions}}, which uses the
#' \code{model_type} field to dispatch hypervolume-specific projection logic
#' via \code{\link[hypervolume]{hypervolume_project}}.
#'
#' @seealso
#' Preprocessing: \code{\link{spatiotemporal_partition}}
#'
#' Modeling: \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
#'   \code{\link{build_temporal_rf}},
#'   \code{\link{generate_spatiotemporal_predictions}}
#'
#' External: \code{\link[hypervolume]{hypervolume_gaussian}},
#'   \code{\link[hypervolume]{hypervolume_svm}}
#'
#' @examples
#' data(tmr_partition_small, package = "TemporalModelR")
#'
#' build_temporal_hv(
#'   partition_result = tmr_partition_small,
#'   model_vars       = c("elevation", "forest_cover"),
#'   method           = "svm",
#'   output_dir       = tempdir(),
#'   create_plot      = FALSE,
#'   verbose          = FALSE
#' )

#' @export
#' @importFrom sf st_drop_geometry
#' @importFrom stats complete.cases var
#' @importFrom grDevices dev.cur dev.off png recordPlot
#' @importFrom graphics plot
#' @importFrom utils combn modifyList write.csv
build_temporal_hv <- function(partition_result,
                              model_vars,
                              method,
                              hypervolume_params = list(),
                              output_dir         = file.path(tempdir(), "Hypervolume_Models"),
                              create_plot        = TRUE,
                              overwrite          = FALSE,
                              plot_palette       = "Dark 2",
                              verbose            = TRUE) {

  ### Package check

  if (!requireNamespace("hypervolume", quietly = TRUE)) {
    stop(paste0(
      "ERROR: The 'hypervolume' package is required to use build_temporal_hv(). ",
      "Install it with: install.packages('hypervolume')"
    ))
  }

  if (missing(partition_result)) {
    stop(paste0("ERROR: 'partition_result' is required. Provide output from ",
                "spatiotemporal_partition() or a path to an .rds file."))
  }

  if (missing(method) || !method %in% c("gaussian", "svm")) {
    stop("ERROR: 'method' is required and must be 'gaussian' or 'svm'.")
  }

  if (missing(model_vars) || !is.character(model_vars) || length(model_vars) == 0) {
    stop("ERROR: 'model_vars' must be a non-empty character vector of predictor names.")
  }

  partition_result <- .load_partition_result(partition_result)
  occurrence_points <- partition_result$points_sf

  if (is.null(occurrence_points) || nrow(occurrence_points) == 0) {
    stop("ERROR: 'partition_result$points_sf' is empty.")
  }
  if (!"fold" %in% names(occurrence_points)) {
    stop(paste0(
      "ERROR: Missing 'fold' column in 'partition_result$points_sf'. Available columns: ",
      paste(names(occurrence_points)[names(occurrence_points) != "geometry"], collapse = ", ")
    ))
  }

  occ_df   <- sf::st_drop_geometry(occurrence_points)
  all_folds <- sort(unique(occ_df$fold[!is.na(occ_df$fold)]))

  if (length(all_folds) == 0) {
    stop("ERROR: All fold values are NA. Re-run spatiotemporal_partition().")
  }

  single_fold_mode <- length(all_folds) == 1
  if (single_fold_mode) {
    if (verbose) message("Single-fold mode: all points used for both training and testing.")
  } else if (length(all_folds) < 2) {
    warning(paste0("Only ", length(all_folds), " unique fold detected."))
  }

  if (verbose) message(paste("Loaded", nrow(occ_df), "points across", length(all_folds), "folds."))

  available_vars <- names(occ_df)[names(occ_df) != "geometry"]
  missing_vars   <- model_vars[!model_vars %in% available_vars]
  if (length(missing_vars) > 0) {
    stop(paste0(
      "ERROR: The following 'model_vars' are not present in the occurrence data: ",
      paste(missing_vars, collapse = ", "),
      ". Available: ", paste(available_vars, collapse = ", ")
    ))
  }

  if (length(hypervolume_params) > 0) {
    valid_gaussian <- c("kde.bandwidth", "quantile.requested", "quantile.requested.type",
                        "chunk.size", "verbose", "samples.per.point")
    valid_svm      <- c("svm.nu", "svm.gamma", "chunk.size", "verbose", "samples.per.point")
    valid_params   <- if (method == "gaussian") valid_gaussian else valid_svm
    invalid        <- setdiff(names(hypervolume_params), valid_params)
    if (length(invalid) > 0) {
      warning(paste0(
        "Unrecognized hypervolume_params for method '", method, "': ",
        paste(invalid, collapse = ", "),
        ". Valid keys: ", paste(valid_params, collapse = ", ")
      ))
    }
    if (method == "gaussian" && "quantile.requested" %in% names(hypervolume_params)) {
      qr <- hypervolume_params$quantile.requested
      if (!is.numeric(qr) || qr <= 0 || qr > 1) {
        warning(paste0("quantile.requested should be between 0 and 1, got: ", qr))
      }
    }
    if (method == "svm" && "svm.nu" %in% names(hypervolume_params)) {
      nu <- hypervolume_params$svm.nu
      if (!is.numeric(nu) || nu <= 0 || nu > 1) {
        warning(paste0("svm.nu should be between 0 and 1, got: ", nu))
      }
    }
  }

  if (!dir.exists(output_dir)) {
    dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
  }

  n_original    <- nrow(occ_df)
  na_counts     <- vapply(model_vars, function(v) sum(is.na(occ_df[[v]])), integer(1))
  if (any(na_counts > 0)) {
    warning(paste0(
      "NA values in: ",
      paste(paste0(names(na_counts[na_counts > 0]), " (", na_counts[na_counts > 0], ")"), collapse = ", ")
    ))
  }
  occ_df    <- occ_df[stats::complete.cases(occ_df[, model_vars, drop = FALSE]), ]
  n_removed <- n_original - nrow(occ_df)
  if (n_removed > 0) {
    warning(paste0("Removed ", n_removed, " rows with NAs. Remaining: ", nrow(occ_df)))
  }
  if (nrow(occ_df) == 0) {
    stop("ERROR: No complete cases remain after removing NA values in model predictors.")
  }

  fold_ids <- sort(unique(occ_df$fold))
  if (verbose) message(paste0(
    "Building hypervolume models (", method, ") for ",
    length(fold_ids), " folds: ", paste(fold_ids, collapse = ", ")
  ))

  default_gaussian <- list(
    quantile.requested      = 0.95,
    quantile.requested.type = "probability",
    chunk.size              = 1000,
    verbose                 = FALSE
  )
  default_svm <- list(
    svm.nu     = 0.01,
    svm.gamma  = 0.5,
    chunk.size = 1000,
    verbose    = FALSE
  )

  combined_file <- file.path(output_dir, paste0("all_hypervolumes_", method, ".rds"))

  model_list      <- list()
  train_data_list <- list()

  if (file.exists(combined_file) && !overwrite) {

    if (verbose) message(paste("Loading existing hypervolume file:", basename(combined_file)))
    saved <- tryCatch(
      readRDS(combined_file),
      error = function(e) stop(paste0("ERROR loading saved hypervolumes. Try overwrite = TRUE: ", e$message))
    )

    if (is.list(saved) && all(c("models", "fold_test_metrics", "model_type") %in% names(saved))) {
      if (verbose) message(paste("Loaded complete result with", length(saved$models), "hypervolumes."))
      class(saved) <- c("TemporalHypervolume", "list")
      return(invisible(saved))
    }

    if (is.list(saved) && "models" %in% names(saved)) {
      model_list      <- saved$models
      train_data_list <- saved$fold_training_data
    } else {
      model_list <- saved
      for (fold in fold_ids) {
        fold_key <- paste0("fold", fold)
        train_data_list[[fold_key]] <- occ_df[occ_df$fold != fold, model_vars, drop = FALSE]
      }
    }

    for (fold in fold_ids) {
      fold_key <- paste0("fold", fold)
      if (!is.null(model_list[[fold_key]]) &&
          (is.null(model_list[[fold_key]]@Name) || model_list[[fold_key]]@Name == "untitled")) {
        model_list[[fold_key]]@Name <- paste("Fold", fold)
      }
    }
    if (verbose) message(paste("Loaded", length(model_list), "hypervolumes from file."))

  } else {

    hv_fn <- if (method == "gaussian") hypervolume::hypervolume_gaussian else hypervolume::hypervolume_svm
    defaults <- if (method == "gaussian") default_gaussian else default_svm

    for (fold in fold_ids) {

      fold_key   <- paste0("fold", fold)
      train_data <- if (single_fold_mode) occ_df[, model_vars, drop = FALSE] else occ_df[occ_df$fold != fold, model_vars, drop = FALSE]
      n_train    <- nrow(train_data)

      if (verbose) message(paste0("Fold ", fold, ": training on ", n_train, " points."))

      if (n_train < 5) {
        stop(paste0("ERROR: Fold ", fold, " has only ", n_train, " training points. Minimum 5 required."))
      }
      if (n_train < 10) {
        warning(paste0("Fold ", fold, " has only ", n_train, " training points. Results may be unreliable."))
      }

      var_check <- vapply(train_data[, model_vars, drop = FALSE], function(x) stats::var(x, na.rm = TRUE), numeric(1))
      zero_var  <- names(var_check[!is.na(var_check) & var_check == 0])
      if (length(zero_var) > 0) {
        warning(paste0("Zero variance in Fold ", fold, " for: ", paste(zero_var, collapse = ", "),
                       ". These predictors will have no effect in this fold."))
      }

      hv_params <- modifyList(defaults, hypervolume_params)

      hv <- tryCatch({
        h <- suppressMessages(do.call(hv_fn, c(list(data = train_data), hv_params)))
        h@Name <- paste("Fold", fold)
        h
      }, error = function(e) {
        stop(paste0("ERROR: Hypervolume construction failed for Fold ", fold, ": ", e$message))
      })

      vol <- tryCatch(
        hypervolume::get_volume(hv),
        error = function(e) NA_real_
      )
      if (verbose) message(paste0("  Volume: ", round(vol, 4)))

      model_list[[fold_key]]      <- hv
      train_data_list[[fold_key]] <- train_data
    }

    tryCatch({
      saveRDS(
        list(models = model_list, fold_training_data = train_data_list),
        combined_file
      )
      if (verbose) message(paste("Saved hypervolumes to:", basename(combined_file)))
    }, error = function(e) {
      warning(paste0("Could not save hypervolume file: ", e$message))
    })
  }

  volumes <- vapply(model_list, function(h) {
    tryCatch(hypervolume::get_volume(h), error = function(e) NA_real_)
  }, numeric(1))

  if (verbose) message("--- Fold Volumes ---")
  for (fold in fold_ids) {
    fold_key <- paste0("fold", fold)
    if (verbose) message(paste0("  Fold ", fold, ": ", round(volumes[fold_key], 4)))
  }

  overlap_stats <- list()

  if (length(model_list) > 1) {
    if (verbose) message("Calculating pairwise hypervolume overlaps...")
    fold_pairs <- utils::combn(names(model_list), 2, simplify = FALSE)

    for (pair in fold_pairs) {
      pair_label <- paste(
        "Fold", gsub("fold", "", pair[1]),
        "vs Fold", gsub("fold", "", pair[2])
      )
      pct <- tryCatch({
        hv_set  <- suppressMessages(hypervolume::hypervolume_set(
          model_list[[pair[1]]], model_list[[pair[2]]],
          check.memory = FALSE, verbose = FALSE
        ))
        stats_out <- suppressMessages(hypervolume::hypervolume_overlap_statistics(hv_set))
        if (is.list(stats_out) && "percent_volume_overlap" %in% names(stats_out)) {
          stats_out$percent_volume_overlap
        } else if (is.numeric(stats_out)) {
          stats_out[1]
        } else NA_real_
      }, error = function(e) {
        warning(paste0("Could not calculate overlap for ", pair_label, ": ", e$message))
        NA_real_
      })
      overlap_stats[[pair_label]] <- pct
      if (verbose) message(paste0("  ", pair_label, ": ", round(pct * 100, 2), "%"))
    }
  }

  plot_list   <- list()
  fold_colors <- .resolve_palette(length(fold_ids), plot_palette)
  names(fold_colors) <- paste0("fold", fold_ids)

  if (create_plot && length(model_vars) >= 2) {
    if (verbose) message("Generating hypervolume plots...")

    for (fold in fold_ids) {
      fold_key  <- paste0("fold", fold)
      plot_file <- file.path(output_dir, paste0("Hypervolume_Fold", fold, ".png"))

      tryCatch({
        grDevices::png(plot_file, width = 1200, height = 1000, res = 150)
        suppressMessages(graphics::plot(
          model_list[[fold_key]],
          pairplot = TRUE, show.3d = FALSE,
          main     = paste("Fold", fold, "Hypervolume"),
          colors   = fold_colors[fold_key]
        ))
        grDevices::dev.off()
        if (verbose) message(paste("Saved Fold", fold, "plot:", basename(plot_file)))
      }, error = function(e) {
        warning(paste0("Could not save plot for Fold ", fold, ": ", e$message))
        if (grDevices::dev.cur() > 1) grDevices::dev.off()
      })

      tryCatch({
        suppressMessages(graphics::plot(
          model_list[[fold_key]],
          pairplot = TRUE, show.3d = FALSE,
          main     = paste("Fold", fold, "Hypervolume"),
          colors   = fold_colors[fold_key]
        ))
        plot_list[[paste0("fold", fold)]] <- grDevices::recordPlot()
      }, error = function(e) {
        warning(paste0("Could not record plot for Fold ", fold, ": ", e$message))
      })
    }

    if (length(model_list) > 1) {
      comp_file <- file.path(output_dir, "Hypervolume_Comparison.png")
      hv_joined <- tryCatch(
        suppressMessages(do.call(hypervolume::hypervolume_join, model_list)),
        error = function(e) {
          warning(paste0("Could not join hypervolumes for comparison plot: ", e$message))
          NULL
        }
      )
      if (!is.null(hv_joined)) {
        tryCatch({
          grDevices::png(comp_file, width = 1400, height = 1200, res = 150)
          suppressMessages(graphics::plot(
            hv_joined,
            pairplot = TRUE, show.3d = FALSE,
            main     = paste("Hypervolume Comparison:", length(model_list), "Folds"),
            colors   = fold_colors
          ))
          grDevices::dev.off()
          if (verbose) message(paste("Saved comparison plot:", basename(comp_file)))
        }, error = function(e) {
          warning(paste0("Could not save comparison plot: ", e$message))
          if (grDevices::dev.cur() > 1) grDevices::dev.off()
        })
        tryCatch({
          suppressMessages(graphics::plot(
            hv_joined,
            pairplot = TRUE, show.3d = FALSE,
            main     = paste("Hypervolume Comparison:", length(model_list), "Folds"),
            colors   = fold_colors
          ))
          plot_list[["comparison"]] <- grDevices::recordPlot()
        }, error = function(e) {
          warning(paste0("Could not record comparison plot: ", e$message))
        })
      }
    }
  }

  if (verbose) message("Computing E-space inclusion metrics on held-out test folds...")

  metric_rows <- list()

  for (fold in fold_ids) {
    fold_key  <- paste0("fold", fold)
    hv_model  <- model_list[[fold_key]]
    test_data <- occ_df[occ_df$fold == fold, model_vars, drop = FALSE]
    test_data <- test_data[stats::complete.cases(test_data), , drop = FALSE]

    tp <- NA_integer_; fn <- NA_integer_
    sensitivity <- NA_real_

    if (!is.null(hv_model) && nrow(test_data) > 0) {
      inclusion <- tryCatch(
        suppressMessages(suppressWarnings(
          hypervolume::hypervolume_inclusion_test(hv_model, test_data)
        )),
        error = function(e) {
          warning(paste0("Inclusion test failed for Fold ", fold, ": ", e$message))
          NULL
        }
      )
      if (!is.null(inclusion)) {
        tp  <- sum(inclusion, na.rm = TRUE)
        fn  <- sum(!inclusion, na.rm = TRUE)
        tot <- tp + fn
        sensitivity <- if (tot > 0) tp / tot else NA_real_
      }
    }

    vol <- tryCatch(hypervolume::get_volume(hv_model), error = function(e) NA_real_)

    metric_rows[[fold_key]] <- data.frame(
      Fold          = fold,
      N_Test        = nrow(test_data),
      E_Volume      = round(vol, 4),
      Testing_TP    = tp,
      Testing_FN    = fn,
      Sensitivity   = round(sensitivity, 4),
      stringsAsFactors = FALSE
    )
  }

  fold_metrics_df           <- do.call(rbind, metric_rows)
  rownames(fold_metrics_df) <- NULL

  metrics_file <- file.path(output_dir, "Fold_Test_Metrics.csv")
  tryCatch(
    utils::write.csv(fold_metrics_df, metrics_file, row.names = FALSE),
    error = function(e) warning(paste0("Could not save fold test metrics: ", e$message))
  )

  if (verbose) message("--- Hypervolume Modeling Complete ---")
  if (verbose) message(paste0("  Method: ", method))
  if (nrow(fold_metrics_df) > 0) {
    if (verbose) message("  Held-out E-space inclusion metrics:")
    if (verbose) message(paste0("  ", paste(sprintf("%-10s",
                                                    c("Fold", "N_Test", "E_Volume", "Testing_TP", "Testing_FN", "Sensitivity")),
                                            collapse = " ")))
    for (i in seq_len(nrow(fold_metrics_df))) {
      r <- fold_metrics_df[i, ]
      if (verbose) message(paste0("  ", paste(sprintf("%-10s", c(
        r$Fold,
        r$N_Test,
        round(r$E_Volume,    3),
        r$Testing_TP,
        r$Testing_FN,
        round(r$Sensitivity, 3)
      )), collapse = " ")))
    }
    if (verbose) message(paste0("  Means      ",
                                paste(sprintf("%-10s", c(
                                  "",
                                  round(mean(fold_metrics_df$N_Test,      na.rm = TRUE), 0),
                                  round(mean(fold_metrics_df$E_Volume,    na.rm = TRUE), 3),
                                  round(mean(fold_metrics_df$Testing_TP,  na.rm = TRUE), 1),
                                  round(mean(fold_metrics_df$Testing_FN,  na.rm = TRUE), 1),
                                  round(mean(fold_metrics_df$Sensitivity, na.rm = TRUE), 3)
                                )), collapse = " ")
    ))
    if (verbose) message(paste("  Saved to:", basename(metrics_file)))
  }

  if (create_plot && nrow(fold_metrics_df) > 0) {
    plot_rate_cols <- c("Sensitivity")
    bar_file <- file.path(output_dir, "Fold_Test_Metrics.png")
    tryCatch({
      grDevices::png(bar_file, width = 300 * length(plot_rate_cols) + 200,
                     height = 700, res = 150)
      .plot_fold_metrics_bars(
        metrics_df  = fold_metrics_df,
        rate_cols   = plot_rate_cols,
        fold_colors = fold_colors,
        title       = "Hypervolume Time Independent Assessment Metrics"
      )
      grDevices::dev.off()
      if (verbose) message(paste("Saved fold metrics plot:", basename(bar_file)))
    }, error = function(e) {
      warning(paste0("Could not save fold metrics plot: ", e$message))
      if (grDevices::dev.cur() > 1) grDevices::dev.off()
    })
    tryCatch({
      .plot_fold_metrics_bars(
        metrics_df  = fold_metrics_df,
        rate_cols   = plot_rate_cols,
        fold_colors = fold_colors,
        title       = "Hypervolume Time Independent Assessment Metrics"
      )
      plot_list[["fold_test_metrics"]] <- grDevices::recordPlot()
    }, error = function(e) {
      warning(paste0("Could not record fold metrics plot: ", e$message))
    })
  }

  result <- list(
    models             = model_list,
    volumes            = volumes,
    overlaps           = overlap_stats,
    method             = method,
    model_vars         = model_vars,
    fold_training_data = train_data_list,
    fold_test_metrics  = fold_metrics_df,
    output_dir         = output_dir,
    model_type         = "hypervolume",
    plots              = plot_list
  )

  class(result) <- c("TemporalHypervolume", "list")

  tryCatch({
    saveRDS(result, combined_file)
    if (verbose) message(paste("Updated saved object with metrics and plots:", basename(combined_file)))
  }, error = function(e) {
    warning(paste0("Could not update saved hypervolume file: ", e$message))
  })

  invisible(result)
}

---
title: "3a. Modeling with a GLM"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Modeling with a GLM}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 5,
  dpi       = 100,
  out.width = "95%"
)
```

<hr>

# Summary

-   [Theory](#sec-theory)
-   [Overview](#sec-overview)
-   [Fitting a temporal GLM](#sec-fit)
-   [Formula syntax](#sec-formula)
-   [Threshold selection](#sec-threshold)
-   [E-space performance](#sec-espace)
-   [Projecting predictions](#sec-project)
-   [G-space performance](#sec-gspace)
-   [Next steps](#sec-next)

<hr>

# Theory {#sec-theory}

TemporalModelR operates across two complementary dimensions to support temporally explicit modeling:

-   `E-space` - (Environmental Space) represents a location in terms of its environmental conditions, independent of geography. A species' niche in E-space is its tolerance for a given set of environmental variables.
-   `G-space` - (Geographic Space) represents a location in terms of where it physically sits on the landscape. Every point in G-space has a corresponding location in E-space.

Because environmental conditions change over time, the same point in G-space may move through E-space across years, decades, or seasons. Traditional, temporally-static workflows assume both G-space and E-space to be stable in time. TemporalModelR assumes only that the niche (i.e., the species' tolerance in E-space) is stable across the study period, while the E-space coordinates of any G-space location may change over time. Species observations are therefore matched to E-space data correct in both space and time, the niche is modeled in time-independent E-space. This is done in the [Preprocessing temporally explicit data](https://cjhughes926.github.io/TemporalModelR/articles/V2_Preprocessing.html) vignette.

By modeling the niche in time-independent E-space using time-independent training data, we can generate a static prediction of a species' environmental tolerances. We can then project these static E-space predictions back onto G-space at explicit time periods, resulting in dynamic, temporally-explicit ENM predictions of species distributions.

<br>

# Overview {#sec-overview}

`build_temporal_glm()` fits one binomial GLM per cross-validation fold as was created during the [Preprocessing temporally explicit data](https://cjhughes926.github.io/TemporalModelR/articles/V2_Preprocessing.html) vignette. Each fold's model is trained on all data outside the fold and evaluated on the held-out fold. The user supplies the right-hand side of the model formula directly in standard R syntax, with this function supporting linear, polynomial, and interactive terms. See [Formula syntax](#sec-formula). The link function defaults to logit but can be set to probit, complementary log-log, or cauchit. A threshold is selected on the training data and applied to the continuous predictions to produce binary suitability output for downstream summarization.

GLMs are fast to fit, transparent to interpret, and tolerant of the kind of small datasets that occurrence records often provide. The trade-off is that nonlinear environmental responses must be specified manually with polynomial terms or transformations. If you want a more flexible response shape, use `build_temporal_gam()` instead.

This vignette runs the seasonal workflow using the bundled `tmr_partition` and `tmr_absences` objects, which are pre-built outputs of the partitioning and pseudoabsence steps produced by the [Preprocessing temporally explicit data](https://cjhughes926.github.io/TemporalModelR/articles/V2_Preprocessing.html) vignette using the same call patterns shown there. The dataset itself is described in [About the Example Dataset](https://cjhughes926.github.io/TemporalModelR/articles/V1_dataset.html). This is the same dataset as is used in each modeling vignette [GLM](https://cjhughes926.github.io/TemporalModelR/articles/V3a_GLM.html), [GAM](https://cjhughes926.github.io/TemporalModelR/articles/V3b_GAM.html), [Random Forest](https://cjhughes926.github.io/TemporalModelR/articles/V3c_RF.html) or [Hypervolume](https://cjhughes926.github.io/TemporalModelR/articles/V3d_HV.html).


```{r}
library(TemporalModelR)
library(terra)

data(tmr_partition, package = "TemporalModelR")

data(tmr_absences,  package = "TemporalModelR")
```

<br>

# Fitting a temporal GLM {#sec-fit}

The minimal call needs the partition, the pseudoabsences, a formula, and the time columns. We fit an additive model in three predictors and let the function select per-fold thresholds via the True Skill Statistic (TSS), though this threshold may also be set manually or using prevalence. See [Threshold selection](#sec-threshold).

`create_plot = TRUE` allows the function to generate visual response curves for each per-fold model result. On the below graphs, response curves are shown per-fold, and the selected threshold is indicated with a horizontal dashed line.

```{r}
glm_out <- build_temporal_glm(
  partition_result     = tmr_partition,
  pseudoabsence_result = tmr_absences,
  model_formula        = ~ elevation + forest_cover + prseas,
  link                 = "logit",
  threshold_method     = "tss",
  output_dir           = file.path(tempdir(), "GLM_Models"),
  create_plot          = TRUE,
  time_cols            = c("year", "season"),
  verbose              = FALSE
)
```

We see that our model correctly identifies a strong positive relationship between both elevation and forest cover and species occurrence. The same `create_plot = TRUE` function also visualizes time independent model assessment metrics relevant to the GLM (above). See [E-space performance](#sec-espace).

```{r}
class(glm_out)

names(glm_out)
```

The returned object is a list of class `"TemporalGLM"` containing the following objects:

-   `$models` - named list of `glm` objects, one per fold (`fold1`, `fold2`, ...). Each is a standard `glm` fit and accepts R's standard `summary()`, `coef()`, and `predict()` calls.
-   `$thresholds` - named numeric vector of probability thresholds used to binarize predictions for each fold.
-   `$threshold_method` - character or numeric value recording how the thresholds were chosen (`"tss"`, `"prevalence"`, or a fixed numeric value).
-   `$model_formula` - the right hand side formula supplied to the function.
-   `$link` - the link function used (`"logit"`, `"probit"`, `"cloglog"`, or `"cauchit"`) supplied to the function.
-   `$model_vars` - character vector of the predictor variable names extracted from the formula. Used by `generate_spatiotemporal_predictions()` to know which raster layers to assemble at projection time.
-   `$fold_training_data` - named list of the training data frames used for each fold, with presences and pseudoabsences combined and environmental columns attached.
-   `$fold_test_metrics` - data frame of per-fold E-space metrics computed on the held-out test set. See [E-space performance](#sec-espace).
-   `$output_dir` - path to the directory where the saved RDS of fitted models and the fold test metrics CSV were written.
-   `$model_type` - `"glm"`, used by `generate_spatiotemporal_predictions()` to choose the correct projection logic.
-   `$plots` - list of recorded base-R plots produced when `create_plot = TRUE` (response curves and the fold metrics bar chart).

<br>

# Formula syntax {#sec-formula}

The response variable `presence` is added automatically to each formula, so only the right-hand side needs to be supplied. Standard R formula operators are valid: `+` for additive terms, `:` for interaction only, `*` for main effects plus interaction, `I()` for arithmetic transformations, `poly()` for orthogonal polynomials, `log()`, `sqrt()`, and so on. Variable names must match column names in the data exactly. These examples that all work:

```{r, eval=FALSE}
### Additive
~ elevation + forest_cover + prseas

### Quadratic term on elevation
~ elevation + I(elevation^2) + forest_cover + prseas

### Orthogonal polynomial of degree 2 on elevation
~ poly(elevation, 2) + forest_cover + prseas

### Forest cover by precipitation interaction
~ elevation + forest_cover * prseas
```

If you choose a link other than logit, set it via the `link` argument.

<br>

# Threshold selection {#sec-threshold}

GLMs produce probabilities on the 0–1 scale. The rest of the TemporalModelR pipeline operates on binary suitability rasters, so a probability threshold must be chosen to convert probabilities into 0/1 predictions. `threshold_method` supports three options:

-   `"tss"` (default) - selects the threshold that maximizes True Skill Statistic (sensitivity + specificity − 1) on the training data. See [E-space performance](#sec-espace). This balances commission and omission.
-   `"prevalence"` - sets the threshold equal to the prevalence (proportion of presences) in the training data for that fold.
-   A numeric value between 0 and 1 - used as a fixed threshold for all folds, e.g. `threshold_method = 0.4`.

The chosen thresholds for our four folds:

```{r}
glm_out$thresholds
```

These are per-fold values because each fold trains on a different subset of data and may end up with different probability distributions. Using a single global threshold across folds risks underclassifying suitability in some folds.

<br>

# E-space performance {#sec-espace}

Each fold's held-out test set provides a set of presence and pseudoabsence points the model has never seen based on the folds defined in the [Preprocessing temporally explicit data](https://cjhughes926.github.io/TemporalModelR/articles/V2_Preprocessing.html) vignette. We compare predicted vs observed at those points and compute confusion-matrix metrics. Because these are evaluated in E-space (the species' environmental tolerance, independent of geography or time), they give a stable picture of how well the model is preforming.

```{r}
glm_out$fold_test_metrics
```

Columns:

-   `Fold` - fold identifier matching the folds in `tmr_partition$points_sf$fold`.
-   `Threshold` - the per-fold probability threshold used to binarize predictions.
-   `Testing_TP` - count of test-set presence points correctly classified as suitable (true positives).
-   `Testing_FN` - count of test-set presence points incorrectly classified as unsuitable (false negatives).
-   `Testing_TN` - count of test-set pseudoabsence points correctly classified as unsuitable (true negatives).
-   `Testing_FP` - count of test-set pseudoabsence points incorrectly classified as suitable (false positives).
-   `Sensitivity` - proportion of test-set presences correctly classified, `Testing_TP / (Testing_TP + Testing_FN)`. Values close to 1 mean the model rarely misses presences.
-   `Specificity` - proportion of test-set pseudoabsences correctly classified, `Testing_TN / (Testing_TN + Testing_FP)`. Values close to 1 mean the model rarely flags pseudoabsence points as suitable.
-   `TSS` - True Skill Statistic, `Sensitivity + Specificity - 1`. Ranges from -1 to 1, with 0 indicating performance no better than random and values above ~0.4 considered useful for most applications.
-   `Kappa` - Cohen's kappa statistic adjusting overall classification accuracy for the accuracy expected by chance. Ranges from -1 to 1 with similar interpretation to TSS.
-   `AUC` - area under the ROC curve, computed on continuous probabilities rather than binarized predictions, so it is threshold-independent. Useful for comparing the models overall performance independent of threshold.

The full ROC curve and each above metric are also graphed as an output of `create_plot = T` in the [Fitting a temporal GLM](#sec-fit) section.

E-space metrics are robust to imbalanced sample sizes across time, because they pool across the time series. They are a good metric for assessing overall model fit. Time-specific (G-space) metrics can also be assessed later when we project the model to specific G-space and time combinations.

<br>

# Projecting predictions {#sec-project}

`generate_spatiotemporal_predictions()` projects each fold's model onto a stack of environmental rasters matching each requested time step, producing one fold-vote raster per time step. The same call works for all four model types; `model_type` in the model object tells the function which projection logic to use.

For this example we project across all 15 years and 4 seasons, producing 60 prediction layers total, each summarizing votes across the four folds:

```{r eval=FALSE}
scaled_dir <- system.file("extdata/rasters_scaled", package = "TemporalModelR")

time_steps <- expand.grid(
  year             = 1:15,
  season           = c("Spring", "Summer", "Autumn", "Winter"),
  stringsAsFactors = FALSE
)

preds <- generate_spatiotemporal_predictions(
  partition_result     = tmr_partition,
  model_result         = glm_out,
  pseudoabsence_result = tmr_absences,
  raster_dir           = scaled_dir,
  variable_patterns    = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR",
    "prseas"       = "prseas_YEAR_SEASON"
  ),
  time_cols            = c("year", "season"),
  time_steps           = time_steps,
  output_dir           = file.path(tempdir(), "GLM_Predictions"),
  overwrite            = TRUE,
  verbose              = FALSE
)
```

```{r echo=FALSE}
preds <- readRDS(system.file("extdata/precomputed/glm/preds.rds",
                             package = "TemporalModelR"))
preds$prediction_files <- list.files(
  system.file("extdata/precomputed/glm/pred_tifs", package = "TemporalModelR"),
  pattern = "\\.tif$", full.names = TRUE
)
```

We can now visualize the 60 prediction rasters as a year-by-season grid. Each cell shows the number of folds (0 to 4) that classified that pixel as suitable. 

`terra::plot()` limits each call to 16 layers, so we render the stack in four blocks of four years each.

```{r fig.width=10, fig.height=5}
pred_stack <- terra::rast(preds$prediction_files)

pred_names    <- basename(preds$prediction_files)
pred_seasons  <- sub(".*_(Spring|Summer|Autumn|Winter)\\.tif$", "\\1", pred_names)
pred_years    <- as.numeric(sub(".*_(\\d+)_(Spring|Summer|Autumn|Winter)\\.tif$",
                                "\\1", pred_names))
season_levels <- c("Spring", "Summer", "Autumn", "Winter")
stack_order   <- order(pred_years, match(pred_seasons, season_levels))

pred_stack <- pred_stack[[stack_order]]
ordered_years   <- pred_years[stack_order]
ordered_seasons <- pred_seasons[stack_order]
names(pred_stack) <- paste0("Y", ordered_years, "_", ordered_seasons)

block1 <- which(ordered_years %in%  1:4)
block2 <- which(ordered_years %in%  5:8)
block3 <- which(ordered_years %in%  9:12)
block4 <- which(ordered_years %in% 13:15)
```

```{r fig.width=10, fig.height=5}
terra::plot(pred_stack[[block1]], nr = 4, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0),  legend = FALSE)
```

```{r fig.width=10, fig.height=5}
terra::plot(pred_stack[[block2]], nr = 4, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0),  legend = FALSE)
```

```{r fig.width=10, fig.height=5}
terra::plot(pred_stack[[block3]], nr = 4, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0),  legend = FALSE)
```

```{r fig.width=10, fig.height=4}
terra::plot(pred_stack[[block4]], nr = 3, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0),  legend = FALSE)
```

Visualizing our predictions across each season and year, we see that there is generally high agreement among models. These rasters represent consensus votes among each of our four folds, with yellow pixels having strong positive consensus among all folds (all four identify the pixel as suitable) and blue pixels having low consensus among folds (few to no folds identify the given pixel as suitable). We also see that the models correctly visually show one of the main temporal trends in the data: loss of habitat through deforestation starting around year 6.

The GLM here does not pick up on the seasonal variability in precipitation patterns driving activity patterns for this hypothetical species, partially because the model itself is missing the relationship between precipitation and occupancy. This is in part due to the lack of low precipitation value background points in our dataset which would correctly identify the gap here and this methods reliance on that background data to make inference. We can explore if other modeling methods are able to fix this hurdle in other vignettes. Alternatively, we could use an alternative method for generating absence points which we describe more in the the [Preprocessing](https://cjhughes926.github.io/TemporalModelR/articles/V2_Preprocessing.html) vignette.

Note that here we show that `time_steps` can handle making predictions from compound variables or variables at different temporal scales so long as their temporal scales are nested. For example here "elevation" has no temporal value and is considered to be static across all time steps. "forest_cover" is measured annually, but is considered to be static across all seasons within a year for the purposes of predictions. "preseason" is measured both by year and season, so resulting seasonal predictions reflect that. However if precipitation was only measured based on aggregate seasons but had no associated year, predictions would fail. Predictions can also be made where all variables share the same time step- for example annual forest cover, annual temperature, and annual precipitation.

Additionally, a plain vector like `time_steps = 1:15` produces one prediction per value of the first time column; the other time columns are filled in with every unique value present in the occurrence data. So a vector input with `time_cols = c("year", "season")` would produce one prediction per year-season combination observed in the data. A data frame like the `expand.grid()` above gives explicit control: any combination of values can be requested, including ones not present in the occurrence data, as long as the matching rasters exist (Such as Summer or Winter predictions).

<br>

# G-space performance {#sec-gspace}

Once predictions are projected, per-time-step metrics become available. These are computed by overlaying the held-out test points (presences plus, optionally, pseudoabsences) on the prediction raster for each time step they fall in and counting correct vs incorrect classifications.

```{r}
head(preds$timestep_metrics)
```

Columns:

-   `Fold` - fold identifier; one row per fold per time step.
-   `Pct_Suitable` - proportion of the study area predicted suitable in that time step.
-   `N_Pres` - number of held-out presence test points falling in that time step for that fold.
-   `TP` - true positives, presences correctly classified as suitable in that time step.
-   `FN` - false negatives, presences incorrectly classified as unsuitable in that time step.
-   `Sensitivity` - `TP / (TP + FN)` for that fold-time step.
-   `CBP` - cumulative binomial probability. Under the null that each test point lands in suitable area at the rate `Pct_Suitable`, this is the probability of observing exactly `TP` correctly classified points by chance. Small values indicate predictions are better than random.
-   `N_Abs` - number of held-out pseudoabsence test points falling in that time step. `NA` for presence-only models.
-   `TN` - true negatives, pseudoabsences correctly classified as unsuitable.
-   `FP` - false positives, pseudoabsences incorrectly classified as suitable.
-   `Specificity` - `TN / (TN + FP)` for that fold-time step.
-   `TSS` - `Sensitivity + Specificity - 1` for that fold-time step.
-   `year` and `season` - the values of `time_cols` for that row, identifying which time step the metrics belong to.

G-space metrics are powerful in time steps with substantial sample sizes but lose meaning when few records exist (as with out example). A fold with only one presence in a given year can score sensitivity 0 or 1 with no real information content. Likewise, the ability to earn a significant G-space CBP score explicitly depends on sample size. Report E-space (from `$fold_test_metrics`) and G-space (from `$timestep_metrics`) together for a complete view of model performance.

The overall summary gives a per-fold summary across all time steps:

```{r}
preds$overall_summary
```

Rapid visual assessment of these metrics can be done using the function `plot_model_assessment()` which generates simple summary plots for each.

```{r}
plot_model_assessment(
  predictions         = preds,
  time_column         = c("year", "season"),
  secondary_time_mode = "combine",
  model_result        = glm_out
)
```

This produces per-fold time series of percent suitable, sensitivity, specificity, CBP, TP/FN, and TN/FP. When `model_result` is also supplied, overall sensitivity and specificity from `$fold_test_metrics` are added as reference lines. When compound time steps are used (such as year and season), you must choose how they are visualized. Choosing `secondary_time_mode = "combine"` will roll them into one continuous x axis. `secondary_time_mode = "facet"` produces stacked plots as seen below, where the first `time_col` is displayed as the x axis, and a different plot is made for each secondary variable (`season` here). By default, the threshold for which CBP is identified as significant is 0.05, but this may also be adjusted manually.

```{r fig.height=20}
plot_model_assessment(
  predictions         = preds,
  time_column         = c("year", "season"),
  secondary_time_mode = "facet",
  model_result        = glm_out,
  cbp_threshold       = 0.001
)
```

<br>

# Next steps {#sec-next}

Now that predictions have been generated, you can assess the model and see if it is preforming satisfactory enough for what your goals are. If this is the case, you can preform post-processing analyses to try to gain additional inference about the spatiotemporal patterns of change in the study region See the [Post-processing predictions](https://cjhughes926.github.io/TemporalModelR/articles/V4_Postprocessing.html) vignette.

For comparison with other algorithms applied to the same dataset:

-   [Modeling with a GAM](https://cjhughes926.github.io/TemporalModelR/articles/V3b_GAM.html) - smooth nonlinear responses.
-   [Modeling with a Random Forest](https://cjhughes926.github.io/TemporalModelR/articles/V3c_RF.html) - flexible ensemble of decision trees for complicated relationships to variables.
-   [Modeling with a Hypervolume](https://cjhughes926.github.io/TemporalModelR/articles/V3d_HV.html) - presence-only modeling using n-dimensional kernel density or one-class SVM.

---
title: "1. About the Example Dataset"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{About the Example Dataset}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 5,
  dpi       = 100,
  out.width = "95%"
)
```

<hr>

# Summary

-   [Description](#description)
-   [Overview](#sec-structure)
-   [Landscape rasters](#sec-predictors)
    -   [Elevation](#sec-tour-elev)
    -   [Forest cover and annual precipitation across years](#sec-tour-annual)
    -   [Seasonal precipitation within a year](#sec-tour-prseas-within)
-   [Occurrence data](#sec-points)
-   [Pre-computed objects and other bundled files](#sec-precomputed)

<hr>

# Description {#description}

To keep the package vignettes self-contained, TemporalModelR ships a small synthetic dataset that the entire workflow can run against in seconds, without requiring you to download external occurrence or environmental data. The dataset is deliberately small but complete, including everything a real temporally explicit SDM workflow would need. The small dataset is meant to represent a simple but changing landscape to visualize the utility of this package and the variety of the types of data that it may be useful for.

This vignette describes the dataset in detail so that the workflow vignettes ([Preprocessing temporally explicit data](https://cjhughes926.github.io/TemporalModelR/articles/V2_Preprocessing.html), [Modeling](https://cjhughes926.github.io/TemporalModelR/articles/V3a_GLM.html), [Post-processing](https://cjhughes926.github.io/TemporalModelR/articles/V4_Postprocessing.html)) can refer back to a single source for what's in `inst/extdata/` and `data()` rather than explaining the dataset through each other vignette. If you're working through the package for the first time, read this first.

<br>

# Overview {#sec-structure}

The included dataset is generated over the following spatial and temporal dimensions:

**Spatial.** A 15 × 30 cell grid at 100 m resolution, giving a 3000 m × 1500 m study area in a custom synthetic local CRS (a Transverse Mercator projection anchored at the equator and prime meridian).

**Temporal.** Fifteen years (labeled 1 through 15) and four seasons (Spring, Summer, Autumn, Winter).

The example landscape has three primary environmental variables driving suitability for our example species: Elevation, Forest Cover, and Precipitation. Elevation is representative of a temporally static variable which will not change over the 15 year study period. Forest cover is representative here of a temporally dynamic variable which changes across time and is measured at a single time step (annually). Precipitation is representative here of a temporally dynamic variable which is measured at compound time steps (here, measurements are made seasonally so that each precipitation measurement is associated with both a year and season). We also include a simplified 'annual precipitation' dataset for alternative simplified examples.

Our 'example species' can be found in mid-high elevations, in areas of high forest cover, and moderate to high precipitation.

Over the time period of the example dataset, we deliberately show an example of deforestation on the landscape in our forest cover dataset, as well as interannual variability and noise in our precipitation dataset. These allow for us to visualize areas of suitability loss over time in addition to the interannual dynamics of suitability over time. These signals are intentionally placed to highlight TemporalModelR's ability to show this spatiotemporal variability on the landscape.

<br>


# Landscape rasters {#sec-predictors}

The bundled raw rasters can be found in `inst/extdata/rasters_raw/` and contain:

-   `elevation.tif` - single static raster (one layer)
-   `forest_cover_<yr>.tif` - 15 annual rasters
-   `prseas_<yr>_<season>.tif` - 60 seasonal rasters (15 years × 4 seasons)
-   `pr_ann_<yr>.tif` - 15 annual rasters, computed as the sum of the four seasonal layers within each year

These can all be loaded from the system for any example analyses:

```{r}
library(TemporalModelR)
library(terra)
library(sf)

raw_dir <- system.file("extdata/rasters_raw",
                       package = "TemporalModelR")
```

Workflow vignettes typically use one of two predictor sets:

-   **Annual workflow:** `elevation`, `forest_cover` (annual), and `pr_ann` (annual precipitation) to illustrate the general utility of each function.
-   **Compound time-step workflow:** `elevation`, `forest_cover` (annual), and `prseas` (seasonal precipitation) to illustrate the function's ability to work with variables measured at more complex compound time steps (precipitation measures associated with specific seasons within each specific year)


## Elevation {#sec-tour-elev}

The elevation surface is fully static across the time series and is the only purely static predictor:

```{r, fig.height=3}
elev <- rast(file.path(raw_dir, "elevation.tif"))

plot(elev, main = "Elevation (m)")
```

## Forest cover and annual precipitation across years {#sec-tour-annual}

Forest cover and annual precipitation are the two dynamic annual predictors. Plotting them side by side with each row representing one year makes the temporal change in each visible at the same time. We visualize every other year below:

```{r, fig.height=14}
years_to_plot <- seq(1, 15, by = 2)

forest_files  <- file.path(raw_dir,
                           paste0("forest_cover_", years_to_plot, ".tif"))
pr_ann_files  <- file.path(raw_dir,
                           paste0("pr_ann_",      years_to_plot, ".tif"))

### Interleave forest and precip so each row of the plot grid is one year
forest_pr_paths        <- c(rbind(forest_files, pr_ann_files))
forest_pr_stack        <- rast(forest_pr_paths)
names(forest_pr_stack) <- c(rbind(paste("Forest_yr", years_to_plot),
                                  paste("Pr_ann_yr", years_to_plot)))

plot(forest_pr_stack, nc = 2)
```

The left column shows forest cover thinning in two locations: a gradual loss on the northeast hill starting around year 4 and a faster loss in a southwest-central patch starting around year 7. The right column shows annual precipitation with a slight overall decline plus the wet (year 3 and year 9) and dry (year 11) years that stand out from their neighbors.

## Seasonal precipitation within a year {#sec-tour-prseas-within}

Seasonal precipitation multiplies the annual base by season: Spring and Autumn are the wettest times of year, Summer is driest, and Winter is intermediate. Year 1 across all four seasons:

```{r, fig.height=6}
season_names <- c("Spring", "Summer", "Autumn", "Winter")

prseas_y1_stack <- rast(file.path(raw_dir,
                                  paste0("prseas_1_",
                                         season_names, ".tif")))

names(prseas_y1_stack) <- season_names

plot(prseas_y1_stack,
     range = c(0, max(values(prseas_y1_stack), na.rm = TRUE)))
```

The spatial structure is preserved across seasons; the seasons differ in overall magnitude.

# Occurrence data {#sec-points}

We also generated an example dataset of 150 'species occurrence locations' across the 15 year / 4 season time frame. The example points represent a high-elevation forest specialist with moderate to high moisture requirements.

First, points are generated for every location/year/season combination above a simple threshold for each variable of interest, with only combinations meeting *all four* environmental filters counting as a candidate occurrence site:

-   Elevation > 1200 m
-   Forest cover > 0.75
-   Annual precipitation > 300 mm
-   Seasonal precipitation > 150 mm (same threshold for Spring, Summer, and Autumn)

Winter is excluded from sampling entirely, so the filter is applied only across the three remaining seasons (Spring, Summer, Autumn) × 15 years = 45 candidate year-season slices.

We apply spatial and temporal autocorrelation to a random sampling algorithm to subset our candidate points across time into only 150 samples, resulting in a clustered, ecologically plausible occurrence dataset distributed across space, year, and season, with realistic survey biases.

The final example points database can be called from the system:

```{r}
pts_file <- system.file("extdata/points/synthetic_occurrence_points.csv",
                        package = "TemporalModelR")
pts <- utils::read.csv(pts_file)

head(pts)


nrow(pts)


table(pts$year, pts$season)
```

To see the distribution of points across both space and time, plot each year-season combination on its own panel. Each row of the grid corresponds to one of the 15 years; each column corresponds to one of the three sampled seasons (Spring, Summer, Autumn). Empty panels indicate year-season combinations with no points:

```{r, fig.height=20}
seasons <- c("Spring", "Summer", "Autumn")
study_extent <- ext(0, 3000, 0, 1500)

opar <- par(no.readonly = TRUE)

par(mfrow = c(15, 3),
    mar   = c(1.5, 1.5, 1.5, 0.5),
    oma   = c(2, 2, 2, 1))

for (yr in 1:15) {
  for (sea in seasons) {
    sub <- pts[pts$year == yr & pts$season == sea, ]

    plot(NULL,
         xlim = c(0, 3000), ylim = c(0, 1500),
         asp  = 1, xaxt = "n", yaxt = "n",
         xlab = "", ylab = "",
         main = paste0("Year ", yr, " - ", sea),
         cex.main = 0.9)

    rect(0, 0, 3000, 1500, border = "grey70")

    if (nrow(sub) > 0) {
      points(sub$x, sub$y, pch = 19, cex = 0.7, col = "darkblue")
    }
  }
}

par(opar)
```

Together, this points dataset and the rasters above make up the landscape and species occurrence data for all of the example applications presented in this package's vignettes.

<br>

# Pre-computed objects and other bundled files {#sec-precomputed}

Alongside the raw inputs, the package ships pre-computed outputs of the full preprocessing and modeling pipelines as `data()` objects to be called into vignettes. Two sets exist, one for the annual workflow and one for the seasonal workflow. The workflow to generate these is shown in the package vignettes, but stable saved copies are included in the package data so users can jump straight to any phase of the workflow without re-running upstream steps.

## Pre-computed `data()` objects

-   **`tmr_partition_annual`** - output of `spatiotemporal_partition()`. A list containing `$folds` (a data frame mapping each occurrence point to one of four cross-validation folds), `$points_sf` (the rarefied and extracted points as an `sf` object, with environmental values attached), `$voronoi_folds` (the spatial Voronoi blocks used to assign folds, also an `sf` object), `$summary` (per-fold point counts), and `$plots` (diagnostic ggplot objects). Built with 2 spatial folds × 2 temporal folds.
-   **`tmr_absences_annual`** - output of `generate_absences()` applied to `tmr_partition_annual`. A list with `$pseudoabsences` (an `sf` object containing 2:1 ratio buffer-sampled pseudoabsence points with environmental values extracted at the matching year), `$plots`, and `$summary`. Use it directly as the `pseudoabsence_result` argument in any of the four presence/absence model builders.
-   **`tmr_glm_annual`** - output of `build_temporal_glm()` applied to `tmr_partition_annual` and `tmr_absences_annual` with formula `~ forest_cover + pr_ann + elevation`, logit link, and TSS threshold selection. A list of class `"TemporalGLM"` containing `$models` (four fitted `glm` objects, one per fold), `$thresholds` (the TSS-optimal threshold per fold), `$model_formula`, `$link`, `$model_vars`, `$fold_training_data`, `$fold_test_metrics` (per-fold AUC, TSS, sensitivity, specificity), and `$plots`. Pass it to `generate_spatiotemporal_predictions()` as the `model_result` argument.
-   **`tmr_predictions_annual`** - output of `generate_spatiotemporal_predictions()` applied to `tmr_glm_annual`, projected across all 15 years (one annual prediction stack per fold). A list with `$timestep_metrics` (per-year, per-fold E-space and G-space evaluation metrics including CBP), `$overall_summary` (across-year aggregates), `$prediction_files` (paths to the per-fold prediction tifs from the build run), and `$model_type`. Useful for `plot_model_assessment()` and for downstream pattern analysis.
-   **`tmr_partition`** - partition built from rarefaction at year-season scale and extraction with `prseas_YEAR_SEASON`. Same list structure as the annual version, but with more points retained because spatiotemporal rarefaction at the seasonal scale preserves multiple observations from the same pixel in different seasons.
-   **`tmr_absences`** - pseudoabsences for `tmr_partition`, generated at the year-season scale so each pseudoabsence is associated with a specific year *and* season and has the corresponding seasonal predictor values attached.
-   **`tmr_glm`** - `build_temporal_glm()` fit with formula `~ forest_cover + prseas + elevation` and `time_cols = c("year", "season")`.
-   **`tmr_predictions`** - predictions from `tmr_glm` projected to all 15 years for the Spring season only (15 prediction layers per fold). The Spring-only projection is what `inst/extdata/predictions/` contains in raster form (see below).

## Intermediate raster and point files

Additionally, `inst/extdata/` contains raster and point files corresponding to intermediate steps throughout various vignettes. These are bundled so that users may call them directly and avoid re-running previous analyses just to produce them. Each subdirectory can be loaded from the system with `system.file()`:

```r
pred_dir <- system.file("extdata/predictions",
                        package = "TemporalModelR")
list.files(pred_dir, pattern = "\\.tif$")
```

The bundled subdirectories are:

-   **`inst/extdata/rasters_aligned/`** - outputs of `raster_align()` on the raw rasters: every layer reprojected and masked to the reference grid.
-   **`inst/extdata/rasters_scaled/`** - z-scored rasters for the seasonal workflow (`forest_cover`, `prseas`, `elevation`), produced by `scale_rasters()`.
-   **`inst/extdata/rasters_scaled_annual/`** - z-scored rasters for the annual workflow (`forest_cover`, `pr_ann`, `elevation`).
-   **`inst/extdata/predictions/`** - 15 per-year fold-vote prediction rasters from the seasonal workflow's `generate_spatiotemporal_predictions()` call. Direct input to `summarize_raster_outputs()`.
-   **`inst/extdata/binary/`** - outputs of `summarize_raster_outputs()` applied to the prediction rasters above:
    -   `consensus_stack.tif` - 15-layer binary consensus stack (one layer per year, suitable where ≥3 of 4 folds agree)
    -   `frequency_raster.tif` - single-layer raster giving the proportion of years each pixel was classified as suitable
-   **`inst/extdata/points/`** - the raw `synthetic_occurrence_points.csv` (and a matching shapefile), plus the intermediate point files from rarefaction, extraction, and scaling for both workflows:
    -   `Pts_annual_*` - rarefied points at the annual scale
    -   `Pts_seasonal_*` - rarefied points at the year-season scale
    -   `extracted_annual_*` - extraction outputs at the annual scale (raw values, scaled values, and scaling parameters)
    -   `extracted_seasonal_*` - extraction outputs at the year-season scale

---
title: "4. Post-processing predictions"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Post-processing predictions}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 5,
  dpi       = 100,
  out.width = "95%"
)
```

<hr>

# Summary

-   [Overview](#sec-overview)
-   [Consensus and frequency summarization](#sec-consensus)
-   [Detecting temporal patterns](#sec-patterns)
-   [Aggregating by spatial unit](#sec-zones)
-   [Summary](#sec-next)

<hr>

# Overview {#sec-overview}

The modeling phase described in the previous four vignettes each produces per-fold consensus prediction rasters across time steps. These are useful, but will often result in stacks of individual predictions (10s to sometimes 100s of rasters) that may be difficult to make straightforward sense of. The goal of the post-processing phase is to take this abundance of data and summarize it in interpretable ways, mainly focused on understanding the overarching temporal dynamics of each pixel, and summarizing these change patterns regionally for intuitive analyses.

Three functions cover this phase, run in order:

1.  `summarize_raster_outputs()` - summarizes fold-vote rasters output from `generate_spatiotemporal_predictions()` into a binary consensus stack (one binary raster per time step) and a frequency raster (the percent of time steps where the binary results indicate suitability). This is important because all subsequent processes require binary rasters as their inputs.
2.  `analyze_temporal_patterns()` - classifies each pixel's trajectory through time via changepoint detection, resulting in a single raster classifying each pixel as either 'stable suitable', 'stable unsuitable', 'increasing in suitability', 'decreasing in suitability', 'fluctuating', or 'no pattern'. It also produces a 'time of first change' raster indicating when those increases or decreases occurred.
3.  `analyze_trends_by_spatial_unit()` - aggregates pixel-level patterns across user-defined polygons. This takes potentially complicated rasters with many pixels and gives regional summaries of the spatiotemporal dynamics of each spatial unit.

This vignette uses the bundled `inst/extdata/predictions/` raster set, which contains 15 per-year fold-vote rasters from the seasonal workflow's `generate_spatiotemporal_predictions()` call, projecting `tmr_glm` across years 1-15 for the Spring season only. These are the same files referenced by `tmr_predictions$prediction_files` and are produced and inspected in the [Modeling with a GLM](https://cjhughes926.github.io/TemporalModelR/articles/V3a_GLM.html) vignette. The underlying dataset is described in [About the Example Dataset](https://cjhughes926.github.io/TemporalModelR/articles/V1_dataset.html).

```{r}
library(TemporalModelR)
library(terra)
library(sf)

pred_dir <- system.file("extdata/predictions", package = "TemporalModelR")

list.files(pred_dir, pattern = "\\.tif$")[1:5]
```

Each pixel of each raster indicates the number of folds which our model predicts that pixel as suitable in that given year. Because this was a four fold model, possible per-pixel values range from 0 (no folds predict this pixel as suitable) to 4 (all folds predict this pixel as suitable).

<br>

# Consensus and frequency summarization {#sec-consensus}

`summarize_raster_outputs()` applies a consensus threshold to convert the fold-count rasters into binary suitability rasters, which are needed for all post-processing analyses. These simplify the fold count rasters by setting a threshold of how many folds must agree for the pixel to be considered suitable in the post-processing analyses. A very conservative modeling process may want all folds to agree for a pixel to be considered suitable (in that case, we would indicate a threshold of 4). A more relaxed modeling process may only require half of the folds to agree for a pixel to be considered suitable (threshold here of 2). The choice depends on the cost of false positives versus false negatives in the given application. Higher consensus is more conservative; lower consensus is more sensitive. Lower consensus thresholds may also introduce additional noise into the post-processing results, which should be considered for this process.

In either case, all pixels which meet or exceed the user set threshold become a binary '1' for suitable, and all others become a binary '0' for unsuitable.

A 'frequency file' is also produced based on the binarized rasters. This is a single raster which indicates the proportion of time steps for which the binary rasters agree a pixel is suitable, and is a good indication of the temporal stability of a pixel's predictions.

```{r}
summary_out <- summarize_raster_outputs(
  predictions_dir = pred_dir,
  output_dir      = file.path(tempdir(), "Binary"),
  consensus       = 2,
  overwrite       = TRUE,
  verbose         = FALSE
)

names(summary_out)
```

The returned object contains the following items:

-   `$consensus_stack` - multi-layer `SpatRaster` with one layer per input time step. For each layer, pixels are 1 where the threshold of folds in agreement set by `consensus` is met and 0 elsewhere.
-   `$frequency_raster` - single-layer `SpatRaster` giving the proportion of time steps each pixel was suitable under the chosen consensus threshold. Values range from 0 (never suitable) to 1 (always suitable).
-   `$consensus` - the consensus threshold value used.
-   `$n_timesteps` - number of time steps processed (equal to the number of input rasters).
-   `$consensus_dir` - path to the directory containing the per-time-step binary consensus rasters that were written to disk.
-   `$frequency_file` - path to the written frequency raster file.

We can visualize the per-year consensus stack to inspect the year-to-year breakdown of projected suitability.

```{r fig.width=10, fig.height=12}
binary_stack   <- summary_out$consensus_stack
frequency_rast <- summary_out$frequency_raster

names(binary_stack) <- paste0("Y", 1:15, "_Spring")

terra::plot(binary_stack, nr = 5, nc = 3,
            mar = c(1.5, 0.5, 1.5, 0.5), legend = FALSE)
```

Suitable pixels (1) appear in yellow and unsuitable pixels (0) in purple. The reduction in suitable area starting around year 6 reflects the deforestation trend present in the underlying forest cover rasters, just as would be expected based on the changing landscape.

The frequency raster is a single layer summarizing all 15 panels into one:

```{r fig.width=8, fig.height=4}
terra::plot(frequency_rast,
            main = "Proportion of years pixel was suitable",
            mar  = c(2.5, 2.5, 2.5, 5.0))
```

Pixels of consistently suitable habitat sit close to 1, indicating they were always predicted as suitable. Pixels that gained or lost suitability across the time series sit between 0 and 1, indicating suitability was inconsistent across time. Surrounding areas sit at 0, indicating they were never suitable. The pixels with values between 0 and 1 are the ones that will carry information for the changepoint detection in the next section.

<br>

# Detecting temporal patterns {#sec-patterns}

`analyze_temporal_patterns()` applies changepoint detection to each pixel's binary suitability time series, classifying it into one of six categories:

1.  **Never suitable** - pixel is unsuitable in every time step.
2.  **Always suitable** - pixel is suitable in every time step.
3.  **No pattern** - pixel varies through time but no significant structured change is detected.
4.  **Increasing suitability** - pixel transitions from unsuitable to suitable at a detectable change point.
5.  **Decreasing suitability** - pixel transitions from suitable to unsuitable at a detectable change point.
6.  **Fluctuating** - pixel shows multiple direction changes through time.

Changepoint detection runs via `fastcpd::fastcpd_binomial()` with each pixel's previous-time-step value included as a predictor to account for temporal autocorrelation. When `spatial_autocorrelation = TRUE`, the proportion of a pixel's first-order neighbors classified as suitable in each time step is also considered as an additional covariate to the changepoint detection model, accounting for spatial autocorrelation in the predictive surface.

The `fastcpd` package is a hard dependency of `analyze_temporal_patterns()` and must be installed before running:

```{r, eval=FALSE}
install.packages("fastcpd")
```

Additionally, unlike other functions in TemporalModelR which may use cyclical or temporally nonlinear time steps (like multiple seasons within a year where predictions are expected to fluctuate), `analyze_temporal_patterns()` expects a linear time series to run the analysis on. In our example, we could have rerun all our examples at an annual scale rather than a seasonal scale and passed that result to this analysis. Alternatively, we may run this analysis on a subset of the data we produced during the previous steps. For example, here we explore the spatiotemporal trends in Springtime suitability in our study region. We define this subset using `time_steps` and the function appropriately selects the matching subset of our predictions from `summarize_raster_outputs()`.

```{r}
time_steps <- expand.grid(
  year             = 1:15,
  season           = "Spring",
  stringsAsFactors = FALSE
)

patterns <- analyze_temporal_patterns(
  binary_stack            = binary_stack,
  summary_raster          = frequency_rast,
  time_steps              = time_steps,
  output_dir              = file.path(tempdir(), "Patterns"),
  spatial_autocorrelation = TRUE,
  alpha                   = 0.05,
  estimate_time           = FALSE,
  overwrite               = TRUE,
  verbose                 = FALSE
)

names(patterns)
```

The returned object contains the following items:

-   `$pattern` - integer `SpatRaster` with values 1 through 6 corresponding to the six categories above.
-   `$time_decrease` - `SpatRaster` showing the time step at which the first significant decrease was detected for pixels classified as decreasing. `NA` elsewhere.
-   `$time_increase` - `SpatRaster` showing the time step at which the first significant increase was detected for pixels classified as increasing. `NA` elsewhere.

This function is computationally complex and may take many hours or days to run on large datasets, and can also be very memory intensive. The following optional arguments help amend memory issues and make estimates about runtime:

-   `n_tiles_x` and `n_tiles_y` - for large rasters, set these greater than 1 to process in spatial tiles and reduce peak memory use. Default is `1` (no tiling). Each tile is processed independently and the results are mosaicked at the end.
-   `estimate_time` - when `TRUE` (default), the function samples a few pixels first and reports an estimated total runtime before committing to the full job. Useful for very long time series or very large or fine-scale rasters. We disable it above since the bundled example is small.

A few additional considerations relevant to this function:

-   The significance level for the changepoint detection defaults to `0.05`. Reducing the threshold of significance can be done by changing the `alpha` parameter. Lower alpha produces fewer false-positive changepoints but may also miss real ones.
-   Time series shorter than ~15 time steps may be too short to detect increases or decreases reliably. The function will return mostly "no pattern" or "never/always suitable" classifications.
-   The classification assumes consecutive rasters representing the same temporal scale (e.g. consecutive years). Mixing scales or skipping time steps will distort the results.
-   The 15-year Spring-only example used here is right at the lower end of what changepoint detection can use. With more years of data the classifications become more reliable.

In the above example, the decreasing pixels show change points clustered in the years where forest cover is reduced, which matches what would be expected based on the simulated data. The time of increase raster is mostly empty because the example data does not include any large-scale habitat gain events.

<br>

# Aggregating by spatial unit {#sec-zones}

`analyze_trends_by_spatial_unit()` overlays user-defined polygons (administrative units, watersheds, ecoregions, etc.) on the prediction and pattern rasters and produces summary tables of habitat composition and trajectory composition per polygon.

For the example dataset, we construct two simple zones: a western half and an eastern half of the study area, and aggregate the patterns within each.

```{r}
study_crs <- sf::st_crs(binary_stack)

zones_sf <- rbind(
  sf::st_sf(ZONE = "West",
            geometry = sf::st_sfc(sf::st_polygon(list(
              matrix(c(0, 0, 1500, 1500, 0,
                       0, 1500, 1500, 0, 0), ncol = 2)
            )), crs = study_crs)),
  sf::st_sf(ZONE = "East",
            geometry = sf::st_sfc(sf::st_polygon(list(
              matrix(c(1500, 1500, 3000, 3000, 1500,
                       0,    1500, 1500, 0,    0),    ncol = 2)
            )), crs = study_crs))
)
```

This is simple for the sake of the example, but the function can handle any number of complex polygons to represent the preferred boundaries of a region.

The function takes the zone polygons plus any subset of the results from `summarize_raster_outputs()` or `analyze_temporal_patterns()`. It creates summary tables and optionally produces figures based on the input information. Different combinations of inputs produce different subsets of the output tables. Our example uses all available inputs from the previously run functions.

```{r}
zone_summary <- analyze_trends_by_spatial_unit(
  shapefile_path       = zones_sf,
  name_field           = "ZONE",
  binary_stack         = binary_stack,
  pattern_raster       = patterns$pattern,
  time_decrease_raster = patterns$time_decrease,
  time_increase_raster = patterns$time_increase,
  time_steps           = time_steps,
  output_dir           = file.path(tempdir(), "ZoneSummary"),
  create_plot          = FALSE,
  verbose              = FALSE
)

names(zone_summary)
```

The returned object contains the following items, with the set of present tables depending on which raster inputs were supplied:

-   `$overall_summary` - pattern composition per spatial unit, with one row per zone and columns for the count of pixels in each of the six pattern categories. Present when `pattern_raster` is supplied.

```{r}
zone_summary$overall_summary
```

-   `$timestep_summary` - suitable pixel count per unit per time step, useful for plotting habitat area through time for each zone. Present when `binary_stack` is supplied.

```{r}
head(zone_summary$timestep_summary)
```

-   `$change_by_timestep` - counts of gain (first increase) and loss (first decrease) events per unit per time step. Present when both change rasters are supplied.

```{r}
head(zone_summary$change_by_timestep)
```

-   `$plots` - recorded plot objects produced when `create_plot = TRUE`.

More robust plots can be manually made from the output tables, but using `create_plot = TRUE` gives quick visual assessments of the summarized spatiotemporal dynamics of the landscape.

```{r}
zone_plots <- analyze_trends_by_spatial_unit(
  shapefile_path       = zones_sf,
  name_field           = "ZONE",
  binary_stack         = binary_stack,
  pattern_raster       = patterns$pattern,
  time_decrease_raster = patterns$time_decrease,
  time_increase_raster = patterns$time_increase,
  time_steps           = time_steps,
  output_dir           = file.path(tempdir(), "ZoneSummary"),
  create_plot          = TRUE,
  verbose              = FALSE
)
```

The resulting per-timestep and per-spatial-unit outputs of this function allow for a robust overarching look at the study region broken down by units we care about. An example application may be looking at habitat availability for a given species by county, understanding regionally where the most suitable habitat is, where the highest losses are occurring, and when those losses are occurring in each spatial unit. Alternatively, an agency may be interested in exploring the effect of habitat restoration projects across a number of parks. That agency could use this tool to identify which parks were seeing the largest gains of habitat in the time frame of interest, when those gains were occurring, and make further management decisions based on that.

Two zones with the same total area predicted suitable in year 15 may have very different histories: one may have been stably suitable throughout, and the other may have lost half its habitat and gained an equivalent area elsewhere. The composition table separates those cases, and the change by timestep table identifies when those changes happened.

<br>

# Summary {#sec-next}

The three post-processing functions together convert the per-fold per-time-step prediction stack from the modeling phase into three interpretable pieces of information:

-   A binary consensus surface per time step.
-   A pattern classification per pixel describing its trajectory.
-   A summary of pattern composition and habitat change per administrative or ecological region.

These outputs show not only where a species is likely to be, but how that is changing and where those changes are occurring, and have many practical applications.

The above workflow can be applied to any of the previously described model outputs: [GLM](https://cjhughes926.github.io/TemporalModelR/articles/V3a_GLM.html), [GAM](https://cjhughes926.github.io/TemporalModelR/articles/V3b_GAM.html), [Random Forest](https://cjhughes926.github.io/TemporalModelR/articles/V3c_RF.html) or [Hypervolume](https://cjhughes926.github.io/TemporalModelR/articles/V3d_HV.html).

---
title: "3d. Modeling with a Hypervolume"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Modeling with a Hypervolume}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 5,
  dpi       = 100,
  out.width = "95%"
)
```

<hr>

# Summary

-   [Theory](#sec-theory)
-   [Overview](#sec-overview)
-   [Fitting a temporal hypervolume](#sec-fit)
-   [Choosing a method](#sec-method)
-   [Tuning hypervolume parameters](#sec-params)
-   [E-space performance](#sec-espace)
-   [Volume and overlap diagnostics](#sec-volume)
-   [Projecting predictions](#sec-project)
-   [G-space performance](#sec-gspace)
-   [Next steps](#sec-next)

<hr>

# Theory {#sec-theory}

TemporalModelR operates across two complementary dimensions to support temporally explicit modeling:

-   `E-space` - (Environmental Space) represents a location in terms of its environmental conditions, independent of geography. A species' niche in E-space is its tolerance for a given set of environmental variables.
-   `G-space` - (Geographic Space) represents a location in terms of where it physically sits on the landscape. Every point in G-space has a corresponding location in E-space.

Because environmental conditions change over time, the same point in G-space may move through E-space across years, decades, or seasons. Traditional, temporally-static workflows assume both G-space and E-space to be stable in time. TemporalModelR assumes only that the niche (i.e., the species' tolerance in E-space) is stable across the study period, while the E-space coordinates of any G-space location may change over time. Species observations are therefore matched to E-space data correct in both space and time, the niche is modeled in time-independent E-space. This is done in the [Preprocessing temporally explicit data](https://cjhughes926.github.io/TemporalModelR/articles/V2_Preprocessing.html) vignette.

By modeling the niche in time-independent E-space using time-independent training data, we can generate a static prediction of a species' environmental tolerances. We can then project these static E-space predictions back onto G-space at explicit time periods, resulting in dynamic, temporally-explicit ENM predictions of species distributions.

<br>

# Overview {#sec-overview}

`build_temporal_hv()` constructs one n-dimensional hypervolume per cross-validation fold as was created during the [Preprocessing temporally explicit data](https://cjhughes926.github.io/TemporalModelR/articles/V2_Preprocessing.html) vignette. Each fold's hypervolume is built from all data outside the fold and evaluated on the held-out fold. A hypervolume is a geometric envelope drawn around the presence points in E-space, and a new point falls inside the envelope (predicted suitable) or outside it (predicted unsuitable). Projecting the hypervolume onto a stack of environmental rasters produces a binary suitability surface for each fold and time step, with no need for a separate thresholding step.

The hypervolume is the only presence-only model in TemporalModelR. It does not need pseudoabsences or background points, which makes it a good choice when reliable absence data is hard to define or when the species' detection probability is highly variable in space. The trade-off is that without negative training data, the model has no way to learn that an environmental gradient separates suitable from unsuitable conditions, it can only describe where presences sit in E-space and project that shape onto the landscape. Therefore the typical E-space metrics around specificity, AUC, and TSS are not available; only sensitivity-based metrics are reported. See [E-space performance](#sec-espace).

This vignette runs the seasonal workflow using the bundled `tmr_partition` object, which is the pre-built output of the partitioning step produced by the [Preprocessing temporally explicit data](https://cjhughes926.github.io/TemporalModelR/articles/V2_Preprocessing.html) vignette using the same call patterns shown there. The dataset itself is described in [About the Example Dataset](https://cjhughes926.github.io/TemporalModelR/articles/V1_dataset.html). Note that no `tmr_absences` object is loaded; the hypervolume builder does not accept a `pseudoabsence_result` argument. Otherwise, this is the same dataset as is used in each modeling vignette [GLM](https://cjhughes926.github.io/TemporalModelR/articles/V3a_GLM.html), [GAM](https://cjhughes926.github.io/TemporalModelR/articles/V3b_GAM.html), [Random Forest](https://cjhughes926.github.io/TemporalModelR/articles/V3c_RF.html) or [Hypervolume](https://cjhughes926.github.io/TemporalModelR/articles/V3d_HV.html).

The `hypervolume` package is a hard dependency of `build_temporal_hv()` and must be installed before running:

```{r, eval=FALSE}
install.packages("hypervolume")
```

```{r}
library(TemporalModelR)
library(terra)

data(tmr_partition, package = "TemporalModelR")
```

<br>

# Fitting a temporal hypervolume {#sec-fit}

The minimal call needs the partition, a character vector of predictor variable names, and a method. We fit a Gaussian kernel density hypervolume across the three predictors. Unlike other modeling functions presented in this package, hypervolume builder does not require a threshold or pseudoabsences, so neither argument appears in the call.

`create_plot = TRUE` allows the function to generate diagnostic pairplots and a combined comparison plot across folds.

```{r}
hv_out <- build_temporal_hv(
  partition_result   = tmr_partition,
  model_vars         = c("elevation", "forest_cover", "prseas"),
  method             = "gaussian",
  hypervolume_params = list(),
  output_dir         = file.path(tempdir(), "Hypervolume_Models"),
  create_plot        = TRUE,
  verbose            = FALSE
)
```

The diagnostic pairplots show the location of presence points and the envelope boundary in each pair of predictor dimensions. These are useful for spotting envelopes that have stretched along one axis due to a few outliers, or that have failed to capture an obvious cluster of presences. Here we see high overlap among all hypervolume folds. 

```{r}
class(hv_out)

names(hv_out)
```

The returned object is a list of class `"TemporalHypervolume"` containing the following objects:

-   `$models` - named list of `Hypervolume` objects, one per fold (`fold1`, `fold2`, ...). Each is a standard `hypervolume::Hypervolume` object and accepts the usual `hypervolume::get_volume()`, `hypervolume::hypervolume_inclusion_test()`, and `hypervolume::hypervolume_project()` calls.
-   `$volumes` - named numeric vector of hypervolume sizes (in units of the standardized predictor space), one per fold. See [Volume and overlap diagnostics](#sec-volume).
-   `$overlaps` - named list of pairwise percent volume overlap statistics between fold hypervolumes.
-   `$method` - the method used (`"gaussian"` or `"svm"`).
-   `$model_vars` - character vector of the predictor variable names supplied to the function. Used by `generate_spatiotemporal_predictions()` to know which raster layers to assemble at projection time.
-   `$fold_training_data` - list of the training data frames used for each fold.
-   `$fold_test_metrics` - data frame of per-fold E-space metrics computed on the held-out test set. See [E-space performance](#sec-espace).
-   `$output_dir` - path to the directory where the saved RDS of fitted hypervolumes and the fold test metrics CSV were written.
-   `$model_type` - `"hypervolume"`, used by `generate_spatiotemporal_predictions()` to choose the correct projection logic.
-   `$plots` - list of recorded base-R plots produced when `create_plot = TRUE` (pairplots and the combined comparison plot).

Unlike GLM/GAM/RF, no `thresholds`, no `threshold_method`, no `model_formula` or `link`. The hypervolume produces a binary inside/outside classification directly, with no probability scale or threshold to choose.

<br>

# Choosing a method {#sec-method}

Two methods are supported, set via the `method` argument:

-   `"gaussian"` - fits a Gaussian kernel density estimate around the presence points in E-space and thresholds it at a requested quantile (default 95%) to define the envelope. The resulting shape is smooth and roughly convex in each dimension. Best when presences are reasonably dense and you want a soft, density-based description of niche.
-   `"svm"` - fits a one-class support vector machine. The SVM finds a boundary that separates the presence data from the rest of E-space using the kernel trick, so the envelope can be highly non-convex and follow the data's geometry tightly. Best when presences are sparse, multi-modal, or arranged in non-elliptical patterns in E-space.

<br>

# Tuning hypervolume parameters {#sec-params}

`hypervolume_params` is a named list passed to `hypervolume::hypervolume_gaussian()` or `hypervolume::hypervolume_svm()` depending on `method`. Anything that would normally be passed to either function can go in `hypervolume_params`. See `?hypervolume::hypervolume_gaussian()` or `hypervolume::hypervolume_svm()` for the full list.


<br>

# E-space performance {#sec-espace}

Each fold's held-out test set provides a set of presence points the model has never seen based on the folds defined in the [Preprocessing temporally explicit data](https://cjhughes926.github.io/TemporalModelR/articles/V2_Preprocessing.html) vignette. We check whether each test point falls inside the fold's hypervolume envelope and count correct vs incorrect classifications. Because the hypervolume is presence-only, only sensitivity-based metrics can be computed, not specificity, AUC, or TSS.

```{r}
hv_out$fold_test_metrics
```

Columns:

-   `Fold` - fold identifier matching the folds in `tmr_partition$points_sf$fold`.
-   `N_Test` - number of held-out presence test points for that fold.
-   `E_Volume` - size of the fitted hypervolume in standardized E-space units. See [Volume and overlap diagnostics](#sec-volume).
-   `Testing_TP` - count of test-set presence points correctly classified as inside the envelope (true positives).
-   `Testing_FN` - count of test-set presence points incorrectly classified as outside the envelope (false negatives).
-   `Sensitivity` - proportion of test-set presences correctly classified, `Testing_TP / (Testing_TP + Testing_FN)`. Values close to 1 mean the model rarely misses presences.
-    `volumes` - reports the size of each fold's hypervolume in standardized E-space units. 
-    `overlaps` - reports pairwise percent volume overlap between fold hypervolumes. High pairwise overlap means the niche estimate is stable across spatially or temporally distinct training subsets.


Sensitivity alone is a one-sided picture of performance: a hypervolume that covers the entire study area would have sensitivity 1 but would be ecologically meaningless. Its usually best to examine sensitivity as it relates to the proportion of the study are being predicted as suitable or the cumulative binomial probability of that sensitivity given the cumulative area predicted as suitable over time. See [G-space performance](#sec-gspace). 

E-space metrics are robust to imbalanced sample sizes across time, because they pool across the time series. They are a good metric for assessing overall model fit. Time-specific (G-space) metrics can also be assessed later when we project the model to specific G-space and time combinations. In the case of hypervolume modeling, G-space metrics must be used to get a complete picture of the model's performance, in absence of specificity metrics.

<br>

# Projecting predictions {#sec-project}

`generate_spatiotemporal_predictions()` projects each fold's hypervolume onto a stack of environmental rasters matching each requested time step, producing one fold-vote raster per time step. The same call works for all four model types; `model_type` in the model object tells the function which projection logic to use. Note that `pseudoabsence_result` is omitted from the call below because the hypervolume workflow does not use pseudoabsences.

For this example we project across all 15 years and 4 seasons, producing 60 prediction layers total, each summarizing votes across the four folds:

```{r eval=FALSE}
scaled_dir <- system.file("extdata/rasters_scaled", package = "TemporalModelR")

time_steps <- expand.grid(
  year             = 1:15,
  season           = c("Spring", "Summer", "Autumn", "Winter"),
  stringsAsFactors = FALSE
)

preds <- generate_spatiotemporal_predictions(
  partition_result  = tmr_partition,
  model_result      = hv_out,
  raster_dir        = scaled_dir,
  variable_patterns = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR",
    "prseas"       = "prseas_YEAR_SEASON"
  ),
  time_cols         = c("year", "season"),
  time_steps        = time_steps,
  output_dir        = file.path(tempdir(), "HV_Predictions"),
  overwrite         = TRUE,
  verbose           = FALSE
)
```

```{r echo=FALSE}
preds <- readRDS(system.file("extdata/precomputed/hv/preds.rds",
                             package = "TemporalModelR"))
preds$prediction_files <- list.files(
  system.file("extdata/precomputed/hv/pred_tifs", package = "TemporalModelR"),
  pattern = "\\.tif$", full.names = TRUE
)
```

We can now visualize the 60 prediction rasters as a year-by-season grid. Each cell shows the number of folds (0 to 4) that classified that pixel as suitable.

`terra::plot()` limits each call to 16 layers, so we render the stack in four blocks of four years each.

```{r fig.width=10, fig.height=5}
pred_stack <- terra::rast(preds$prediction_files)

pred_names    <- basename(preds$prediction_files)
pred_seasons  <- sub(".*_(Spring|Summer|Autumn|Winter)\\.tif$", "\\1", pred_names)
pred_years    <- as.numeric(sub(".*_(\\d+)_(Spring|Summer|Autumn|Winter)\\.tif$",
                                "\\1", pred_names))
season_levels <- c("Spring", "Summer", "Autumn", "Winter")
stack_order   <- order(pred_years, match(pred_seasons, season_levels))

pred_stack <- pred_stack[[stack_order]]
ordered_years   <- pred_years[stack_order]
ordered_seasons <- pred_seasons[stack_order]
names(pred_stack) <- paste0("Y", ordered_years, "_", ordered_seasons)

block1 <- which(ordered_years %in%  1:4)
block2 <- which(ordered_years %in%  5:8)
block3 <- which(ordered_years %in%  9:12)
block4 <- which(ordered_years %in% 13:15)
```

```{r fig.width=10, fig.height=5}
vote_cols <- c("#440154", "#3b528b", "#21918c", "#5ec962", "#fde725")

terra::plot(pred_stack[[block1]], nr = 4, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0), legend = FALSE,
            range = c(0, 4), type = "classes",
            levels = 0:4, col = vote_cols)
```

```{r fig.width=10, fig.height=5}
terra::plot(pred_stack[[block2]], nr = 4, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0), legend = FALSE,
            range = c(0, 4), type = "classes",
            levels = 0:4, col = vote_cols)
```

```{r fig.width=10, fig.height=5}
terra::plot(pred_stack[[block3]], nr = 4, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0), legend = FALSE,
            range = c(0, 4), type = "classes",
            levels = 0:4, col = vote_cols)
```

```{r fig.width=10, fig.height=4}
terra::plot(pred_stack[[block4]], nr = 3, nc = 4,
            mar = c(1.0, 1.0, 1.5, 3.0), legend = FALSE,
            range = c(0, 4), type = "classes",
            levels = 0:4, col = vote_cols)
```

Visualizing our predictions across each season and year, we see that there is generally high agreement among models. These rasters represent consensus votes among each of our four folds, with yellow pixels having strong positive consensus among all folds (all four identify the pixel as suitable) and blue pixels having low consensus among folds (few to no folds identify the given pixel as suitable). We also see that the models correctly visually show one of the main temporal trends in the data: loss of habitat through deforestation starting around year 6.

Unlike predictions from previous model methodologies, hypervolume status as a presence only modeling method means that it is not sensitive to absence data. Instead, it creates predictions exclusively based on the ranges of our input data. As a result, our projections from hypervolume models correctly identify precipitations seasonal role in driving suitability, almost universally identifying no suitability in the dry summer months, and correctly visualizing constrictions in suitability during drought years.

Note that here we show that `time_steps` can handle making predictions from compound variables or variables at different temporal scales so long as their temporal scales are nested. For example here "elevation" has no temporal value and is considered to be static across all time steps. "forest_cover" is measured annually, but is considered to be static across all seasons within a year for the purposes of predictions. "preseason" is measured both by year and season, so resulting seasonal predictions reflect that. However if precipitation was only measured based on aggregate seasons but had no associated year, predictions would fail. Predictions can also be made where all variables share the same time step- for example annual forest cover, annual temperature, and annual precipitation.

Additionally, a plain vector like `time_steps = 1:15` produces one prediction per value of the first time column; the other time columns are filled in with every unique value present in the occurrence data. So a vector input with `time_cols = c("year", "season")` would produce one prediction per year-season combination observed in the data. A data frame like the `expand.grid()` above gives explicit control: any combination of values can be requested, including ones not present in the occurrence data, as long as the matching rasters exist (Such as Summer or Winter predictions).

<br>

# G-space performance {#sec-gspace}

Once predictions are projected, per-time-step metrics become available. These are computed by overlaying the held-out test points (presences only for the hypervolume workflow) on the prediction raster for each time step they fall in and counting correct vs incorrect classifications.

```{r}
head(preds$timestep_metrics)
```

Columns:

-   `Fold` - fold identifier; one row per fold per time step.
-   `Pct_Suitable` - proportion of the study area predicted suitable in that time step.
-   `N_Pres` - number of held-out presence test points falling in that time step for that fold.
-   `TP` - true positives, presences correctly classified as suitable in that time step.
-   `FN` - false negatives, presences incorrectly classified as unsuitable in that time step.
-   `Sensitivity` - `TP / (TP + FN)` for that fold-time step.
-   `CBP` - cumulative binomial probability. Under the null that each test point lands in suitable area at the rate `Pct_Suitable`, this is the probability of observing exactly `TP` correctly classified points by chance. Small values indicate predictions are better than random.
-   `year` and `season` - the values of `time_cols` for that row, identifying which time step the metrics belong to.

G-space metrics are powerful in time steps with substantial sample sizes but lose meaning when few records exist (as with out example). A fold with only one presence in a given year can score sensitivity 0 or 1 with no real information content. Likewise, the ability to earn a significant G-space CBP score explicitly depends on sample size. Report E-space (from `$fold_test_metrics`) and G-space (from `$timestep_metrics`) together for a complete view of model performance.

The overall summary gives a per-fold summary across all time steps:

```{r}
preds$overall_summary
```

Rapid visual assessment of these metrics can be done using the function `plot_model_assessment()` which generates simple summary plots for each.

```{r}
plot_model_assessment(
  predictions         = preds,
  time_column         = c("year", "season"),
  secondary_time_mode = "combine",
  model_result        = hv_out
)
```

This produces per-fold time series of percent suitable, sensitivity, CBP, and TP/FN. Specificity and TN/FP panels are not produced for hypervolume predictions because pseudoabsences are not part of the workflow. When `model_result` is also supplied, overall sensitivity from `$fold_test_metrics` is added as a reference line. When compound time steps are used (such as year and season), you must choose how they are visualized. Choosing `secondary_time_mode = "combine"` will roll them into one continuous x axis. `secondary_time_mode = "facet"` produces stacked plots as seen below, where the first `time_col` is displayed as the x axis, and a different plot is made for each secondary variable (`season` here). By default, the threshold for which CBP is identified as significant is 0.05, but this may also be adjusted manually.

```{r fig.height=20}
plot_model_assessment(
  predictions         = preds,
  time_column         = c("year", "season"),
  secondary_time_mode = "facet",
  model_result        = hv_out,
  cbp_threshold       = 0.001
)
```

In both visualizations, we can clearly see the constrictions in suitable area during the summer in the G-space predictions per time step graphs.

<br>

# Next steps {#sec-next}

Now that predictions have been generated, you can assess the model and see if it is preforming satisfactory enough for what your goals are. If this is the case, you can preform post-processing analyses to try to gain additional inference about the spatiotemporal patterns of change in the study region See the [Post-processing predictions](https://cjhughes926.github.io/TemporalModelR/articles/V4_Postprocessing.html) vignette.

For comparison with other algorithms applied to the same dataset:

-   [Modeling with a GLM](https://cjhughes926.github.io/TemporalModelR/articles/V3a_GLM.html) - linear and polynomial responses.
-   [Modeling with a GAM](https://cjhughes926.github.io/TemporalModelR/articles/V3b_GAM.html) - smooth nonlinear responses.
-   [Modeling with a Random Forest](https://cjhughes926.github.io/TemporalModelR/articles/V3c_RF.html) - flexible ensemble of decision trees.

---
title: "2. Preprocessing temporally explicit data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Preprocessing temporally explicit data}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 5,
  dpi       = 100,
  out.width = "95%"
)
```

<hr>

# Summary

-   [Overview](#sec-overview)
-   [Aligning rasters](#sec-align)
-   [Spatiotemporal rarefaction](#sec-rarefy)
-   [Temporally explicit extraction](#sec-extract)
-   [Scaling rasters](#sec-scale)
-   [Spatiotemporal partitioning](#sec-partition)
-   [Generating pseudoabsences](#sec-absences)
-   [Outputs ready for modeling](#sec-outputs)

<hr>

# Overview {#sec-overview}

The preprocessing phase of TemporalModelR takes raw occurrence records and environmental rasters and produces the structured inputs that the modeling functions expect. Six functions cover this phase, and they are typically run in order:

1.  `raster_align()` - reproject and mask all rasters to a common reference grid.
2.  `spatiotemporal_rarefaction()` - subset occurrence points to one per pixel per time step.
3.  `temporally_explicit_extraction()` - extract environmental values matched to each point's observation time.
4.  `scale_rasters()` - z-score rasters using the per-variable means and SDs from extraction. This step is optional, but should be applied for models sensitive to consistent scaling in predictor variables. If not, raw values and aligned rasters may be used. 
5.  `spatiotemporal_partition()` - assign points to single fold, spatial, temporal, spatiotemporal, or random cross-validation folds.
6.  `generate_absences()` - produce fold-stratified pseudoabsences for presence/absence models.

This vignette walks through each step using the bundled synthetic dataset described in the [About the Example Dataset](https://cjhughes926.github.io/TemporalModelR/articles/V1_dataset.html) vignette. Read that vignette first if you have not already, as it explains the structure of the raw rasters and occurrence points used here. We use the **seasonal (compound time-step) workflow** throughout, with `forest_cover` (annual), `prseas` (year-season), and `elevation` (static) as predictors. The annual workflow is identical apart from substituting `pr_ann` for `prseas` and dropping `season` from the time columns.

```{r}
library(TemporalModelR)
library(terra)
library(sf)

raw_dir  <- system.file("extdata/rasters_raw", package = "TemporalModelR")

pts_file <- system.file("extdata/points/synthetic_occurrence_points.csv",
                        package = "TemporalModelR")

ref_file <- file.path(raw_dir, "elevation.tif")
study_crs <- sf::st_crs(terra::rast(ref_file))
```

<br>

# Aligning rasters {#sec-align}

`raster_align()` standardizes every raster in a directory to a single reference grid by reprojecting, resampling, and masking. This is essential because downstream functions assume all environmental layers share identical CRS, resolution, and extent. Even small differences are enough to misalign extraction at the pixel level. With the bundled dataset the raw rasters are already on the same grid, but we run the alignment step to show the call pattern.

```{r, eval=FALSE}
aligned_dir <- file.path(tempdir(), "rasters_aligned")

raster_align(
  input_dir        = raw_dir,
  output_dir       = aligned_dir,
  reference_raster = ref_file,
  resample_method  = "bilinear",
  overwrite        = TRUE
)
```

`resample_method = "bilinear"` is appropriate for continuous variables like elevation and precipitation. Use `"near"` for categorical rasters such as land cover classes; bilinear interpolation between class codes would incorrectly generate codes that do not exist.

We now have a directory of aligned rasters which can be used in downstream analyses. 

```{r}
aligned_dir <- system.file("extdata/rasters_aligned", package = "TemporalModelR")

list.files(aligned_dir, pattern = "\\.tif$")[1:6]
```

<br>

# Spatiotemporal rarefaction {#sec-rarefy}

`spatiotemporal_rarefaction()` reduces sampling bias and pseudoreplication by retaining one occurrence per pixel per unique time-step combination. The produces an only spatially rarefied table (one point per pixel, regardless of time), and when `time_cols` is provided, it additionally produces a spatiotemporally rarefied table that preserves multiple observations from the same pixel when they fall in different time steps.

This distinction matters for temporally explicit modeling. A static workflow would discard the second observation at the same pixel as redundant. A temporally explicit workflow keeps it, correctly recognizing that the environment at that pixel may have changed between the two observation times, making the second record carry unique information about the species' niche.

```{r}
rare_dir <- file.path(tempdir(), "rarefied")

rare_out <- spatiotemporal_rarefaction(
  points_sp        = pts_file,
  output_dir       = rare_dir,
  reference_raster = ref_file,
  time_cols        = c("year", "season"),
  xcol             = "x",
  ycol             = "y",
  points_crs       = study_crs,
  output_prefix    = "Pts_seasonal",
  verbose          = FALSE
)

rare_out$input_points

rare_out$spatial_points

rare_out$spatiotemporal_points
```

The spatial-only count is the number of unique pixels containing at least one record, retaining only 84 points of the starting 150. The spatiotemporal count is the number of unique pixel-year-season combinations and so will typically be larger whenever the dataset contains repeat visits to the same pixel in different time steps. In this instance, all 150 starting points contain unique pixel-year-season combinations. 

Two CSVs are written to `output_dir`. The spatiotemporal one is what we pass forward; the spatial-only file is retained for reference and for users who want a static comparison:

```{r}
rare_out$files_created
```

If we wanted to focus on annual variation rather than both annual and seasonal variation, we could instead run this focusing only 'year' as our time_col of interest. This would instead save all unique pixel-year combinations, but assume locations from the same year and location but different seasons share no new data.

```{r}
rare_dir <- file.path(tempdir(), "rarefied")

rare_out_annual <- spatiotemporal_rarefaction(
  points_sp        = pts_file,
  output_dir       = rare_dir,
  reference_raster = ref_file,
  time_cols        = "year",
  xcol             = "x",
  ycol             = "y",
  points_crs       = study_crs,
  output_prefix    = "Pts_ann",
  verbose          = FALSE
)

rare_out_annual$input_points

rare_out_annual$spatial_points

rare_out_annual$spatiotemporal_points
```

This still retains many more points (144) than just the spatially explicit extraction.

<br>

# Temporally explicit extraction {#sec-extract}

`temporally_explicit_extraction()` temporally aligns the raster dataset with the temporal columns of the now rarefied species occurrence dataset. For each occurrence record it extracts environmental values from the raster matched to that point's *observation time*, rather than from a long-term average. The pairing between point and raster is driven by `variable_patterns`: clean variable names on the left, filename patterns with time placeholders on the right.

```{r}
ext_dir <- file.path(tempdir(), "extracted")

ext_out <- temporally_explicit_extraction(
  points_sp           = rare_out$files_created$spatiotemporal,
  raster_dir          = aligned_dir,
  variable_patterns   = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR",
    "prseas"       = "prseas_YEAR_SEASON"
  ),
  time_cols           = c("year", "season"),
  xcol                = "X",
  ycol                = "Y",
  points_crs          = study_crs,
  output_dir          = ext_dir,
  output_prefix       = "extracted_seasonal",
  save_raw            = TRUE,
  save_scaled         = TRUE,
  save_scaling_params = TRUE,
  verbose             = FALSE
)

head(ext_out$raw_values)
```

A few details worth highlighting:

-   `"elevation" = "elevation"` is a static pattern with no placeholder, so every record is matched to the single `elevation.tif` raster.
-   `"forest_cover" = "forest_cover_YEAR"` substitutes each record's `year` value for `YEAR`, so a record where the year column has a value of 7 is matched to the raster `forest_cover_7.tif`.
-   `"prseas" = "prseas_YEAR_SEASON"` substitutes both `year` and `season`, so a record where the year column is 3 and the season column = spring is matched to the raster `prseas_3_Spring.tif`.
-   We could also include `"pr_ann" = "pr_ann_YEAR"` to match with annual measures of precipitation.


The same call generates three outputs:

```{r}
ext_out$files_created
```

The raw values file is what models that don't need scaling will consume. The scaled file applies a z-score per variable using the means and SDs computed across all records. The scaling parameters file records those means and SDs so the same transformation can be applied to the prediction rasters, if necessary. 

```{r}
ext_out$scaling_params
```

Z-scoring matters when an algorithm is sensitive to variable scale, such as those that rely on Euclidean distances and will misbehave if one variable's units dwarf the others (e.g. Temperature running from -10 to 38 C compared to precipitation running from 0 to 3000 mm. Use the scaling parameters CSV when running `scale_rasters()` so the rasters and the extracted point data share the same transformation.

<br>

# Scaling rasters {#sec-scale}

`scale_rasters()` z-scores every raster in a directory using the per-variable means and SDs from extraction. The transformation is `(value - mean) / sd`, applied pixel-wise.

```{r, eval=FALSE}
scaled_dir <- file.path(tempdir(), "rasters_scaled")

scale_rasters(
  input_dir           = aligned_dir,
  output_dir          = scaled_dir,
  scaling_params_file = ext_out$files_created$scaling_params,
  variable_patterns   = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR",
    "prseas"       = "prseas_YEAR_SEASON"
  ),
  time_cols           = c("year", "season"),
  overwrite           = TRUE,
  verbose             = FALSE
)
```

`variable_patterns` mirrors the extraction call. For each variable in the scaling parameters file, every matching raster in `input_dir` is z-scored with the same mean and SD, this guarantees the rasters and the points share a consistent transformation.

```{r}
scaled_dir <- system.file("extdata/rasters_scaled", package = "TemporalModelR")

list.files(scaled_dir, pattern = "\\.tif$")[1:6]
```

We now have a directory of scaled rasters which we'll use for future analyses. 

<br>

# Spatiotemporal partitioning {#sec-partition}

`spatiotemporal_partition()` assigns each occurrence record to a cross-validation fold which will be used to create training and testing datasets to build and assess our temporally explicit models. The function supports five modes:

-   **Single fold.** All points used for both training and testing. No held-out validation. Useful for a final production model after cross validated quality has been established, or when sample sizes are too small for splitting.
-   **Random.** Folds are random shuffles with no spatial or temporal structure. Useful as a naive baseline.
-   **Spatial-only.** Recursive k-d tree bisection produces geographically distinct contiguous regions, one per fold. Use when you want to test geographic transferability and don't have time structure to exploit.
-   **Temporal-only.** Each fold spans the full study area but covers a distinct slice of the time series, defined by quantile-based breaks. Use to test temporal transferability.
-   **Spatiotemporal.** Combines the two: `n_spatial_folds` are geographically distinct folds and `n_temporal_folds` are time-restricted folds. Tests both kinds of transferability simultaneously.

We use the spatiotemporal design with two folds in each pool. The function requires the scaled (or raw) extracted points as input, a polygon defining the study area, and a single `time_cols` column (unlike the multi-column placeholders elsewhere; see `?spatiotemporal_partition` for the rationale).

```{r}
ext_scaled_file <- system.file(
  "extdata/points/extracted_seasonal_Scaled_Values.csv",
  package = "TemporalModelR"
)

study_area_sf <- sf::st_as_sf(sf::st_as_sfc(
  sf::st_bbox(c(xmin = 0, xmax = 3000, ymin = 0, ymax = 1500),
              crs = study_crs)
))

partition <- spatiotemporal_partition(
  reference_shapefile_path = study_area_sf,
  points_file_path         = ext_scaled_file,
  xcol                     = "x",
  ycol                     = "y",
  points_crs               = study_crs,
  time_cols                = "year",
  n_spatial_folds          = 2,
  n_temporal_folds         = 2,
  max_attempts = 10,
  max_imbalance = 0.15,
  create_plot              = TRUE,
  verbose                  = FALSE
)

partition$summary
```

The returned `points_sf` carries a `fold` column alongside the original predictor columns. Two of the four folds are geographically restricted but span the full time series; the other two overlap spatially and span the remainder of the study area but are restricted to a distinct time slice each.

For partitioning purposes, time information must be encoded in a single column based on what temporal scale is relevant for the desired partitioning. Both example workflows in this package use the coarsest time scale for partitioning `time_cols = "year"` at this step (the season is preserved in the point data for later modeling and prediction, just not used for fold construction). If both year and season are conceptually meaningful for fold breaks, encode them into a single ordered numeric column before partitioning. Alternatively, partitioning could be done across seasons to view the environmental transferability of seasonal environmental values. 

<br>

# Generating pseudoabsences {#sec-absences}

`generate_absences()` produces fold-stratified pseudoabsence/background points for presence/absence modeling. Pseudoabsences measure the environmental characteristics of locations not otherwise characterized by species occurrences and are essential for GLMs, GAMs, and random forests when no 'true absence' points exist. The hypervolume method is presence-only and does not need them.

Four generation methods are supported:

-   **Random.** Uniform sampling across the study area with a small buffer around presences to avoid exact overlap.
-   **Buffer.** Sampling within a user-specified radius of presence locations. Useful when you want pseudoabsences from the same general environment as the presences, on the assumption that nearby unsampled locations are more likely to represent true absence than locations far from any survey effort.
-   **Environmental.** Cells outside the species' environmental tolerance envelope are identified as candidates, then k-means clustering selects a spatially representative subset. When `buffer_distance` is also supplied the environmental filtering is restricted to within that buffer, implementing the three-step approach of Senay et al. (2013).
-   **User data.** A user-supplied set of points is used directly as the absence locations rather than generating them. Environmental values are extracted internally at each point's matched time step using the same `raster_dir` and `variable_patterns` as the other methods. Fold assignment is performed automatically by spatial join against the partition's fold boundaries, with any points outside the spatial folds routed to temporal folds by their time value. This is the appropriate method when you have true absence records, or target-group background points from a separate survey.

The function distributes pseudoabsences across folds and across time steps in proportion to the presences in each fold-time combination, so the potential environmental conditions experienced by each pseudoabsence match those experienced by the presences it complements.

## Buffer method

We use the buffer method with a 300 m radius (three pixels at the synthetic landscape's 100 m resolution) and a 2:1 pseudoabsence-to-presence ratio:

```{r}
absences <- generate_absences(
  partition_result         = partition,
  reference_shapefile_path = study_area_sf,
  raster_dir               = scaled_dir,
  variable_patterns        = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR",
    "prseas"       = "prseas_YEAR_SEASON"
  ),
  method                   = "buffer",
  buffer_distance          = 300,
  ratio                    = 2,
  time_cols                = c("year", "season"),
  create_plot              = TRUE,
  plot_by_fold             = TRUE,
  verbose                  = FALSE
)

absences$summary
```

Note that `variable_patterns` and `time_cols` here use both `year` and `season`. The pseudoabsences are stamped with year-season values just like the presences, and the environmental values attached to them are pulled from the matching time-specific raster. This consistency is what keeps the temporally explicit workflow intact through the model fitting step.

## User data method

The `generate_absences` function alternatively allows for users to process a set of defined absences for use with downstream operations in this package. These points may represent true absence records, such as negative occupancy survey locations, or may represent pseudoabsence locations, such as occupancy locations for similar taxa used as a proxy for survey effort. When using `method = "user_data"`, the function assigns each point to a fold, then extracts temporally matched environmental values exactly as it would for generated pseudoabsences. The output is identical except for the fact that the points were not generated internally, and this output can subsequently be used in any function requiring outputs from `generate_absences`.

We consider `synthetic_user_presences.csv` as a second related species to our species of interest. We assume that the survey methodology is similar for both species, and any surveys conducted for this second species would also have subsequently detected our species of interest if it was present.

```{r}
user_pts_file <- system.file(
  "extdata/points/synthetic_user_presences.csv",
  package = "TemporalModelR"
)

user_pts <- utils::read.csv(user_pts_file)

head(user_pts)

nrow(user_pts)
```

The file has the same `x`, `y`, `year`, `season` columns as the primary occurrence data. Before passing it to `generate_absences()`, we apply the same spatiotemporal rarefaction step used for the primary presences. This ensures that if the second species' dataset contains multiple records in the same pixel and time step due to repeat surveys or aggregated data, only one is retained, preventing pseudoreplication in the absence set for the same reason it is prevented in the presence set.

```{r}
user_rare_dir <- file.path(tempdir(), "rarefied_user")

user_rare_out <- spatiotemporal_rarefaction(
  points_sp        = user_pts_file,
  output_dir       = user_rare_dir,
  reference_raster = ref_file,
  time_cols        = c("year", "season"),
  xcol             = "x",
  ycol             = "y",
  points_crs       = study_crs,
  output_prefix    = "Pts_user",
  verbose          = FALSE
)

user_rare_out$input_points

user_rare_out$spatiotemporal_points
```

We pass the rarefied file to `generate_absences()` via `user_absence_data`, along with the coordinate column names and CRS so that it can be processed as pseudoabsences for future modeling.

```{r}
absences_user <- generate_absences(
  partition_result         = partition,
  reference_shapefile_path = study_area_sf,
  raster_dir               = scaled_dir,
  variable_patterns        = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR",
    "prseas"       = "prseas_YEAR_SEASON"
  ),
  method                   = "user_data",
  user_absence_data        = user_rare_out$files_created$spatiotemporal,
  xcol                     = "X",
  ycol                     = "Y",
  points_crs               = study_crs,
  time_cols                = c("year", "season"),
  create_plot              = TRUE,
  plot_by_fold             = TRUE,
  verbose                  = FALSE
)

absences_user$summary
```

The result has the same structure as any other `generate_absences()` output and can be passed directly to any of the model builders. 

<br>

# Outputs ready for modeling {#sec-outputs}

At this point the three key downstream inputs are in hand:

-   `scaled_dir` - the directory of scaled (or aligned in `rasters_aligned`, if scaling was skipped) rasters from which environmental conditions will be projected across space and time.
-   `partition` - the rarefied, fold-assigned, time explicit environment data from species occurrence points.
-   `absences` - the fold-stratified pseudoabsences with matching environmental attributes.

These are passed directly to any of the four model builders covered in the following parallel vignettes:

-   [Modeling with a GLM](https://cjhughes926.github.io/TemporalModelR/articles/V3a_GLM.html)
-   [Modeling with a GAM](https://cjhughes926.github.io/TemporalModelR/articles/V3b_GAM.html)
-   [Modeling with a Random Forest](https://cjhughes926.github.io/TemporalModelR/articles/V3c_RF.html)
-   [Modeling with a Hypervolume](https://cjhughes926.github.io/TemporalModelR/articles/V3d_HV.html)

Once a model is fitted, predictions are projected through space and time using `generate_spatiotemporal_predictions()`, and the resulting prediction stack is summarized in the [Post-processing predictions](https://cjhughes926.github.io/TemporalModelR/articles/V4_Postprocessing.html) vignette.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_documentation.R
\docType{data}
\name{tmr_partition_annual}
\alias{tmr_partition_annual}
\title{Pre-built spatiotemporal partition (annual workflow)}
\format{
A list as returned by \code{\link{spatiotemporal_partition}}.
}
\usage{
tmr_partition_annual
}
\description{
Annual variant of \code{tmr_partition}. Built from points rarefied with
\code{time_cols = "year"} only (one point per pixel per year) and
extracted against annual predictors. Predictors attached to each point
are \code{forest_cover}, \code{pr_ann} (annual precipitation), and
\code{elevation}, all z-scored. Uses
\code{n_spatial_folds = 2} and \code{n_temporal_folds = 2} for four
folds. Loaded with \code{data(tmr_partition_annual)}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_documentation.R
\docType{data}
\name{tmr_absences}
\alias{tmr_absences}
\title{Pre-built pseudoabsence result (seasonal workflow)}
\format{
A list as returned by \code{\link{generate_absences}}, containing
  \code{$pseudoabsences} (an sf object with attached predictor columns),
  \code{$plots}, and \code{$summary}.
}
\usage{
tmr_absences
}
\description{
Output from \code{\link{generate_absences}} using the buffer method with
a 300 m buffer (3 pixels at the synthetic landscape's 100 m resolution)
and a 2:1 pseudoabsence-to-presence ratio. Pseudoabsences are stratified
by fold from \code{tmr_partition} and have \code{forest_cover},
\code{prseas}, and \code{elevation} extracted at each location's
year-season combination. Loaded with \code{data(tmr_absences)}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatiotemporal_partition.R
\name{spatiotemporal_partition}
\alias{spatiotemporal_partition}
\title{Spatiotemporal Cross-Validation Partitioning}
\usage{
spatiotemporal_partition(reference_shapefile_path, points_file_path,
                         time_cols = NULL, xcol = NULL, ycol = NULL,
                         points_crs = NULL, n_spatial_folds = 0,
                         n_temporal_folds = 0, n_balanced_folds = 0,
                         n_random_folds = 0, single_fold= FALSE,
                         max_imbalance = 0.05, max_attempts = 10,
                         create_plot = TRUE, plot_palette = "Dark 2",
                         output_file= NULL, verbose    = TRUE)
}
\arguments{
\item{reference_shapefile_path}{Character or sf object. Path to a polygon
file or an \code{sf} polygon object defining the study area.}

\item{points_file_path}{Character, sf object, sfc object, Spatial object,
or data frame. Path to occurrence data (\code{.csv}, \code{.shp},
\code{.geojson}, \code{.gpkg}) or a spatial object.}

\item{time_cols}{Character. Name of a single column containing temporal
values (e.g. year). Used to define temporal blocks. Required when using
temporal folds. Must be a single column name; does not support more than one
time column unlike other functions in this package. Compound time representations
(e.g. year + season) should be encoded into a single ordered numeric column
before partitioning, or only one (e.g. year) should be used.}

\item{xcol}{Character. Name of the x-coordinate column. Required when
\code{points_file_path} is a CSV file or data frame.}

\item{ycol}{Character. Name of the y-coordinate column. Required when
\code{points_file_path} is a CSV file or data frame.}

\item{points_crs}{Character or CRS object. CRS of the input points.
Required when \code{points_file_path} is a CSV file or data frame.}

\item{n_spatial_folds}{Integer. Number of spatially explicit folds. Ignored
when using random folds. Default is \code{0}.}

\item{n_temporal_folds}{Integer. Number of temporally explicit folds.
When used alone (with \code{n_spatial_folds = 0}), creates temporal-only
folds where each fold spans the full study area but covers a distinct
slice of the time series. When combined with \code{n_spatial_folds},
creates a spatiotemporal design. Ignored when using random folds.
Default is \code{0}.}

\item{n_balanced_folds}{Integer. Reserved for future use. Default is
\code{0} (disabled).}

\item{n_random_folds}{Integer. Number of random folds with no spatial or
temporal structure. Overrides all other fold parameters. Default is
\code{0}.}

\item{single_fold}{Logical. If \code{TRUE}, bypasses all partitioning and
assigns all points to a single fold (fold 1). In this mode all points are
used for both training and testing, producing a single model trained on
the full dataset. All downstream functions accept the result identically
to a standard multi-fold partition. Overrides all fold count parameters.
Default is \code{FALSE}.}

\item{max_imbalance}{Numeric. Maximum allowed fold size imbalance as a
proportion between 0 and 1. Default is \code{0.05}.}

\item{max_attempts}{Integer. Maximum number of partitioning attempts for
spatiotemporal and balanced modes. Each attempt re-runs the spatial block
construction; the attempt with the lowest imbalance is returned. Ignored
for random and spatial-only modes. Default is \code{10}.}

\item{create_plot}{Logical. If \code{TRUE} (default), generates
diagnostic plots showing fold distributions.}

\item{plot_palette}{Character. Name of an HCL or RColorBrewer palette used
to color folds in diagnostic plots. Accepts any HCL palette name (see
\code{\link[grDevices]{hcl.pals}}) or, if \pkg{RColorBrewer} is installed,
any Brewer palette name. Default is \code{"Dark 2"}.}

\item{output_file}{Character. Optional path to save the result as an
\code{.rds} file. The parent directory will be created if it does not
exist. Default is \code{NULL}.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress
messages during processing. Includes the partition mode, fold
structure, per-fold point counts, and file-save confirmation.}
}
\value{
Invisibly returns a list containing:
\itemize{
  \item \code{folds}: Data frame of fold assignments with a \code{fold}
    column identifying each point's cross-validation fold.
  \item \code{points_sf}: sf object of occurrence points with assigned
    folds.
  \item \code{voronoi_folds}: sf object of Voronoi polygons representing
    the spatial fold boundaries. \code{NULL} for random folds, temporal-only
    folds, and single-fold mode.
  \item \code{summary}: Data frame of partitioning summary statistics.
  \item \code{plots}: Named list of recorded plot objects when
    \code{create_plot = TRUE}. Empty list in single-fold mode.
}
}
\description{
Preprocesses species occurrence data by partitioning it into spatially and
temporally structured folds for cross-validation. Supports creation of
spatial-only folds, temporal-only folds, and random folds.
}
\details{
Works better with smaller numbers of folds and may have difficulties
creating even folds for large numbers of groups or where sample sizes
are very small.


The function partitions data into folds using one of five modes:
\itemize{
  \item \strong{Single fold}: All points are assigned to fold 1 and used
    for both training and testing. This produces a single model trained on
    the full dataset with no held-out validation. Useful when sample sizes
    are too small for cross-validation, or as a final production model step
    after cross-validation has already established model quality. Set
    \code{single_fold = TRUE}. All downstream functions accept the result
    identically to standard multi-fold output.

  \item \strong{Random}: Points are assigned to folds by random shuffling
    with no spatial or temporal structure. Each fold is a simple random
    sample of the full dataset, intended as a naive baseline that makes no
    attempt to reduce spatial or temporal autocorrelation between training
    and test sets. Use \code{n_random_folds}.

  \item \strong{Spatial-only}: The study area is divided into \eqn{k}
    contiguous spatial regions using a recursive k-d tree bisection
    algorithm. At each step the point set is split along its longest
    spatial axis, recursively halving until the target number of folds
    is reached. A centroid reassignment pass then refines boundaries
    to improve balance. Each region becomes one fold, so training
    always occurs on data from geographically distinct areas relative
    to the test fold. No temporal separation is imposed, meaning that
    points from any time period may appear in any fold. Use
    \code{n_spatial_folds} alone.

  \item \strong{Temporal-only}: Each fold covers the full spatial extent
    of the study area but is restricted to a distinct, non-overlapping
    slice of the time series. The global time series is divided into
    \code{n_temporal_folds} equal intervals using quantile-based breaks,
    and all points within each interval form one fold. This design tests
    model transferability across time while retaining full spatial coverage
    in every fold. Use \code{n_temporal_folds} alone (with
    \code{n_spatial_folds = 0}). Requires \code{time_cols}.

  \item \strong{Spatiotemporal}: Folds are assigned using the same
    recursive k-d tree bisection as spatial-only mode, operating on
    the full point set to produce spatially contiguous groups. The
    resulting groups are then split into a spatial pool
    (\code{n_spatial_folds} folds drawn from geographically distinct
    regions) and a temporal pool (\code{n_temporal_folds} folds each
    restricted to a distinct slice of the time series but spanning
    the full study area). Together the two pools assess both geographic
    and temporal transferability in a single cross-validation design.
    Use \code{n_spatial_folds} and \code{n_temporal_folds} together.
    Requires \code{time_cols}.
}

Fold assignment uses a recursive k-d tree bisection algorithm that splits
points along their longest spatial axis at each step, followed by a
centroid reassignment pass to improve boundary regularity and point-count
balance. Voronoi tessellation on fold centroids is used only for
visualisation of the resulting spatial boundaries. For temporal
mode, temporal blocks are defined by dividing the global time series
into equal intervals using quantile-based breaks. For spatiotemporal mode,
the typical spatial assignment is done, but with one larger spatial block
made with enough points to represent all of the temporal folds, then the
temporal blocking is applied to those points.

Partitioned datasets are suitable for cross-validation in modeling
workflows, ensuring spatial and/or temporal independence between folds.
}
\examples{
pts_file <- system.file(
  "extdata/points/extracted_seasonal_Scaled_Values.csv",
  package = "TemporalModelR"
)

ref_file <- system.file("extdata/rasters_raw/elevation.tif",
                        package = "TemporalModelR")

study_crs <- sf::st_crs(terra::rast(ref_file))

study_area_sf <- sf::st_as_sf(sf::st_as_sfc(
  sf::st_bbox(c(xmin = 0, xmax = 3000, ymin = 0, ymax = 1500),
              crs = study_crs)
))

spatiotemporal_partition(
  reference_shapefile_path = study_area_sf,
  points_file_path         = pts_file,
  xcol                     = "x",
  ycol                     = "y",
  points_crs               = study_crs,
  time_cols                = "year",
  n_spatial_folds          = 2,
  n_temporal_folds         = 2,
  create_plot              = FALSE,
  verbose                  = FALSE
)
}
\seealso{
Preprocessing: \code{\link{spatiotemporal_rarefaction}},
  \code{\link{temporally_explicit_extraction}},
  \code{\link{generate_absences}}

Modeling: \code{\link{build_temporal_hv}},
  \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
  \code{\link{build_temporal_rf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyze_trends_by_spatial_unit.R
\name{analyze_trends_by_spatial_unit}
\alias{analyze_trends_by_spatial_unit}
\title{Summarize Temporal Patterns and Trends by Spatial Unit}
\usage{
analyze_trends_by_spatial_unit(shapefile_path, name_field, binary_stack = NULL,
                               pattern_raster = NULL, time_decrease_raster = NULL,
                               time_increase_raster = NULL, time_steps = NULL,
                               output_dir = NULL, overwrite = FALSE,
                               create_plot = TRUE, pie_scale = 0.15,
                               verbose = TRUE)
}
\arguments{
\item{shapefile_path}{Character, sf object, or sfc object. Path to a
shapefile or directory containing one, an sf object, or an sfc geometry.
Shapefiles spatial units will be used as the units for the data summary.}

\item{name_field}{Character. Attribute field to use as spatial unit labels.}

\item{binary_stack}{\code{SpatRaster} or character. Optional stack of binary
prediction rasters across time. Required for per-unit habitat summaries
and time series plots.}

\item{pattern_raster}{\code{SpatRaster} or character. Optional pattern
classification raster from \code{\link[TemporalModelR]{analyze_temporal_patterns}}.
Required for pattern composition summaries and scatterpie map.}

\item{time_decrease_raster}{\code{SpatRaster} or character. Optional raster
of first decrease time step from \code{\link[TemporalModelR]{analyze_temporal_patterns}}.}

\item{time_increase_raster}{\code{SpatRaster} or character. Optional raster
of first increase time step from \code{\link[TemporalModelR]{analyze_temporal_patterns}}.}

\item{time_steps}{Vector. Time labels for layers in \code{binary_stack}.
Required when \code{binary_stack} or change rasters are provided.}

\item{output_dir}{Character or \code{NULL}. Directory for CSV outputs and
plots. If \code{NULL}, results are returned in memory only and no files
are written. Default is \code{NULL}.}

\item{overwrite}{Logical. If \code{TRUE}, recomputes and overwrites existing
CSVs. Default is \code{FALSE}.}

\item{create_plot}{Logical. If \code{TRUE} (default), generates and saves
plots.}

\item{pie_scale}{Numeric in (0, 1]. Largest pie's radius as a fraction of
the smaller map dimension. Default is \code{0.15}, meaning the biggest
pie spans roughly 30\% of the smaller map dimension (radius = 15\%, so
diameter = 30\%). Smaller pies scale so their area is proportional to
the unit's pixel count. The fraction is converted internally to
coordinate units, so this argument works identically across CRSes
(degrees, meters, etc.) without manual rescaling.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress
messages during processing. Includes details on raster extraction and
per-spatial-unit summaries.}
}
\value{
Invisibly returns a list containing:
\itemize{
  \item \code{overall_summary}: Pattern composition per spatial unit
    (present when \code{pattern_raster} is supplied).
  \item \code{timestep_summary}: Suitable pixel counts per unit per time step
    (present when \code{binary_stack} is supplied).
  \item \code{change_by_timestep}: Gain and loss pixel counts per unit per time
    step (present when both change rasters are supplied).
  \item \code{plots}: Named list of recorded plot objects (present when
    \code{create_plot = TRUE}).
}
}
\description{
Aggregates temporal pattern classifications and change metrics
across user-defined spatial units (e.g., states, counties, watersheds).
Returns summary tables and optionally generates simple visualizations.
}
\details{
Summarizes results from modeling and post-processing at the scale of specific
spatial blocks to allow for a nuanced look at spatiotemporal patterns.
}
\examples{
con_file <- system.file("extdata/binary/consensus_stack.tif",
                        package = "TemporalModelR")

binary_stack <- terra::rast(con_file)

study_crs <- sf::st_crs(binary_stack)

zones_sf <- rbind(
  sf::st_sf(ZONE = "West",
            geometry = sf::st_sfc(sf::st_polygon(list(
              matrix(c(0, 0, 1500, 1500, 0,
                       0, 1500, 1500, 0, 0), ncol = 2)
            )), crs = study_crs)),
  sf::st_sf(ZONE = "East",
            geometry = sf::st_sfc(sf::st_polygon(list(
              matrix(c(1500, 1500, 3000, 3000, 1500,
                       0,    1500, 1500, 0,    0),    ncol = 2)
            )), crs = study_crs))
)

time_steps <- expand.grid(
  year             = 1:15,
  season           = "Spring",
  stringsAsFactors = FALSE
)

analyze_trends_by_spatial_unit(
  shapefile_path = zones_sf,
  name_field     = "ZONE",
  binary_stack   = binary_stack,
  time_steps     = time_steps,
  create_plot    = FALSE,
  verbose        = FALSE
)
}
\seealso{
Post-processing: \code{\link{summarize_raster_outputs}},
  \code{\link[TemporalModelR]{analyze_temporal_patterns}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scale_rasters.R
\name{scale_rasters}
\alias{scale_rasters}
\title{Scale Environmental Rasters Using Species Occurrence Data}
\usage{
scale_rasters(input_dir, output_dir, scaling_params_file,
              variable_patterns, time_cols = NULL, output_suffix = "_Scaled",
              overwrite = FALSE, verbose = TRUE)
}
\arguments{
\item{input_dir}{Character. Directory containing aligned input .tif raster files,
typically the output of \code{\link{raster_align}}.}

\item{output_dir}{Character. Directory where scaled rasters will be saved.}

\item{scaling_params_file}{Character. Path to CSV file with columns:
variable, mean, sd. Typically generated by
\code{\link{temporally_explicit_extraction}}.}

\item{variable_patterns}{Named character vector mapping clean variable names
to raster filename patterns. For time-varying variables include the time
placeholder in the pattern (e.g. \code{"forest_cover" = "forest_cover_YEAR"});
for static variables omit it (e.g. \code{"elevation" = "elevation"}). Time
placeholders must match entries in \code{time_cols}.}

\item{time_cols}{Character vector of time placeholders for dynamic variables.
Default is \code{NULL}.}

\item{output_suffix}{Character. Suffix to append to output raster filenames.
Default is \code{"_Scaled"}.}

\item{overwrite}{Logical. If \code{TRUE}, overwrites existing output files.
If \code{FALSE} (default), existing files are skipped.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress
messages during processing. Includes per-variable scaling progress.}
}
\value{
Invisibly returns a list containing:
\itemize{
  \item \code{n_scaled}: Integer. Total number of variables successfully
    scaled (static plus dynamic).
  \item \code{output_dir}: Character. Path to the directory containing scaled
    rasters.
}
}
\description{
Preprocessing function that standardizes raster files using precomputed mean
and standard deviation values from species occurrence data. Supports both
temporally static and dynamic rasters.
}
\details{
Applies scaling to rasters via z-score transformation: (value - mean) / sd
as is calculated during \code{\link{temporally_explicit_extraction}}. Scaled
rasters are written to output_dir.

Scaled rasters should use the same scaling parameters derived from species
occurrence data to ensure consistent standardization between training data
and prediction layers.
}
\examples{
aln_dir     <- system.file("extdata/rasters_aligned",
                           package = "TemporalModelR")

params_file <- system.file(
  "extdata/points/extracted_seasonal_Scaling_Parameters.csv",
  package = "TemporalModelR"
)

out_dir <- file.path(tempdir(), "scaled")

scale_rasters(
  input_dir           = aln_dir,
  output_dir          = out_dir,
  scaling_params_file = params_file,
  variable_patterns   = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR",
    "prseas"       = "prseas_YEAR_SEASON"
  ),
  time_cols           = c("year", "season"),
  overwrite           = TRUE,
  verbose             = FALSE
)
}
\seealso{
Preprocessing: \code{\link{temporally_explicit_extraction}},
  \code{\link{raster_align}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_documentation.R
\docType{data}
\name{tmr_absences_user}
\alias{tmr_absences_user}
\title{Pre-built user_data pseudoabsences (seasonal workflow)}
\format{
A list as returned by \code{\link{generate_absences}}, containing
  \code{$pseudoabsences} (an sf object with attached predictor columns),
  \code{$plots}, and \code{$summary}.
}
\usage{
tmr_absences_user
}
\description{
Output of \code{\link{generate_absences}} run with
\code{method = "user_data"} against \code{\link{tmr_partition}}, using
\code{inst/extdata/points/synthetic_user_presences.csv} as the supplied
absence locations. The CSV contains occurrence-formatted points for a
'related speecies' to that data of interest to be used as a proxy for survey
effort for the species of interest, serving as an alternative method for
generating pseudoabsences.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_spatiotemporal_predictions.R
\name{generate_spatiotemporal_predictions}
\alias{generate_spatiotemporal_predictions}
\title{Generate Spatiotemporal Predictions from Temporal Models}
\usage{
generate_spatiotemporal_predictions(partition_result, model_result,
                                    pseudoabsence_result = NULL, raster_dir,
                                    variable_patterns, time_cols, time_steps,
                                    output_dir = file.path(tempdir(), "Predictions"),
                                    overwrite = FALSE,
                                    verbose = TRUE)
}
\arguments{
\item{partition_result}{List or character. Output from
\code{\link{spatiotemporal_partition}} or path to an \code{.rds}
file containing that output.}

\item{model_result}{List or character. Output from any of
\code{\link{build_temporal_hv}}, \code{\link{build_temporal_glm}},
\code{\link{build_temporal_gam}}, or \code{\link{build_temporal_rf}}, or
a path to an \code{.rds} file. Model type is detected automatically from
the \code{model_type} field.}

\item{pseudoabsence_result}{List, character, or \code{NULL}. Optional.
Output from \code{\link{generate_absences}} or path to an
\code{.rds} file. When supplied for presence/absence models (GLM, GAM, RF),
the held-out pseudoabsence test points for each fold are filtered to the
current time step and used alongside presence test points to compute
per-timestep TN, FP, Specificity, and TSS. These columns are added to the
timestep metrics table when pseudoabsences are available. Ignored for
hypervolume models. Default is \code{NULL}.}

\item{raster_dir}{Character. Directory containing environmental raster
files (\code{.tif}), typically the output of
\code{\link{raster_align}} or \code{\link{scale_rasters}}. File names
must follow the patterns supplied in \code{variable_patterns}, with any
time placeholder substituted for the corresponding value from
\code{time_cols}.}

\item{variable_patterns}{Named character vector mapping clean variable names
to raster filename patterns. For time-varying variables include the time
placeholder in the pattern (e.g. \code{"forest_cover" = "forest_cover_YEAR"});
for static variables omit it (e.g. \code{"elevation" = "elevation"}). Time
placeholders must match entries in \code{time_cols}.}

\item{time_cols}{Character. Name of the column(s) containing year or time
step values in the occurrence data. Must match \code{time_cols} used in
\code{\link{spatiotemporal_partition}} and the time placeholders used in
\code{variable_patterns}.}

\item{time_steps}{Vector, data frame, or matrix of time periods for which
to generate predictions.}

\item{output_dir}{Character. Directory to write prediction rasters and the
assessment metrics CSV. Default is \code{file.path(tempdir(), "Predictions")}.}

\item{overwrite}{Logical. If \code{TRUE}, overwrites existing output files.
If \code{FALSE} (default), existing files are skipped.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress
messages during processing. Includes per-time-step prediction progress.}
}
\value{
Invisibly returns a list containing:
\itemize{
  \item \code{timestep_metrics}: data frame of per-fold per-time-step
    metrics. Columns always present: \code{Fold}, the time column(s),
    \code{Pct_Suitable}, \code{N_Pres}, \code{TP}, \code{FN},
    \code{Sensitivity}, \code{CBP}. When \code{pseudoabsence_result} is
    supplied for parametric models, additionally: \code{N_Abs}, \code{TN},
    \code{FP}, \code{Specificity}, \code{TSS}. Saved as
    \code{Timestep_Assessment_Metrics.csv}.
  \item \code{overall_summary}: data frame of pooled metrics per fold.
    Columns: \code{Fold}, \code{N_Timesteps}, \code{Mean_Pct_Suitable},
    \code{Total_TP}, \code{Total_FN}, \code{Overall_Sensitivity},
    \code{Overall_CBP}. When pseudoabsences available: additionally
    \code{Total_TN}, \code{Total_FP}, \code{Overall_Specificity},
    \code{Overall_TSS}.
  \item \code{prediction_files}: character vector of paths to saved
    prediction rasters.
  \item \code{model_type}: character string recording the model type used.
}
}
\description{
Generates temporally explicit habitat suitability predictions by projecting
fitted models onto environmental raster stacks matching each time period.
Accepts output from any of the four TemporalModelR modeling functions.
Produces per-time-step assessment metrics showing how predictions and
test point coverage vary over time.
}
\details{
G-space predictions are produced as rasters for each time-step and fold of an
input model.

Per-timestep metrics are computed per fold per time step by extracting
raster predictions at the held-out presence points that fall within that
time step. \code{Pct_Suitable} records the proportion of the study area
predicted suitable. \code{Sensitivity} shows how consistently test points are
captured. \code{CBP} (cumulative binomial probability) tests whether the
observed number of correctly predicted test points is better than expected
by random placement: under the null, each point has \code{Pct_Suitable}
probability of falling in suitable area, so
\code{dbinom(TP, N_Test_Pts, Pct_Suitable)} gives the probability of
observing exactly \code{TP} correct by chance. Small values indicate
predictions are better than random.
}
\examples{
data(tmr_partition, package = "TemporalModelR")

data(tmr_glm,       package = "TemporalModelR")

data(tmr_absences,  package = "TemporalModelR")

scl_dir    <- system.file("extdata/rasters_scaled",
                          package = "TemporalModelR")

time_steps <- expand.grid(
  year             = 1:15,
  season           = "Spring",
  stringsAsFactors = FALSE
)

generate_spatiotemporal_predictions(
  partition_result     = tmr_partition,
  model_result         = tmr_glm,
  pseudoabsence_result = tmr_absences,
  raster_dir           = scl_dir,
  variable_patterns    = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR",
    "prseas"       = "prseas_YEAR_SEASON"
  ),
  time_cols            = c("year", "season"),
  time_steps           = time_steps,
  output_dir           = tempdir(),
  overwrite            = TRUE,
  verbose              = FALSE
)
}
\seealso{
Preprocessing: \code{\link{scale_rasters}},
  \code{\link{spatiotemporal_partition}}

Modeling: \code{\link{build_temporal_hv}},
  \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
  \code{\link{build_temporal_rf}}

Post-processing: \code{\link{summarize_raster_outputs}},
  \code{\link{plot_model_assessment}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatiotemporal_rarefaction.R
\name{spatiotemporal_rarefaction}
\alias{spatiotemporal_rarefaction}
\title{Spatiotemporal Rarefaction of Species Occurrence Data}
\usage{
spatiotemporal_rarefaction(points_sp, output_dir, reference_raster,
                           time_cols = NULL, xcol = NULL, ycol = NULL,
                           points_crs = NULL, output_prefix = "Pts_Database",
                           verbose = TRUE)
}
\arguments{
\item{points_sp}{Input point data. Accepts an sf object, data frame,
SpatialPointsDataFrame, or file path to a \code{.csv}, \code{.shp},
\code{.geojson}, or \code{.gpkg} file.}

\item{output_dir}{Character. Directory where output CSV files will be saved.}

\item{reference_raster}{Character, SpatRaster, or RasterLayer. Raster used
to define pixel boundaries for rarefaction. Accepts a file path,
RasterLayer, or SpatRaster.}

\item{time_cols}{Character vector. Column names in \code{points_sp} defining
temporal grouping for spatiotemporal rarefaction. When \code{NULL}
(default), only spatial rarefaction is performed.}

\item{xcol}{Character. Name of the x-coordinate column. Required when
\code{points_sp} is a CSV file or data frame.}

\item{ycol}{Character. Name of the y-coordinate column. Required when
\code{points_sp} is a CSV file or data frame.}

\item{points_crs}{Character or CRS object. CRS of the input points.
Required when \code{points_sp} is a CSV file or data frame.}

\item{output_prefix}{Character. Prefix for output file names. Default is
\code{"Pts_Database"}. Output files are named
\code{<output_prefix>_OnePerPix.csv} and, when \code{time_cols} are
provided, \code{<output_prefix>_OnePerPixPerTimeStep.csv}.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress
messages during processing. Includes file loading, missing-data removal
counts, rarefaction summaries, and a final comparison of spatial
vs spatiotemporal point counts.}
}
\value{
Invisibly returns a list containing:
\itemize{
  \item \code{input_points}: Integer. Number of input points after CRS
    alignment and removal of rows with missing time column values.
  \item \code{spatial_points}: Integer. Number of points retained after
    spatial-only rarefaction (one per pixel).
  \item \code{spatiotemporal_points}: Integer. Number of points retained
    after spatiotemporal rarefaction (one per pixel per time combination).
    \code{NA} when \code{time_cols} is not provided.
  \item \code{time_cols_used}: Character vector of time columns used.
    \code{NULL} when \code{time_cols} is not provided.
  \item \code{spatial_table}: Data frame of spatially rarefied points with
    columns \code{pixel_id}, \code{X}, \code{Y}, and any \code{time_cols}.
  \item \code{spatiotemporal_table}: Data frame of spatiotemporally rarefied
    points with columns \code{pixel_id}, \code{X}, \code{Y}, and
    \code{time_cols}. \code{NULL} when \code{time_cols} is not provided.
  \item \code{files_created}: Named list of file paths written. Always
    contains \code{$spatial}; additionally contains
    \code{$spatiotemporal} when \code{time_cols} are provided.
}
}
\description{
Preprocessing function that rarefies species occurrence data to one point per
raster pixel, optionally accounting for temporal components. Reduces sampling
bias and spatial autocorrelation in occurrence datasets.
}
\details{
The function assigns each point to a raster pixel using the resolution and
extent of \code{reference_raster}, then performs:
\itemize{
  \item Spatial rarefaction: retains one point per pixel, written to
    \code{<output_prefix>_OnePerPix.csv}.
  \item Spatiotemporal rarefaction (when \code{time_cols} are provided):
    retains one point per pixel per unique combination of time column values,
    written to \code{<output_prefix>_OnePerPixPerTimeStep.csv}.
}

Output CSV files are suitable as direct input to
\code{\link{temporally_explicit_extraction}}.
}
\examples{
pts_file <- system.file(
  "extdata/points/synthetic_occurrence_points.csv",
  package = "TemporalModelR"
)

ref_file <- system.file("extdata/rasters_raw/elevation.tif",
                        package = "TemporalModelR")

out_dir  <- file.path(tempdir(), "rarefied")

spatiotemporal_rarefaction(
  points_sp        = pts_file,
  output_dir       = out_dir,
  reference_raster = ref_file,
  time_cols        = c("year", "season"),
  xcol             = "x",
  ycol             = "y",
  points_crs       = terra::crs(terra::rast(ref_file)),
  verbose          = FALSE
)
}
\seealso{
Preprocessing: \code{\link{temporally_explicit_extraction}},
  \code{\link{spatiotemporal_partition}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_temporal_rf.R
\name{build_temporal_rf}
\alias{build_temporal_rf}
\title{Build Temporal Random Forest Models Across Cross-Validation Folds}
\usage{
build_temporal_rf(partition_result, pseudoabsence_result, model_vars,
                  rf_params = list(), threshold_method = "tss",
                  output_dir = file.path(tempdir(), "RF_Models"),
                  create_plot = TRUE, plot_palette = "Dark 2",
                  overwrite = FALSE, time_cols = NULL, verbose = TRUE)
}
\arguments{
\item{partition_result}{List or character. Output from
\code{\link{spatiotemporal_partition}} or path to an \code{.rds}
file containing that output.}

\item{pseudoabsence_result}{List or character. Output from
\code{\link{generate_absences}} or path to an \code{.rds} file
containing that output.}

\item{model_vars}{Character vector. Names of predictor columns to include in
the Random Forest. All variables must be present as columns in both the
presence and pseudoabsence data.}

\item{rf_params}{Named list. Additional arguments passed to
\code{\link[randomForest]{randomForest}}, such as \code{ntree} (number of
trees, default 500), \code{mtry} (number of variables tried at each split,
default \code{floor(sqrt(length(model_vars)))}), and \code{nodesize}
(minimum node size, default 1 for classification). Default is an empty
list, which uses \pkg{randomForest} defaults.}

\item{threshold_method}{Character or numeric. Method used to convert
continuous predicted probabilities to binary suitability. Accepted values:
\itemize{
  \item \code{"prevalence"}: Sets threshold equal to the prevalence
    (proportion of presences) in the training data for that fold.
  \item \code{"tss"}: Selects the threshold that maximizes the True Skill
    Statistic (sensitivity + specificity - 1) on the training data.
    Default.
  \item A numeric value between 0 and 1 (e.g. \code{0.4}): Uses that
    value as a fixed threshold for all folds directly.
}}

\item{output_dir}{Character. Directory to write output files including saved
model objects and plots. Default is \code{file.path(tempdir(), "RF_Models")}.}

\item{create_plot}{Logical. If \code{TRUE}, generates a per-fold variable
importance plot, partial dependence curves for each predictor, and a
combined ROC curve summary. Default is \code{TRUE}.}

\item{plot_palette}{Character. Name of an HCL or RColorBrewer palette used
to color folds in diagnostic plots. Accepts any HCL palette name (see
\code{\link[grDevices]{hcl.pals}}) or, if \pkg{RColorBrewer} is installed,
any Brewer palette name. Default is \code{"Dark 2"}.}

\item{overwrite}{Logical. If \code{TRUE}, overwrites existing saved model
files. If \code{FALSE}, loads existing files when available. Default is
\code{FALSE}.}

\item{time_cols}{Character. Name of the column(s) containing year or time
step values in the occurrence data. Must match \code{time_cols} used in
\code{\link{spatiotemporal_partition}}. Default is \code{NULL}.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress
messages during processing. Includes per-fold training summaries and
file-saved messages.}
}
\value{
A list with class \code{"TemporalRF"} containing:
\itemize{
  \item \code{models}: Named list of fitted \code{randomForest} objects,
    one per fold.
  \item \code{thresholds}: Named numeric vector of probability thresholds
    used for binary classification, one per fold.
  \item \code{threshold_method}: Character string recording the thresholding
    method used.
  \item \code{model_vars}: Character vector of predictor names used.
  \item \code{variable_importance}: Named list of importance data frames,
    one per fold, with mean decrease in accuracy for each predictor.
  \item \code{fold_training_data}: Named list of training data frames used
    to fit each fold model, retained for downstream prediction.
  \item \code{fold_test_metrics}: Data frame of held-out test fold metrics
    per fold: \code{Threshold}, \code{AUC}, \code{TSS}, \code{Kappa},
    \code{Sensitivity}, and \code{Specificity}. Also written to
    \code{Fold_Test_Metrics.csv} in \code{output_dir}.
  \item \code{output_dir}: Path to the output directory.
  \item \code{model_type}: Character string \code{"rf"}, used by
    \code{\link{generate_spatiotemporal_predictions}}.
  \item \code{plots}: Named list of recorded plot objects when
    \code{create_plot = TRUE}. Plots can be replayed with
    \code{grDevices::replayPlot()}.
}
}
\description{
Modeling function that constructs Random Forest classification models for
each cross-validation fold using presence and pseudoabsence data. Each model
reserves one fold as testing data and uses the remaining folds as training
data. The user specifies predictors as a character vector.
Predicted probabilities of presence are extracted from the
out-of-bag or in-bag vote fractions and thresholded to produce binary
suitability classifications. Variable importance is recorded for each fold.
The returned object follows the same
structure as \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
and \code{\link{build_temporal_hv}}, and is accepted directly by
\code{\link{generate_spatiotemporal_predictions}}.
}
\details{
Random Forests are fit using \code{\link[randomForest]{randomForest}} from
the \pkg{randomForest} package. The response is treated as a factor
(\code{0}/\code{1}) so the model runs in classification mode, which produces
class vote fractions used as predicted probabilities. Importance is computed
with \code{importance = TRUE} and \code{type = 1} (mean decrease in
accuracy).

Predicted probabilities are the vote fraction for class \code{1} from
\code{predict(..., type = "prob")[, "1"]}. These are used for threshold
selection and ROC curve construction.

Diagnostic plots include: a variable importance bar chart (mean decrease in
accuracy across folds), partial dependence curves for each predictor showing
the marginal effect of each variable while averaging over all others (with
rug marks for presences and pseudoabsences), and a combined ROC curve panel.

The returned object is recognized by
\code{\link{generate_spatiotemporal_predictions}}, which uses the
\code{model_type} field to use the correct prediction and evaluation
logic.
}
\examples{
data(tmr_partition, package = "TemporalModelR")

data(tmr_absences,  package = "TemporalModelR")

build_temporal_rf(
  partition_result     = tmr_partition,
  pseudoabsence_result = tmr_absences,
  model_vars           = c("elevation", "forest_cover", "prseas"),
  rf_params            = list(ntree = 100),
  threshold_method     = "tss",
  output_dir           = tempdir(),
  create_plot          = FALSE,
  time_cols            = c("year", "season"),
  verbose              = FALSE
)
}
\seealso{
Preprocessing: \code{\link{spatiotemporal_partition}},
  \code{\link{generate_absences}}

Modeling: \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
  \code{\link{build_temporal_hv}},
  \code{\link{generate_spatiotemporal_predictions}}

External: \code{\link[randomForest]{randomForest}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_documentation.R
\docType{data}
\name{extdata}
\alias{extdata}
\title{Bundled rasters, point files, and prediction outputs}
\description{
Several non-R objects ship in \code{inst/extdata/} for use in function
examples and the package vignette. They cannot be portably serialized
as \code{.rda} files, so they are stored as GeoTIFF and CSV files and
loaded via \code{\link[base]{system.file}}.
}
\details{
\describe{
  \item{\code{extdata/rasters_raw/}}{Raw synthetic environmental rasters
    before alignment. Includes \code{elevation.tif},
    \code{forest_cover_1.tif} through \code{forest_cover_15.tif},
    \code{pr_ann_1.tif} through \code{pr_ann_15.tif}, and
    \code{prseas_<year>_<season>.tif} for each of 15 years and 4
    seasons (60 files).}
  \item{\code{extdata/rasters_aligned/}}{Same rasters after
    \code{\link{raster_align}} has reprojected and masked them to the
    reference grid.}
  \item{\code{extdata/rasters_scaled/}}{Aligned rasters z-scored using
    the scaling parameters from the seasonal extraction. Used by
    examples and modeling functions that work with the seasonal
    predictor set.}
  \item{\code{extdata/rasters_scaled_annual/}}{Aligned rasters z-scored
    using the scaling parameters from the annual extraction. Used by
    examples and modeling functions that work with the annual
    predictor set (\code{pr_ann} instead of \code{prseas}).}
  \item{\code{extdata/points/synthetic_occurrence_points.csv}}{The raw
    synthetic presence dataset: 150 points with \code{x}, \code{y},
    \code{year}, \code{season}, and \code{pres = 1}.}
  \item{\code{extdata/points/synthetic_occurrence_points.shp}}{Same
    points as a shapefile.}
  \item{\code{extdata/points/synthetic_user_presences.csv}}{A second-species
    presence dataset for demonstrating \code{method = "user_data"} in
    \code{\link{generate_absences}}. Points were derived from buffer-
    constrained environmental pseudoabsences of the primary species and
    reformatted as presences (\code{pres = 1}). Same column structure as
    \code{synthetic_occurrence_points.csv} (\code{x}, \code{y},
    \code{year}, \code{season}, \code{pres}).}
  \item{\code{extdata/points/extracted_seasonal_*.csv}}{Outputs from
    \code{\link{temporally_explicit_extraction}} using the seasonal
    predictor set: \code{_Raw_Values.csv}, \code{_Scaled_Values.csv},
    and \code{_Scaling_Parameters.csv}.}
  \item{\code{extdata/points/extracted_annual_*.csv}}{Same outputs but
    using the annual predictor set.}
  \item{\code{extdata/predictions/}}{Per-timestep fold-vote rasters
    from the seasonal workflow's
    \code{\link{generate_spatiotemporal_predictions}} call. Fifteen
    files (\code{Prediction_<year>_Spring.tif}) suitable for direct
    input to \code{\link{summarize_raster_outputs}}.}
  \item{\code{extdata/binary/consensus_stack.tif}}{Multi-layer
    GeoTIFF of binary suitable / unsuitable rasters produced by
    \code{\link{summarize_raster_outputs}} with \code{consensus = 3}.
    Fifteen layers, one per year, ordered 1 through 15.}
  \item{\code{extdata/binary/frequency_raster.tif}}{Companion
    single-layer raster giving the proportion of years each pixel was
    classified as suitable.}
  \item{\code{extdata/precomputed/}}{Precomputed prediction outputs read
    directly by the modeling vignettes (V3a-V3d) so that
    \code{\link{generate_spatiotemporal_predictions}} does not need to be
    rerun at vignette build time. One subdirectory per model type
    (\code{glm/}, \code{gam/}, \code{rf/}, \code{hv/}), each containing
    \code{preds.rds} and a \code{pred_tifs/} folder. \code{preds.rds} is
    the list returned by
    \code{\link{generate_spatiotemporal_predictions}} for that model
    (\code{$timestep_metrics}, \code{$overall_summary},
    \code{$prediction_files}, and \code{$model_type}), with
    \code{$prediction_files} reduced to bare file names rather than
    absolute build-time paths. \code{pred_tifs/} holds the 60
    per-timestep fold-vote rasters
    (\code{Prediction_<year>_<season>.tif}, 15 years x 4 seasons)
    projected from that model. The vignettes load \code{preds.rds} and
    rebuild \code{$prediction_files} from \code{pred_tifs/} via
    \code{\link[base]{system.file}}.}
}

Example load patterns:
\preformatted{
  ### Aligned raster directory
  aln_dir <- system.file("extdata/rasters_aligned",
                         package = "TemporalModelR")

  ### Consensus stack (multi-layer)
  binary_stack <- terra::rast(system.file(
    "extdata/binary/consensus_stack.tif", package = "TemporalModelR"
  ))

  ### Frequency raster
  frequency_rast <- terra::rast(system.file(
    "extdata/binary/frequency_raster.tif", package = "TemporalModelR"
  ))

  ### Per-timestep prediction directory
  pred_dir <- system.file("extdata/predictions",
                          package = "TemporalModelR")

  ### Precomputed GLM prediction object and its rasters
  glm_preds <- readRDS(system.file(
    "extdata/precomputed/glm/preds.rds", package = "TemporalModelR"
  ))
  glm_pred_files <- list.files(
    system.file("extdata/precomputed/glm/pred_tifs",
                package = "TemporalModelR"),
    pattern = "\\.tif$", full.names = TRUE
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_documentation.R
\docType{data}
\name{tmr_absences_annual}
\alias{tmr_absences_annual}
\title{Pre-built pseudoabsence result (annual workflow)}
\format{
A list as returned by \code{\link{generate_absences}}.
}
\usage{
tmr_absences_annual
}
\description{
Annual variant of \code{tmr_absences}. Buffer method with 300 m buffer
and 2:1 ratio, generated against \code{tmr_partition_annual} with the
annual predictor set (\code{forest_cover}, \code{pr_ann},
\code{elevation}). Loaded with \code{data(tmr_absences_annual)}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_documentation.R
\docType{data}
\name{tmr_partition_small}
\alias{tmr_partition_small}
\title{Pre-built spatiotemporal partition, small version (seasonal workflow)}
\format{
A list with the same structure as \code{\link{tmr_partition}},
  containing \code{$folds}, \code{$points_sf}, \code{$voronoi_blocks},
  \code{$voronoi_folds}, \code{$summary}, and \code{$plots}.
}
\usage{
tmr_partition_small
}
\description{
A subsampled version of \code{\link{tmr_partition}} retaining
approximately half the points per fold, for use in examples and tests
where runtime is a concern. Built by sampling
\code{ceiling(n / 2)} rows from each fold of \code{tmr_partition$points_sf}.
Loaded with \code{data(tmr_partition_small)}.
}
\seealso{
\code{\link{tmr_partition}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarize_raster_outputs.R
\name{summarize_raster_outputs}
\alias{summarize_raster_outputs}
\title{Summarize Prediction Rasters into Consensus Outputs}
\usage{
summarize_raster_outputs(predictions_dir, output_dir = NULL,
                         consensus = 1, file_pattern = "Prediction_.*\\\.tif$",
                         overwrite = FALSE, verbose = TRUE)
}
\arguments{
\item{predictions_dir}{Character. Directory containing prediction raster
files, typically the \code{output_dir} used in
\code{\link{generate_spatiotemporal_predictions}}.}

\item{output_dir}{Character. Directory for output files. Defaults to
\code{predictions_dir} if \code{NULL}.}

\item{consensus}{Integer. Minimum number of folds that must agree on
suitability for a pixel to be classified as suitable in the binary output.
For example, \code{consensus = 1} marks any pixel suitable if at least one
fold predicts it suitable; \code{consensus = 4} requires at least four folds to
agree. Must be between 1 and the maximum vote count in the rasters (number of
folds). Default is \code{1}.}

\item{file_pattern}{Character. Regular expression to match prediction raster
files. Default is \code{"Prediction_.*\\.tif$"}.}

\item{overwrite}{Logical. If \code{TRUE}, overwrites existing output files.
If \code{FALSE} (default), existing files are skipped.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress
messages during processing.}
}
\value{
Invisibly returns a list containing:
\itemize{
  \item \code{consensus_stack}: \code{SpatRaster} stack of per-time-step
    binary consensus rasters (one layer per input file).
  \item \code{frequency_raster}: \code{SpatRaster} showing the proportion of
    time steps during which each pixel met the consensus threshold. Values
    range from 0 (never suitable) to 1 (always suitable).
  \item \code{consensus}: the consensus threshold used.
  \item \code{n_timesteps}: number of time steps processed.
  \item \code{consensus_dir}: path to the directory containing per-time-step
    binary consensus rasters.
  \item \code{frequency_file}: path to the written frequency raster file.
}
}
\description{
Post-processing function that synthesizes per-time-step fold-vote rasters
(output from \code{\link{generate_spatiotemporal_predictions}}) into binary
consensus predictions and a temporal frequency summary. Each input raster
contains integer vote counts per pixel the number of cross-validation
folds that classified that pixel as suitable. The \code{consensus} threshold
controls how many folds must agree for a pixel to be classified as suitable
in the output binary rasters.
}
\details{
Input rasters are fold-vote-count rasters produced by
\code{\link{generate_spatiotemporal_predictions}}, where each pixel value is
the number of cross-validation fold models that predicted suitability at that
location for that time step. The \code{consensus} threshold is applied as:
\code{binary = as.integer(vote_count >= consensus)}.

The frequency raster (proportion of time steps suitable under the chosen
consensus) serves as input to \code{\link[TemporalModelR]{analyze_temporal_patterns}} for
identifying long-term trends in suitability.
}
\examples{
pred_dir <- system.file("extdata/predictions",
                        package = "TemporalModelR")

summarize_raster_outputs(
  predictions_dir = pred_dir,
  output_dir      = tempdir(),
  consensus       = 3,
  overwrite       = TRUE,
  verbose         = FALSE
)
}
\seealso{
Upstream: \code{\link{generate_spatiotemporal_predictions}}

Downstream: \code{\link[TemporalModelR]{analyze_temporal_patterns}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/temporally_explicit_extraction.R
\name{temporally_explicit_extraction}
\alias{temporally_explicit_extraction}
\title{Extract Time-Aligned Environmental Values at Species Occurrences}
\usage{
temporally_explicit_extraction(points_sp, raster_dir, variable_patterns,
                               time_cols, xcol = NULL, ycol = NULL,
                               points_crs = NULL, output_dir,
                               output_prefix = "temp_explicit_df",
                               save_raw = TRUE, save_scaled = TRUE,
                               save_scaling_params = TRUE,
                               verbose = TRUE)
}
\arguments{
\item{points_sp}{sf object, SpatialPointsDataFrame, file path to
.csv/.shp/.geojson/.gpkg, or data frame with coordinate columns.}

\item{raster_dir}{Character. Directory containing environmental raster
files (\code{.tif}), typically the output of \code{\link{raster_align}}.
File names must follow the patterns supplied in \code{variable_patterns}, with any
time placeholder substituted for the corresponding value from
\code{time_cols}.}

\item{variable_patterns}{Named character vector mapping clean variable names
to raster filename patterns. For time-varying variables include the time
placeholder in the pattern (e.g. \code{"forest_cover" = "forest_cover_YEAR"});
for static variables omit it (e.g. \code{"elevation" = "elevation"}). Time
placeholders must match entries in \code{time_cols}.}

\item{time_cols}{Character vector of time column names present in the point
data (e.g., c("YEAR"), c("YEAR", "MONTH")).}

\item{xcol}{Character. Name of the x-coordinate column. Required when
\code{points_sp} is a CSV file or data frame.}

\item{ycol}{Character. Name of the y-coordinate column. Required when
\code{points_sp} is a CSV file or data frame.}

\item{points_crs}{Character or CRS object. CRS of the input points.
Required when \code{points_sp} is a CSV file or data frame.}

\item{output_dir}{Character. Directory to write output files.}

\item{output_prefix}{Character. Prefix for output filenames. Default is
"temp_explicit_df".}

\item{save_raw}{Logical. If \code{TRUE} (default), writes raw extracted values
CSV. If \code{FALSE}, skips raw values output.}

\item{save_scaled}{Logical. If \code{TRUE} (default), writes z-scaled values
CSV. If \code{FALSE}, skips scaled values output.}

\item{save_scaling_params}{Logical. If \code{TRUE} (default), writes CSV of
per-variable means and standard deviations. If \code{FALSE}, skips scaling
parameters output.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress
messages during processing. Includes file loading, extraction
progress, and file-save confirmation.}
}
\value{
Invisibly returns a list containing:
\itemize{
  \item \code{raw_values}: Data frame of raw extracted values at each
    occurrence record (when \code{save_raw = TRUE}; \code{NULL} otherwise).
  \item \code{scaled_values}: Data frame of z-scaled extracted values
    (when \code{save_scaled = TRUE}; \code{NULL} otherwise).
  \item \code{scaling_params}: Data frame of per-variable means and standard
    deviations used for scaling (when \code{save_scaling_params = TRUE};
    \code{NULL} otherwise). Pass this to \code{\link{scale_rasters}}.
  \item \code{files_created}: Named list of file paths written, with elements
    \code{raw}, \code{scaled}, and \code{scaling_params} (each \code{NULL}
    when the corresponding save flag is \code{FALSE}).
}
}
\description{
Preprocessing function that extracts raster values to species occurrence
records based on temporal components. Matches environmental layers to
occurrence timestamps and optionally computes scaling parameters for standardization.
}
\details{
Extracts raster values to species occurrence records based on matched temporal components.
Matches environmental layers to occurrence timestamps and optionally computes scaling
parameters for standardization.

Output CSV files are written to output_dir containing raw values, scaled
values, and scaling parameters.

Scaling parameters (mean and standard deviation) are optionally computed across
all occurrence records for each variable. These parameters should be used with
\code{\link{scale_rasters}} to standardize prediction layers.
}
\examples{
pts_file <- system.file(
  "extdata/points/synthetic_occurrence_points.csv",
  package = "TemporalModelR"
)

aln_dir  <- system.file("extdata/rasters_aligned",
                        package = "TemporalModelR")

ref_file <- system.file("extdata/rasters_raw/elevation.tif",
                        package = "TemporalModelR")

out_dir  <- file.path(tempdir(), "extracted")

temporally_explicit_extraction(
  points_sp           = pts_file,
  raster_dir          = aln_dir,
  variable_patterns   = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR",
    "prseas"       = "prseas_YEAR_SEASON"
  ),
  time_cols           = c("year", "season"),
  xcol                = "x",
  ycol                = "y",
  points_crs          = terra::crs(terra::rast(ref_file)),
  output_dir          = out_dir,
  save_raw            = TRUE,
  save_scaled         = FALSE,
  save_scaling_params = TRUE,
  verbose             = FALSE
)
}
\seealso{
Preprocessing: \code{\link{spatiotemporal_rarefaction}},
  \code{\link{scale_rasters}}, \code{\link{spatiotemporal_partition}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_temporal_gam.R
\name{build_temporal_gam}
\alias{build_temporal_gam}
\title{Build Temporal GAM Models Across Cross-Validation Folds}
\usage{
build_temporal_gam(partition_result, pseudoabsence_result, model_formula,
                   link = "logit", gam_params = list(method = "REML"),
                   threshold_method  = "tss",
                   output_dir = file.path(tempdir(), "GAM_Models"),
                   create_plot = TRUE, plot_palette = "Dark 2",
                   overwrite = FALSE, time_cols = NULL, verbose = TRUE)
}
\arguments{
\item{partition_result}{List or character. Output from
\code{\link{spatiotemporal_partition}} or path to an \code{.rds}
file containing that output.}

\item{pseudoabsence_result}{List or character. Output from
\code{\link{generate_absences}} or path to an \code{.rds} file
containing that output.}

\item{model_formula}{Formula or character. The right-hand side of the model
formula supplied as either a formula object or a character string. The
response variable (\code{presence}) is always added automatically on the
left-hand side, so only the right-hand side needs to be provided. Both of
the following are accepted and equivalent:
\itemize{
  \item \code{~ s(Var1) + s(Var2) + Var3}
  \item \code{"~ s(Var1) + s(Var2) + Var3"}
}
Standard \pkg{mgcv} formula syntax applies. Smooth terms are specified
with \code{s()} for univariate smooths, \code{te()} for tensor product
smooths of two or more variables, and \code{ti()} for tensor product
interaction terms. Parametric terms can be included alongside smooth terms
using \code{+}. The basis type and dimension can be controlled via
arguments to \code{s()}, e.g. \code{s(Var1, k = 5, bs = "tp")}. All
predictor names referenced in the formula must be present as columns in
both the presence and pseudoabsence data.}

\item{link}{Character. The link function for the binomial GAM. One of
\code{"logit"} (default), \code{"probit"}, \code{"cloglog"}, or
\code{"cauchit"}. See \code{\link[stats]{binomial}} for details on each
link function.}

\item{gam_params}{Named list. Additional arguments passed to
\code{\link[mgcv]{gam}}, such as \code{method} for the smoothing
parameter estimation method (e.g. \code{"REML"}) or
\code{select} for additional shrinkage. Default is
\code{list(method = "REML")}.}

\item{threshold_method}{Character or numeric. Method used to convert
continuous predicted probabilities to binary suitability. Accepted values:
\itemize{
  \item \code{"prevalence"}: Sets threshold equal to the prevalence
    (proportion of presences) in the training data for that fold.
  \item \code{"tss"}: Selects the threshold that maximizes the True Skill
    Statistic (sensitivity + specificity - 1) on the training data.
    Default.
  \item A numeric value between 0 and 1 (e.g. \code{0.4}): Uses that
    value as a fixed threshold for all folds directly.
}}

\item{output_dir}{Character. Directory to write output files including saved
model objects and plots. Default is \code{file.path(tempdir(), "GAM_Models")}.}

\item{create_plot}{Logical. If \code{TRUE}, generates per-fold response
curve plots and a combined ROC curve summary. Default is \code{TRUE}.}

\item{plot_palette}{Character. Name of an HCL or RColorBrewer palette used
to color folds in diagnostic plots. Accepts any HCL palette name (see
\code{\link[grDevices]{hcl.pals}}) or, if \pkg{RColorBrewer} is installed,
any Brewer palette name. Default is \code{"Dark 2"}.}

\item{overwrite}{Logical. If \code{TRUE}, overwrites existing saved model
files. If \code{FALSE}, loads existing files when available. Default is
\code{FALSE}.}

\item{time_cols}{Character. Name of the column(s) containing year or time
step values in the occurrence data. Must match \code{time_cols} used in
\code{\link{spatiotemporal_partition}}. Default is \code{NULL}.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress
messages during processing. Includes per-fold training summaries and
file-saved messages. The completion summary and metrics table are
always printed regardless of this setting.}
}
\value{
A list with class \code{"TemporalGAM"} containing:
\itemize{
  \item \code{models}: Named list of fitted \code{gam} objects, one per
    fold.
  \item \code{thresholds}: Named numeric vector of probability thresholds
    used for binary classification, one per fold.
  \item \code{threshold_method}: Character string recording the thresholding
    method used.
  \item \code{model_formula}: The formula object as passed to the fitting
    function.
  \item \code{link}: Character string recording the link function used.
  \item \code{model_vars}: Character vector of predictor names extracted
    from the formula right-hand side.
  \item \code{fold_training_data}: Named list of training data frames used
    to fit each fold model, retained for downstream prediction.
  \item \code{fold_test_metrics}: Data frame of held-out test fold metrics
    per fold: \code{Threshold}, \code{AUC}, \code{TSS}, \code{Kappa},
    \code{Sensitivity}, and \code{Specificity}. Also written to
    \code{Fold_Test_Metrics.csv} in \code{output_dir}.
  \item \code{output_dir}: Path to the output directory.
  \item \code{model_type}: Character string \code{"gam"}, used by
    \code{\link{generate_spatiotemporal_predictions}}.
  \item \code{plots}: Named list of recorded plot objects when
    \code{create_plot = TRUE}. Plots can be replayed with
    \code{grDevices::replayPlot()}.
}
}
\description{
Modeling function that constructs binomial generalized additive models (GAMs)
for each cross-validation fold using presence and pseudoabsence data. Each
model reserves one fold as testing data and uses the remaining folds as
training data. The user supplies the model formula directly using standard
\pkg{mgcv} formula syntax, including smooth terms such as \code{s()},
\code{te()}, and \code{ti()}. Supports automatic or manual probability
thresholding for converting continuous predictions to binary suitability
classifications necessary for downstream analyses. The returned object follows the same
structure as \code{\link{build_temporal_glm}}, \code{\link{build_temporal_hv}},
and \code{\link{build_temporal_rf}}, and is accepted directly by
\code{\link{generate_spatiotemporal_predictions}}.
}
\details{
GAMs are fit using \code{\link[mgcv]{gam}} from the \pkg{mgcv} package with
\code{family = binomial(link = link)}. Smooth terms default to thin plate
regression splines (\code{bs = "tp"}) with the basis dimension \code{k}
chosen automatically by \pkg{mgcv} unless specified in the formula. Smoothing
parameters are estimated by REML by default.

The returned object is recognized by
\code{\link{generate_spatiotemporal_predictions}}, which uses the
\code{model_type} field to use the correct prediction and evaluation
logic.
}
\examples{
data(tmr_partition, package = "TemporalModelR")

data(tmr_absences,  package = "TemporalModelR")

build_temporal_gam(
  partition_result     = tmr_partition,
  pseudoabsence_result = tmr_absences,
  model_formula        = ~ s(elevation) + s(forest_cover) + s(prseas),
  threshold_method     = "tss",
  output_dir           = tempdir(),
  create_plot          = FALSE,
  time_cols            = c("year", "season"),
  verbose              = FALSE
)
}
\seealso{
Preprocessing: \code{\link{spatiotemporal_partition}},
  \code{\link{generate_absences}}

Modeling: \code{\link{build_temporal_glm}}, \code{\link{build_temporal_rf}},
  \code{\link{build_temporal_hv}},
  \code{\link{generate_spatiotemporal_predictions}}

External: \code{\link[mgcv]{gam}}, \code{\link[mgcv]{s}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyze_temporal_patterns.R
\name{analyze_temporal_patterns}
\alias{analyze_temporal_patterns}
\title{Analyze Temporal Patterns in Binary Raster Time Series}
\usage{
analyze_temporal_patterns(
  binary_stack,
  summary_raster,
  time_steps,
  fastcpd_params = list(),
  output_dir = NULL,
  n_tiles_x = 1,
  n_tiles_y = 1,
  alpha = 0.05,
  spatial_autocorrelation = TRUE,
  verbose = TRUE,
  estimate_time = TRUE,
  overwrite = FALSE
)
}
\arguments{
\item{binary_stack}{RasterStack, RasterBrick, or character. Stack of binary
raster layers across time, or path to directory containing binary rasters.
Typically from \code{\link{summarize_raster_outputs}}.}

\item{summary_raster}{RasterLayer. Per-pixel proportion of time periods where
a pixel is suitable. From \code{\link{summarize_raster_outputs}}.}

\item{time_steps}{Integer vector. Time labels corresponding to raster layers
(same length as number of layers).}

\item{fastcpd_params}{List. Named list of parameters passed to fastcpd
changepoint detection function. Default is empty list. Supports parameterization
from \code{\link[fastcpd]{fastcpd_binomial}}.}

\item{output_dir}{Character. Optional. Output directory for pattern rasters.
When \code{NULL} (default), rasters are written to temporary files and not
saved persistently. Provide a path to write named output files to disk.}

\item{n_tiles_x}{Integer. Number of tiles in the x direction for tiled
processing. Default is \code{1}. Increase to reduce peak memory use for
large rasters and prevent crashes.}

\item{n_tiles_y}{Integer. Number of tiles in the y direction for tiled
processing. Default is \code{1}. Increase to reduce peak memory use for
large rasters and prevent crashes.}

\item{alpha}{Numeric. Significance level for changepoint detection. Default
is \code{0.05}.}

\item{spatial_autocorrelation}{Logical. If \code{TRUE} (default), includes a
neighbor variable in the changepoint analysis to account for spatial
autocorrelation.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress
messages during processing.}

\item{estimate_time}{Logical. If \code{TRUE} (default), estimates runtime from
a sample of pixels before full processing begins. If \code{FALSE}, proceeds
directly to processing without a time estimate.}

\item{overwrite}{Logical. If \code{TRUE}, overwrites existing output files.
If \code{FALSE} (default), existing files are skipped.}
}
\value{
Invisibly returns a list containing:
\itemize{
  \item \code{pattern}: \code{SpatRaster} classifying pixels as integer
    values 1-6, corresponding to "Never Suitable", "Always Suitable",
    "No Pattern", "Increasing Suitability", "Decreasing Suitability",
    or "Fluctuating".
  \item \code{time_decrease}: \code{SpatRaster} showing the time step of
    first significant decrease for pixels classified as decreasing.
  \item \code{time_increase}: \code{SpatRaster} showing the time step of
    first significant increase for pixels classified as increasing.
}
}
\description{
Postprocessing function that applies changepoint detection methods to
identify temporal trends in habitat suitability across consecutive
predictions. Classifies pixels as stable, increasing in quality, or
decreasing in quality, and identifies time periods of significant change.
}
\details{
Applies changepoint detection using fastcpd to identify significant temporal
shifts in spatial suitability. Accounts for spatial and temporal
autocorrelation when \code{spatial_autocorrelation = TRUE}. The
\code{fastcpd_params} list allows customization of the changepoint detection
algorithm, and is passed through to \code{\link[fastcpd]{fastcpd.binomial}}
(e.g. \code{list(cost_adjustment = "BIC", trim = 0.025)}).

Pattern classifications enable identification of expanding, contracting, or
stable g-space distributions over time or site level assessments of directional
change in suitability.

Classification assumes consecutive rasters. Time periods
shorter than ~15 time steps may be too short to classify increases or decreases.
}
\examples{
\donttest{
  con_file <- system.file("extdata/binary/consensus_stack.tif",
                          package = "TemporalModelR")

  frq_file <- system.file("extdata/binary/frequency_raster.tif",
                          package = "TemporalModelR")

  binary_stack   <- terra::rast(con_file)

  summary_raster <- terra::rast(frq_file)

  time_steps <- expand.grid(
    year             = 1:15,
    season           = "Spring",
    stringsAsFactors = FALSE
  )

  analyze_temporal_patterns(
    binary_stack            = binary_stack,
    summary_raster          = summary_raster,
    time_steps              = time_steps,
    output_dir              = tempdir(),
    spatial_autocorrelation = FALSE,
    overwrite               = TRUE,
    estimate_time           = FALSE,
    verbose                 = FALSE
  )
}
}
\seealso{
Postprocessing: \code{\link{summarize_raster_outputs}}

External: \code{\link[fastcpd]{fastcpd}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_documentation.R
\docType{data}
\name{tmr_glm}
\alias{tmr_glm}
\title{Pre-built GLM result (seasonal workflow)}
\format{
A list of class \code{"TemporalGLM"} as returned by
  \code{\link{build_temporal_glm}}, containing \code{$models},
  \code{$thresholds}, \code{$threshold_method}, \code{$model_formula},
  \code{$link}, \code{$model_vars}, \code{$fold_training_data},
  \code{$fold_test_metrics}, \code{$output_dir}, \code{$model_type},
  and \code{$plots}.
}
\usage{
tmr_glm
}
\description{
Output from \code{\link{build_temporal_glm}} fit to \code{tmr_partition}
and \code{tmr_absences} with the formula
\code{~ forest_cover + prseas + elevation}, a logit link, and TSS-based
threshold selection. One model per fold. Loaded with
\code{data(tmr_glm)}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster_align.R
\name{raster_align}
\alias{raster_align}
\title{Align and Standardize Raster Files to a Reference Raster}
\usage{
raster_align(input_dir, output_dir, reference_raster,
             output_suffix = "_Masked_Updated", pattern = ".*\\\\.tif$",
             resample_method = "bilinear",
             overwrite = FALSE, verbose = TRUE)
}
\arguments{
\item{input_dir}{Character. Directory containing the input raster files.}

\item{output_dir}{Character. Directory where processed rasters will be saved.}

\item{reference_raster}{Character, SpatRaster, or RasterLayer. File path or
raster object used as the alignment reference.}

\item{output_suffix}{Character. Suffix appended to output filenames. Default
is \code{"_Masked_Updated"}. Output files are named
\code{<original_name><output_suffix>.tif}.}

\item{pattern}{Character. Regular expression used to match raster files
within \code{input_dir}. Default is \code{".*\\.tif$"}.}

\item{resample_method}{Character. Resampling method passed to
\code{\link[terra]{resample}}. Default is \code{"bilinear"}, which is
appropriate for continuous variables. Use \code{"near"} for categorical
rasters (e.g. land cover classes) to avoid interpolating between class
codes. Other accepted values include \code{"cubic"} and \code{"lanczos"}.}

\item{overwrite}{Logical. If \code{TRUE}, overwrites existing output files.
If \code{FALSE} (default), existing files are skipped.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress
messages during processing. Includes file counts, overwrite mode,
per-file progress, and a completion summary.}
}
\value{
Invisibly returns a list containing:
\itemize{
  \item \code{output_files}: Character vector of file paths written to
    \code{output_dir}.
  \item \code{n_processed}: Integer. Number of rasters successfully processed
    in this call (excludes skipped files when \code{overwrite = FALSE}).
}
}
\description{
Preprocessing function that aligns a batch of raster files to a specified
reference raster by performing reprojection, resampling, and masking
operations. Ensures all rasters share identical CRS, resolution, and spatial
extent for later analyses.
}
\details{
For each raster in \code{input_dir} matching \code{pattern}, the function:
\enumerate{
  \item Reprojects to the CRS of the reference raster
  \item Resamples to match the reference resolution using \code{resample_method}
  \item Masks values outside the reference raster's non-NA extent
  \item Saves the result as a GeoTIFF to \code{output_dir}
}

This preprocessing step ensures spatial alignment before applying other downstream
analyses such as \code{\link{temporally_explicit_extraction}} or \code{\link{scale_rasters}}.
}
\examples{
raw_dir <- system.file("extdata/rasters_raw", package = "TemporalModelR")

ref     <- file.path(raw_dir, "elevation.tif")

out_dir <- file.path(tempdir(), "aligned")

raster_align(
  input_dir        = raw_dir,
  output_dir       = out_dir,
  reference_raster = ref,
  overwrite        = TRUE,
  verbose          = FALSE
)
}
\seealso{
Preprocessing: \code{\link{scale_rasters}},
  \code{\link{temporally_explicit_extraction}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_documentation.R
\docType{data}
\name{tmr_predictions}
\alias{tmr_predictions}
\title{Pre-built spatiotemporal predictions (seasonal workflow)}
\format{
A list as returned by
  \code{\link{generate_spatiotemporal_predictions}}, containing
  \code{$timestep_metrics}, \code{$overall_summary},
  \code{$prediction_files}, and \code{$model_type}.
}
\usage{
tmr_predictions
}
\description{
Output from \code{\link{generate_spatiotemporal_predictions}} projecting
\code{tmr_glm} across 15 years for the Spring season only (one
prediction layer per year-season combination, 15 total). Loaded with
\code{data(tmr_predictions)}.
}
\details{
Note that \code{$prediction_files} stores the absolute paths used at
build time. To work with the per-timestep rasters on a user machine,
regenerate them via \code{\link{generate_spatiotemporal_predictions}}
or use the bundled prediction set in \code{inst/extdata/predictions/}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_documentation.R
\docType{data}
\name{tmr_predictions_annual}
\alias{tmr_predictions_annual}
\title{Pre-built spatiotemporal predictions (annual workflow)}
\format{
A list as returned by
  \code{\link{generate_spatiotemporal_predictions}}.
}
\usage{
tmr_predictions_annual
}
\description{
Annual variant of \code{tmr_predictions}. Projects \code{tmr_glm_annual}
across 15 years on a single (year-only) time axis. Loaded with
\code{data(tmr_predictions_annual)}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_documentation.R
\docType{data}
\name{tmr_partition}
\alias{tmr_partition}
\title{Pre-built spatiotemporal partition (seasonal workflow)}
\format{
A list as returned by \code{\link{spatiotemporal_partition}},
  containing \code{$folds}, \code{$points_sf}, \code{$voronoi_blocks},
  \code{$voronoi_folds}, \code{$summary}, and \code{$plots}.
}
\usage{
tmr_partition
}
\description{
Output from \code{\link{spatiotemporal_partition}} run on the synthetic
occurrence dataset bundled in \code{inst/extdata/}. Built with
\code{n_spatial_folds = 2} and \code{n_temporal_folds = 2}, producing
four cross-validation folds (2 spatial + 2 temporal). Points span 15
years and the seasons Spring, Summer, and Autumn on a 15 x 30 cell
study area (3000 m x 1500 m, 100 m pixels) in a custom synthetic local
CRS. Predictors attached to each point are \code{forest_cover},
\code{prseas} (seasonal precipitation), and \code{elevation}, all
z-scored. Loaded with \code{data(tmr_partition)}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_temporal_hv.R
\name{build_temporal_hv}
\alias{build_temporal_hv}
\title{Build Temporal Hypervolume Models Across Cross-Validation Folds}
\usage{
build_temporal_hv(partition_result, model_vars, method,
                  hypervolume_params = list(),
                  output_dir = file.path(tempdir(), "Hypervolume_Models"),
                  create_plot = TRUE, overwrite = FALSE,
                  plot_palette = "Dark 2", verbose = TRUE)
}
\arguments{
\item{partition_result}{List or character. Output from
\code{\link{spatiotemporal_partition}} or path to an \code{.rds}
file containing that output.}

\item{model_vars}{Character vector. Names of predictor columns to use in
hypervolume construction. All variables must be present as columns in
the occurrence data produced by \code{\link{temporally_explicit_extraction}}.}

\item{method}{Character. Hypervolume method. One of \code{"gaussian"} for
Gaussian kernel density estimation or \code{"svm"} for one-class support
vector machine. Required.}

\item{hypervolume_params}{Named list. Additional parameters passed to
\code{\link[hypervolume]{hypervolume_gaussian}} or
\code{\link[hypervolume]{hypervolume_svm}}. For Gaussian models, valid
keys are \code{kde.bandwidth}, \code{quantile.requested},
\code{quantile.requested.type}, \code{chunk.size}, \code{verbose}, and
\code{samples.per.point}. For SVM models, valid keys are \code{svm.nu},
\code{svm.gamma}, \code{chunk.size}, \code{verbose}, and
\code{samples.per.point}. Default is an empty list, which uses the
built-in defaults.}

\item{output_dir}{Character. Directory to write output files including saved
hypervolume objects and plots. Default is \code{file.path(tempdir(), "Hypervolume_Models")}.}

\item{create_plot}{Logical. If \code{TRUE}, generates pairplot visualisations
of each fold's hypervolume and a combined comparison plot. Default is
\code{TRUE}.}

\item{overwrite}{Logical. If \code{TRUE}, overwrites existing saved
hypervolume files. If \code{FALSE}, loads existing files when available.
Default is \code{FALSE}.}

\item{plot_palette}{Character. Name of an HCL or RColorBrewer palette used
to color folds in diagnostic plots. Accepts any HCL palette name (see
\code{\link[grDevices]{hcl.pals}}) or, if \pkg{RColorBrewer} is installed,
any Brewer palette name. Default is \code{"Dark 2"}.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress
messages during processing. Includes per-fold hypervolume construction
progress, volume summaries, and overlap statistics.}
}
\value{
A list with class \code{"TemporalHypervolume"} containing:
\itemize{
  \item \code{models}: Named list of fitted \code{Hypervolume} objects, one
    per fold, named \code{fold1}, \code{fold2}, etc. This naming convention
    matches the other \code{build_temporal_*()} functions and is required by
    \code{\link{generate_spatiotemporal_predictions}}.
  \item \code{volumes}: Named numeric vector of hypervolume sizes (units
    depend on the number of dimensions and bandwidth).
  \item \code{overlaps}: Named list of pairwise percent volume overlaps
    between all fold combinations.
  \item \code{method}: Character string recording the method used.
  \item \code{model_vars}: Character vector of predictor names used.
  \item \code{fold_training_data}: Named list of training data frames used
    to fit each fold model, retained for consistency with the other model
    types and for downstream use.
  \item \code{fold_test_metrics}: Data frame of E-space inclusion metrics
    computed on the held-out test points for each fold. Columns: \code{fold},
    \code{n_test}, \code{volume}, \code{tp}, \code{fn}, \code{sensitivity},
    \code{Sensitivity}. Printed as a summary table at the end of
    model building.
  \item \code{output_dir}: Path to the output directory.
  \item \code{model_type}: Character string \code{"hypervolume"}, used by
    \code{\link{generate_spatiotemporal_predictions}}.
  \item \code{plots}: List of recorded plot objects (if
    \code{create_plot = TRUE}). Plots can be replayed with
    \code{grDevices::replayPlot()}.
}
}
\description{
Modeling function that constructs hypervolume models for each
cross-validation fold using either Gaussian kernel density estimation or
one-class SVM. Each hypervolume reserves one fold as testing data and uses
the remaining folds as training data. The returned object follows the same
structure as \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
and \code{\link{build_temporal_rf}}, and is accepted directly by
\code{\link{generate_spatiotemporal_predictions}}.
}
\details{
For N folds, constructs N hypervolumes where each fold reserves one group
of points for testing and uses the remaining N-1 groups for training.
Pairwise overlap statistics quantify similarity between hypervolumes across
folds, with low overlap indicating that the environmental space sampled
differs substantially across folds.

Hypervolume objects are saved as a combined \code{.rds} file in
\code{output_dir}. If the file already exists and \code{overwrite = FALSE},
the saved file is loaded rather than re-fitting.

The returned object is accepted directly by
\code{\link{generate_spatiotemporal_predictions}}, which uses the
\code{model_type} field to dispatch hypervolume-specific projection logic
via \code{\link[hypervolume]{hypervolume_project}}.
}
\examples{
data(tmr_partition_small, package = "TemporalModelR")

build_temporal_hv(
  partition_result = tmr_partition_small,
  model_vars       = c("elevation", "forest_cover"),
  method           = "svm",
  output_dir       = tempdir(),
  create_plot      = FALSE,
  verbose          = FALSE
)
}
\seealso{
Preprocessing: \code{\link{spatiotemporal_partition}}

Modeling: \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
  \code{\link{build_temporal_rf}},
  \code{\link{generate_spatiotemporal_predictions}}

External: \code{\link[hypervolume]{hypervolume_gaussian}},
  \code{\link[hypervolume]{hypervolume_svm}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_absences.R
\name{generate_absences}
\alias{generate_absences}
\title{Generate Temporally Explicit Pseudoabsence Points}
\usage{
generate_absences(partition_result, reference_shapefile_path, raster_dir,
                  variable_patterns, method = "random", ratio = 1,
                  buffer_distance = NULL, env_percentile = 0.05,
                  time_cols = NULL, pseudoabsence_times = NULL,
                  min_points_per_timestep = 1, user_absence_data = NULL,
                  xcol = NULL, ycol = NULL, points_crs = NULL,
                  create_plot = TRUE, plot_by_fold = FALSE,
                  plot_palette = "Dark 2", output_file = NULL,
                  verbose = TRUE)
}
\arguments{
\item{partition_result}{List or character. Output from
\code{\link{spatiotemporal_partition}} or path to an \code{.rds}
file containing that output.}

\item{reference_shapefile_path}{Character or sf object. Path to a polygon
file or an \code{sf} polygon object defining the study area.}

\item{raster_dir}{Character. Directory containing environmental raster
files (\code{.tif}), typically the output of
\code{\link{raster_align}} or \code{\link{scale_rasters}}. File names
must follow the patterns supplied in \code{variable_patterns}, with any
time placeholder substituted for the corresponding value from
\code{time_cols}. Required for all methods.}

\item{variable_patterns}{Named character vector mapping clean variable names
to raster filename patterns. For time-varying variables include the time
placeholder in the pattern (e.g. \code{"forest_cover" = "forest_cover_YEAR"});
for static variables omit it (e.g. \code{"elevation" = "elevation"}). Time
placeholders must match entries in \code{time_cols}.}

\item{method}{Character. Pseudoabsence generation method. One of
\code{"random"}, \code{"buffer"}, \code{"environmental"}, or
\code{"user_data"}. Default is \code{"random"}. When \code{"user_data"}
is specified, \code{user_absence_data} is used as the source of absence
locations instead of generating them. \code{ratio}, \code{buffer_distance},
\code{env_percentile}, and \code{pseudoabsence_times} are ignored.
\code{raster_dir} and \code{variable_patterns} are required for all
methods, including \code{"user_data"}, for temporally-matched environmental
extraction at the absence points.}

\item{ratio}{Numeric. Number of pseudoabsence points to generate per
presence point. Default is \code{1}. Values of 2, 10, 50, etc. are
accepted. Points are always distributed proportionally across time steps
within each fold. Set to \code{0} to disable proportional allocation and
use a fixed number of points per time step instead, in which case
\code{min_points_per_timestep} must be greater than 0. \code{ratio} and
\code{min_points_per_timestep} cannot both be 0.}

\item{buffer_distance}{Numeric. Distance in the units of the CRS (typically
meters for projected CRS) within which pseudoabsence points are sampled.
Required when \code{method = "buffer"}. When \code{method =
"environmental"}, supplying a value automatically applies a spatial buffer
constraint before environmental profiling, following the three-step approach
of Senay et al. (2013). If \code{NULL} for the environmental method, no
spatial constraint is applied. Default is \code{NULL}.}

\item{env_percentile}{Numeric between 0 and 1. Quantile threshold used to
define the boundary of the known environmental tolerance when
\code{method = "environmental"}. Environmental cells within this quantile
range across all variables are excluded from pseudoabsence sampling.
Default is \code{0.05} (5th to 95th percentile envelope).}

\item{time_cols}{Character or character vector. Name of the column(s)
containing the time step values. Must match \code{time_cols} used in
\code{\link{spatiotemporal_partition}} and the time placeholders used
in \code{variable_patterns}. Default is \code{NULL}.}

\item{pseudoabsence_times}{Vector. Optional vector of specific time step
values (for the first time column) at which to generate pseudoabsences.
When \code{NULL} (default), all time steps present in the occurrence data
are used.}

\item{min_points_per_timestep}{Integer. Minimum number of pseudoabsence
points to generate per time step per fold. Default is \code{1}. When
\code{ratio = 0}, this value sets the exact (fixed) number of points
generated per time step per fold, independent of the number of
presence points. \code{ratio} and \code{min_points_per_timestep} cannot
both be 0.}

\item{user_absence_data}{Character, sf object, sfc object, Spatial object,
or data frame. Path to occurrence data (\code{.csv}, \code{.shp},
\code{.geojson}, \code{.gpkg}) or a spatial object to be processed as absence
data for downstream operations. Required when \code{method = "user_data"}.
Should be preprocessed with \code{\link{spatiotemporal_rarefaction}} in the
same formatting as presence data if not already thinned.}

\item{xcol}{Character. Name of the x-coordinate column in \code{user_absence_data}.
Required when when \code{method = "environmental"} and \code{user_absence_data}
is a CSV file or data frame.}

\item{ycol}{Character. Name of the y-coordinate column in \code{user_absence_data}.
Required when when \code{method = "environmental"} and \code{user_absence_data}
is a CSV file or data frame.}

\item{points_crs}{Character or CRS object. CRS of the \code{user_absence_data}.
Required when when \code{method = "environmental"} and
\code{user_absence_data} is a CSV file or data frame.}

\item{create_plot}{Logical. If \code{TRUE} (default), generates diagnostic
plots showing the spatial and temporal distribution of generated
pseudoabsence points alongside presence points.}

\item{plot_by_fold}{Logical. If \code{TRUE}, generates one map per fold. If
\code{FALSE} (default), generates a single combined map.}

\item{plot_palette}{Character. Name of an HCL or RColorBrewer palette used
to color folds in diagnostic plots. Accepts any HCL palette name (see
\code{\link[grDevices]{hcl.pals}}) or, if \pkg{RColorBrewer} is installed,
any Brewer palette name. Default is \code{"Dark 2"}.}

\item{output_file}{Character. Optional path to save the result as an
\code{.rds} file. The parent directory will be created if it does not
exist. Default is \code{NULL}.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress
messages during processing. Includes per-fold and per-time-step
pseudoabsence counts.}
}
\value{
Invisibly returns a list containing:
\itemize{
  \item \code{pseudoabsences}: An sf object of all generated pseudoabsence
    points with columns \code{fold}, \code{temporal_block}, \code{presence}
    (always 0), the time column(s) if provided, and extracted environmental
    variable values matched to each point's time step.
  \item \code{plots}: A named list of recorded plot objects when
    \code{create_plot = TRUE}. Contains \code{temporal_distribution} and
    either \code{spatial_combined} or one \code{spatial_fold_N} entry per
    fold. Plots can be replayed with \code{grDevices::replayPlot()}.
  \item \code{summary}: A data frame summarising points generated per fold
    with columns \code{fold}, \code{n_presences}, \code{n_pseudoabsences},
    and \code{ratio_achieved}.
}
}
\description{
Generates pseudoabsence or background points for each fold produced by
\code{\link{spatiotemporal_partition}}, distributed across time steps
proportionally to the number of presence points in each time step within each
fold. Three generation methods are supported: random sampling within the
study area, buffer-constrained sampling around presence points, and
environmentally biased sampling that targets areas outside the known
environmental tolerance of the species. Additionally, this function can be used
to process user defined absences to work with downflow operations. This may be
any list of points for which a user wants to count as 'absences' in downflow
operations, including negative occupancy surveys, or alternatively surveys for
similar species which may act as pseudoabsences for the presence of the species
of interest.
}
\details{
Generates sets of background data based on user-specified methodology that
can be used as pseudoabsence data for the purposes of training
presence/absence models.

The four generation methods differ in how the absence locations are obtained:
\itemize{
  \item \strong{Random}: Points are sampled uniformly at random from the
    full study area, excluding a negligible buffer around presence locations
    to prevent exact overlap.
  \item \strong{Buffer}: Points are sampled within buffers of radius
    \code{buffer_distance} drawn around all fold presences, clipped to the
    reference shapefile boundary.
  \item \strong{Environmental}: Raster cells whose values fall outside the
    species tolerance envelope in at least one variable are identified as
    candidates. K-means clustering then selects a spatially
    representative subset. If \code{buffer_distance} is supplied the
    environmental filtering is applied only within that buffered region,
    implementing the full three-step approach of Senay et al. (2013).
  \item \strong{User data}: Absence locations are taken directly from
    \code{user_absence_data} to be used when user has a predefined set of
    absense points. Points are assigned to folds by spatial join against the
    partition fold boundaries, with unmatched points routed to
    temporal folds by time value. Environmental values are then extracted
    at the supplied locations using the same time-matched logic as the
    generated methods.
}
}
\examples{
data(tmr_partition_small, package = "TemporalModelR")

scl_dir   <- system.file("extdata/rasters_scaled",
                         package = "TemporalModelR")

ref_file  <- system.file("extdata/rasters_raw/elevation.tif",
                         package = "TemporalModelR")

study_crs <- sf::st_crs(terra::rast(ref_file))

study_area_sf <- sf::st_as_sf(sf::st_as_sfc(
  sf::st_bbox(c(xmin = 0, xmax = 3000, ymin = 0, ymax = 1500),
              crs = study_crs)
))

generate_absences(
  partition_result         = tmr_partition_small,
  reference_shapefile_path = study_area_sf,
  raster_dir               = scl_dir,
  variable_patterns        = c(
    "elevation"    = "elevation",
    "forest_cover" = "forest_cover_YEAR"
  ),
  method                   = "random",
  ratio                    = 1,
  time_cols                = c("year"),
  create_plot              = FALSE,
  verbose                  = FALSE
)
}
\references{
Senay SD, Worner SP, Ikeda T (2013) Novel Three-Step Pseudo-Absence
Selection Technique for Improved Species Distribution Modeling.
PLoS ONE 8(8): e71218.
}
\seealso{
Preprocessing: \code{\link{spatiotemporal_partition}},
  \code{\link{temporally_explicit_extraction}}

Modeling: \code{\link{build_temporal_hv}},
  \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
  \code{\link{build_temporal_rf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_model_assessment.R
\name{plot_model_assessment}
\alias{plot_model_assessment}
\title{Visualize Model Assessment Metrics Across Time}
\usage{
plot_model_assessment(predictions, time_column,
                      secondary_time_mode = "combine", model_result = NULL,
                      cbp_threshold = 0.05, plot_palette = "Dark 2",
                      verbose = TRUE)
}
\arguments{
\item{predictions}{List returned by
\code{\link{generate_spatiotemporal_predictions}}, or a named list with
at least a \code{timestep_metrics} element. The \code{timestep_metrics}
element may also be a path to a \code{Timestep_Assessment_Metrics.csv}
file produced by \code{\link{generate_spatiotemporal_predictions}}.}

\item{time_column}{Character. Name of the primary time column in
\code{timestep_metrics} to use as the x axis (e.g. \code{"year"}).
When predictions span multiple time columns (e.g. \code{"year"} and
\code{"season"}), provide all relevant column names as a character
vector and control how secondary columns are handled via
\code{secondary_time_mode}.}

\item{secondary_time_mode}{Character. How to handle secondary time columns
when \code{time_column} has length > 1. One of:
\itemize{
  \item \code{"combine"} (default): secondary time values are appended to
    the primary value to form a single ordered x-axis label
    (e.g. \code{1_Spring}, \code{1_Summer}, \code{2_Spring}, ...).
  \item \code{"facet"}: a separate plot is produced for each unique
    combination of secondary time values, with the primary time column
    as the x axis on every panel.
}}

\item{model_result}{List or character. Optional. Output from a
\code{build_temporal_*()} function or path to its \code{.rds} file. When
supplied, overall sensitivity and specificity from
\code{model_result$fold_test_metrics} are added as per-fold reference
lines. Default is \code{NULL}.}

\item{cbp_threshold}{Numeric. Significance threshold for CBP. Default is
\code{0.05}.}

\item{plot_palette}{Character. Name of an HCL or RColorBrewer palette used
to color folds in diagnostic plots. Accepts any HCL palette name (see
\code{\link[grDevices]{hcl.pals}}) or, if \pkg{RColorBrewer} is installed,
any Brewer palette name. Default is \code{"Dark 2"}.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress
messages during processing.}
}
\value{
Invisibly returns a named list containing:
\itemize{
  \item \code{pct_suitable}: Recorded plot of proportion of study area
    predicted suitable per time step.
  \item \code{sensitivity}: Recorded plot of per-timestep sensitivity.
  \item \code{specificity}: Recorded plot of per-timestep specificity
    (only present when pseudoabsence data were used).
  \item \code{cbp}: Recorded plot of cumulative binomial probability per
    time step on a log scale.
  \item \code{tp_fn}: Recorded plot of true positives and false negatives
    per time step.
  \item \code{tn_fp}: Recorded plot of true negatives and false positives
    per time step (only present when pseudoabsence data were used).
  \item \code{timestep_summary}: Data frame of per-time-step cross-fold
    mean and SD for each metric.
  \item \code{overall_summary}: Data frame from
    \code{predictions$overall_summary}, when present.
}
}
\description{
Generates diagnostic plots from the per-timestep assessment table produced
by \code{\link{generate_spatiotemporal_predictions}}, optionally overlaying
overall reference values from the model result object.
}
\details{
Plots per-fold and per-timestep diagnostic plots for data produced by
\code{\link{generate_spatiotemporal_predictions}}. These quick visuals can
be used by users to assess model performance and significance and decide
if the model's performance warrants further interpretation of the results
through post-processing analyses.
}
\examples{
data(tmr_predictions, package = "TemporalModelR")

plot_model_assessment(
  predictions         = tmr_predictions,
  time_column         = c("year", "season"),
  secondary_time_mode = "combine",
  verbose             = FALSE
)
}
\seealso{
Preprocessing: \code{\link{spatiotemporal_partition}},
  \code{\link{generate_absences}}

Modeling: \code{\link{build_temporal_glm}}, \code{\link{build_temporal_gam}},
  \code{\link{build_temporal_rf}}, \code{\link{build_temporal_hv}},

Post-processing: \code{\link{generate_spatiotemporal_predictions}},
  \code{\link{summarize_raster_outputs}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_documentation.R
\docType{data}
\name{tmr_glm_annual}
\alias{tmr_glm_annual}
\title{Pre-built GLM result (annual workflow)}
\format{
A list of class \code{"TemporalGLM"}.
}
\usage{
tmr_glm_annual
}
\description{
Annual variant of \code{tmr_glm}. Fit with the formula
\code{~ forest_cover + pr_ann + elevation}, a logit link, and
TSS-based threshold selection. One model per fold. Loaded with
\code{data(tmr_glm_annual)}.
}
\keyword{datasets}
